
package sam.cipher;

import sam.toolkit.*;
import sam.io.*;
import sam.cipher.Cipher;
import sam.ui.UserInterface;
import sam.core.Settings;

import java.io.*;


/** Executable class for the registered cipher implementations.
 **/
public class Benchmark {

		public static void main(String[] args) {
			sam.core.Settings.Startup.load();
			UserInterface.get().log().setLevel(5);
			if (args.length != 4) {
				System.err.println("Syntax is: java " + Benchmark.class.getName() + " sample-length-in-kB number-of-ciphers " +
													 "vertical-threads horizontal-threads");
				System.exit(3);
			}
			UserInterface.get().log().log(0, Benchmark.class, "Initializing bechmark of " + Settings.Environment.programDescription() );
			try {
				System.exit( benchmark(Integer.valueOf(args[0]).intValue() * 1024, Integer.valueOf(args[1]).intValue(),
															 Integer.valueOf(args[2]).intValue(), Integer.valueOf(args[3]).intValue() ) ? 0 : 1);
			} catch (NumberFormatException e) {
				System.err.println("Invalid number format !");
				main(new String[0]);  // This displays syntax ...
			} catch (Throwable t) {
				System.err.println("Fatal error occured: " + Text.getMessage(t));
				System.exit(2);
			}
			UserInterface.get().log().log(0, Benchmark.class, "Benchmark completed, exiting.");
		}

		private static boolean benchmark(int length, int ciphers, int verticalThreads, int horizontalThreads) throws Throwable {
			checkRange(length, "Attempt to build sample of negative size");
			checkRange(ciphers, "Attempt to use a negative number of ciphers");
			checkRange(verticalThreads, "Attempt to use a negative number of vertical threads");
			checkRange(horizontalThreads, "Attempt to use a negative number of horizontal threads");
			UserInterface.get().log().log(1, Benchmark.class, "Initializing " + length + " bytes of random data");
			final byte[] sourceData = new byte[length];
			for (int i=0; i<sourceData.length; )
				sourceData[ i++ ] = sam.toolkit.Random.nextByte();
			final String[] availableCiphers = Cipher.REGISTRY.getCiphers();
			UserInterface.get().log().log(1, Benchmark.class, "Building a stack of " + ciphers + " ciphers out of " + availableCiphers.length);
			final Cipher[] encryptingCiphers = new Cipher[ciphers];
			final Cipher[] decryptingCiphers = new Cipher[ciphers];
			// BEWARE: decrypting ciphers are stored in <b>reversed</b> order !
			for (int i=0, l=decryptingCiphers.length-1; l >= 0; ++i, --l) {
				final String cipherName = availableCiphers[ i % availableCiphers.length ];
				UserInterface.get().log().log(2, Benchmark.class, "Adding the " + cipherName + " algorithm ");
				final long[] key = new long[64];  // Which means 4096 bits ...
				for (int j=0; j<key.length; )
					key[j++] = sam.toolkit.Random.nextLong();
				encryptingCiphers[i] = Cipher.REGISTRY.newCipher(cipherName, key, Cipher.ENCRYPT);
				decryptingCiphers[l] = Cipher.REGISTRY.newCipher(cipherName, key, Cipher.DECRYPT);
			}
			UserInterface.get().log().log(1, Benchmark.class, "Timing encryption");
			long startTime = System.currentTimeMillis();
			ByteArrayOutputStream encodedByteArrayOutput = new ByteArrayOutputStream(sourceData.length);
			final long encryptionDuration = encode(sourceData, encryptingCiphers, encodedByteArrayOutput, verticalThreads, horizontalThreads);
			final byte[] encryptedData = encodedByteArrayOutput.toByteArray();
			logRate(sourceData.length, ciphers, encryptionDuration, "encrypted", horizontalThreads * verticalThreads);
			UserInterface.get().log().log(1, Benchmark.class, "Timing decryption");
			ByteArrayOutputStream decodedByteArrayOutput = new ByteArrayOutputStream(sourceData.length);
			final long decryptionDuration = encode(encryptedData, decryptingCiphers, decodedByteArrayOutput, verticalThreads, horizontalThreads);
			byte[] decryptedData = decodedByteArrayOutput.toByteArray();
			logRate(decryptedData.length, ciphers, decryptionDuration, "decrypted", horizontalThreads * verticalThreads);
			UserInterface.get().log().log(1, Benchmark.class, "Verifying data integrity:");
			final boolean exitCode = compare(sourceData, decryptedData);
			logRate(sourceData.length + decryptedData.length, ciphers, encryptionDuration + decryptionDuration, "total processed",
							horizontalThreads * verticalThreads);
			return exitCode;
		}

		private static boolean compare(final byte[] source, final byte[] target) {
			final int p = source.length;
			final int q = target.length;
			final int n = (p < q) ? p : q;
			if (p != q)
				UserInterface.get().log().log(2, Benchmark.class, "WARNING, source is " + p + " bytes long whereas target is " + q + " !");
			UserInterface.get().log().log(2, Benchmark.class, "Comparing " + n + " bytes");
			for (int i=0; i<n; ++i)
				if (source[i] != target[i]) {
					UserInterface.get().log().log(3, Benchmark.class, "source[" + i + "]=0x" + Integer.toHexString(source[i]) +
																	" differs from target[" + i +
									"]=0x" + Integer.toHexString(target[i]) );
					return false;
				}
			UserInterface.get().log().log(2, Benchmark.class, "Everything fine.");
			return true;
		}

		private static void checkRange(int number, String message) {
			if (number > 0)
				return;
			UserInterface.get().log().log(1, Benchmark.class, message + ": " + number + " !");
		}

		private static void logRate(long length, int numberOfCiphers, long duration, String message, int numberOfThreads) {
			final int rate = (int) Math.ceil( 1024.0 * length / duration);
			final int perThreadPerCipherRate = numberOfCiphers * rate / numberOfThreads;
			UserInterface.get().running().print("Statistics", length + " Bytes " +  message + " in " + (duration / 1000) + " seconds at" +
							rate + "B/s (" + perThreadPerCipherRate + "B/s per thread par cipher" + "): " +
																		(float) (CYCLE_DURATION * perThreadPerCipherRate) + " bytecode-ops per Byte per thread" );
		}


		/** Encodes <CODE>input</CODE> to <CODE>encoded</CODE> and returns the duration of the process in milliseconds **/
		private static long encode(byte[] sourceData, Cipher[] ciphers, OutputStream target, int verticalThreads, int horizontalThreads)
				throws IOException, InterruptedException {
			InputStream source = new ByteArrayInputStream(sourceData);
			ParallelIO parallelIO = null;
			if (verticalThreads > 1) {
				parallelIO = new ParallelIO();
				target = parallelIO.new GatherOutputStream(target);
			}
			for (int i=0; i<ciphers.length; i++) {
				ByteArrayFilter filter = ciphers[i].new ByteArrayFilter();
				if (horizontalThreads > 1)
					filter = new MultiThreadedByteArrayFilter(filter, new Workshop(horizontalThreads), 4096);
				target = filter.new OutputStream(16, target);
			}
			if (verticalThreads > 1)
				target = parallelIO.new ScatterOutputStream(target, new Workshop(verticalThreads));
			final long startTime = System.currentTimeMillis();
			sam.io.IO.bufferedDump(source, target);
			target.flush();
			target.close();
			return System.currentTimeMillis() - startTime;
		}

		private static final double CYCLE_DURATION = computeCycleDuration();

		private static final double computeCycleDuration() {
			UserInterface.get().log().log(0, Benchmark.class, "Benchmarkgin virtual machine");
			final long LOOP = 16 * 1024;
			double duration = 0.0;
			for (int i=0; i<10; i++) {
				final long start = System.currentTimeMillis();
				for (int j=0; j<LOOP; ++j)
					j = 2 * j - j;  // Non-optimizable calculation for <CODE>j = j</CODE>
				final long end = System.currentTimeMillis();
				duration += (double) (end - start) / (double) LOOP;
			}
			return duration / (10.0 * 2);  // 10 loops and 2 operations per iteration
		}


}

