
package sam.cipher;

import sam.toolkit.Text;
import sam.ui.UserInterface;
import sam.core.Settings;
import sam.io.IO;

import java.io.*;


/** General purpose ciphering class.
 **/
public abstract class Cipher {

  /** Exception class for ciphering functions.
	** Extending <CODE>java.lang.Error</CODE>, this has <b>not</b> to be declared in methods declarations.
	** <p>Thrown only while dealing with registration.
	** <p>Subclasses should throw instances of this class or a subclass
	** rather than standard or dedicated <CODE>Exception</CODE>s.
	**/
	public static class Error extends java.lang.Error {
		protected Error(String message) {	super(message);	}
		protected Error(Throwable pending) { super( Text.getMessage(pending) );	}
	}

	/** Encryption specifier for instantiation */
	public static final boolean ENCRYPT = true;

	/** Decryption specifier for instantiation */
	public static final boolean DECRYPT = false;

	/** Decryption specifier for instantiation */
	public static final boolean INVERSE = DECRYPT;

	/** Central registry of all known cipher implementations. **/
	public static final Registry REGISTRY = new Registry();
		
	/** Actual key assigned to this cipher **/
  protected transient final Key key;  // Key is NOT to be included during serialzation

  /** Builds a new <CODE>Cipher</CODE>
	 ** @see Key
	 **/
  protected Cipher(Key key) {	this.key = key;	}

	/** Main ciphering code.
	 ** @see cipher
	 ** @deprecated
	 **/
  protected void cipher(byte[] data) { cipher(data, 0); }

	/** Main ciphering code.
	 ** @see cipher
	 ** @deprecated
	 **/
	private void cipher(byte[] data, int offset) {
		final int size = getPreferredSize(data.length);
		final int length = data.length - offset;
		if (length <= size)
			cipher(data, offset, key);
		else {
			if (length % size != 0)
				throw new Error("Can NOT cipher block of " + length + " bytes in chunks of " + size + " bytes !");
			for (int i=offset; i<data.length; i += size)
				cipher(data, i, key);
		}
  }


  /** Either encrypts or decrypts data according to the key direction.
   ** @see Cipher.Key.direction
   ** @see Cipher.encrypt
   ** @see Cipher.decrypt
	 **/
	protected void cipher(byte[] data, int offset, Key key) {
	  if (Settings.Constants.Key.IGNORE_EMPTY && key.isEmpty())
			return;
		if (key.direction == ENCRYPT)  // Which reduces to <CODE>if (key.direction)</CODE> at <b>compile</b> time
			encrypt(data, offset, key);
		else
			decrypt(data, offset, key);
  }

	/** Called in case of encrypting key **/
  abstract protected void encrypt(byte[] data, int offset, Key key);

	/** Called in case of decrypting key **/
  abstract protected void decrypt(byte[] data, int offset, Key key);

  /** Empty cipher implementation, does nothing.
	 ** This is usefull for debugging purpose.
	 **/
  public static final Cipher EMPTY = new Cipher(null) {  // Do NOT register this "toy" class ...
		public void encrypt(byte[] data, int offset, Key key) {  }
		public void decrypt(byte[] data, int offset, Key key) {  }
		public int getPreferredSize(int size) {	return 0;  }  // Do not reserve any unused memory ...
  };

	/** <CODE>byte[]</CODE> array filter implementation based on <CODE>this</CODE> cipher.
	 ** This is useful for tranparent input/output-stream filtering
	 **/
	public class ByteArrayFilter extends sam.io.ByteArrayFilter {
		/** Builds a new <CODE>byte[]</CODE> array filter **/
		public ByteArrayFilter() { this(Settings.Constants.BYTE_ARRAY_FILTER_MAXIMUM_BUFFER_LENGTH); }
		/** Builds a new <CODE>byte[]</CODE> array filter of specified filter block length **/
		public ByteArrayFilter(int length) { super( getPreferredSize(length) ); }
		protected final void filter(byte[] data, int offset) { cipher(data, offset, key); }
	}

	/** Tests this cipher's implementation.
	 ** <p><b>Warning</b>Note that a cipher might be working even if this return false,
	 ** in particular in case of block/key size inconsistencies or inner <CODE>Key</CODE> class.
	 **  However the cipher
	 ** is <em>guaranteed</em> to work properly if this returns <CODE>true</CODE>
	 ** @return true if the cipher is working properly with current key.
	 **/
	protected boolean test() {
		final byte[] source = new byte[ getPreferredSize(64 * 1024) ];
		fill(source, 0, source.length);
		final byte[] data = (byte[]) source.clone();
		try {
			encrypt(data, 0, key);
			if (sam.toolkit.Text.compare(source, data))
				UserInterface.get().log().log(0, Cipher.class, "WARNING: no-op ciphering");
			decrypt(data, 0, key);
			return sam.toolkit.Text.compare(source, data);
		} catch (Throwable t) {
			t.printStackTrace();
			UserInterface.get().error().handleException("WARNING: exception \"" + t.getMessage() +
																									"\" occured while testing " + this.getClass().getName(), t);
			return false;
		}
	}

	/** Tests a cipher through the complete stream-based mechanism.
	 ** @param cipherName the supposedly registered cipher to be tested
	 ** @param length the sample data length
	 ** @param password the password to be used
	 ** @return wether the test was successfull or not
	 **/
	public static boolean streamTest(String cipherName, int length, String password) throws IOException, InterruptedException {
		UserInterface.get().log().log(1, Cipher.class, "Starting cipher \"" + cipherName
																		+ "\" byte[" + length + "] with password \"" + password + "\"");
		byte[] source = new byte[length];
		Cipher cipher = Cipher.REGISTRY.newCipher(cipherName, password, true);
		UserInterface.get().log().log(1, Cipher.class, "Filling byte[" + length + "] randomly");
		cipher.fill(source, 0, source.length);
			
		UserInterface.get().log().log(1, Cipher.class, "Ciphering stream of " + source.length + " bytes");
		ByteArrayOutputStream byteArrayOutput = new ByteArrayOutputStream();
		OutputStream output = cipher.new ByteArrayFilter().new OutputStream(byteArrayOutput);
		IO.bufferedDump(new ByteArrayInputStream(source), output);
		output.flush();
		output.close();
		byte[] encoded = byteArrayOutput.toByteArray();
		UserInterface.get().log().log(2, Cipher.class, "Got encoded byte[" + encoded.length + "]");

		UserInterface.get().log().log(1, Cipher.class, "Deciphering stream of " + encoded.length + " bytes");
		byteArrayOutput = new ByteArrayOutputStream();
		IO.bufferedDump( Cipher.REGISTRY.newCipher(cipherName, password, false).new ByteArrayFilter()
										 .new InputStream( new ByteArrayInputStream(encoded) ), byteArrayOutput);
		byte[] decoded = byteArrayOutput.toByteArray();
		UserInterface.get().log().log(2, Cipher.class, "Got decoded byte[" + decoded.length + "]");

		boolean success = sam.toolkit.Text.compare(source, decoded);
		if (! success)
			UserInterface.get().error().warn("Non-inversible cipher !");
		UserInterface.get().log().log(1, Cipher.class, "Done.");
		return success;
	}

	/** Tests all registered ciphers implementations.
	 ** Syntax is <CODE>java sam.cipher.Cipher [password] [cipher1 [cipher2 [...]]]</CODE>
	 ** If none specified then all registered ciphers are used.
	 **/
	public static void main(String[] args) {
		final int length = 64 * 1024;
		try {
			Settings.Startup.load();
			String[] ciphers = (args.length > 1) ? (String[]) Text.subArray(args, 1, args.length - 1) : REGISTRY.getCiphers(); 
			String password = (args.length > 0) ? args[0] : "abcdefghijklmnopqrstuvwxyz0123456789";
			byte[] data = new byte[length];
			System.out.println("Testing " + Text.toString(ciphers) + " on " + data.length + " bytes with password \"" + password + "\"");
			fill(data, 0, data.length);
			for (int i=0; i<ciphers.length; ++i) {
				System.out.print("> cipher \"" + ciphers[i] + "\": direct test ");
				try {
					byte[] encrypted = (byte[]) data.clone();
					REGISTRY.newCipher(ciphers[i], password, ENCRYPT).cipher(encrypted);
					REGISTRY.newCipher(ciphers[i], password, DECRYPT).cipher(encrypted);
					System.out.print( sam.toolkit.Text.compare(data, 0, data.length, encrypted, 0, encrypted.length) ? "successfull" : "FAILED !");
					System.out.print(", stream: ");
					System.out.println( streamTest(ciphers[i], length, password) ? "successfull" : "FAILED !");
				} catch (Error e) {
					System.err.println("could not be set up: " + e.getMessage() );
					e.printStackTrace();
				}
			}
		} catch (Throwable t) {
			UserInterface.get().error().handleException("Cipher test", t);
		} finally {
			sam.ui.LogInterface.stop();
		}
	}



  /** Called whenever the whole buffer can not be filled up with fresh data.
   ** <p>To be overriden as required.
   ** <p>This implementation fills up remaining bytes with random data.
   **/
  protected static void fill(byte[] data, int start, int end) {
		for (int i=start; i<end; )
			data[i++] = sam.toolkit.Random.nextByte();  // Not so efficient actually ...
  }


   /** Specifiy preferred buffer size.
		** <p>May return a fixed value, ignoring suggested size (e.g. <CODE>return 16;</CODE>).
		** <p>By default, this implementation simply returns its argument unchanged.
		** @param size proposed buffer size (in bytes)
		** @return adjusted buffer size (in bytes)
		**/
  protected int getPreferredSize(int size) {
		return size;
  }

   /** Stores one <CODE>int</CODE> into the first <CODE>4</CODE> cells starting at <CODE>byteOffset</CODE>
		** of a <CODE>byte[]</CODE> array.
		**/
  protected static void store(byte[] data, int byteOffset, int a) {
		data[byteOffset + 0] = (byte) (a) ;
		data[byteOffset + 1] = (byte) (a >>> 8) ;
		data[byteOffset + 2] = (byte) (a >>> 16) ;
		data[byteOffset + 3] = (byte) (a >>> 24) ;
  }

   /** Stores one <CODE>int</CODE> into the first <CODE>4</CODE> cells starting at <CODE>byteOffset</CODE>
		** of a <CODE>byte[]</CODE> array.
		**/
  protected static void store(byte[] data, int byteOffset, long a) {
		data[byteOffset + 0] = (byte) (a) ;
		data[byteOffset + 1] = (byte) (a >>> 8) ;
		data[byteOffset + 2] = (byte) (a >>> 16) ;
		data[byteOffset + 3] = (byte) (a >>> 24) ;
		data[byteOffset + 4] = (byte) (a >>> 32) ;
		data[byteOffset + 5] = (byte) (a >>> 40) ;
		data[byteOffset + 6] = (byte) (a >>> 48) ;
		data[byteOffset + 7] = (byte) (a >>> 56) ;
  }

	/** Rebuilds one <CODE>int</CODE> from specified <CODE>byte[]</CODE> array starting at <CODE>byteOffset</CODE>.
	 **/
  protected static int unstoreInt(byte[] data, int byteOffset) {
		return
			(data[byteOffset + 0] & 0xFF)       |
			(data[byteOffset + 1] & 0xFF) <<  8 |
			(data[byteOffset + 2] & 0xFF) << 16 |
			(data[byteOffset + 3] & 0xFF) << 24;
  }

	/** Rebuilds one <CODE>long</CODE> from specified <CODE>byte[]</CODE> array starting at <CODE>byteOffset</CODE>.
	 **/
  protected static long unstoreLong(byte[] data, int byteOffset) {
		return
			(data[byteOffset + 0] & 0xFFL)       |
			(data[byteOffset + 1] & 0xFFL) <<  8 |
			(data[byteOffset + 2] & 0xFFL) << 16 |
			(data[byteOffset + 3] & 0xFFL) << 24 |
			(data[byteOffset + 4] & 0xFFL) << 32 |
			(data[byteOffset + 5] & 0xFFL) << 40 |
			(data[byteOffset + 6] & 0xFFL) << 48 |
			(data[byteOffset + 7] & 0xFFL) << 56;
  }

	/** Rebuilds one <CODE>long</CODE> from specified <CODE>byte[]</CODE> array starting at <CODE>byteOffset</CODE>.
	 **/
  protected static long unstoreReversedLong(byte[] data, int byteOffset) {
		return
			(data[byteOffset + 0] & 0xFFL) << 56 |
			(data[byteOffset + 1] & 0xFFL) << 48 |
			(data[byteOffset + 2] & 0xFFL) << 40 |
			(data[byteOffset + 3] & 0xFFL) << 32 |
			(data[byteOffset + 4] & 0xFFL) << 24 |
			(data[byteOffset + 5] & 0xFFL) << 16 |
			(data[byteOffset + 6] & 0xFFL) <<  8 |
			(data[byteOffset + 7] & 0xFFL);
  }

   /** Stores one <CODE>long</CODE> into the first <CODE>4</CODE> cells starting at <CODE>byteOffset</CODE>
		** of a <CODE>byte[]</CODE> array.
		**/
  protected static void storeReversed(byte[] data, int byteOffset, long a) {
		data[byteOffset + 0] = (byte) (a >>> 56) ;
		data[byteOffset + 1] = (byte) (a >>> 48) ;
		data[byteOffset + 2] = (byte) (a >>> 40) ;
		data[byteOffset + 3] = (byte) (a >>> 32) ;
		data[byteOffset + 4] = (byte) (a >>> 24) ;
		data[byteOffset + 5] = (byte) (a >>> 16) ;
		data[byteOffset + 6] = (byte) (a >>>  8) ;
		data[byteOffset + 7] = (byte) (a >>>  0) ;
  }

}
