
package sam.cipher;

import sam.ui.UserInterface;
import sam.core.Settings;


/** General-purpose <CODE>Cipher</CODE> factory.
 ** This is an abstraction for instantiating various ciphering algorithm and parsing keys.
 **/
public abstract class Factory {

	/** The code name of this cipher **/
	public final String name;

	/** A human-readable description of this cipher.
	 ** This should include any copyright notice.
	 **/
	public final String description;

	/** Builds a new <CODE>Cipher</CODE> factory
	 ** @param name the code name of this cipher
	 ** @param description a human-readable description of this cipher.
	 **/
	protected Factory(String name, String description) {
		this.name = name;
		this.description = description;
		UserInterface.get().log().log(2, Factory.class, "Loaded " + name + " factory");
	}

	/** Instantiates a new <CODE>Cipher</CODE> implementation.
	 ** @param key the key as full <CODE>long</CODE>s, its length having been retrofitted by the
	 ** <CODE>getPreferredKeySize</CODE> method.
	 ** @param mode <CODE>true</CODE> if encrypting, <CODE>false</CODE> otherwise (decrypting).
	 ** @throws Error in case of impossible instantiation
	 ** @see Cipher#ENCRYPT
	 ** @see Cipher#DECRYPT
	 **/
	abstract protected Cipher make(long[] key, boolean mode) throws Error;

	/** Creates a new <CODE>Cipher</CODE> instance with specified key.
	 ** First fits the key length to cipher's specfied length.
	 ** @throw Error if passed array is empty and empty arays are not allowed.
	 ** @see #make
	 ** @see #getPreferredKeySize
	 */
	Cipher newInstance(long[] key, boolean mode) throws Error {
		UserInterface.get().log().log(2, Factory.class, "Initializing " + name + " " + ((mode == Cipher.ENCRYPT) ? "cipher" : "uncipher") );
		return make( Key.repack(resizeKeyToPreferredSize(Key.expand(key, 64)), 64), mode );
	}

	/** Specifies preferred key size, <b>in bits</b>.
	 ** <b>This shall be overridden</b> to match legal key sizes of specific algorithms.
	 ** @param suggestedKeySize the suggested key size as computed from the password length
	 ** @return a key size suited to this cipher implementation
	 ** @see Key
	 ** @see #highestMultiple
	 **/
	protected int getPreferredKeySize(int suggestedKeySize) { return suggestedKeySize; }

	/** Gets the highest integer <b>less than</b> or equal to <CODE>number</CODE> that is a multiple
	 ** of <CODE>root</CODE>,
	 ** and that is higher than <CODE>root</CODE>. Simply <CODE>root</CODE> if no such integer.
	 **/
	protected static int highestMultiple(int number, int root) {
		int result = root * (int) Math.floor((double)number / (double) root);
		return (result < root) ? root : result;
	}


	/** Gets the highest integer <b>less than</b> or equal to <CODE>suggested</CODE> from exhaustive list of
	 ** valid values,
	 ** or minimum of those if such a number does not exist.
	 **/
	protected static int highestIn(int suggested, int[] authorized) throws Error {
		int highest = -1;
		int min = Integer.MAX_VALUE;
		for (int i=0; i<authorized.length; i++) {
			if (authorized[i] < min)
				min = authorized[i];
			if ( (authorized[i] < suggested) && (authorized[i] > highest))
				highest = authorized[i];
		}
		return (highest < 0) ? min : highest;
	}


	/** Ensures a size fits in specified range.
	 ** @return <CODE>suggested</CODE> if it fits in specified interval, or bound otherwise.
	 ** @throws Error if <CODE>max</CODE> is less than <CODE>min</CODE>.
	 **/
	protected static int ranged(int suggested, int min, int max) throws Error {
		if (min > max)
			throw new Error("No integer can fit between " + min + " and " + max + " !");
		return (suggested < min) ? min : (suggested > max) ? max : suggested;
	}

	/** Resizes <CODE>key</CODE> to its desired length.
	 ** @see #getPreferredKeySize
	 **/
	protected boolean[] resizeKeyToPreferredSize(boolean[] key) throws Error {
		return (key == null) ? null : resizeKey(key, getPreferredKeySize(key.length) );
	}

	/** Resizes <CODE>key</CODE> to specified length.
	 ** If the current size is larger, the key is truncated.
	 ** If it is smaller, the key is concatenated to itself until it is large enough.
	 ** @param bits the key material to be resized
	 ** @param keySize the desired new key size
	 ** @throws Error if this key is empty and expected size is not
	 **/
	protected static boolean[] resizeKey(boolean[] key, int keySize) throws Error {
		if (keySize == key.length)
			return key;
		boolean[] newKey = new boolean[keySize];
		if (keySize == 0) {
			if (! Settings.Constants.Key.ALWAYS_SILENT)
				if (Settings.Constants.Key.WARN_EMPTY)
					UserInterface.get().error().warn("Empty key built from " + key.length + " bits");
				else
					UserInterface.get().log().log(11, Factory.class, "Empty key built from " + key.length + " bits");
		}
		if (keySize < key.length) {
			if (! Settings.Constants.Key.ALWAYS_SILENT) {
				if (Settings.Constants.Key.LOG_TRUNCATIONS)
					UserInterface.get().log().log(11, Factory.class, "Truncating key of size " + key.length + " to " + keySize + " bits");
				if (Settings.Constants.Key.WARN_TRUNCATIONS)
					UserInterface.get().error().warn("Truncating key of size " + key.length + " to " + keySize + " bits");
			}
			System.arraycopy(key, 0, newKey, 0, keySize);
		} else {  // (keySize > key.length)
			if (key.length == 0)
				throw new Error("Empty key can not be expanded to a " + keySize + " bits wide key !");
			if (! Settings.Constants.Key.ALWAYS_SILENT) {
				if (Settings.Constants.Key.LOG_EXPANSIONS)
					UserInterface.get().log().log(11, Factory.class, "Expanding key of size " + key.length + " to " + keySize + " bits");
				if (Settings.Constants.Key.WARN_EXPANSIONS)
					UserInterface.get().error().warn("Expanding key of size " + key.length + " to " + keySize + " bits");
			}
			int offset;
			for (offset = 0; offset < keySize - key.length; offset += key.length)
				System.arraycopy(key, 0, newKey, offset, key.length);
			System.arraycopy(key, 0, newKey, offset, keySize - offset);
		}
		return newKey;
	}

}

