
/*
  This file is distributed under the implicit authorization of the author, due to its web plublication.
  This whole code remains the exclusive property of its original author, through his responsibility for the
  use of this file may in no case be engaged.
  Note that this file is NOT distributed under the terms of the General Public License.

*/

package sam.cipher;


/** Implementation of the <a href="http://www.tecapro.com/aesfrog.htm">Frog</a> cipher.
 *
 * Retrofitted for personal use by <a href=mailto:sam@glacote.com>Guillaume Lacte</a>.
 *
 * <p>
 *                                                              *
 * Portions of this code are <b>Copyright</b> &copy; 1997, 1998
 * <a href="http://www.systemics.com/">Systemics Ltd</a> on behalf of the
 * <a href="http://www.systemics.com/docs/cryptix/">Cryptix Development Team</a>.
 * <br>All rights reserved.<p>
 *
 * <b>$Revision: $</b>
 * @author  Raif S. Naffah
 */
public class Frog extends Cipher {

    static { 
			Cipher.REGISTRY.register( new ReflexiveFactory( Frog.class, "Frog",
																						 "Frog is a block cipher with key lengths ranging from 40 up to 1000 bits\n\n" +
																						 "Candidate algorithm for the NIST AES Quest.\n" +
																						 "See http://www.tecapro.com/aesfrog.htm for further details.\n") {
						public int getPreferredKeySize(int size) {
							return ranged(size, 40, 1000);
						}
			} );
		}

    public Frog(sam.cipher.Key key) {
			super(key);
    }

		protected int getPreferredSize(int size) { return BLOCK_SIZE;	}

		/** Key implementation that is adapted to the {@link Frog} cipher. **/
    public static class Key extends sam.cipher.Key {
				private final Entry[] internalKey, cipherBits;

				/** Mandatory constructor implementation.
				 ** @see sam.cipher.ReflexiveFactory
				 **/
				public Key(long[] longs) {
					this(longs, true);
				}

				public sam.cipher.Key inverse() {
					return new Key(bits, ! direction);
				}

				public boolean isEmpty() {
					return (internalKey.length == 0) || (cipherBits.length == 0) || super.isEmpty();
				}

				private Key(long[] longs, boolean direction) {
					super(longs, 64, direction);
					internalKey = hashKey( getBytes() );
					cipherBits = makeInternalKey(direction, internalKey );
				}

				void set(int index, int value) {
					internalKey[index / Key.Entry.size()].set( index % Key.Entry.size(), value );
				}

				int get(int index) {
					return internalKey[index / Key.Entry.size()].get( index % Key.Entry.size() );
				}

				protected static class Entry {
						int[] xorBuffer;
						int[] SubstPermu;
						int[] BombPermu;

						protected Entry() {
								xorBuffer = new int[BLOCK_SIZE];
								SubstPermu = new int[256];
								BombPermu = new int[BLOCK_SIZE];
						}

						public static int size() {
								return BLOCK_SIZE * 2 + 256;
						}

						public void set(int index, int value) {
								if (value < 0)
										value = 256 + value;
								if (index < BLOCK_SIZE) 
										xorBuffer[index] = value;
								else if (index < BLOCK_SIZE + 256) 
										SubstPermu[index-BLOCK_SIZE] = value;
								else
										BombPermu[index-BLOCK_SIZE-256] = value;  
						}

						public int get(int index) {
								if (index < BLOCK_SIZE) 
										return xorBuffer[index];
								else if (index < BLOCK_SIZE + 256) 
										return SubstPermu[index-BLOCK_SIZE];
								else
										return BombPermu[index-BLOCK_SIZE-256];  
						}

						public void CopyFrom(Entry source) {
								int i;
								for (i=0; i<source.xorBuffer.length; i++)
										xorBuffer[i] = source.xorBuffer[i];
								for (i=0; i<source.SubstPermu.length; i++)
										SubstPermu[i] = source.SubstPermu[i];
								for (i=0; i<source.BombPermu.length; i++)
										BombPermu[i] = source.BombPermu[i];
						}

				}


				/*
					Processes unstructured internalKey into a valid internalKey.
					Reference Text: section B.1.2
				*/
				private static Key.Entry[] makeInternalKey(boolean isEncrypting, Key.Entry[] source) {
						byte[] used = new byte[BLOCK_SIZE];
						int ite, j, i, k, l;
						int posi;
						byte change;
						if (source.length == 0)
								return new Key.Entry[0];
						Key.Entry[] key = new Key.Entry[ ROUNDS ];
						for ( i=0; i < ROUNDS; i++ ) {
								key[i] = new Key.Entry();
								key[i].CopyFrom( source[i] );
						}
						posi = 1;
						for ( ite = 0; ite < ROUNDS; ite ++ ) {
								key[ite].SubstPermu = makePermutation( key[ite].SubstPermu );
								if ( ! isEncrypting )
										key[ite].SubstPermu = invertPermutation(key[ite].SubstPermu);
								// See B.1.1a
								key[ite].BombPermu = makePermutation(key[ite].BombPermu );
								/* Join smaller cycles in BombPermu into one cycle
									 (See B.1.1b for rational and B.1.4 for code) */
								for ( i=0; i<BLOCK_SIZE; used[i]=0, i++ )
										;
								j = 0;
								for ( i=0; i<BLOCK_SIZE-1; i++ ) {
										if ( key[ite].BombPermu[j] == 0 ) {
												k = j;
												do {
														k = ( k + 1 ) % BLOCK_SIZE;
												} while ( used[k]!=0 );
												key[ite].BombPermu[j] = k;
												l = k;
												while ( key[ite].BombPermu[l]!=k ) l = key[ite].BombPermu[l];
												key[ite].BombPermu[l] = 0;
										}
										used[j] = 1;
										j = key[ite].BombPermu[j];
								}

								/* Remove references to next element within BombPermu.
									 See B.1.1c for rational and B.1.4.b for code. */
								for (i = 0; i < BLOCK_SIZE; i++) {
										j = (i == BLOCK_SIZE-1) ? 0 : i + 1;
										if (key[ite].BombPermu[i] == j) {
												k = (j == BLOCK_SIZE-1) ? 0 : j + 1;
												key[ite].BombPermu[i] = k;
										}
								}
						}
						return key;
				}
  
				private static Key.Entry[] hashKey(byte[] binaryKey) {
						/* Hash binaryKey of keyLen bytes into randomKey
							 Reference Text: section B.1.2 */
						int keyLen = binaryKey.length;
						if (keyLen == 0)
								return new Key.Entry[0];
						byte[] buffer = new byte[BLOCK_SIZE];
						Key.Entry[] simpleKey = new Key.Entry[ ROUNDS ];
						Key.Entry[] internalKey = new Key.Entry[ ROUNDS ];
						int iSeed, iFrase;
						int sizeKey,i,posi,size, last;
						for ( i = 0; i < ROUNDS; i++ )
								simpleKey[i] = new Key.Entry();
						for ( i = 0; i < ROUNDS; i++ )
								internalKey[i] = new Key.Entry();
						sizeKey = Key.Entry.size() * ROUNDS;
						/* Initialize SimpleKey with user supplied key material and random seed.
							 See B.1.2a */
						iSeed = 0; iFrase = 0;
						for ( i = 0; i < sizeKey; i++ )	{
								simpleKey[i / Key.Entry.size()].set(i % Key.Entry.size(), randomSeed[iSeed] ^ binaryKey[iFrase]);
								if ( iSeed<250 )
										iSeed++;
								else
										iSeed = 0;
								if ( iFrase<keyLen-1 )
										iFrase++;
								else
										iFrase = 0;
						}
						// Convert simpleKey into a valid internal key (see B.1.2b)
						simpleKey = makeInternalKey(true, simpleKey);
						for ( i = 0; i < BLOCK_SIZE; buffer[i++] = 0 )
								;
						// Initialize IV vector (see B.1.2c)
						last = keyLen - 1;
						if ( last > BLOCK_SIZE )
								last = BLOCK_SIZE-1;
						for ( i = 0; i <= last; buffer[i] ^= binaryKey[i], i++ )
								;
						buffer[0] ^= keyLen;
						posi = 0;
						/* Fill randomKey with the cipher texts produced successive
							 encryptions (see B.1.2.c) */
						do {
								encrypt(buffer, 0, simpleKey);
								size = sizeKey - posi;
								if ( size > BLOCK_SIZE ) size = BLOCK_SIZE;
								for (i=0;i<BLOCK_SIZE;i++)
										if ( buffer[i] < 0 )
												internalKey[(posi+i)/Key.Entry.size()].set((posi+i)%Key.Entry.size(), buffer[i]+256 ); 
										else
												internalKey[(posi+i)/Key.Entry.size()].set((posi+i)%Key.Entry.size(), buffer[i] ); 
								posi += size;
						} while ( posi != sizeKey );
						return internalKey;
				}

    }



    /**
     * Encrypt exactly one block of plaintext.
     *
     * @param data          The plaintext.
     * @param sessionKey  The session key to use for encryption.
     */
    protected void encrypt (byte[] data, int offset, sam.cipher.Key key) {
			encrypt(data, offset, ((Key) key).cipherBits);
    }

    private static void encrypt (byte[] data, int offset, Key.Entry[] key) {
			for (int ite=0; ite<ROUNDS; ite++ )
				for (int ib=0; ib < BLOCK_SIZE; ++ib ) {
					data[offset + ib] ^= key[ite].xorBuffer[ib];
					data[offset + ib] = (byte) key[ite].SubstPermu[ (data[offset + ib] < 0) ? (data[offset + ib] + 256) : data[offset + ib] ];
					data[offset + ((ib < BLOCK_SIZE - 1) ? (ib + 1) : 0)] ^= data[offset + ib];
					data[offset +  key[ite].BombPermu[ib] ] ^= data[offset + ib];
				}
    }

    /**
     * Decrypt exactly one block of ciphertext.
     *
     * @param data          The ciphertext.
     * @param sessionKey  The session key to use for decryption.
     */
		protected void decrypt (byte[] data, int offset, sam.cipher.Key aKey) {
			Key.Entry[] key = ((Key) aKey).cipherBits;
			for (int ite = ROUNDS - 1; ite >= 0; ite-- )
				for (int ib = BLOCK_SIZE - 1; ib >= 0; ib-- ) {
					data[offset + key[ite].BombPermu[ib] ] ^= data[offset + ib];
					data[offset + ((ib < BLOCK_SIZE - 1) ? (ib + 1) : 0)] ^= data[offset + ib];
					data[offset + ib] = (byte) key[ite].SubstPermu[ (data[offset + ib] < 0) ? (data[offset + ib] + 256) : data[offset + ib] ];
					data[offset + ib] ^= key[ite].xorBuffer[ib];
				}
    }

    // -------------------------------
    // Utilities and constants
    private static final int BLOCK_SIZE = 16; // bytes in a data-block

    private static final int ROUNDS = 8;

    // Values defined from RAND Corporation's "A Million Random Digits"
    private static int[] randomSeed = {
				113, 21,232, 18,113, 92, 63,157,124,193,166,197,126, 56,229,229,
				156,162, 54, 17,230, 89,189, 87,169,  0, 81,204,  8, 70,203,225,
				160, 59,167,189,100,157, 84, 11,  7,130, 29, 51, 32, 45,135,237,
				139, 33, 17,221, 24, 50, 89, 74, 21,205,191,242, 84, 53,  3,230,
				231,118, 15, 15,107,  4, 21, 34,  3,156, 57, 66, 93,255,191,  3,
				85,135,205,200,185,204, 52, 37, 35, 24, 68,185,201, 10,224,234,
				7,120,201,115,216,103, 57,255, 93,110, 42,249, 68, 14, 29, 55,
				128, 84, 37,152,221,137, 39, 11,252, 50,144, 35,178,190, 43,162,
				103,249,109,  8,235, 33,158,111,252,205,169, 54, 10, 20,221,201,
				178,224, 89,184,182, 65,201, 10, 60,  6,191,174, 79, 98, 26,160,
				252, 51, 63, 79,  6,102,123,173, 49,  3,110,233, 90,158,228,210,
				209,237, 30, 95, 28,179,204,220, 72,163, 77,166,192, 98,165, 25,
				145,162, 91,212, 41,230,110,  6,107,187,127, 38, 82, 98, 30, 67,
				225, 80,208,134, 60,250,153, 87,148, 60, 66,165, 72, 29,165, 82,
				211,207,  0,177,206, 13,  6, 14, 92,248, 60,201,132, 95, 35,215,
				118,177,121,180, 27, 83,131, 26, 39, 46, 12};
  
    /**	Receives an arbitrary byte array of (lastElem -1) elements and
				returns a permutation with values between 0 and lastElem.
				Reference Text: section B.1.3
		*/
    private static int[] makePermutation( int[] permu ) {
				int use[] = new int[256];
				int i, j, k, count, last;
				int lastElem = permu.length - 1;

				// Initialize use array
				for (i=0; i<=lastElem; use[i]=i, i++)
						;
				last = lastElem;
				j = 0;
				// Fill permutation with non-sequencial, unique values
				for (i=0; i<lastElem; i++ ) {
						j = (j+permu[i]) % (last+1);
						permu[i] = use[j];
						// Remove use[index] value from use array
						if ( j < last )
								for ( k=j; k <= last-1; use[k] = use[k+1], k++ );
						last--;
						if ( j > last ) j = 0;
				}
				permu[lastElem] = use[0];
				return permu;
    }

    /** 
				Inverts a permutation
    */
    private static int[] invertPermutation( int[] orig ) {
				int invert[] = new int[256];
				int i, lastElem = orig.length-1;
				for ( i=0; i<=lastElem; i++ )
						invert[orig[i]] = i;
				return invert;
    }
  

    /**
       moves an entire block of size bytes 1 bit to the left
    */
    private static byte[] shif1bitLeft( byte[] buffer ) {
				byte[] result = new byte[ BLOCK_SIZE ];
				int i;
				for ( i = BLOCK_SIZE-1; i >= 0 ; i-- )
					result[i] = buffer[i];
				for ( i = BLOCK_SIZE-1; i >= 0 ; i-- ) {
					result[i] = (byte) (result[i] << 1);
					if (i > 0) 
						result[i] |= result[i-1] >> 7;
				}
				return result;
    }

}

