
package sam.cipher;

import sam.core.Settings;
import sam.ui.UserInterface;


/** General purpose key class.
 ** <p>This class's data is immutable, and so should be subclasses.
 ** <p>You may create a sub-class in the following way:
 ** <pre><codeblock>
 ** class MyKey extends Cipher.Key {
 **   public MyKey(long[] longs) {
 **     super( makeKey(longs), 27 );  //
 **   }
 **   private static long[] makeKey(long[] longs) {
 **      long[] key = new long[39];
 **      ...  // Use 64 bits of 'longs[i]' to generate 27 bits in 'key[j]'
 **    return key;
 **   }
 **   public Key inverse() { return new MyKey(...); }
 ** }
 ** </codeblock></pre>
*/
public class Key {

		/** One char issues no single bit */
		public static final Key EMPTY = new Key(new long[0], (char) 0);

		/** One char issues 16 bits */
		public static final Mode UNICODE = new Mode(16, (char) 0);
		/** One char issues 8 bits */
		public static final Mode ASCII   = new Mode(8, (char) 0);
		/** One char issues 6 bits.
				<p>Default mode for keyboard-entered passworsd.*/
		public static final Mode PLAIN   = new Mode(6, '0');
		/** One char (<CODE>0,1</CODE>) issues 1 bit */
		public static final Mode BINARY  = new Mode(1, '0');
		/** One char (<CODE>0,1,..,7</CODE>) issues 3 bits */
		public static final Mode OCTAL   = new Mode(3, '0');
		/** One char (<CODE>0,1,..9,A,..F</CODE>) issues 4 bits */
		public static final Mode HEXA    = new Mode(4, '0');
		/** One char issues no bit at all */
		public static final Mode IGNORE = new Mode(0, (char) 0);

		/**
		   Parsing mode for converting characters to significant bits.
		*/
		public static final class Mode {
				private final int sourceBits;
				private final int offset;
				private Mode(int sourceBits, char offset) {
						this.sourceBits = (sourceBits >= 0) ? (sourceBits <= 16) ? sourceBits : 16 : 0;
						this.offset = offset;
				}
		}

		/**  Reports wether this is an encryption- or a decryption-key.	*/
		public final boolean direction;

		/**  Actual bits.	*/
		protected final long[] bits;

		/** Number of significant bits per <CODE>long</CODE> of <CODE>bits</CODE>.	*/
		protected final int bitsPerLong;

		/**  Mask for keeping only the significant bits of a <CODE>long</CODE>.	*/
		protected final int mask;

		protected final boolean[] asBooleans;
		protected final byte[]    asBytes;
		protected final short[]   asShorts;
		protected final int[]     asInts;
		protected final long[]    asLongs;


		/** Builds a new key.
		 ** @param password the plain text passphrase
		 ** @param mode parsing mode
		 ** @param bitsPerLong desired number of bits to be stored per int (max 32).
		 ** @see parse
		 **/
		public Key(String password, Mode mode, int bitsPerLong) {	this(parse(password, mode, bitsPerLong), bitsPerLong); }

		/** Builds a new key.
		 ** @param bits the <CODE>boolean[]</CODE> array of bits composing the key material
		 **/
		protected Key(boolean[] bits) {	this( repack(bits, 64) ); }  // Since 64 is the default bits-per-<CODE>long</CODE>

		/** Builds a new key.
		 ** @param bits the <CODE>long[]</CODE> array of 64-bit groups composing the key material
		 ** Note that this constructor <b>must</b> be public as <CODE>sam.cipher.Key</CODE> might
		 ** be the default key used.
		 **/
		public Key(long[] bits) { this(bits, 64); }  // Set 64 as the default bits-per-<CODE>long</CODE>

		/** Builds a new key.
		 ** @param bits the <CODE>long[]</CODE> array of bit groups composing the key material
		 ** @param bitsPerLong the number of significant bits per 64-bits <CODE>long</CODE> to be used
		 **/
		protected Key(long[] bits, int bitsPerLong) {	this(bits, bitsPerLong, true); }

		/** Builds a new key.
		 ** @param bits the <CODE>long[]</CODE> array of bit groups composing the key material
		 ** @param bitsPerLong the number of significant bits per 64-bits <CODE>long</CODE> to be used
		 ** @param direction either {@link Cipher#ENCRYPT} or {@link Cipher#DECRYPT}
		 **/
		protected Key(long[] bits, int bitsPerLong, boolean direction) {
			this.bits = bits;
			this.bitsPerLong = bitsPerLong;
			this.mask = (1 << bitsPerLong) - 1;
			this.direction = direction;
			if ( (! Settings.Constants.Key.ALWAYS_SILENT) && (Settings.Constants.Key.LOG_KEY_INSTANTIATION) )
				UserInterface.get().log().log(4, Key.class, "Created new " + getClass().getName() +
																			((direction == Cipher.ENCRYPT) ? " encrypting" : " decrypting")	+ " key" +
																			(Settings.Constants.Key.LOG_KEY_MATERIAL_LENGTH
																			 ? (" with " + (bits.length * bitsPerLong) + "bits") : "") 
																			);

			this.asBooleans = expandToBooleans();
			this.asBytes    = expandToBytes();
			this.asShorts   = expandToShorts();
			this.asInts     = expandToInts();
			this.asLongs    = expandToLongs();
		}

		/** Wether this key is empty **/
		public boolean isEmpty() {	return (bits.length == 0) || (bitsPerLong < 1); 	}

		/** Inverses this key.
		 ** This turns a ciphering key into a de-ciphering one, and vice-versa.
		 ** To be overriden as required.
		 ** <p>This implementation simply returns <CODE>this</CODE>.
		 **/
		public Key inverse() {
			if (Settings.Constants.Key.WARN_SUBCLASSES_INVERSION && ( ! Key.class.equals( getClass() ) ))
				UserInterface.get().error().warn("Inversing " + getClass().getName() + " key as a base " +
																				 Key.class.getName() + " key; key material might be lost");
			return new Key(bits, bitsPerLong, ! direction);
		}

		/** Gets a human-readable description of this key.
		 ** <b>Warning:</b> this methods returns the actual key material
		 ** @return description String of the form <CODE>"Key(8|0xff)[0011011000101]"</CODE>.
		 ** @see #toPassword
		 **/
		public String toString() {
			StringBuffer buffer =
				new StringBuffer().append("Key(").append(bitsPerLong).append("|0x").append( Integer.toHexString(mask) ).append(")[");
			for (int i=0; i<bits.length; i++)
				buffer.append( Long.toBinaryString(bits[i]) );
			return buffer.append(']').toString();
		}

		/** Unparses a key to a passphrase.
		 ** <b>Warning:</b> this methods returns the actual key material
		 ** @param mode the parsing mode.
		 ** @return the password.
		 **/
		public String toPassword(Mode mode) {
				StringBuffer buffer = new StringBuffer();
				long[] repacked = repack( expand(bits, bitsPerLong), mode.sourceBits);
				for (int i=0; i<repacked.length; i++) {
						buffer.append( (char) (repacked[i] + mode.offset) );
				}
				return buffer.toString();
		}


		/** Unparses a key to a passphrase.
		 ** The default unparsing mode is <CODE>PLAIN</CODE>
		 ** <b>Warning:</b> this methods returns the actual key material
		 ** @param mode the parsing mode.
		 ** @return the password.
		 **/
		public String toPassword() { return toPassword(PLAIN);	}

		/** @see #getBooleans	**/
		public boolean[] get1()          { return getBooleans();	}
		/** @see #getBooleans	**/
		public boolean[] getBits()       { return getBooleans();	}
		/**  @return all significant bits repacked in a one-bit <CODE>boolean[]<CODE> array. **/
		public boolean[] getBooleans()   { return asBooleans; }

		private boolean[] expandToBooleans()   { return expand(bits, bitsPerLong); }

		/** @see #getBytes	**/
		public byte[] get8() {	return getBytes();	}
		/**  @return all significant bits repacked in a 8-bits <CODE>byte[]<CODE> array. **/
		public byte[] getBytes() { return asBytes; }

		public byte[] expandToBytes() {
			final long[] longs8 = repack(expand(bits, bitsPerLong), 8);
			byte[] bytes = new byte[longs8.length];
			for (int i=0; i<longs8.length; i++)
				bytes[i] = (byte) longs8[i];
			return bytes;
		}

		/** @see #getShorts	**/
		public short[] get16() { return getShorts(); }
		/**  @return all significant bits repacked in a 16 bits <CODE>short[]<CODE> array. **/
		public short[] getShorts() { return asShorts; }

		public short[] expandToShorts() {
				final long[] longs16 = repack(expand(bits, bitsPerLong), 16);
				short[] shorts = new short[longs16.length];
				for (int i=0; i<longs16.length; i++)
						shorts[i] = (short) longs16[i];
				return shorts;
		}

		/**  @see #getInts	**/
		public int[] get32() { return getInts(); }
		/** @return all significant bits repacked in a 32 bits <CODE>int[]<CODE> array. **/
		public int[] getInts() { return asInts; }

		public int[] expandToInts() {
				final long[] longs32 = repack(expand(bits, bitsPerLong), 32);
				int[] ints = new int[longs32.length];
				for (int i=0; i<longs32.length; i++)
						ints[i] = (int) longs32[i];
				return ints;
		}

		/**  @see #getLongs	**/
		public long[] get64() {	return getLongs(); }
		/** @return all significant bits repacked in a 64 bits <CODE>long[]<CODE> array. **/
		public long[] getLongs() { return asLongs; }

		/** @return all significant bits repacked in a 64 bits <CODE>long[]<CODE> array. **/
		public long[] expandToLongs() { return repack(expand(bits, bitsPerLong), 64);	}


		/**
		   General-purpose.
		   <p>Required to be overriden by subclasses only if specific instance data are declared.
		*/
		public boolean equals(Object object) {
				if (object == this)
						return true;
				if ( ! (object instanceof Key))
						return false;
				Key other = (Key) object;
				if (bitsPerLong != other.bitsPerLong)
						return false;
				for (int i=0; i<bits.length; i++)
						if (bits[i] != other.bits[i])
								return false;
				return true;
		}
	

		/** Parses a <CODE>String</CODE> passphrase to a <CODE>long[]</CODE> array key material.
		 ** @param password the password to be parsed
		 ** @param mode the parsing mode
		 ** @param bitsPerLong the desired number of significant bits to be stored in each resulting <CODE>long</CODE>.
		 ** <p>Note that the actual number of significant bits of the password is part of the <CODE>mode</CODE> parameter,
		 ** the resulting array is thus sized accordingly to the <CODE>bitsPerLong</CODE> parameter.
		 **/
		protected static long[] parse(String password, Mode mode, int bitsPerLong) {
			return repack( expand( toLongArray(password, mode.offset), mode.sourceBits), bitsPerLong);
		}

		/** Converts an array of <CODE>long</CODE>s to an array of <CODE>byte</CODE>s, without any loss of data.
		 ** @param longs the source array of 64-bits full <CODE>long</CODE>s.
		 ** @return an four times larger array of 8-bits full <CODE>byte</CODE>s.
		 **/
		protected static byte[] toBytes(long[] longs) {
			long[] expanded = repack(expand(longs, 64), 8);
			byte[] bytes = new byte[expanded.length];
			for (int i=0; i<expanded.length; i++)
				bytes[i] = (byte) expanded[i];
			return bytes;
		}

		/** Converts an array of <CODE>boolean</CODE>s to an array of <CODE>long</CODE>s, without
		 ** any loss of data.
		 ** @param booleans the source array of 1-bit full <CODE>boolean</CODE>s.
		 ** @return an 64 times smaller array of 64-bits full <CODE>long</CODE>s.
		 **/
		protected static long[] toLongs(boolean[] booleans) {
			long[] longs = new long[ booleans.length ];
			for (int i=0; i<booleans.length; i++)
				longs[i] = (long) (booleans[i] ? 1 : 0);
			return longs;
		}

		/** Converts an array of <CODE>byte</CODE>s to an array of <CODE>long</CODE>s, without
		 ** any loss of data.
		 ** @param bytes the source array of 8-bits full <CODE>byte</CODE>s.
		 ** @return an 8 times smaller array of 64-bits full <CODE>long</CODE>s.
		 **/
		protected static long[] toLongs(byte[] bytes) {
				long[] longs = new long[ bytes.length ];
				for (int i=0; i<bytes.length; i++)
						longs[i] = (long) bytes[i];
				return longs;
		}

		/** Converts an array of <CODE>short</CODE>s to an array of <CODE>long</CODE>s, without
		 ** any loss of data.
		 ** @param shorts the source array of 16-bits full <CODE>short</CODE>s.
		 ** @return an 4 times smaller array of 64-bits full <CODE>long</CODE>s.
		 **/
		protected static long[] toLongs(short[] shorts) {
				long[] longs = new long[ shorts.length ];
				for (int i=0; i<shorts.length; i++)
						longs[i] = (long) shorts[i];
				return longs;
		}

		/** Converts an array of <CODE>char</CODE>s to an array of <CODE>long</CODE>s, without
		 ** any loss of data.
		 ** @param shorts the source array of 16-bits full <CODE>char</CODE>s.
		 ** @return an 4 times smaller array of 64-bits full <CODE>long</CODE>s.
		 **/
		protected static long[] toLongs(char[] chars) {
				long[] longs = new long[ chars.length ];
				for (int i=0; i<chars.length; i++)
						longs[i] = (long) chars[i];
				return longs;
		}

		/** Converts an array of <CODE>int</CODE>s to an array of <CODE>long</CODE>s, without
		 ** any loss of data.
		 ** @param ints the source array of 32-bits full <CODE>int</CODE>s.
		 ** @return an twice smaller array of 64-bits full <CODE>long</CODE>s.
		 **/
		protected static long[] toLongs(int[] ints) {
				long[] longs = new long[ ints.length ];
				for (int i=0; i<ints.length; i++)
						longs[i] = (long) ints[i];
				return longs;
		}


		/** Wraps a <CODE>String</CODE> of <CODE>char</CODE>s to an array of <CODE>long</CODE>s, without
		 ** any loss of data.
		 ** @param password the <CODE>char</CODE> <CODE>String</CODE>
		 ** @param offset the constant offset to be arithmetically substracted to each <CODE>char</CODE>
		 ** to turn it to a <CODE>long</CODE>
		 ** @return a <CODE>long[]</CODE> array of same size
		 **/
		static long[] toLongArray(String password, int offset) {
			long[] array = new long[ password.length() ];
			for (int i=0; i<array.length; i++)
				array[i] = (long) password.charAt(i) - offset;
			return array;
		}

		/** Expands a <CODE>long[]</CODE> array to a <CODE>boolean[]</CODE> array, with possibly loss of data.
		 ** @param data the <CODE>long[]</CODE> array to be expanded
		 ** @param bits the number of significant bits to be kept from each <CODE>long</CODE>
		 **/
		static boolean[] expand(long[] data, int bits) {
			boolean[] expanded = new boolean[ bits * data.length ];
			for (int i=0; i<expanded.length; i++)
				expanded[i] = ( (data[i / bits] & (1 << (i % bits))) != 0);
			return expanded;
		}

		/** Repacks a <CODE>boolean[]</CODE> array to a <CODE>long[]</CODE> array, without loss of data.
		 ** @param data the <CODE>boolean[]</CODE> array to be repacked
		 ** @param bits the number of significant bits to be kept from each <CODE>long</CODE>
		 **/
		static long[] repack(boolean[] data, int bits) {
			if (bits == 0)
				return new long[0];
			int size = data.length / bits;
			if (data.length % bits != 0)
				size ++;
			long[] repacked = new long[ size ];
			for (int i = data.length - 1; i >= 0 ; i--)
				repacked[ i / bits  ] = 2 * repacked[ i / bits] + ( data[i] ? 1 : 0);
			return repacked;
		}

}



