
/*
  This file is distributed under the implicit authorization of the author, due to its web plublication.
  This whole code remains the exclusive property of its original author, through his responsibility for the
  use of this file may in no case be engaged.
  Note that this file is NOT distributed under the terms of the General Public License.

*/


package sam.cipher;


/** Implementation of the <a href="http://www.adfa.oz.au/~lpb/research/loki97/">Loki97</a>
 *
 * Retrofitted for personal use by <a href=mailto:sam@glacote.com>Guillaume Lacte</a>.
 *
 * LOKI97 is a 128-bit symmetric block cipher with a 256-bit key schedule,
 * which may be initialised from 128, 192, or 256-bit keys. It uses 16 rounds
 * of data computation using a balanced feistel network with a complex
 * function f which incorporates two S-P layers. The 256-bit key schedule
 * uses 33 rounds of an unbalanced feistel network using the same complex
 * function f to generate the subkeys.<p>
 *
 * LOKI97 was written by Lawrie Brown (ADFA), Josef Pieprzyk, and Jennifer
 * Seberry (UOW) in 1997.<p>
 *
 * <b>Copyright</b> &copy; 1997, 1998 jointly by <a href="mailto:Lawrie.Brown@adfa.oz.au">
 * Lawrie Brown</a> & ITRACE, and
 * <a href="http://www.systemics.com/">Systemics Ltd</a> on behalf of the
 * <a href="http://www.systemics.com/docs/cryptix/">Cryptix Development Team</a>.
 * <br>All rights reserved.<p>
 *
 * <b>$Revision: $</b>
 * @author  Lawrie Brown
 * @author  Raif S. Naffah
 */
public class Loki extends Cipher {

    static { 
				Cipher.REGISTRY.register( new ReflexiveFactory(Loki.class, "Loki97",
																							 "Loki97 is a 16-rounds 128-bit symetric block cipher with key lengths of" +
																							 " 128, 192 and 256 bits\n\n" +
																							 "Candidate algorithm for the NIST AES Quest.\n" +
																							 "See http://www.adfa.oz.au/~lpb/research/loki97/ for further details.\n" ) {
								protected int getPreferredKeySize(int size) {
										return highestIn(size, new int[] { 128, 192, 256 });
								}
						} );
		}


    public Loki(sam.cipher.Key key) {
				super(key);
    }

		protected int getPreferredSize(int size) { return BLOCK_SIZE;	}


		/** Key implementation that is adapted to the {@link Loki} cipher. **/
    public static class Key extends sam.cipher.Key {

				/** Mandatory constructor implementation.
				 ** @see sam.cipher.ReflexiveFactory
				 **/
				public Key(long[] longs) {
					super( format( toBytes(longs) ), 64, true);
				}


				/**
				 * Expand a user-supplied key material into a LOKI97 session key.
				 *
				 * @param key  The 128/192/256-bit user-key to use.
				 * @exception  InvalidKeyException if the key is invalid.
				 */
				private static long[] format (byte[] k) {
						if (k == null)
								return new long[0];
						int length = 8 * (k.length / 8);
						if (length == 0)
								return new long[0];
						long[] SK = new long[NUM_SUBKEYS];	// array of subkeys
						long deltan = DELTA;			// multiples of delta
						int i = 0;				// index into key input
						long k4, k3, k2, k1;			// key schedule 128-bit entities
						long f_out;				// fn f output value for debug

						// pack key into 128-bit entities: k4, k3, k2, k1
						k4 = (k[i++] & 0xFFL) << 56 | (k[i++] & 0xFFL) << 48 |
								(k[i++] & 0xFFL) << 40 | (k[i++] & 0xFFL) << 32 |
								(k[i++] & 0xFFL) << 24 | (k[i++] & 0xFFL) << 16 |
								(k[i++] & 0xFFL) <<  8 | (k[i++] & 0xFFL);
						k3 = (k[i++] & 0xFFL) << 56 | (k[i++] & 0xFFL) << 48 |
								(k[i++] & 0xFFL) << 40 | (k[i++] & 0xFFL) << 32 |
								(k[i++] & 0xFFL) << 24 | (k[i++] & 0xFFL) << 16 |
								(k[i++] & 0xFFL) <<  8 | (k[i++] & 0xFFL);
						if (length == 16) {   // 128-bit key - call fn f twice to gen 256 bits
								k2 = f(k3, k4);
								k1 = f(k4, k3);
						} else {                // 192 or 256-bit key - pack k2 from key data
								k2 = (k[i++] & 0xFFL) << 56 | (k[i++] & 0xFFL) << 48 |
										(k[i++] & 0xFFL) << 40 | (k[i++] & 0xFFL) << 32 |
										(k[i++] & 0xFFL) << 24 | (k[i++] & 0xFFL) << 16 |
										(k[i++] & 0xFFL) <<  8 | (k[i++] & 0xFFL);
								if (length == 24) // 192-bit key - call fn f once to gen 256 bits
										k1 = f(k4, k3);
								else                // 256-bit key - pack k1 from key data
										k1 = (k[i++] & 0xFFL) << 56 | (k[i++] & 0xFFL) << 48 |
												(k[i++] & 0xFFL) << 40 | (k[i++] & 0xFFL) << 32 |
												(k[i++] & 0xFFL) << 24 | (k[i++] & 0xFFL) << 16 |
												(k[i++] & 0xFFL) <<  8 | (k[i++] & 0xFFL);
						}


						// iterate over all LOKI97 rounds to generate the required subkeys
						for (i = 0; i < NUM_SUBKEYS; i++) {
								f_out = f(k1 + k3 + deltan, k2);
								SK[i] = k4 ^ f_out;		// compute next subkey value using fn f
								k4 = k3;			// exchange the other words around
								k3 = k2;
								k2 = k1;
								k1 = SK[i];
								deltan += DELTA;		// next multiple of delta
						}

						return SK;
				}


    }


    /**
     * Encrypt exactly one block of plaintext.
     *
     * @param data          The plaintext.
     * @param key  The session key to use for encryption.
     */
    protected void encrypt (byte[] data, int offset, sam.cipher.Key key) {
        long[] SK = key.getLongs();
				int inOffset = 0;
				long L = unstoreReversedLong(data, offset + 0);
				long R = unstoreReversedLong(data, offset + 8);
        // compute all rounds for this 1 block
        long nR, f_out;
        int k = 0;
        for (int i = 0; i < ROUNDS; i++) {
            nR = R + SK[k++];
            f_out = f(nR, SK[k++]);
            nR += SK[k++];
            R = L ^ f_out;
            L = nR;
        }
				storeReversed(data, offset + 0, R);
				storeReversed(data, offset + 8, L);
    }


    /**
     * Decrypt exactly one block of ciphertext.
     *
     * @param data          The plaintext.
     * @param key  The session key to use for decryption.
     */
    protected void decrypt (byte[] data, int offset , sam.cipher.Key key) {
        long[] SK = key.getLongs();
				int inOffset = 0;
				long L = unstoreReversedLong(data, offset + 0);
				long R = unstoreReversedLong(data, offset + 8);
        // compute all rounds for this 1 block
        long nR, f_out;
        int k = NUM_SUBKEYS - 1;
        for (int i = 0; i < ROUNDS; i++) {
            nR = R - SK[k--];
            f_out = f(nR, SK[k--]);
            nR -= SK[k--];
            R = L ^ f_out;
            L = nR;
        }
				storeReversed(data, offset + 0, R);
				storeReversed(data, offset + 8, L);
    }


    // ------------------------------------------
    // Utilities and constants


    /**
     * Returns a byte residue of base b to power 3 mod g in GF(2^n).
     *
     * @param b  Base of exponentiation, the exponent being always 3.
     * @param g  Irreducible polynomial generating Galois Field (GF(2^n)).
     * @param n  Size of the galois field.
     * @return (b ** 3) mod g.
     */
    private static final byte exp3 (int b, int g, int n) {
        if (b == 0)
            return 0;
        int r = b;            // r = b ** 1
        b = mult(r, b, g, n); // r = b ** 2
        r = mult(r, b, g, n); // r = b ** 3
        return (byte) r;
    }

    /**
     * Returns the product of two binary numbers a and b, using the
     * generator g as the modulus: p = (a * b) mod g. g Generates a
     * suitable Galois Field in GF(2^n).
     *
     * @param a  First multiplicand.
     * @param b  Second multiplicand.
     * @param g  Irreducible polynomial generating Galois Field.
     * @param n  Size of the galois field.
     * @return (a * b) mod g.
     */
    private static final int mult (int a, int b, int g, int n) {
        int p = 0;
        while (b != 0) {
            if ((b & 0x01) != 0)
                p ^= a;
            a <<= 1;
            if (a >= n)
                a ^= g;
            b >>>= 1;
        }
        return p;
    }

    /**
     * Complex highly non-linear round function
     * f(A,B) = Sb(P(Sa(E(KP(A,hi(B))))),lo(B))
     *
     * @param A  64-bit input A
     * @param B  64-bit input B
     * @return The resulting 64-bit function value
     */
    private static long f (long A, long B) {
        // Intermediate values in the computation are:
        //   d = KP(A,B)
        //   e = P(Sa(d))
        //   f = Sb(e,B)

        // Compute d = KP(A,B), where KP is a keyed permutation used to 
        //    exchange corresponding bits in 32-bit words [Al,Ar] 
        //    based on the lower half of B (called Br) (swap if B bit is 1)
        //    KP(A,B) = ((Al & ~Br)|(Ar & Br)) | ((Ar & ~Br)|(Al & Br))

        int Al = (int)(A >>> 32);
        int Ar = (int) A;
        int Br = (int) B;
        long d = ((long)((Al & ~Br) | (Ar & Br)) << 32) |
						((long)((Ar & ~Br) | (Al & Br)) & 0xFFFFFFFFL);

        // Compute e = P(Sa(d))
        //    mask out each group of 12 bits for E
        //    then compute first S-box column [S1,S2,S1,S2,S2,S1,S2,S1]
        //    permuting output through P (with extra shift to build full P)

        long e = P[S1[(int)((d >>> 56 | d << 8) & 0x1FFF)] & 0xFF] >>> 7 |
						P[S2[(int)((d >>> 48)          &  0x7FF)] & 0xFF] >>> 6 |
						P[S1[(int)((d >>> 40)          & 0x1FFF)] & 0xFF] >>> 5 |
						P[S2[(int)((d >>> 32)          &  0x7FF)] & 0xFF] >>> 4 |
						P[S2[(int)((d >>> 24)          &  0x7FF)] & 0xFF] >>> 3 |
						P[S1[(int)((d >>> 16)          & 0x1FFF)] & 0xFF] >>> 2 |
						P[S2[(int)((d >>>  8)          &  0x7FF)] & 0xFF] >>> 1 |
						P[S1[(int)( d                  & 0x1FFF)] & 0xFF];

        // Compute f = Sb(e,B)
        //    where the second S-box column is [S2,S2,S1,S1,S2,S2,S1,S1]
        //    for each S, lower bits come from e, upper from upper half of B

        long f =
            (S2[(int)(((e>>>56) & 0xFF) | ((B>>>53) &  0x700))] & 0xFFL) << 56 |
            (S2[(int)(((e>>>48) & 0xFF) | ((B>>>50) &  0x700))] & 0xFFL) << 48 |
            (S1[(int)(((e>>>40) & 0xFF) | ((B>>>45) & 0x1F00))] & 0xFFL) << 40 |
            (S1[(int)(((e>>>32) & 0xFF) | ((B>>>40) & 0x1F00))] & 0xFFL) << 32 |
            (S2[(int)(((e>>>24) & 0xFF) | ((B>>>37) &  0x700))] & 0xFFL) << 24 |
            (S2[(int)(((e>>>16) & 0xFF) | ((B>>>34) &  0x700))] & 0xFFL) << 16 |
            (S1[(int)(((e>>> 8) & 0xFF) | ((B>>>29) & 0x1F00))] & 0xFFL) <<  8 |
            (S1[(int)(( e       & 0xFF) | ((B>>>24) & 0x1F00))] & 0xFFL);

        return f;
    }


    /** Number of bytes in a data-block. */
    private static final int BLOCK_SIZE = 16;

    /** Number of rounds for the algorithm. */
    private static final int ROUNDS = 16;

    /** Number of subkeys used by the algorithm. */
    private static final int NUM_SUBKEYS = 3 * ROUNDS;

    /** Constant value for Delta which is used in the key schedule */
    private static final long DELTA = 0x9E3779B97F4A7C15L;

    /** Generator polynomial for S-box S1, in GF(2<sup>13</sup>). */
    private static final int S1_GEN = 0x2911;

    /** Size of S-box S1, for 13-bit inputs. */
    private static final int S1_SIZE = 0x2000;

    /** Table of pre-computed S-box S1 values. */
    private static final byte[] S1 = new byte[S1_SIZE];

    /** Generator polynomial for S-box S2, in GF(2<sup>11</sup>). */
    private static final int S2_GEN = 0xAA7;

    /** Size of S-box S2, for 11-bit inputs. */
    private static final int S2_SIZE = 0x800;

    /** Table of pre-computed S-box S2 values. */
    private static final byte[] S2 = new byte[S2_SIZE];

    /**
     * Table specifying the pre-computed permutation P. Note:<ul>
     * <li> precompute permutations for lowest 8 bits only, since the
     *      remainder of the permutation is related to it by successive
     *      right shifts for each successive input byte.
     * <li> value of P is a 64-bit wide (long) mask of the permuted input
     *      value.</ul>
     */
    private static final long[] P = new long[0x100];

    // Static code - to intialise the S-box and permutation tables
    //...........................................................................

    static {
        // precompute S-box tables for S1 and S2
        final int S1_MASK = S1_SIZE - 1;  // mask to select S1 input bits
        final int S2_MASK = S2_SIZE - 1;  // mask to select S2 input bits
        int i = 0; // index into S-box
        int b; // S-box fn input
        for (i = 0; i < S1_SIZE; i++) { // for all S1 inputs
            b = i ^ S1_MASK; // compute input value
            S1[i] = exp3(b, S1_GEN, S1_SIZE); // compute fn value
        }
        for (i = 0; i < S2_SIZE; i++) { // for all S2 inputs
            b = i ^ S2_MASK; // compute input value
            S2[i] = exp3(b, S2_GEN, S2_SIZE); // compute fn value
        }
        //
        // initialising expanded permutation P table (for lowest byte only)
        //   Permutation P maps input bits [63..0] to outputs bits:
        //   [56, 48, 40, 32, 24, 16,  8, 0,
        //    57, 49, 41, 33, 25, 17,  9, 1,
        //    58, 50, 42, 34, 26, 18, 10, 2,
        //    59, 51, 43, 35, 27, 19, 11, 3,
        //    60, 52, 44, 36, 28, 20, 12, 4,
        //    61, 53, 45, 37, 29, 21, 13, 5,
        //    62, 54, 46, 38, 30, 22, 14, 6,
        //    63, 55, 47, 39, 31, 23, 15, 7]  <- this row only used for table
        //  However, since it is so regular, can construct it on the fly
        long pval; // constructed permutation output value
        for (i = 0; i < 0x100; i++) { // loop over all 8-bit inputs
            pval = 0L;
            // for each input bit permute to specified output position
            for (int j = 0, k = 7; j < 8; j++, k += 8)
                pval |= (long)((i >>> j) & 0x1) << k;
            P[i] = pval;
        }

    }


}
