
package sam.cipher;

import sam.toolkit.Text;

import java.lang.reflect.*;



/** Implementation of a <CODE>Cipher</CODE> factory based on reflection.
       Registers a new <CODE>Cipher</CODE> class.
       Any specialized <CODE>Key</CODE> class associated with the cipher must be declared innerly with a
       <CODE>long[]</CODE> constructor.
       <p>To be called typically in a static initializer of a subclass, as in:
       <pre><codeblock>
	   public class MyCipher extends Cipher {
	       static {
			 Cipher.register( MyCipher.class, "My cipher", "This is my private key cipher" );  // May throw a Cipher.Error
    	   }

          ...

		  public class Key extends Cipher.Key {
   		      public Key(long[] bits) {
          		  super(bits);
            	   ...
           	   }
      	   }
	   }
       </pre></codeblock>
	   or alternatively
       <pre><codeblock>
	   public class MyCipher extends Cipher {
	       static {
			 Cipher.register( new ReflexiveFactory("Mycipher", "This is my own cipher class") {
			     public int getPreferredKeySize(int suggestedSize) {
				     return highestMultiple(suggestedSize, 8);  // My cipher requires key lengths be multiple of 8 bits
				 }
			 } );  // May throw a Cipher.Error
    	   }
          ...
	   }
       </pre></codeblock>
 **/
public class ReflexiveFactory extends Factory {

		/** Cipher's one-argument(<CODE>public Cipher(Cipher.Key)</CODE>) constructor */
		private final Constructor cipherConstructor;
		/** Key's one-argument(<CODE>public Key(long[])</CODE>) constructor. */
		private final Constructor keyConstructor;

		/** Builds a new reflexive <CODE>Cipher</CODE> factory based on the <b>supposedly</b> existing outer class.
		 ** Typical use would be to declare a static inner subclass of <CODE>ReflexiveFactory</CODE> inside your
		 ** <CODE>Cipher</CODE> subclass.
		 ** @param description a human-readable description of this cipher implementation
		 ** Note that the registration name is automatically inferred from the ciphering class name.
		 ** <b>Warning</b> using automatic naming based on class name can cause registration name collision !
		 ** <b>Warning: this is not a <CODE>ReflexiveFactory(String className)</CODE> constructor !</b>
		 **/
		public ReflexiveFactory(String description) throws Cipher.Error {	this(null, null, description); }

		/** Builds a new reflexive <CODE>Cipher</CODE> factory based on the <b>supposedly</b> existing outer class.
		 ** Typical use would be to declare a static inner subclass of <CODE>ReflexiveFactory</CODE> inside your
		 ** <CODE>Cipher</CODE> subclass.
		 ** @param name the registration name for this cipher implementation.
		 ** @param description a human-readable description of this cipher implementation
		 **/
		public ReflexiveFactory(String name, String description) throws Cipher.Error { this(null, name, description);	}

		/** Makes a new registration entry for a cipher.
		 ** <p>The main cipher constructor <CODE>public Cipher(Cipher.Key)</CODE> is obtained by reflection.
		 ** <p>The <CODE>public Key(long[])</CODE> constructor is obtained by reflection, either from a declared
		 ** <b>inner-class</b> of the <CODE>cipher</CODE> class, or from the default <CODE>Cipher.Key</CODE> class.
		 ** @param cipher the cipher class, or <CODE>null</CODE> if to be retrieved automatically through reflection.
		 ** @param name the registration name of the cipher, or <CODE>null</CODE> if default name.
		 ** @param description the description to be associated with the cipher, or <CODE>null</CODE> if none.
		 ** @throws Cipher.Error in case of security or instantiation exceptions
		 ** @see java.lang.SecurityManager
		*/
		public ReflexiveFactory(Class cipher, String name, String description) throws Cipher.Error {
			super( getBestName(name, cipher), getBestDescription(description, cipher) );
				if (cipher == null)
						cipher = getOuterCipherClass();
				if ( ! Cipher.class.isAssignableFrom(cipher) )
						throw new Cipher.Error("Not a cipher class: " + cipher.getName() );
				try {
					Constructor keyConstructor = getKeyConstructor( cipher.getDeclaredClasses() );
					if (keyConstructor == null)
						keyConstructor = getKeyConstructor( cipher.getClasses() );
					if (keyConstructor == null)
						keyConstructor = KEY_CONSTRUCTOR;
					this.cipherConstructor = sam.toolkit.Process.getConstructor(cipher, new Class[] { Key.class } );
					this.keyConstructor = keyConstructor;
				} catch (java.lang.Exception e) {
					throw new Cipher.Error(e);
				}
		}

		/** Builds a new cipher instance through <CODE>constructor</CODE>s found by introspection.
		 **/
		protected Cipher make(long[] keyBits, boolean mode) throws Cipher.Error {
			try {
				Key key = (Key) keyConstructor.newInstance( new Object[] { keyBits } );
				return (Cipher) cipherConstructor.newInstance( new Object[] { mode ? key : key.inverse() } );
			} catch (InvocationTargetException e) {
				throw new Cipher.Error( "Could not create " + (mode ? "inverse" : "") + " key for \"" +
																name + "\":\n" + Text.getMessage( e.getTargetException() ) );
			} catch (java.lang.Exception e) {
				throw new Cipher.Error(e);
			}
		}


		/** Gets the best name for specified cipher class.
		 ** This is the supplied <CODE>name</CODE> if it is neither <CODE>null</CODE> nor equal to <CODE>""</CODE>;
		 ** otherwise the last part of the class name is used, which includes dot-separated outer-classes
		 ** but <b>excludes package specification</b>.
		 ** <b>Warning</b> using automatic naming based on class name can cause registration name collision !
		 **/
		protected static String getBestName(String name, Class clazz) {
			return ((name == null) || ("".equals(name))) ? Text.getLastClassName(clazz) : name;
		}

		/** Gets the best name for specified cipher class.
		 ** This is the supplied <CODE>name</CODE> if it is neither <CODE>null</CODE> nor equal to <CODE>""</CODE>;
		 ** otherwise a default string is used stating that ... no information is available.
		 **/
		protected static String getBestDescription(String description, Class clazz) {
			return ((description == null) || ("".equals(description))) ? ("(No information about " + clazz.getName() + ")") : description;
		}


		/** Gets the nearest outer class that is a subclass of <CODE>sam.cipher.Cipher</CODE>.
		 **/
		private Class getOuterCipherClass() {
			Class clazz = getClass();
			while ( (clazz != null) && (! Cipher.class.isAssignableFrom(clazz)) )
					{ 
					clazz = clazz.getDeclaringClass();
					System.err.println(" > Considering " + clazz);
					if (clazz != null)
						System.err.println(": is a cipher = " + Cipher.class.isAssignableFrom(clazz));
					}
			return clazz;
		}


		private static Constructor getKeyConstructor(Class[] classes) throws Cipher.Error {
				try {
						for (int i=0; i<classes.length; i++)
								if ( Key.class.isAssignableFrom(classes[i]) ) {
										try {
											return sam.toolkit.Process.getConstructor(classes[i], new Class[] { long[].class } );
										} catch (NoSuchMethodException e) {
												throw new Cipher.Error(e);
										}
								}
						return null;
				} catch (java.lang.Exception e) {
						throw new Cipher.Error(e);
				}
		}

		private static Constructor KEY_CONSTRUCTOR = getKeyConstructor( new Class[] { Key.class } );

}

