
/*
  This file is distributed under the implicit authorization of the author, due to its web plublication.
  This whole code remains the exclusive property of its original author, through his responsibility for the
  use of this file may in no case be engaged.
  Note that this file is NOT distributed under the terms of the General Public License.

*/

package sam.cipher;

import sam.ui.UserInterface;

/** Implementation of the <a href=http://csrc.nist.gov/encryption/aes/rijndael>Rijndael</a> cipher.
 *
 * Retrofitted for personal use by <a href=mailto:sam@glacote.com>Guillaume Lacte</a>.
 *
 * Rijndael --pronounced Reindaal-- is a variable block-size (128-, 192- and
 * 256-bit), variable key-size (128-, 192- and 256-bit) symmetric cipher.<p>
 *
 * Rijndael was written by <a href="mailto:rijmen@esat.kuleuven.ac.be">Vincent
 * Rijmen</a> and <a href="mailto:Joan.Daemen@village.uunet.be">Joan Daemen</a>.<p>
 *
 * Portions of this code are <b>Copyright</b> &copy; 1997, 1998
 * <a href="http://www.systemics.com/">Systemics Ltd</a> on behalf of the
 * <a href="http://www.systemics.com/docs/cryptix/">Cryptix Development Team</a>.
 * <br>All rights reserved.<p>
 *
 * <b>$Revision: $</b>
 * @author  Raif S. Naffah
 * @author  Paulo S. L. M. Barreto
 */
public class Rijndael extends Cipher{

    static {
				Cipher.REGISTRY.register( new ReflexiveFactory(Rijndael.class, "Rijndael",
																							 "Rijndael is a variable block-size, variable key-size symetric cipher.\n\n" +
																							 "See http://www.esat.kuleuven.ac.be/~rijmen/rijndael/ for further details.\n" ) {
								public int getPreferredKeySize(int size) { return highestMultiple(size, 8); }
						} );
		}

		protected int getPreferredSize(int size) {
			return 16;
		}

    public Rijndael(sam.cipher.Key key) {
			super(key);
    }

		/** Key implementation that is adapted to the {@link Rijndael} cipher. **/
		public static class Key extends sam.cipher.Key {
				private final int[][] Ke, Kd;

				private Key(long[] bits, boolean direction, int[][] Ke, int[][] Kd) {
					super(bits, 64, direction);
					this.Ke = Ke;
					this.Kd = Kd;
				}

				public sam.cipher.Key inverse() {
					return new Key(bits, ! direction, Ke, Kd);
				}

				public boolean isEmpty() {
					return (Ke == null) || (Kd == null);
				}

				/** Mandatory constructor implementation.
				 ** @see sam.cipher.ReflexiveFactory
				 **/
				public Key(long[] longs) {
					super(longs, 64, true);
					int[] ints = new sam.cipher.Key(longs).getInts();
					int KC = 2 * (ints.length / 2); // 8 * (ints.length / 8);
					if (KC == 0) {
						UserInterface.get().log().log(13, Rijndael.class, "Built empty key");
						Ke = Kd = null;
						return;
					}
					int ROUNDS = getRounds(KC, BLOCK_SIZE);
					int[] tk = new int[4 * (ROUNDS + 1)];
					int limit = (ints.length < tk.length) ? ints.length : tk.length;
					System.arraycopy(ints, 0, tk, 0, limit);

					int BC = BLOCK_SIZE / 4;
					Ke = new int[ROUNDS + 1][BC]; // encryption round keys
					Kd = new int[ROUNDS + 1][BC]; // decryption round keys
					int ROUND_KEY_COUNT = (ROUNDS + 1) * BC;
					int i, j;

					// copy values into round key arrays
					int t = 0;
					for (j = 0; (j < KC) && (t < ROUND_KEY_COUNT); j++, t++) {
							Ke[t / BC][t % BC] = tk[j];
							Kd[ROUNDS - (t / BC)][t % BC] = tk[j];
					}
					int tt,
							rconpointer = 0;
					while (t < ROUND_KEY_COUNT) {
							// extrapolate using phi (the round key evolution function)
							tt = tk[KC - 1];
							tk[0] ^= (S[(tt >>> 16) & 0xFF] & 0xFF) << 24 ^
									(S[(tt >>>  8) & 0xFF] & 0xFF) << 16 ^
									(S[ tt         & 0xFF] & 0xFF) <<  8 ^
									(S[(tt >>> 24) & 0xFF] & 0xFF)       ^
									(rcon[rconpointer++]   & 0xFF) << 24;
							if (KC != 8)
									for (i = 1, j = 0; i < KC; ) tk[i++] ^= tk[j++];
							else {
									for (i = 1, j = 0; i < KC / 2; ) tk[i++] ^= tk[j++];
									tt = tk[KC / 2 - 1];
									tk[KC / 2] ^= (S[ tt         & 0xFF] & 0xFF)       ^
											(S[(tt >>>  8) & 0xFF] & 0xFF) <<  8 ^
											(S[(tt >>> 16) & 0xFF] & 0xFF) << 16 ^
											(S[(tt >>> 24) & 0xFF] & 0xFF) << 24;
									for (j = KC / 2, i = j + 1; i < KC; ) tk[i++] ^= tk[j++];
							}
							// copy values into round key arrays
							for (j = 0; (j < KC) && (t < ROUND_KEY_COUNT); j++, t++) {
									Ke[t / BC][t % BC] = tk[j];
									Kd[ROUNDS - (t / BC)][t % BC] = tk[j];
							}
					}
					for (int r = 1; r < ROUNDS; r++)    // inverse MixColumn where needed
							for (j = 0; j < BC; j++) {
									tt = Kd[r][j];
									Kd[r][j] = U1[(tt >>> 24) & 0xFF] ^
											U2[(tt >>> 16) & 0xFF] ^
											U3[(tt >>>  8) & 0xFF] ^
											U4[ tt         & 0xFF];
							}

				}

    }




    /**
     * Convenience method to encrypt exactly one block of plaintext, assuming
     * Rijndael's default block size (128-bit).
     *
     * @param  in         The plaintext.
     * @param  sessionKey The session key to use for encryption.
     */
    protected void encrypt(byte[] data, int offset, sam.cipher.Key sessionKey) {
        int[][] Ke = ((Key) sessionKey).Ke;
				if (Ke == null)
					return;
        int ROUNDS = Ke.length - 1;
        int[] Ker = Ke[0];

        int t0   = unstoreInt(data, offset + 0) ^ Ker[0];
        int t1   = unstoreInt(data, offset + 4) ^ Ker[1];
        int t2   = unstoreInt(data, offset + 8) ^ Ker[2];
        int t3   = unstoreInt(data, offset + 12) ^ Ker[3];

        int a0, a1, a2, a3;
        for (int r = 1; r < ROUNDS; r++) {          // apply round transforms
            Ker = Ke[r];
            a0   = (T1[(t0 >>> 24) & 0xFF] ^
                    T2[(t1 >>> 16) & 0xFF] ^
                    T3[(t2 >>>  8) & 0xFF] ^
                    T4[ t3         & 0xFF]  ) ^ Ker[0];
            a1   = (T1[(t1 >>> 24) & 0xFF] ^
                    T2[(t2 >>> 16) & 0xFF] ^
                    T3[(t3 >>>  8) & 0xFF] ^
                    T4[ t0         & 0xFF]  ) ^ Ker[1];
            a2   = (T1[(t2 >>> 24) & 0xFF] ^
                    T2[(t3 >>> 16) & 0xFF] ^
                    T3[(t0 >>>  8) & 0xFF] ^
                    T4[ t1         & 0xFF]  ) ^ Ker[2];
            a3   = (T1[(t3 >>> 24) & 0xFF] ^
                    T2[(t0 >>> 16) & 0xFF] ^
                    T3[(t1 >>>  8) & 0xFF] ^
                    T4[ t2         & 0xFF]  ) ^ Ker[3];
            t0 = a0;
            t1 = a1;
            t2 = a2;
            t3 = a3;
        }

        // last round is special
        Ker = Ke[ROUNDS];
        int tt = Ker[0];
        data[offset +  3] = (byte)(S[(t0 >>> 24) & 0xFF] ^ (tt >>> 24));
        data[offset +  2] = (byte)(S[(t1 >>> 16) & 0xFF] ^ (tt >>> 16));
        data[offset +  1] = (byte)(S[(t2 >>>  8) & 0xFF] ^ (tt >>>  8));
        data[offset +  0] = (byte)(S[ t3         & 0xFF] ^  tt        );
        tt = Ker[1];
        data[offset +  7] = (byte)(S[(t1 >>> 24) & 0xFF] ^ (tt >>> 24));
        data[offset +  6] = (byte)(S[(t2 >>> 16) & 0xFF] ^ (tt >>> 16));
        data[offset +  5] = (byte)(S[(t3 >>>  8) & 0xFF] ^ (tt >>>  8));
        data[offset +  4] = (byte)(S[ t0         & 0xFF] ^  tt        );
        tt = Ker[2];
        data[offset + 11] = (byte)(S[(t2 >>> 24) & 0xFF] ^ (tt >>> 24));
        data[offset + 10] = (byte)(S[(t3 >>> 16) & 0xFF] ^ (tt >>> 16));
        data[offset +  9] = (byte)(S[(t0 >>>  8) & 0xFF] ^ (tt >>>  8));
        data[offset +  8] = (byte)(S[ t1         & 0xFF] ^  tt        );
        tt = Ker[3];
        data[offset + 15] = (byte)(S[(t3 >>> 24) & 0xFF] ^ (tt >>> 24));
        data[offset + 14] = (byte)(S[(t0 >>> 16) & 0xFF] ^ (tt >>> 16));
        data[offset + 13] = (byte)(S[(t1 >>>  8) & 0xFF] ^ (tt >>>  8));
        data[offset + 12] = (byte)(S[ t2         & 0xFF] ^  tt        );
    }

    /**
     * Convenience method to decrypt exactly one block of plaintext, assuming
     * Rijndael's default block size (128-bit).
     *
     * @param  data         The ciphertext.
     * @param  sessionKey The session key to use for decryption.
     */
    protected void decrypt(byte[] data, int offset, sam.cipher.Key sessionKey) {
        int[][] Kd = ((Key) sessionKey).Kd;
				if (Kd == null)
					return;
        int ROUNDS = Kd.length - 1;
        int[] Kdr = Kd[0];

        int t0   = unstoreInt(data, offset + 0) ^ Kdr[0];
        int t1   = unstoreInt(data, offset + 4) ^ Kdr[1];
        int t2   = unstoreInt(data, offset + 8) ^ Kdr[2];
        int t3   = unstoreInt(data, offset + 12) ^ Kdr[3];

        int a0, a1, a2, a3;
        for (int r = 1; r < ROUNDS; r++) {          // apply round transforms
            Kdr = Kd[r];
            a0   = (T5[(t0 >>> 24) & 0xFF] ^
                    T6[(t3 >>> 16) & 0xFF] ^
                    T7[(t2 >>>  8) & 0xFF] ^
                    T8[ t1         & 0xFF]  ) ^ Kdr[0];
            a1   = (T5[(t1 >>> 24) & 0xFF] ^
                    T6[(t0 >>> 16) & 0xFF] ^
                    T7[(t3 >>>  8) & 0xFF] ^
                    T8[ t2         & 0xFF]  ) ^ Kdr[1];
            a2   = (T5[(t2 >>> 24) & 0xFF] ^
                    T6[(t1 >>> 16) & 0xFF] ^
                    T7[(t0 >>>  8) & 0xFF] ^
                    T8[ t3         & 0xFF]  ) ^ Kdr[2];
            a3   = (T5[(t3 >>> 24) & 0xFF] ^
                    T6[(t2 >>> 16) & 0xFF] ^
                    T7[(t1 >>>  8) & 0xFF] ^
                    T8[ t0         & 0xFF]  ) ^ Kdr[3];
            t0 = a0;
            t1 = a1;
            t2 = a2;
            t3 = a3;
        }

        // last round is special
        Kdr = Kd[ROUNDS];
        int tt = Kdr[0];
        data[offset +  3] = (byte)(Si[(t0 >>> 24) & 0xFF] ^ (tt >>> 24));
        data[offset +  2] = (byte)(Si[(t3 >>> 16) & 0xFF] ^ (tt >>> 16));
        data[offset +  1] = (byte)(Si[(t2 >>>  8) & 0xFF] ^ (tt >>>  8));
        data[offset +  0] = (byte)(Si[ t1         & 0xFF] ^  tt        );
        tt = Kdr[1];
        data[offset +  7] = (byte)(Si[(t1 >>> 24) & 0xFF] ^ (tt >>> 24));
        data[offset +  6] = (byte)(Si[(t0 >>> 16) & 0xFF] ^ (tt >>> 16));
        data[offset +  5] = (byte)(Si[(t3 >>>  8) & 0xFF] ^ (tt >>>  8));
        data[offset +  4] = (byte)(Si[ t2         & 0xFF] ^  tt        );
        tt = Kdr[2];
        data[offset + 11] = (byte)(Si[(t2 >>> 24) & 0xFF] ^ (tt >>> 24));
        data[offset + 10] = (byte)(Si[(t1 >>> 16) & 0xFF] ^ (tt >>> 16));
        data[offset +  9] = (byte)(Si[(t0 >>>  8) & 0xFF] ^ (tt >>>  8));
        data[offset +  8] = (byte)(Si[ t3         & 0xFF] ^  tt        );
        tt = Kdr[3];
        data[offset + 15] = (byte)(Si[(t3 >>> 24) & 0xFF] ^ (tt >>> 24));
        data[offset + 14] = (byte)(Si[(t2 >>> 16) & 0xFF] ^ (tt >>> 16));
        data[offset + 13] = (byte)(Si[(t1 >>>  8) & 0xFF] ^ (tt >>>  8));
        data[offset + 12] = (byte)(Si[ t0         & 0xFF] ^  tt        );
    }


    // Rijndael own methods
    //...........................................................................


    /**
     * Return The number of rounds for a given Rijndael's key and block sizes.
     *
     * @param keySize    The size of the user key material in bytes.
     * @param blockSize  The desired block size in bytes.
     * @return The number of rounds for a given Rijndael's key and
     *      block sizes.
     */
    public static int getRounds (int keySize, int blockSize) {
        switch (keySize) {
        case 16:
            return blockSize == 16 ? 10 : (blockSize == 24 ? 12 : 14);
        case 24:
            return blockSize != 32 ? 12 : 14;
        default: // 32 bytes = 256 bits
            return 14;
        }
    }


    // Constants and variables

    private static final int BLOCK_SIZE = 16; // default block size in bytes

    private static final int[] alog = new int[256];
    private static final int[] log =  new int[256];

    private static final byte[] S =  new byte[256];
    private static final byte[] Si = new byte[256];
    private static final int[] T1 = new int[256];
    private static final int[] T2 = new int[256];
    private static final int[] T3 = new int[256];
    private static final int[] T4 = new int[256];
    private static final int[] T5 = new int[256];
    private static final int[] T6 = new int[256];
    private static final int[] T7 = new int[256];
    private static final int[] T8 = new int[256];
    private static final int[] U1 = new int[256];
    private static final int[] U2 = new int[256];
    private static final int[] U3 = new int[256];
    private static final int[] U4 = new int[256];
    private static final byte[] rcon = new byte[30];

    private static final int[][][] shifts = new int[][][] {
        { {0, 0}, {1, 3}, {2, 2}, {3, 1} },
        { {0, 0}, {1, 5}, {2, 4}, {3, 3} },
        { {0, 0}, {1, 7}, {3, 5}, {4, 4} }
    };


    // Static code - to intialise S-boxes and T-boxes

    static {
        int ROOT = 0x11B;
        int i, j = 0;

        // produce log and alog tables, needed for multiplying in the
        // field GF(2^m) (generator = 3)
        alog[0] = 1;
        for (i = 1; i < 256; i++) {
            j = (alog[i-1] << 1) ^ alog[i-1];
            if ((j & 0x100) != 0) j ^= ROOT;
            alog[i] = j;
        }
        for (i = 1; i < 255; i++) log[alog[i]] = i;
        byte[][] A = new byte[][] {
            {1, 1, 1, 1, 1, 0, 0, 0},
            {0, 1, 1, 1, 1, 1, 0, 0},
						{0, 0, 1, 1, 1, 1, 1, 0},
						{0, 0, 0, 1, 1, 1, 1, 1},
						{1, 0, 0, 0, 1, 1, 1, 1},
						{1, 1, 0, 0, 0, 1, 1, 1},
						{1, 1, 1, 0, 0, 0, 1, 1},
						{1, 1, 1, 1, 0, 0, 0, 1}
				};
        byte[] B = new byte[] { 0, 1, 1, 0, 0, 0, 1, 1};

        // substitution box based on F^{-1}(x)
        int t;
        byte[][] box = new byte[256][8];
        box[1][7] = 1;
        for (i = 2; i < 256; i++) {
            j = alog[255 - log[i]];
            for (t = 0; t < 8; t++)
                box[i][t] = (byte)((j >>> (7 - t)) & 0x01);
        }

        // affine transform:  box[i] <- B + A*box[i]
        byte[][] cox = new byte[256][8];
        for (i = 0; i < 256; i++)
            for (t = 0; t < 8; t++) {
                cox[i][t] = B[t];
                for (j = 0; j < 8; j++)
                    cox[i][t] ^= A[t][j] * box[i][j];
            }

        // S-boxes and inverse S-boxes
        for (i = 0; i < 256; i++) {
            S[i] = (byte)(cox[i][0] << 7);
						for (t = 1; t < 8; t++)
								S[i] ^= cox[i][t] << (7-t);
            Si[S[i] & 0xFF] = (byte) i;
        }

        // T-boxes
        byte[][] G = new byte[][] {
            {2, 1, 1, 3},
            {3, 2, 1, 1},
            {1, 3, 2, 1},
            {1, 1, 3, 2}
        };
        byte[][] AA = new byte[4][8];
        for (i = 0; i < 4; i++) {
            for (j = 0; j < 4; j++) AA[i][j] = G[i][j];
            AA[i][i+4] = 1;
        }
        byte pivot, tmp;
        byte[][] iG = new byte[4][4];
        for (i = 0; i < 4; i++) {
            pivot = AA[i][i];
            if (pivot == 0) {
                t = i + 1;
                while ((AA[t][i] == 0) && (t < 4))
                    t++;
                if (t == 4)
                    throw new RuntimeException("G matrix is not invertible");
                else {
                    for (j = 0; j < 8; j++) {
                        tmp = AA[i][j];
                        AA[i][j] = AA[t][j];
                        AA[t][j] = (byte) tmp;
                    }
                    pivot = AA[i][i];
                }
            }
            for (j = 0; j < 8; j++)
                if (AA[i][j] != 0)
                    AA[i][j] = (byte)
                        alog[(255 + log[AA[i][j] & 0xFF] - log[pivot & 0xFF]) % 255];
            for (t = 0; t < 4; t++)
                if (i != t) {
                    for (j = i+1; j < 8; j++)
                        AA[t][j] ^= mul(AA[i][j], AA[t][i]);
                    AA[t][i] = 0;
                }
        }
        for (i = 0; i < 4; i++)
            for (j = 0; j < 4; j++) iG[i][j] = AA[i][j + 4];

        int s;
        for (t = 0; t < 256; t++) {
            s = S[t];
            T1[t] = mul4(s, G[0]);
            T2[t] = mul4(s, G[1]);
            T3[t] = mul4(s, G[2]);
            T4[t] = mul4(s, G[3]);

            s = Si[t];
            T5[t] = mul4(s, iG[0]);
            T6[t] = mul4(s, iG[1]);
            T7[t] = mul4(s, iG[2]);
            T8[t] = mul4(s, iG[3]);

            U1[t] = mul4(t, iG[0]);
            U2[t] = mul4(t, iG[1]);
            U3[t] = mul4(t, iG[2]);
            U4[t] = mul4(t, iG[3]);
        }
        //
        // round constants
        //
        rcon[0] = 1;
        int r = 1;
        for (t = 1; t < 30; )
						rcon[t++] = (byte) (r = mul(2, r));

    }

    // multiply two elements of GF(2^m)
    private static final int mul (int a, int b) {
        return (a != 0 && b != 0) ?
            alog[(log[a & 0xFF] + log[b & 0xFF]) % 255] :
            0;
    }

    // convenience method used in generating Transposition boxes
    private static final int mul4 (int a, byte[] b) {
        if (a == 0) return 0;
        a = log[a & 0xFF];
        int a0 = (b[0] != 0) ? alog[(a + log[b[0] & 0xFF]) % 255] & 0xFF : 0;
        int a1 = (b[1] != 0) ? alog[(a + log[b[1] & 0xFF]) % 255] & 0xFF : 0;
        int a2 = (b[2] != 0) ? alog[(a + log[b[2] & 0xFF]) % 255] & 0xFF : 0;
        int a3 = (b[3] != 0) ? alog[(a + log[b[3] & 0xFF]) % 255] & 0xFF : 0;
        return a0 << 24 | a1 << 16 | a2 << 8 | a3;
    }


}
