
package sam.cipher;

import sam.toolkit.Text;


/** Class specifying a cipher with its associated password
 ** This basically reduces to a pair <CODE>cipher name, password</CODE>.
 **/
public class Specification {

	/** The choosen cipher name **/
	public final String name;

	/** The choosen password **/
	public final transient String password;

	/** Built a new specification without password.
	 ** @param cipher the name of a registered cipher
	 **/
	public Specification(String cipher) { this(cipher, null); }

	/** Built a new specification with specified password.
	 ** @param cipher the name of a registered cipher
	 ** @param password the associated passphrase
	 **/
	public Specification(String cipher, String password) {
		this.name = (cipher == null) ? "" : cipher;
		this.password = (password == null) ? "" : password;
	}

	/** Wether this specifications is empty, i.e. has an empty cipher name or password **/
	public boolean isEmpty() {
		return ("".equals(name)) || ("".equals(password));
	}

	/** Instantiates this specification to a proper <CODE>Cipher</CODE>
	 ** @param mode wether the cipher should be a ciphering or a deciphering cipher
	 ** @return <CODE>null</CODE> if passwords is <CODE>null</CODE> or empty
	 **/
	public Cipher getCipher(boolean mode) throws Error {
		return isEmpty() ? null : Cipher.REGISTRY.newCipher(name, password, mode);
	}

	/** Human-readable specification.
	 ** <b>Warning:</b> this includes the <CODE>password</CODE> in plain text !
	 **/
	public String toString() { return "Cipher(" + name + ':' + password +")";	}

	/** Gets the human-readable description of this cipher.
	 **/
	public String getDescription() {
		return Cipher.REGISTRY.getDescription(name);
	}


	/** Instantiates a specification list to a proper <CODE>Ciphers</CODE> list.
	 ** This automatically drops ciphers whose password is null or empty
	 ** @param mode wether the cipher should be a ciphering or a deciphering cipher
	 ** @return <CODE>null</CODE> any instantiation fails due to an <CODE>Error</CODE>
	 **/
	public static Cipher[] getCiphers(Specification[] specifications, boolean mode) {
		if (specifications == null)
			return null;
		Cipher[] allCiphers = new Cipher[specifications.length];
		for (int i=0; i<specifications.length; ++i)
			try {
				allCiphers[i] = specifications[i].getCipher(mode);
			} catch (Error e) {
				sam.ui.UserInterface.get().error().handleException("Could not instantiate cipher " + specifications[i].name, e);
				return null;
			}
		return (Cipher[]) Text.dropNulls(allCiphers);  // This cast will always succeed
	}

	/** @return wether all required passwords are specified among passed specifications,
	 ** or if <CODE>specifications</CODE> is <CODE>null</CODE>
	 **/
	public static boolean areFullySpecified(Specification[] specifications) {
		if (specifications == null)
			return false;
		for (int i=0; i<specifications.length; ++i)
			if (specifications[i].isEmpty())
				return false;
		return true;
	}

	/** Gets the list of all currently registerd ciphers as a <CODE>Specification[]</CODE> list.
	 ** No default password is ever supplied.
	 ** @see Registry.getCiphers
	 **/
	public static Specification[] getRegisteredCiphers() {
		return fillIn(Cipher.REGISTRY.getCiphers(), new String[0]);
	}

	/** Drops all specifications whose password is null or empty
	 **/
	public static Specification[] dropEmpty(Specification[] specifications) {
		if (specifications == null)
			return null;
		Specification[] nonEmptySpecifications = new Specification[specifications.length];
		for (int i=0; i<specifications.length; ++i)
			nonEmptySpecifications[i] = ((specifications[i] == null) || (specifications[i].isEmpty())) ? null : specifications[i];
		return (Specification[]) Text.dropNulls(nonEmptySpecifications);  // This cast will always succeed
	}

	/** Gets the names of a stack of specifications. **/
	public static String[] getNames(Specification[] specifications) {
		if (specifications == null)
			return null;
		String[] names = new String[ specifications.length ];
		for (int i=0; i<names.length; ++i)
			names[i] = specifications[i].name;
		return names;
	}

	/** Fills cipher and password specifications into an array of <CODE>Specification</CODE>s.
	 ** Cipher and passwords are matched one-by-one from the first element of each array.
	 ** Note that some passwords are allowed to be missing in which case they will get empty;
	 ** on the other hand there should <em>not</em> be more passwords than ciphers.
	 ** @param ciphers the ordered list of ciphers
	 ** @param passwords the ordered list of passwords
	 **/
	public static Specification[] fillIn(String[] ciphers, String[] passwords) {
		if (ciphers == null)
			ciphers = new String[0];
		if (passwords == null)
			passwords = new String[0];
		if (ciphers.length < passwords.length)
			return null;
		Specification[] specifications = new Specification[ ciphers.length ];
		for (int i=0; i<specifications.length; ++i)
			specifications[i] = new Specification(ciphers[i], (i < passwords.length) ? passwords[i] : null);
		return specifications;
	}

}



