
package sam.cipher;

import sam.ui.UserInterface;
import sam.core.Settings;

import java.io.*;


/** Generic variable-size {@link sam.cipher.War} factory.
 ** This allows for dynamic instantiation of <CODE>War</CODE> of various sizes.
 **
 ** <p>War stands for <em>Wide Area Rotator</em>
 ** <p>This class implements a cipher that given a key of size <CODE>n*p</CODE> bits shuffles a data-buffer
 ** of size <CODE>4^p</CODE> bytes <CODE>n</CODE> times. For the actual implementation, <CODE>p = 8</CODE>
 ** so that the buffer is 65536 bytes wide.
 ** <p>Easy to break out by brute force (due to small keys: for example <CODE>n*8</CODE>bits,
 ** implying only <CODE>n*65536</CODE> attack cases, which is not so big after all),
 ** it is <b>not</b> intended to be used solely, but rather as a destructuring step in the ciphering stack.
 **/
public class WarFactory extends sam.cipher.Factory {

	static {
		for (int i=0; i<Settings.Constants.KNOWN_WAR_SIZES.length; )
			new WarFactory( Settings.Constants.KNOWN_WAR_SIZES[ i++ ] );
	}

	/** Number of bits per key element **/
	public final int keyBits;

	/** Size of the square composing the buffer **/
	public final int size;

	/** Total buffer length, i.e. <CODE>size^2</CODE> **/
	public final int length;

	/** Builds a variable-size new <CODE>War</CODE> factory.
	 ** @param keyBits Number of bits per key element
	 ** Note that this also registers thie factory
	 ** @see Registry#register
	 **/
	public WarFactory(int keyBits) {
		super("War" + keyBits, keyBits + "-bits Wide Area Rotator is an asymetric private key cipher of arbitrarily" +
					" large keys.\nCurrent implementation uses shuffles a " +	(1 << (2 * keyBits)) + "-bytes wide buffer\n" );
		if (keyBits < 1)
			throw new IllegalArgumentException("Attempt to create War factory of non-positive size " + keyBits);
		if (keyBits > 16)
			throw new IllegalArgumentException("Attempt to create incredibly huge War factory of size " + keyBits);
		this.keyBits = keyBits;
		this.size = 1 << keyBits;
		this.length = size * size;
		Cipher.REGISTRY.register(this);
	}

	public int getPreferredKeySize(int size) {
		return keyBits * (size / keyBits);
	}

	/** Instantiates a new {@link WarFactory.War} cipher. **/
	protected Cipher make(long[] material, boolean direction) throws Error {
		Key key = new Key( resize(material) );
		return new War( (direction == Cipher.ENCRYPT) ? key : (Key) key.inverse() );
	}


	/** General purpose key class.
	 ** Keys are parsed from <CODE>String</CODE> using specified number of significant bits.
	 **/
	protected class Key extends sam.cipher.Key {

		/** The number of rounds issuable through these bits **/
		protected final int numberOfRounds;

		/** Builds a new <CODE>War</CODE> with specified key material
		 ** @param bits the <CODE>boolean[]</CODE> array composing the key material.
		 ** Note that it shall have been resized to a the uppest lower multiple of {@link WarFactory#keyBits}
		 **/
		protected Key(boolean[] bits) {
			super(bits);
			this.numberOfRounds = bits.length / keyBits;
		}

		/** Inverses this key.
		 ** <p>Please note that this calculation is dependant of the cipher algorithm actually used,
		 ** which is the <CODE>WAR</CODE> algorithm here.
		 ** @see War
		 **/
		public sam.cipher.Key inverse() {
			boolean[] inversed = new boolean[asBooleans.length];
			for (int i=0, j=asBooleans.length - keyBits; j >= 0; i += keyBits, j -= keyBits)
				for (int k=0; k<keyBits; ++k)
					inversed[j + k] = ! asBooleans[i + k];
			return new Key(inversed);
		}

		/** Gets the number of rounds to be processed by this key.
		 ** @return the number of rounds issuable through these bits
		 **/
		public int numberOfRounds() { return numberOfRounds; }

	}

	/** Resizes <CODE>key</CODE> to the uppest lower multiple of <CODE>keyBits</CODE>
	 ** If the current size is larger, the key is truncated.
	 ** If it is smaller, the key is concatenated to itself until it is large enough.
	 ** @param bits the key material to be resized
	 ** @param keySize the desired new key size
	 ** @throws Error if this key is empty and expected size is not
	 **/
	protected boolean[] resize(long[] bits) throws Error {
		return (keyBits <= 0) || (bits == null) ? new boolean[0] : resizeKey(Key.expand(bits, 64), keyBits * (bits.length * 64) / keyBits );
	}

	/** Actual <CODE>War</CODE> implementation **/
	public class War extends Cipher {

		/** Builds a new <CODE>War</CODE> cipher.
		 ** @param key the fixed key to be used (either encryption or decryption key)
		 ** @see Key
		 **/
		protected War(Key key) { super(key); }

		public int getPreferredSize(int size) {	return length; }  // <!>

		/** Encrypts exactly one block of plaintext.
		 ** @param data          The plaintext.
		 ** @param key  The session key to use for encryption.
		 **/
		protected void encrypt (byte[] data, int offset, sam.cipher.Key key) {
			rotate(data, offset, key.getBits() );
		}

		/** Decrypts exactly one block of ciphertext.
		 ** @param data          The plaintext.
		 ** @param key  The session key to use for decryption.
		 **/
		protected void decrypt (byte[] data, int offset, sam.cipher.Key key) {
			rotate(data, offset, key.getBits() );
		}

		/** Main permutation routine.
		 **
		 ** <p>This works as follows: given a square
		 **	   <pre><codeblock>
		 **	     A B
		 **	     C D
		 **	   </codeblock></pre>
		 **	   it first rotates block either clockwise
		 **	   <pre><codeblock>
		 **	     C A
		 **	     D B
		 **	   </codeblock></pre>
		 **	   or counter-clockwise
		 **	   <pre><codeblock>
		 **	     B D
		 **	     A C
		 **	   </codeblock></pre>
		 **	   according to <CODE>directions</CODE>'s lowest bit.
		 **	   It then recurses into each of the four squares for the next bits.
		 **/
		protected void rotate(byte[] data, int offset, boolean[] directions) {
			for (int i=0; i<directions.length; i += keyBits)
				rotate(data, directions, i, 0, size, offset, offset + size);
		}

		/**
		 ** Implementation notes:
		 ** <p>Using a buffer for expanding the permutation once into an <CODE>int[]</CODE> array, and then only
		 ** copying the source data using one more level of indirection would be certainly more efficient.
		 ** <p>Not implemented by now however: memory, developpement time, etc.
		 ** <p>Just do it !
		 **/
		private void rotate(byte[] data, boolean[] directions, int directionsIndex, int down, int up, int left, int right) {
			final int halfHeight = (up - down) / 2,
				halfWidth = (right - left) / 2;
			if ( ( halfHeight <= 1) || (halfWidth <= 1) )
				return;
			if (directionsIndex >= directions.length) {
				// The code should never get here:
				// the previous test should return first as long as <CODE>4^directions.length == length</CODE>
				UserInterface.get().error().warn("Terminating at directions index " + directionsIndex + "/" + directions.length +
																			 " while still at (" + down + "," + left + ")x(" + up + "," + right + ")");
				return;
			}
			final int midHeight = (down + up) / 2,
					midWidth = (left + right) / 2;
			final boolean mode = directions[directionsIndex];
			++ directionsIndex;
			if (mode) {  // Counter-clockwise operation
				// First cycle right
				for (int i=down; i < midHeight; i++)
					for (int j=left; j < midWidth; j++)
						right(data, i, j, halfHeight, halfWidth);
				// Then recurse
				rotate( data, directions, directionsIndex, down, midHeight, left, midWidth  );  // up-left
				rotate( data, directions, directionsIndex, down, midHeight, midWidth, right );  // down-left
				rotate( data, directions, directionsIndex, midHeight, up,   left, midWidth  );  // up-right
				rotate( data, directions, directionsIndex, midHeight, up,   midWidth, right );  // down-right
			} else {  // Clockwise operation
				// First recurse
					rotate( data, directions, directionsIndex, midHeight, up,   midWidth, right );  // down-right
					rotate( data, directions, directionsIndex, midHeight, up,   left, midWidth  );  // up-right
					rotate( data, directions, directionsIndex, down, midHeight, midWidth, right );  // down-left
					rotate( data, directions, directionsIndex, down, midHeight, left, midWidth  );  // up-left
					// Then cycle left
					for (int i = midHeight - 1; i >= down; i--)
						for (int j = midWidth - 1; j >= left; j--)
							left(data, i, j, halfHeight, halfWidth);
			}
		}

		private void right(byte[] data, int x, int y, int height, int width) {
			int i, j;
			byte buffer = data[ i = i(x, y ) ];
			data[i] = data[ j = i(x+height, y) ];
			data[j] = data[ i = i(x+height, y+width) ];
			data[i] = data[ j = i(x, y+width) ];
			data[j] = buffer;
		}

		private void left(byte[] data, int x, int y, int height, int width) {
			int i, j;
			byte buffer = data[ i = i(x, y ) ];
			data[i] = data[ j = i(x, y+width) ];
			data[j] = data[ i = i(x+height, y+width) ];
			data[i] = data[ j = i(x+height, y) ];
			data[j] = buffer;
		}

		/** Basic rectangular-to-linear accessor
		 ** Note that this method will most likely be inlined at compile-time.
		 **/
		private final int i(int x, int y) {	return size * x + y; }

	}

}

