
/* This whole file is distributed under the terms of the General Public License.
 * Last revision date: 2002/03/06 by glacote
 */

package sam.core;

import sam.entry.Entry;
import sam.io.IO;
import sam.toolkit.Text;
import sam.core.Settings;
import sam.ui.UserInterface;

import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.util.jar.*;


/**
 ** This class defines the internal <CODE>Jar</CODE> archive layout.
 ** @see Resource
 **/
public class ArchiveLayout {
	private ArchiveLayout() { }  // Purely static class

	/** The path separator to be used in <CODE>Entry</CODE> **/
	public static final char ENTRY_PATH_SEPARATOR = '/';

	/** SAM file format version, currently 0.90 (encoded as the <CODE>90</CODE> integer) */
	public static final int LAYOUT_VERSION_NUMBER = 90;

	/** Specific resources in the jar archive. **/
	public static class Resource {
		private static final String BASE = "data/";  // Everything in a subdirectory (cleaner) excepts the disclaimer

		/** The entry name of this resource **/
		public final String name;

		/** A human-readable description of this resource **/
		public final String title;

		/** Wether this resource must always be present in the archive **/
		public final boolean isOptionnal;

		/** The preferred compression method for this resource
		 ** @see java.util.zip.Zip.ZipEntry
		 **/
		protected final int compressionMethod; 

		public String toString() { return name; }

		/** Builds a new resource specification.
		 ** By default this resource is assumed to be always present in the archive
		 ** @param name the entry name of this resource
		 ** @param title a human-readable description of this resource
		 **/
		private Resource(String name, String title) { this(name, title, false); }

		/** Builds a new resource specification
		 ** @param name the entry name of this resource
		 ** @param title a human-readable description of this resource
		 ** @param isOptionnal wether this resource must always be present in the archive
		 **/
		private Resource(String name, String title, boolean isOptionnal) {
			this(name, title, isOptionnal, false);
		}

		/** Builds a new resource specification
		 ** @param name the entry name of this resource
		 ** @param title a human-readable description of this resource
		 ** @param isOptionnal wether this resource must always be present in the archive
		 ** @param doNotCompress wether to avoid compressing this resource
		 ** while storing it into the <CODE>Jar</CODE> archive
		 **/
		private Resource(String name, String title, boolean isOptionnal, boolean doNotCompress) {
			this.name = BASE + name;
			this.title = title;
			this.isOptionnal = isOptionnal;
			this.compressionMethod = doNotCompress ? 0 : Settings.Defaults.JAR_COMPRESSION_LEVEL;
		}

		/** Opens this <CODE>Resource</CODE> from the boot-time jar file assumed to be the archive **/
		public InputStream get() throws IOException { return get(null); }

		/** Opens this <CODE>Resource</CODE> from specified archive
		 ** @param archive the archive to read from
		 **/
		public InputStream get(ZipFile archive) throws IOException {
			UserInterface.get().log().log(2, ArchiveLayout.class, "Recovering " + title + " from " + name);
			InputStream stream;
			if (archive == null)
			 stream = Settings.Environment.BOOT_CLASS.getClassLoader().getResourceAsStream(name);
			else {
				ZipEntry entry = archive.getEntry(name);
				stream = (entry == null) ? null : archive.getInputStream(entry);
			}
			if ( (stream == null) &&  (! isOptionnal) )
				UserInterface.get().error().warn("Missing resource \"" + name + "\": " + title);
			return stream;
		}

		/** Reads this <CODE>Resource</CODE> as a <CODE>String</CODE>
		 ** @param archive the archive to read from
		 ** @return <CODE>null</CODE> in case of I/O error
		 **/
		public String getString(ZipFile archive) {
			try {
				return IO.getFromStream( get(archive) );
			} catch (IOException e) {
				UserInterface.get().error().handleException("Could not recover " + name + " from " + archive, e);
				return null;
			}
		}

		/** Reads this <CODE>Resource</CODE> line-by-line as a <CODE>String[]</CODE> array
		 ** @param archive the archive to read from
		 ** @return <CODE>new String[0]</CODE> in case of I/O error
		 **/
		public String[] getStrings(ZipFile archive) {	return Text.splitLines( getString(archive) );	}

		/** Reads this <CODE>Resource</CODE> as a <CODE>long</CODE> number
		 ** @param archive the archive to read from
		 ** @return <CODE>0</CODE> in case of I/O error
		 **/
		public long getLong(ZipFile archive) throws NumberFormatException {
			String value = getString(archive);
			if ((value == null) || ( "".equals(value) ))
					if (isOptionnal)
							return 0;
					else {
						UserInterface.get().error().warn("Could recover not integral number \"" + name + "\" from " + archive
																						 + ", defaulting to " + Settings.Constants.ARCHIVE_RESOURCE_INTEGRAL_VALUE);
						return Settings.Constants.ARCHIVE_RESOURCE_INTEGRAL_VALUE;
					}
			else
				return Text.fromHexa(value);
		}

		/** Reads this <CODE>Resource</CODE> as a <CODE>int</CODE> number
		 ** @param archive the archive to read from
		 ** @return <CODE>0</CODE> in case of I/O error
		 **/
		public int getInt(ZipFile archive) throws NumberFormatException {
			return (int) getLong(archive);
		}

		/** Reads this <CODE>Resource</CODE> as a <CODE>byte[]</CODE> array
		 ** @param archive the archive to read from
		 ** @return <CODE>null</CODE> in case of I/O error
		 **/
		public byte[] getBytes(ZipFile archive) {
			try {
				return IO.readFully( get(archive) );
			} catch (IOException e) {
				UserInterface.get().error().handleException("Could not read \"" + name + "\" fully", e);
				return null;  // Fake for compiler
			}
		}

		/** Opens this <CODE>Resource</CODE> from specified archive
		 ** @param archive the archive write to
		 **/
		public OutputStream put(JarOutputStream jar) {
			try {
				return IO.putNextEntry(jar, name, compressionMethod);
			} catch (IOException e) {
				UserInterface.get().error().handleException("Could not set " + title + " as \"" + name + "\"", e);
				return null;  // Fake for compiler
			}
		}

		/** Writes this <CODE>Resource</CODE> as a <CODE>long</CODE> number
		 ** @param archive the archive to write to
		 **/
		public void putLong(long content, JarOutputStream jar) { putString(Text.toHexa(content), jar); }

		/** Writes this <CODE>Resource</CODE> as a <CODE>int</CODE> number
		 ** @param archive the archive to write to
		 **/
		public void putInt(int content, JarOutputStream jar) { putLong( (long) content, jar); }

		/** Writes this <CODE>Resource</CODE> as a <CODE>String</CODE>
		 ** @param archive the archive to write to
		 **/
		public void putString(String content, JarOutputStream jar) { putInputStream( IO.getAsStream(content), jar);	}

		/** Writes this <CODE>Resource</CODE> line-by-line as a <CODE>String[]</CODE> array
		 ** @param archive the archive to write to
		 **/
		public void putStrings(String[] content, JarOutputStream jar) {	putInputStream( IO.getAsStream(content), jar); }

		/** Writes this <CODE>Resource</CODE> as an <CODE>InputStream</CODE>
		 ** @param archive the archive to write to
		 **/
		public void putInputStream(InputStream content, JarOutputStream jar) {
			UserInterface.get().log().log(2, ArchiveLayout.class, "Setting " + title + " as " + name);
			try {
				sam.io.IO.jar(content, name, jar);
			} catch (IOException e) {
				UserInterface.get().error().handleException("Could not set " + title + " as \"" + name + "\"", e);
			}
		}

		/** Writes this <CODE>Resource</CODE> line-by-line as a <CODE>byte[]</CODE> array
		 ** @param archive the archive to write to
		 **/
		public void putBytes(byte[] content, JarOutputStream jar) {
			UserInterface.get().log().log(2, ArchiveLayout.class, "Setting " + title + " as " + name);
			try {
				sam.io.IO.jar(content, name, jar);
			} catch (IOException e) {
				UserInterface.get().error().handleException("Could not set \"" + title + "\"", e);
			}
		}

	}

	/** Loads the full catalog of specified archive, including files contained in imported archives.
	 ** <p>Note that not all of these entries are actually stored in this archive, as some might be
	 ** stored in such a previously imported archive.
	 ** <p>However this is the list of all entries of <em>this</em> archive if it does not import
	 **  any other one.
	 **/
	public static sam.entry.Enumerator loadFullCatalog(File archive) {
		if (archive == null)
			return sam.entry.Enumerator.EMPTY;
		try {
			UserInterface.get().log().log(2, ArchiveLayout.class, "Loading full catalog from archive \"" + archive + "\"");
			ZipFile file = new ZipFile(archive);
			sam.entry.List imported, catalog;
			UserInterface.get().log().log(3, ArchiveLayout.class, "Loading archive catalog");
			catalog = sam.entry.Enumerator.load( CATALOG.get(file) ).statify();
			String importedArchive = null;
			UserInterface.get().log().log(3, ArchiveLayout.class, "Loading archive imports");
			try {
				importedArchive = IO.getFromStream( IMPORTED_ARCHIVE.get(file) );
			} catch (IOException e) {
				UserInterface.get().error().fail("Could not load imports archive reference", e);
			}
			if (importedArchive == null)
					UserInterface.get().log().log(4, ArchiveLayout.class, "No imports found in archive");
			imported = (importedArchive == null) ? null : loadFullCatalog(new File(importedArchive)).statify();
			if (imported != null)
				UserInterface.get().log().log(3, ArchiveLayout.class, "Merging " + imported.count() + " imports with " +
																			catalog.count() + " included entries");
			return (imported == null) ? catalog : imported.appendNewerFrom(catalog);
		} catch (IOException e) {
			UserInterface.get().error().fail("Could not load full catalog from archive \"" + archive + "\"", e);
			return null;  // Fake for compiler
		}
	}

	/** Manifest used for archive.
	 ** <p>The most important is the <CODE>Main-Class: Setting.Environment.BOOT_CLASS</CODE> entry.
	 ** @see Setting.Environment.BOOT_CLASS
	 **/
	public static final Manifest MANIFEST =
		makeManifest( "Manifest-Version: 1.0\nMain-Class: " + Settings.Environment.BOOT_CLASS.getName() +
									"\nName: Sam - Secure Archive Manager\nSealed: true\n" +
									"Specification-Title: \"General purpose encryptor\"\nSpecification-Version: \"" + LAYOUT_VERSION_NUMBER + "\"\n" +
									"Specification-Vendor: \"Partially free software (partial-GPL)\"\n" +
									"Implementation-Title: \"sam\"\nImplementation-Version: \"" + Settings.Environment.PROGRAM_VERSION + "\"\n" +
									"Implementation-Vendor: \"" + Settings.Environment.AUTHOR + "(mailto:" + Settings.Environment.EMAIL + ")" + "\"" );

	private static Manifest makeManifest(String manifest) {
		try {
			return new Manifest( IO.getAsStream(manifest) );
		} catch (IOException e) {
			return new Manifest();
		}
	}

		/** The <CODE>long</CODE> version number of the archive format */
		public static final Resource SAM_VERSION = new Resource("VERSION.txt", "archive version");

		/** The optionnal imported archive */
		public static final Resource IMPORTED_ARCHIVE = new Resource("import.text", "imported archive", true);

		/** The ordered list of all ciphers, in plain ASCII with new line as delimiter.	*/
		public static final Resource CIPHERS = new Resource("ciphers.txt", "ciphers list");

		/** The plain text name of the packer used **/
		public static final Resource PACKER = new Resource("packer.txt", "packer");

		/** The plain text name of the tracker used **/
		public static final Resource TRACKER = new Resource("tracker.txt", "tracker");

		/** The block size used @see sam.Packer */
		public static final Resource BLOCK_SIZE = new Resource("block-size.txt", "block size");

		/** An optionnal comment associated with the archive, in plain utf. */
		public static final Resource COMMENT = new Resource("comment.txt", "comment", true);

		/** The <CODE>(int:"fileLength",int:"date",utf:"fileName")*</CODE> binary list of all archived files */
		public static final Resource CATALOG = new Resource("data.cat", "archive catalog");

		/** The <CODE>(long:"garbageBytes",long:"dataBytes")*</CODE> Mixing trace */
		public static final Resource MIXER = new Resource("mixing.cat", "mixing trace");

		/** The serial number of this archive */
		public static final Resource SERIAL_NUMBER = new Resource("serial-number.txt", "Serial number");

		/** The user creating the archive  **/
		public static final Resource USER = new Resource("user.txt", "user who created the archive");

		/** The group of the user creating the archive  **/
		public static final Resource GROUP = new Resource("group.txt", "group of the user who created the archive");

		/** The compressed raw data of all orignal files first appended together and then filtered through the ciphers,
		 ** and randomly mixed with random-sized randomly unbalanced data chunks.
		 **/
		public static final Resource DATA = new Resource("data.sam", "main ciphered data", false, true);

		public static final LayoutVersion LAYOUT_VERSION = new LayoutVersion();

		public static class LayoutVersion extends Resource {
			private LayoutVersion() { super("LAYOUT.txt", "archive layout version"); }

			public void putVersion(JarOutputStream jar) { putLong(LAYOUT_VERSION_NUMBER, jar); }

			public void check(ZipFile archive) {
				final int archiveVersion = getInt(archive);
				if (archiveVersion != LAYOUT_VERSION_NUMBER)
					UserInterface.get().error().incompatibleArchiveLayout(archiveVersion);
			}

		}

		/** To be included runtime, source and documentation files */
		public static final Runtime RUNTIME =	new Runtime( Settings.Distribution.SDK );

		public static class Runtime extends Resource {

			private final String[] specifications;

			Runtime(String[] specifications) {
				super("sam/", "runtime", true);
				this.specifications = specifications;
			}

			public void putRuntime(JarOutputStream jar) {
				try {
					UserInterface.get().log().log(2, ArchiveLayout.class, "Scanning for runtime" );
					final sam.entry.List files = Settings.Distribution.scan(specifications).statify().sortAndRemoveDuplicates();
					UserInterface.get().log().log(2, ArchiveLayout.class, "Including runtime: " + files.count() + " files and "
																				+ Text.getSizeString(files.size()) );
					while ( files.hasMoreElements() ) {
						Entry next = files.nextEntry();
						UserInterface.get().log().log(3, ArchiveLayout.class, "Including " + next.verbose() );
						sam.io.IO.jar( next.getInputStream(), next.length, next.name, jar);
					}
				} catch (IOException e) {
					UserInterface.get().error().handleException("Could not include runtime into archive \"" + title + "\"", e);
				}
			}

		}

}

