
package sam.core;

import sam.ui.UserInterface;
import sam.entry.Enumerator;
import sam.toolkit.Text;
import sam.cipher.*;

import java.io.File;


/** General-purpose central repository of <em>all</em> constants and default settings of the program.
 **/
public class Settings {

	/** Repository of various environment-dependant values. **/
	public static class Environment {

		/** <a href=mailto:sam@glacote.com>Author</a> of this program as a <CODE>String</CODE> **/
		public static final String AUTHOR = "Guillaume Lacte";

		/** <a href=mailto:sam@glacote.com>Email address</a> of the author, as a <CODE>String</CODE> **/
		public static final String EMAIL = "sam@glacote.com";

		/** <a href=glacote.com/sam>Home page</a> of the program, as a <CODE>String</CODE> **/
		public static final String HOME_PAGE = "glacote.com/sam";

  	public static final String PROGRAM_NAME = "SAM";

  	/** SAM version, currently 0.92 (encoded as the <CODE>92</CODE> integer) **/
	  public static final int PROGRAM_VERSION = 92;

  	/** SAM version, as a human-readable <CODE>String</CODE>
		 ** @see #PROGRAM_VERSION
		 **/
		public static String programVersion() { return "" + (PROGRAM_VERSION / 100.0) ; }

		/** The user running the current VM **/
		public static final String USER = getProperty("user.name", "sam");

		/** The home directory of the user running the current VM **/
		public static final String HOME = getProperty("user.home", File.separator);

		/** The group of the user running the current VM **/
		public static final String GROUP = getProperty("user.name", "sam");

		/** The operating system the current VM is running under **/
		public static final String OPERATING_SYSTEM = getProperty("os.name", "non-conventional operating system");

		/** The operating system version the current VM is running under **/
		public static final String OPERATING_SYSTEM_VERSION = getProperty("os.version", "undefined");

		/** The hardware architecture the current VM is running on **/
		public static final String COMPUTER_ARCHITECTURE = getProperty("os.arch", "non-conventional hardware architecture");

		/** Wether the standard input is synchronous, or not
		 ** @see sam.ui.text.UserInterface
		 **/
		public static final boolean STDIN_IS_SYNCRHONOUS = false;

		/** Gets a greeting string, suitable to display e.g. at startup **/
		public static final String greetings() { return "Welcome to " + PROGRAM_NAME + " v" + programVersion() + " by " + AUTHOR ; }

		/** The main runnable class **/
		public static final Class BOOT_CLASS = sam.main.class;

		/** The command-line prefix to run the program
		 ** This looks like <CODE>"java sam.main"</CODE>
		 **/
  	public static final String PROGRAM_BINARY = "java " + BOOT_CLASS.getName();

		/** A human-readable description of the program and its environment **/
  	public static String programDescription() {
			return "Safe Archive Manager version " + programVersion() + "  by " + AUTHOR	+ " (" + EMAIL + ")\n"
					+ "java version " + System.getProperty("java.version", "?") + " (jvm version " + System.getProperty("java.vm.version", "?")
					+ ") on " + OPERATING_SYSTEM + " version " + OPERATING_SYSTEM_VERSION	+ " for " + COMPUTER_ARCHITECTURE ;
		}

		/** Checks wether the VM is running under a <CODE>Microsoft</CODE> operating system. **/
		public static boolean isRunningOnMicrosoftOS() {
			return OPERATING_SYSTEM_VERSION.toLowerCase().indexOf("microsoft") >= 0;
		}

		/** Security disclaimer for users of <CODE>Microsoft</CODE> operating system. **/
		public static final String MICROSOFT_DISLAIMER = "Warning:\n" +
				"You are running SAM under " + OPERATING_SYSTEM_VERSION + ".\n" +
				OPERATING_SYSTEM_VERSION + " is a HIGHLY UNSECURE operating system,\n" +
				"full of dramatic security whole whose correction depends\n" +
				"on the good will of the Microsoft(tm) Company.\n\n" +
				"It makes NO SENSE to use such an operating system to handle\n" +
				"confidential data.\n" +
				"You should consider installing an TRUE operating system,\n" +
				"e.g. Linux (www.linux.com)" ;

		/** Prints a security disclaimer if running under a <CODE>Microsoft</CODE> operating system.
		 ** @see #MICROSOFT_DISLAIMER
		 **/
		public static void warnMicrosoftOperatingSystemOwners() {
			if ( isRunningOnMicrosoftOS() )
				UserInterface.get().running().print("Important note", MICROSOFT_DISLAIMER);
		}

		/** Appends a class path to the current class path
		 **/
		public static void appendClassPath(String classPath) {
			if (classPath == null)
				return;
			try {
				String oldClassPath = System.getProperty("java.classpath");
				System.setProperty("java.classpath", (oldClassPath != null) ? classPath : (classPath + ':' + oldClassPath) );
			} catch (SecurityException e) {
				UserInterface.get().error().handleException("Could not expand class path by \"" + classPath +	"\" due to security restrictions: ",
																										e);
			}
		}

		/** System properties keys associated to a valid temporary directory **/
		private static final String[] TEMPORARY_DIRECTORIES = { "TMP", "TEMP", "tmp", "temp", "user.home"  };

		/** Gets the default temporary directory **/
		public static String getDefaultTemporary() {
			String directory;
			for (int i=0; i<TEMPORARY_DIRECTORIES.length; i++)
				if ( (directory = System.getProperty(TEMPORARY_DIRECTORIES[i])) != null)
					return directory;
			return null;
		}

		public static final String TEMPORARY_FILES_SUFFIX = ".tmp";

		private static String getProperty(String property, String defaultValue) {
			try {
				String value = System.getProperty(property, defaultValue);
				return (value == null) ? defaultValue : value;
			} catch (SecurityException e) {
				return defaultValue;
			}
		}

	}

	/** Repository for various files distribution policies.
	 **/
	public static class Distribution {

		/** The base directory of all source files **/
		public static final File BASE_DIRECTORY = sam.toolkit.Process.getClassDirectory(Environment.BOOT_CLASS).getParentFile();

		/** String[] of all binaries files **/
		public static final String[] BINARIES = { "sam/*.class" };

		/** String[] of all documentation files **/
		public static final String[] DOCUMENTATION = { "*.html", "stylesheet.css" };

		/** String[] of all license files **/
		public static final String[] LICENSE = { "GPL.txt", "README.html"};

		/** String[] of all freely distributable source files **/
		public static final String[] FREE_SOURCES = {
				"sam/io", "sam/event", "sam/ui", "sam/core", "sam/rules", "sam/entry", "sam/toolkit",
				"sam/cipher/Cipher.java", "sam/cipher/Benchmark.java", "sam/cipher/XOR.java", "sam/cipher/WarFactory.java" };

		/** String[] of all source files ruled by a dedicated license **/
		public static final String[] RESTRICTED_SOURCES = {
				"sam/cipher/DES.java", "sam/cipher/Deal.java", "sam/cipher/Frog.java", "sam/cipher/Loki.java", "sam/cipher/Mars.java",
				"sam/cipher/RC6.java", "sam/cipher/Rijndael.java", "sam/cipher/Serpent.java", "sam/cipher/Twofish.java" };

		/** String[] of all source files, were they restricted of freely distributable **/
		public static final String[] SOURCES = Text.merge(new String[][] { FREE_SOURCES, RESTRICTED_SOURCES } );

		/** String[] of all files composing the Sam Runtime Environment **/
		public static final String[] SRE = Text.merge(new String[][] { BINARIES, LICENSE } );

		/** String[] of all files composing the Sam Development Kit **/
		public static final String[] SDK = Text.merge(new String[][] { SRE, SOURCES, DOCUMENTATION } );;

		/** Scans all files macthing <CODE>specifications</CODE> from the runtime base directory **/
		public static Enumerator scan(String[] specifications) { return sam.entry.FileSystem.scan(BASE_DIRECTORY, specifications); }

		/** Builds a manifest file set to boot on the main boot class.
		 ** @see sam.core.Settings.Environment#BOOT_CLASS
		 **/
		public static java.util.jar.Manifest makeManifest() { return makeManifest(Environment.BOOT_CLASS); }

		/** Builds a manifest file set to boot on specified boot class.
		 ** @param bootClass the class to boot from
		 ** See <a href=http://java.sun.com/products/jdk/1.2/docs/guide/jar/manifest.html>Official specifications</a> for details.
		 **/
		public static java.util.jar.Manifest makeManifest(Class bootClass) {
			try {
				return new java.util.jar.Manifest( sam.io.IO.getAsStream("Manifest-Version: 1.0\nMain-Class: " + bootClass.getName()
																					+ "\nName: Sam - Secure Archive Manager\nSealed: true\n"
																					+ "Specification-Title: \"General purpose encryptor\"\nSpecification-Version: \""
																					+ "0." + sam.core.ArchiveLayout.LAYOUT_VERSION_NUMBER + "\"\n"
																					+ "Specification-Vendor: \"Partially free software (partial-GPL)\"\n"
																					+ "Implementation-Title: \"sam\"\nImplementation-Version: \""
																					+ "0." + Environment.PROGRAM_VERSION + "\"\n"
																					+ "Implementation-Vendor: \"" + Environment.AUTHOR + "\""
																					+ "Implementation-URL: \"" + Environment.HOME_PAGE + "\""
																					) );
			} catch (java.io.IOException e) {
				UserInterface.get().error().handleException("Could not create proper manifest, archive might not be executable although valid", e);
				return new java.util.jar.Manifest();
			}
		}

	}

	/** Repository  of various constants
	 ** Changing them does not modify the program semantic but might change its performance/visual behaviour
	 **/
	public static class Constants {

		/** Lifetime of the random consumer thread.
		 ** Valid values are any positive value, or <CODE>0</CODE> to disable consumer threads.
		 ** @see sam.toolkit.Random
		 **/
		public static final int RANDOM_GENERATOR_CONSUMER_LIFETIME = 10;

		/** <CODE>String</CODE> that erases the previous character when displayed. **/
		public static final String BACKSPACE = "\b";

		/** Alias to the list of all registered ciphers **/
		public static final String CIPHER_WILDCARD = "*";

		/** Prefix for file alias command-line argument
		 ** @see sam.rules.Syntax
		 ** @see sam.rules.SyntaxSet
		 **/
		public static final char COMMAND_LINE_FILE_ALIAS = '@';

		/** Size of I/O dump buffers in bytes
		 ** @see sam.io.IO
		 **/
		public static final int DUMP_BUFFER_SIZE = 4096;

		/** Size of <CODE>byte[]</CODE> array filters.
		 ** @see sam.io.ByteArrayFilter
		 ** @see sam.core.Storer
		 ** @see sam.core.Unstorer
		 **/
		 public static final int FILTER_BLOCK_BUFFER_SIZE = 16;

		/** Block size to be used for parallel I/O computations.
		 ** @see sam.io.ParallelIO
		 ** @see sam.core.Storer
		 ** @see sam.core.Unstorer
		 **/
		public static final int MULTI_THREADED_BLOCK_SIZE = 1024;

		/** The granularity of plain data commitment.
		 ** @see sam.event.ProgressEventNotifier
		 **/
		public static final long PROGRESS_PLAIN_GRANULARITY = 32 * 1024;

		/** The granularity of encoded data commitment.
		 ** @see sam.event.ProgressEventNotifier
		 **/
		public static final long PROGRESS_ENCODED_GRANULARITY = 128 * 1024;

		/** Scheduling priority of mailbox post agents
		 ** @see sam.toolkit.Mailbox
		 ** @see java.lang.Thread#NORM_PRIORITY
		 **/
		public static final int MAIL_POSTMASTER_PRIORITY = Thread.NORM_PRIORITY;

		/** Wether the <CODE>ProgressEvent</CODE> postmaster(s) can stopped several times.
		 ** This is safe although computationnally expensive and theoretically useless.
		 ** @see sam.toolkit.Mailbox
		 **/
		public static final boolean ALLOW_MULTIPLE_MAIL_STOPS = true;

		/** The minimum log level
		 ** @see sam.ui.LogInterface
		 **/
		public static final int MINIMUM_LOG_LEVEL = -1;

		/** The maximum log level
		 ** @see sam.ui.LogInterface
		 **/
		public static final int MAXIMUM_LOG_LEVEL = 100;

		/** Wether to set the log level as soon as it has been parsed from command-line
		 ** or to wait for the whole command-line to have been parsed ans set it then.
		 ** @see sam.ui.LogInterface
		 **/
		public static final boolean EARLY_LOG_LEVEL_SETTING = true;;

		/** Tabulation string to be used to indent log messages according to their depth.
		 ** @see sam.ui.LogInterface
		 **/
		public static final String LOG_INDENTATION = "  ";

		/** Wether a log reader should always be started statically
		 ** @see sam.ui.LogInterface
		 **/
		public static final boolean ALWAYS_STARTS_LOG = true;

		/** The <CODE>long</CODE> value returned in case of broken archive.
		 ** @see sam.core.ArchiveLayout
		 **/
		public static final long ARCHIVE_RESOURCE_INTEGRAL_VALUE = 0;

		/** Conventionnal file name designating the standard output **/
		public static final String STDOUT_NAME = "stdout";

		/** Conventionnal file name designating the standard error **/
		public static final String STDERR_NAME = "stderr";

		/** Default output **/
		public static final String DEFAULT_TRACE = STDOUT_NAME;

		/** Wether each newly registered class should automatically be set the active class.
		 ** @see sam.toolkit.ClassRegistry
		 **/
		public static final boolean NEWLY_REGISTERED_CLASS_IS_AUTOMATICALLY_SET = false;

		/** Repository of settings related to user interfaces. **/
		public static class UI {

			/** The name of a failsafe user interface
			 ** Its initialization should <b>never</b> fail.
			 ** @see sam.ui.basic.UserInterface
			 **/
			public static String FAILSAFE = "basic";

			/** The decreasing ordered list of the preferred user interfaces **/
			public static String[] PREFERENCES = new String[] { "awt", "text", "basic" };

			/** Repository of all settings related to the text-based user interfaces. **/
			public static class Text {

				/** The displayed name length of entries in the text-based user interface
				 ** @see sam.ui.text.UserInterface
				 **/
				public static final int ENTRY_SHORT_NAME_LENGTH = 20;

				/** The displayed name length of entries in the text-based user interface
				 ** @see sam.ui.text.UserInterface
				 **/
				public static final int PROGRESS_BAR_LENGTH = 50;

				/** Maximum length of a text-base progress event line
				 ** @see sam.ui.text.UserInterface
				 **/
				public static final int LINE_LENGTH = ENTRY_SHORT_NAME_LENGTH + PROGRESS_BAR_LENGTH + 60;

			}

			/** Repository of all settings related to the text-based user interfaces. **/
			public static class AWT {

				/** Default life time of dialog spawned by the <CODE>awt</CODE>-based user interface
				 ** @see sam.ui.awt.InfoBox
				 **/
				public static final long DIALOG_DEFAULT_LIFETIME = 5;

				/** Immutable dimension of the main progress window. **/
				public static final java.awt.Dimension PROGRESS_WINDOW_SIZE = new java.awt.Dimension(280, 140);

				/** Immutable dimension of the main progress window. **/
				public static final java.awt.Dimension LOG_WINDOW_SIZE = new java.awt.Dimension(600, 280);

				/** Default location of components freshly spawned by the <CODE>awt</CODE>-based user interface
				 ** @see sam.ui.awt.InfoBox
				 **/
				public static final java.awt.Point DEFAULT_LOCATION = new java.awt.Point(200, 150);

				/** Default log window position relative to the main window.
				 ** @see java.awt.BorderLayout
				 ** @see sam.ui.awt.LogInterface
				 **/
				public static final String DEFAULT_LOG_POSITION = java.awt.BorderLayout.EAST;

				/** Default maximal length of log buffer of the <CODE>awt</CODE>-based user interface
				 ** @see sam.ui.awt.InfoBox
				 **/
				public static final int LOG_LENGTH = 4000;

				/** Wether opening the log window should also enable log **/
				public static final boolean SHOWING_LOG_ENABLES_IT = false;

				/** Wether opening the log window should also enable log **/
				public static final boolean HIDING_LOG_DISABLES_IT = false;

				/** Length of entry names in progress window **/
				public static final int ENTRY_NAME_LENGTH = 20;

				/** Wether to print compression ratio **/
				public static final boolean PRINT_COMPRESSION_RATIO = true;

				/** Plain progress bar color
				 ** @see sam.ui.awt.ProgressBar
				 **/
				public static final java.awt.Color PLAIN_BAR_COLOR = java.awt.Color.blue.darker();

				/** Encoded progress bar color
				 ** @see sam.ui.awt.ProgressBar
				 **/
				public static final java.awt.Color ENCODED_BAR_COLOR = java.awt.Color.cyan.darker();

				/** Text progress bar color
				 ** @see sam.ui.awt.ProgressBar
				 **/
				public static final java.awt.Color TEXT_BAR_COLOR = java.awt.Color.lightGray.darker();

			}

		}

		/** Number of digits to be used to report entry length.
		 ** This sets an upper bound to the largest entry able to be listed
		 ** (but does <b>not</b> affect other operations like storing or unstoring).
		 ** @see sam.entry.Entry#list
		 **/
		public static final int ENTRY_LENGTH_DIGITS = 12;	

		/** The maximum block size in kilobytes **/
		public static final int MAXIMUM_BLOCK_SIZE = 4 * 1024;  // Which is 4 MBytes ...

		/** Wether the cipher/password must not always be rentered by user
		 ** whenever it is adequately set in command-line.
		 **/
		public static final boolean DO_NOT_REENTER_FULLY_SPECIFIED_CIPHERS = true;

		/** Wether command-line ciphers are ginored in extracting mode and replaced
		 ** by those reloaded from archive.
		 ** Note that command-line passwords are <b>not</b> dropped, in turn.
		 ** <p>Although this is semantically arguable it allows the same
		 ** <CODE>"-pciphers:passwords"</CODE> be used both in archving and 
		 ** in extracting mode.
		 **/
		public static final boolean SUPPLIED_CIPHERS_IGNORED_IN_EXTRACTING = true;

		/** Wether to pre-extract data from archive before unstoring.
		 ** This is a bug-fix: the <CODE>Java</CODE> API seems to handle inner <CODE>Zip</CODE>
		 ** input streams incorrectly. In particular, due to internal native methods the
		 ** buffering appears to be skewed, i.e. it does not always read as many bytes as required
		 ** at each call. Thus it seems necessary to pre-extract ciphered data to a temporary
		 ** file before handling it.
		 **/
		public static final boolean PREEXTRACT_DATA = true;

		/** Central repository of settings related to cipher keys.
		 ** @see sam.cipher.Key
		 **/
		public static class Key {

			/** Wether to be totally silent about key material.
			 ** This is a paranoa setting.
			 ** This setting takes precedence on all other key-related log settings.
			 **/
			public static final boolean ALWAYS_SILENT = false;

			/** Wether to issue a log whenever a new key is created. **/
			public static final boolean LOG_KEY_INSTANTIATION = true;

			/** Wether to include effective key material length while logging key instantiation.
			 ** Note that this setting has no effect unless {@link #LOG_KEY_INSTANTIATION} is <CODE>true</CODE>.
			 **/
			public static final boolean LOG_KEY_MATERIAL_LENGTH = true;

			/** Wether to issue a log whenever a key is turncated to match a ciphers specific key size. **/
			public static final boolean LOG_TRUNCATIONS = true;

			/** Wether to issue a warning whenever a key is turncated to match a ciphers specific key size. **/
			public static boolean WARN_TRUNCATIONS = false;

			/** Wether to issue a log whenever a key is turncated to match a ciphers specific key size.
			 ** <b>Warning:</b> key expansion is inherently unsafe as it reduces the key relative entropy.
			 **/
			public static final boolean LOG_EXPANSIONS = true;
				
			/** Wether to issue a warning whenever a key is turncated to match a ciphers specific key size.
			 ** <b>Warning:</b> key expansion is inherently unsafe as it reduces the key relative entropy.
			 **/
			public static final boolean WARN_EXPANSIONS = true;

			/** Wether to discard ciphers whose <CODE>Key</CODE> is empty.
			 ** <b>Warning:</b> it is necessary for <CODE>Key</CODE> subclasses to override the <CODE>Key.isEmpty()</CODE>
			 ** method adequately to avoid not-really empty keys be discarded.
			 ** @see sam.cipher.Key#isEmpty
			 **/
			public static final boolean IGNORE_EMPTY = true;
				
			/** Wether to issue a warning if a key is built empty. **/
			public static final boolean WARN_EMPTY = true;
			
				
			/** Wether to issue a warning whenever a <CODE>sam.cipher.Key</CODE> subclass
			 ** is inversed through the <CODE>sam.cipher.Key.inverse()</CODE> method
			 ** which returns an new instance of <CODE>sam.cipher.Key</CODE> instead of
			 ** the subclass.
			 **/
			public static final boolean WARN_SUBCLASSES_INVERSION = true;

		}

		/** List of all registered WAR ciphers.
		 ** Valid values are any list of non-negative integers.
		 ** Remember that as the buffer size will be <CODE>4 ^ bits</CODE>
		 ** this <CODE>bits</CODE> value should remain below 16.
		 ** @see sam.cipher.WarFactory
		 **/
		public static final int[] KNOWN_WAR_SIZES = new int[] { 8 };  // { 4, 8, 16 }

		/** Maximum length of a <CODE>ByteArrayFilter</CODE> buffer
		 ** @see sam.cipher.Cipher
		 **/
		public static final int BYTE_ARRAY_FILTER_MAXIMUM_BUFFER_LENGTH = 1024 * 1024;

		/** Wether to issue a warning whenever an archive is extracted from a different
		 ** user than the one who created it.
		 **/
		public static final boolean WARN_DIFFERENT_USER = true;

		/** Wether to issue a warning whenever an archive is extracted from a user
		 ** in a different group than that of the user who created it.
		 **/
		public static final boolean WARN_DIFFERENT_GROUP = true;

		/** Wether to ask the user for confirmation before overwriting a local file
		 ** during extraction.
		 **/
		public static final boolean ASK_BEFORE_OVERWRITING_LOCAL_FILES = true;

		/** Wether to add expensive but failsafe tests during parallel I/O operations.
		 ** This <b>must</b> set if the application is expected to run accross multiple
		 ** processors and the JVM uses aggressive value caching
		 ** (<a href=java.sun.com/j2se>Sun's</a> does).
		 **/
		public static final boolean PARALLELIO_PARANOIAC = true;

	}

	/** Central repository of various default values.
	 ** These values are queried whenever a required field is not supplied directly by the user.
	 ** A value of <CODE>null</CODE> means that <b>no</b> default value is provided and the user
	 ** should be interactively prompted to supply one.
	 **/
	public static class Defaults {

		/** Default user interface. 
		 ** Example: <CODE>"basic"</CODE> or <CODE>null</CODE>
		 ** Currently <CODE>"basic"</CODE>
		 **/
		public static final String USER_INTERFACE = "basic";

		/** Default archive name. 
		 ** Example: <CODE>"MyArchive.jar"</CODE> or <CODE>null</CODE>
		 ** Currently <CODE>null</CODE>
		 **/
		public static final File ARCHIVE = null;

		/** Default log level.
		 ** Example: anything between <CODE>-1</CODE> and <CODE>Settings.Constants.MAXIMUM_LOG_LEVEL</CODE>
		 ** Currently <CODE>-1</CODE> (i.e. no logging at all)
		 **/
		public static final int LOG_LEVEL = -1;

		/** Default cipher stack.
		 ** Example: <CODE>null</CODE> or <CODE>new Specification[] { new Specification("Rijndael", "default password") }</CODE>
		 ** or even <CODE>new Specification[] { new Specification("Rijndael") }</CODE>
		 **/
		public static Specification[] CIPHER_SPECIFICATIONS = null;

		/** Prefix of all log messages.
		 ** This is usefull for output post-processing (<CODE>grep</CODE> or alike)
		 ** Example: <CODE>"Log"</CODE> or <CODE>null</CODE>
		 ** Currently none.
		 **/
		public static final String LOG_PREFIX = null;  // e.g. , ">"

		/** Gets the default trace for specific type.
		 ** @param type a single-word description of the trace type, e.g. "log" or "mismatching_files"
		 ** In this implementation this always return <CODE>null</CODE>.
		 ** @see sam.rules.UnixLikeSyntaxSet#makeWritableTrace
		 **/
		public static final String TRACE(String type) { return true ? null : (((type == null) ? "trace" : type) + ".log"); }


		/**	Default block size <b>in kiloBytes</b>.
		 ** Example: anything from <CODE>0</CODE> (unlimited) to <CODE>Settings.Constants.MAXIMUM_BLOCK_SIZE</CODE>
		 ** Currently <CODE>64</CODE>
		 **/
		public static final int BLOCK_SIZE = 64;  // In kilo-bytes

		/**	Default packer.
		 ** Example: any valid <CODE>Packer.get("name")</CODE> or <CODE>null</CODE>
		 ** Currently <CODE>null</CODE>
		 **/
		public static final sam.io.Packer PACKER = sam.io.Packer.get("Huffman");

		/**	Default tracker.
		 ** Example: any valid <CODE>Tracker.get("name")</CODE> or <CODE>null</CODE>
		 ** Currently <CODE>null</CODE>
		 **/
		public static final sam.io.Tracker TRACKER = sam.io.Tracker.get("Checksum");

		/** Default number of vertical threads.
		 ** Example: anything above <CODE>1</CODE>
		 ** Currently <CODE>1</CODE>
		 **/
		public static final int VERTICAL_THREADS = 1;

		/** Default number of horizontal threads.
		 ** Example: anything above <CODE>1</CODE>
		 ** Currently <CODE>1</CODE>
		 **/
		public static final int HORIZONTAL_THREADS = 1;

		/** Default comment.
		 ** Example: <CODE>"This is a <a href=http://glacote.com/sam>SAM</a> archive"</CODE> or <CODE>null</CODE>
		 ** Currently <CODE>null</CODE>
		 **/
		public static final String COMMENT = null;

		/** Default file specification for to-be-archived files.
		 ** Example: <CODE>"."</CODE> (current directory), <CODE>"*.java"</CODE> or <CODE>null</CODE>
		 ** Currently <CODE>null</CODE>
		 **/
		public static final String TO_BE_ARCHIVED_FILES = ".";  // Default is to add current directory (wilcards <b>not</b> allowed here)

		/** Default file specification for to-be-extracted files.
		 ** Example: <CODE>"."</CODE> (current directory), <CODE>"*.java"</CODE> or <CODE>null</CODE>
		 ** Currently <CODE>null</CODE>
		 **/
		public static final String TO_BE_EXTRACTED_FILES = "*"; // Default is to extract all files

		/** Default <CODE>Jar</CODE> compression level
		 ** Example: anything from <CODE>0</CODE> (no compression) to <CODE>9</CODE> (best and slowest compression).
		 ** Currently <CODE>9</CODE>
		 **/
		public static final int JAR_COMPRESSION_LEVEL = 9;

		/** Default temporary directory
		 ** Might <b>not</b> be <CODE>null</CODE>
		 **/
		public static final File TEMPORARY_DIRECTORY = new File(File.separator + "tmp");
	}

	/** Placeholder class for forcing various <CODE>static</CODE> initializers as required by the program.
	 ** This is for example the case for static registration of cipher or user interface implementations.
	 **/
	public static class Startup {

		/** The user interface to be loaded at startup **/
		public static final String BOOTSTRAP_USER_INTERFACE = "basic";

		private static boolean isLoaded = false;

		/** Loads various classes (and thus executes their <CODE>static</CODE> initializers
		 ** @return wether the loading was successfull
		 **/
		public static boolean load() {
			if (isLoaded)
				return true;
			if (Constants.ALWAYS_STARTS_LOG)
				sam.ui.LogInterface.start();
			try {
				Object dummy = new Class[] {
						// User interface implementations
						sam.ui.basic.UserInterface.class, sam.ui.text.UserInterface.class, sam.ui.awt.UserInterface.class,
						// Cipher implementations
						XOR.class, DES.class, Deal.class, Twofish.class, RC6.class, Frog.class,
						Loki.class, Rijndael.class, Serpent.class, Mars.class, WarFactory.class
				};
			} catch (Throwable t) {
				UserInterface.get().error().handleException("Failed to load libraries", t);
				return false;
			}
			Environment.warnMicrosoftOperatingSystemOwners();
			isLoaded = true;
			return true;
		}

	}

	/** Placeholder class for holding various static finalizations. **/
	public static class Stop {

		/** Wether application aws already unloaded. **/
		private static boolean isUnloaded = false;

		/** Finalizes the application. **/
		public static void unload() {
			if (isUnloaded)
				return;
			try {
				if ( ! Constants.ALWAYS_STARTS_LOG )
					sam.ui.LogInterface.start();
				sam.ui.LogInterface.stop();
				UserInterface.disposeFrames();
				isUnloaded = true;
			} catch (Throwable t) {
				System.err.println("Failed to unload libraries: " + Text.getMessage(t) );
			}
		}

	}

}
