
package sam.core;

import sam.rules.RuleSet;
import sam.rules.ModusOperandi;
import sam.entry.Entry;
import sam.io.*;
import sam.cipher.Cipher;
import sam.toolkit.Text;
import sam.toolkit.Workshop;
import sam.ui.UserInterface;
import sam.event.*;
import sam.toolkit.Random;

import java.io.*;
import java.util.jar.JarOutputStream;


/** General-purpose class implementing the storing procedure.
 **/
public abstract class Storer extends ModusOperandi.Opus {

	protected Storer(String name) { super(name); }

	protected JarOutputStream process(ModusOperandi modusOperandi) {
		try {
			// Should report modusOperandi.removedFiles()
			final sam.entry.Enumerator toBeProcessedfiles = modusOperandi.setToBeProcessedFiles( modusOperandi.newerFiles() );
			if ( toBeProcessedfiles.isEmpty() ) {
				UserInterface.get().running().print("Empty archive", "No newer file, exiting");
				return null;
			}
			// Check OS
			final JarOutputStream archive = createArchive( modusOperandi.archive() );
			OutputStream output = open(modusOperandi, archive);
			final CloseAvoiderOutputStream filtered = new CloseAvoiderOutputStream(output);
			final Mixer mixer = modusOperandi.packer().getMixer();
			mixer.setBlockSize( modusOperandi.blockSize() );
			mixer.setTracker( modusOperandi.tracker() );
			output = mixer.new OutputStream(filtered);
			output = modusOperandi.notifier().progress().new PlainOutputStream(output);
			store(toBeProcessedfiles, output, modusOperandi.notifier() );  // Here the hard job
			close(output, modusOperandi.packer(), filtered);
			return archive;
		} catch (IOException e) {
			UserInterface.get().error().handleException("I/O error while storing files: ", e);
			return null;
		}
	}

	/** Creates the archive.
	 ** @param archive the archive file to be created of overwritten.
	 **/
	protected JarOutputStream createArchive(File archive) {
		try {
			UserInterface.get().log().log(1, Storer.class, "Creating Jar archive \"" + archive + "\"");
			return new JarOutputStream(new FileOutputStream(archive), Settings.Distribution.makeManifest() );
		} catch (IOException e) {
			UserInterface.get().error().fail("Fatal error while creating archive file \"" + archive + "\"", e);
			return null; // Fake for compiler
		}
	}

	/** Opens an entry of the archive to write plain data into.
	 ** This includes properly ciphering and packing data, and all this possibly in among pultiple threads.
	 ** <p><b>WARNING:</b> the returned <CODE>OutputStream</CODE> must be closed before any subsequent write
	 ** to the archive, in order for the <CODE>JarEntry</CODE> to be properly closed.
	 **/
	protected OutputStream open(ModusOperandi modusOperandi, JarOutputStream archive) throws IOException {
		UserInterface.get().log().log(2, Unstorer.class, "Initializing deciphering stack");
		OutputStream output = ArchiveLayout.DATA.put(archive);  // Opens the data Zip entry of the archive
		output = modusOperandi.notifier().progress().new EncodedOutputStream(output);  // Transparently reports written bytes
		ParallelIO parallelIO = null;
		if (modusOperandi.verticalThreads() > 1) {  // Initializes adequate number of vertical threads
			UserInterface.get().log().log(3, Unstorer.class, "Enabling " + modusOperandi.verticalThreads() + " vertical threads");
			parallelIO = new ParallelIO();
			output = parallelIO.new GatherOutputStream(output);
		}
		Cipher[] ciphers = modusOperandi.ciphers();
		for (int i=0; i<ciphers.length; ++i) {  // Initializes cascaded ciphering filters
			ByteArrayFilter filter = ciphers[i].new ByteArrayFilter();
			if (modusOperandi.horizontalThreads() > 1) {
				UserInterface.get().log().log(3, Unstorer.class, "Enabling " + modusOperandi.horizontalThreads() + " horizontal threads");
				filter = new MultiThreadedByteArrayFilter(filter, new Workshop(modusOperandi.horizontalThreads()),
																									Settings.Constants.MULTI_THREADED_BLOCK_SIZE);
			}
			output = filter.new OutputStream(Settings.Constants.FILTER_BLOCK_BUFFER_SIZE, output);
		}
		if (modusOperandi.verticalThreads() > 1)
			output = parallelIO.new ScatterOutputStream(output, new Workshop(modusOperandi.verticalThreads()));
		UserInterface.get().log().log(3, Unstorer.class, "Initializing " + modusOperandi.packer().name + " packer");
		output = modusOperandi.packer().filterOutput(output);  // Packs written bytes on-the-fly
		return output;
	}

	/** Closes the archive entry.
	 ** @param output the entry <CODE>OutputStream</CODE>
	 ** @param filtered the intermediate filtered stream
	 **/
	protected void close(OutputStream output, Packer packer, CloseAvoiderOutputStream filtered) throws IOException {
		UserInterface.get().log().log(3, Unstorer.class, "Finalizing archive");
		output.flush();  // Warning: recall that <CODE>ByteArrayFilter.OutputStream</CODE>s do <b>not</b> flush on <CODE>flush()</CODE>
		output.close();  // However this forces them to do it without <em>really</em> closing the base stream
		packer.finishFilter(filtered);
		filtered.reallyDoClose();  // This <em>really</em> does close the base stream
	}

	/** Stores specified files
	 ** @return the actual files catalogthe mixing trace
	 **/
	protected void store(sam.entry.Enumerator files, OutputStream output, EventNotifier eventNotifier) throws IOException {
		UserInterface.get().log().log(3, Unstorer.class, "Storing entries");
		while ( files.hasMoreElements() ) {
			Entry entry;
			try {
				entry = files.nextEntry();
			} catch (java.util.NoSuchElementException e) {
				UserInterface.get().error().handleException("Missing file since scanning directory.", e);
				continue;
			}
			eventNotifier.progress().notifyNextEntry(entry);
			eventNotifier.file().notifyAll(FileEventNotifier.PROCESSED, entry, "");
			java.io.InputStream input =	new BufferedInputStream(entry.getInputStream());
			try {
				UserInterface.get().log().log(4, Unstorer.class, name + " " + entry);
				storeEntry(input, output, entry);
			} catch (IOException e) {
				eventNotifier.file().notifyAll(FileEventNotifier.FAILED, entry, "I/O Error: " + e.getMessage() );
				UserInterface.get().error().handleException("I/O error while processing \"" + entry + "\"", e);
			} finally {
				input.close();
			}
		}
	}

	protected abstract void storeEntry(InputStream source, OutputStream archive, Entry entry) throws IOException ;

}
