
package sam.core;

import sam.rules.RuleSet;
import sam.rules.ModusOperandi;
import sam.entry.Entry;
import sam.io.*;
import sam.cipher.Cipher;
import sam.toolkit.Text;
import sam.toolkit.Workshop;
import sam.ui.UserInterface;
import sam.event.*;

import java.io.*;
import java.util.zip.CRC32;
import java.util.zip.ZipFile;
import java.util.jar.JarOutputStream;


/** General-purpose class implementing the unstoring procedure.
 **/
public abstract class Unstorer extends ModusOperandi.Opus {

	protected Unstorer(String name) { super(name); }

	protected JarOutputStream process(ModusOperandi modusOperandi) {
		try {
			if (modusOperandi.comment() != null)
				printComment( modusOperandi.comment() );
			if (modusOperandi.importedArchive() != null)
				UserInterface.get().running().print("This archive is part of a multiple set, you should extract "
																						+ modusOperandi.importedArchive() + " first");
			checkUserAndGroup(modusOperandi.creatorUser(), modusOperandi.creatorGroup(), Settings.Environment.USER, Settings.Environment.GROUP);
			final sam.entry.Enumerator toBeProcessedfiles = modusOperandi.setToBeProcessedFiles( modusOperandi.files() );
			InputStream cipheredData = openArchive(modusOperandi.archive(), modusOperandi.temporaryDirectory() );
			InputStream input = open(modusOperandi, cipheredData);
			Mixer mixer = modusOperandi.packer().getMixer();
			mixer.setBlockSize( modusOperandi.blockSize() );
			mixer.setTracker( modusOperandi.tracker() );
			unstore(toBeProcessedfiles, modusOperandi.wildcards(), modusOperandi.outputDirectory(), input, modusOperandi.notifier());
			input.close();
			return null;
		} catch (IOException e) {
			UserInterface.get().error().fail("I/O error while unstoring files: ", e);
			return null;
		}
	}

	/** Prints the comment, if any.
	 ** @param comment the <b>non-<CODE>null</CODE></b> comment
	 **/
	protected void printComment(String comment) {
		UserInterface.get().running().print("Informations", comment);
	}

	/** Checks user and group information.
	 ** @param creatorUser the system-dependent user who created the archive
	 ** @param creatorGroup the system-dependent group of the user who created the archive
	 ** @param currentUser the system-dependent user who is currently running the application
	 ** @param currentGroup the system-dependent group of the user who is currently running the application
	 **/
	protected void checkUserAndGroup(String creatorUser, String creatorGroup, String currentUser, String currentGroup) {
		if (Settings.Constants.WARN_DIFFERENT_GROUP && (! currentGroup.equalsIgnoreCase(creatorGroup)) )
			UserInterface.get().error().warn("You are in group \"" + currentGroup + "\" while this archive was created by a user of the group \""
																			 + creatorGroup + "\"");
		else
			if (Settings.Constants.WARN_DIFFERENT_USER && (! currentUser.equalsIgnoreCase(creatorUser)) )
				UserInterface.get().error().warn("You are \"" + currentUser + "\" while this archive was created by \"" + creatorUser + "\"");
	}

	/** Opens the encoded data from the archive.
	 ** @param archive the archive
	 ** @param temporaryDirectory a writable temporary directory in case of data pre-extraction
	 **/
	protected InputStream openArchive(File archive, File temporaryDirectory) {
		try {
			UserInterface.get().log().log(1, Storer.class, "Opening Jar archive \"" + archive + "\" for reading");
			InputStream data = ArchiveLayout.DATA.get( new ZipFile(archive.getPath()) );
			if (Settings.Constants.PREEXTRACT_DATA) {
				UserInterface.get().log().log(2, Unstorer.class, "Pre-extracting ciphered data; this might take a long time ...");
				data = IO.preextract(data, temporaryDirectory);
			}
			return data;
		} catch (IOException e) {
			UserInterface.get().error().fail("Fatal error while recovering main ciphered data from archive \"" + archive + "\"", e);
			return null; // Fake for compiler
		}
	}

	/** Opens the encoded data and turns it to a plain data stream
	 ** @param modusOperandi the modus operandi containing the application settings
	 ** @param input the encoded data
	 **/
	protected InputStream open(ModusOperandi modusOperandi, InputStream input) throws IOException {
		UserInterface.get().log().log(2, Unstorer.class, "Initializing deciphering stack");
		EventNotifier notifier = modusOperandi.notifier();
		input = notifier.progress().new EncodedInputStream(input);
		ParallelIO parallelIO = null;
		if (modusOperandi.verticalThreads() > 1) {
			UserInterface.get().log().log(3, Unstorer.class, "Enabling " + modusOperandi.verticalThreads() + " vertical threads");
			parallelIO = new ParallelIO();
			input = parallelIO.new GatherInputStream(input);
		}
		Cipher[] ciphers = modusOperandi.ciphers();
		for (int i=0; i<ciphers.length; i++) {
			ByteArrayFilter filter = ciphers[i].new ByteArrayFilter();
			if (modusOperandi.horizontalThreads() > 1) {
				UserInterface.get().log().log(3, Unstorer.class, "Enabling " + modusOperandi.horizontalThreads() + " horizontal threads");
				filter = new MultiThreadedByteArrayFilter(filter, new Workshop(modusOperandi.horizontalThreads()), 1024);
			}
			input = filter.new InputStream(Settings.Constants.FILTER_BLOCK_BUFFER_SIZE, input);
		}
		if (modusOperandi.verticalThreads() > 1)
			input = parallelIO.new ScatterInputStream(input, new Workshop(modusOperandi.verticalThreads() ));
		UserInterface.get().log().log(3, Unstorer.class, "Initializing " + modusOperandi.packer().name + " packer");
		input = modusOperandi.packer().filterInput(input);
		input = modusOperandi.packer().getMixer().new InputStream(input);
		input = notifier.progress().new PlainInputStream(input);
		return input;
	}

	/** Unstores specified entries.
	 ** @param files the catalog of all entries contained in the archive
	 ** @param wildcards the filename filter specifying which entries to extract
	 ** @param actualDirectory the <em>base</em> directory to extract files into
	 ** @param input the <b>decoded</b> data stream
	 ** @param eventNotifier the event notifier to handle events
	 **/
	protected void unstore(sam.entry.Enumerator files, sam.entry.List wildcards, File actualDirectory,
												 InputStream input, EventNotifier eventNotifier) throws IOException {
		UserInterface.get().log().log(3, Unstorer.class, "Unstoring entries");
		while ( files.hasMoreElements() ) {
			Entry entry;
			try {
				entry = files.nextEntry();
			} catch (java.util.NoSuchElementException e) {
				UserInterface.get().error().handleException("Missing file since scanning archive content.", e);
				continue;
			}
			if (wildcards.specifies(entry)) {
				if ( notifyProgress() )
					eventNotifier.progress().notifyNextEntry(entry);
				eventNotifier.file().notifyAll(FileEventNotifier.PROCESSED, entry, "");
				try {
					UserInterface.get().log().log(4, Unstorer.class, name + " " + entry);
					unstoreEntry(input, entry, actualDirectory, eventNotifier.file() );
				} catch (IOException e) {
					eventNotifier.file().notifyAll(FileEventNotifier.FAILED, entry, "I/O Error: " + e.getMessage() );
				} finally {
					input.close();
				}
			} else {
				eventNotifier.file().notifyAll(FileEventNotifier.SKIPPED, entry, "");
				try {
					UserInterface.get().log().log(4, Unstorer.class, "Skipping " + entry);
					IO.bufferedDump(input, IO.NULL_OUTPUT, entry.length);
				} catch (IOException e) {
					UserInterface.get().error().warn("I/O error while skipping file " + entry.name + ": " + e.getMessage() );
				} finally {
					input.close();
				}
			}
		}
	}

	/** Unstores an entry.
	 ** @param data the <b>decoded</b> data stream
	 ** @param entry the entry to be processed
	 ** @param actualDirectory the <em>base</em> directory to extract files into
	 ** @param eventNotifier the event notifier to handle events
	 **/
	protected abstract void unstoreEntry(InputStream data, Entry entry, File actualDirectory, FileEventNotifier fileEventNotifier)
		throws IOException;


	/** Wether to notify progress information.
	 ** This is overridden to return <CODE>false</CODE> by special unstorers that do
	 ** not want that, like during the catalog listing of an archive.
	 **/
	protected boolean notifyProgress() { return true; }

}
