
package sam.entry;

import java.util.StringTokenizer;


/** Comparator over <CODE>Entry</CODE> instances */
public class Comparator implements java.util.Comparator {

	/** Lexicographical comparator over <CODE>Entry</CODE> extensions, then name, then path */
	public static final Comparator LEXICOGRAPHICAL = new Comparator( java.text.Collator.getInstance() ).new Extension(":/.");

		private final java.util.Comparator pathComparator;

		public Comparator(java.util.Comparator pathComparator) {
			this.pathComparator = pathComparator;
		}

		protected int compare(String a, String b) {
			if ( a.equals(b) )
				return 0;
			return pathComparator.compare(a, b);
		}


		/** Note: this comparator imposes orderings that are inconsistent with equals, for which concerns length and date. **/
		public int compare(Object a, Object b) throws ClassCastException {
			if (a == null)
				return (b == null) ? 0 : -1;
			if (b == null)
				return (a == null) ? 0 : +1;
			if ( a.equals(b) )
				return 0;
			return compare( ((Entry) a).name, ((Entry) b).name );
		}



		/** This class implements a reverse lexicographic order on tokenized named. **/
		public class Tokenized extends Comparator {
			final String separators;

			public Tokenized(String separators) {
				super(Comparator.this);
				this.separators = separators;
			}

			protected int compare(String a, String b) {
				String[] x = getTokens(a), y = getTokens(b);
				return compare(x, x.length - 1, y, y.length - 1);
			}

			private int compare(String[] a, int x, String[] b, int y) {
				if (x < 0)
					return (y < 0) ? 0 : -1;  // is 'b' also exhausted ?
				if (y < 0)
					return +1;  // 'a' has more tokens
				int comparison = Comparator.this.compare(a[x], b[y]);
				return (comparison != 0) ? comparison : compare(a, x-1, b, y-1);
			}

			String[] getTokens(String a) {
				StringTokenizer tokenizer = new StringTokenizer(a, separators);
				String[] tokens = new String[ tokenizer.countTokens() ];
				for (int i=0; i<tokens.length; i++)
					tokens[i] = tokenizer.nextToken();
				return tokens;
			}

		}


		/** This class implements an extension/name/path ordering. **/
		public class Extension extends Tokenized {

			public Extension(String separators) {
				super(separators);
			}

			String[] getTokens(String a) {
				String[] tokens = new String[ separators.length() + 1];
				for (int i = tokens.length - 1; i > 0; i--) {
					int cut = a.lastIndexOf( separators.charAt(i-1) );
					if (cut >= 0) {
						tokens[i] = a.substring(cut + 1);
						a = a.substring(0, cut);
					} else
						tokens[i] = "";
				}
				tokens[0] = a;
				return tokens;
			}

		}


}
