
package sam.entry;

import sam.toolkit.Text;
import sam.io.IO;
import sam.event.FileEventNotifier;
import sam.ui.UserInterface;
import sam.core.Settings;
import sam.core.ArchiveLayout;
import sam.rules.ModusOperandi;

import java.io.*;
import java.util.*;


/** This class is a basic abstraction over files for the purpose of archiving.
 ** <p>Instances of this class are fully immutable.
 ** Note also that this class has <b>no public constructor</b> as it should only
 ** be instantiated through an <CODE>Enumerator</CODE> sub-class.
 ** @see Enumerator
 ** @see FileSystem
 **/
public class Entry {

	/** Dummy entry for testing/debugging purposes **/
	public static final Entry DUMMY = new Entry("dummy");

	private final File directory;

	/** Actual total number of bytes contained by this entry. */
	public final long length;

	/** Actual name of this entry.
	 ** The patch separator is specified in {@link sam.core.ArchiveLayout#ENTRY_PATH_SEPARATOR}
	 ** <b>Warning:</b> native <CODE>File</CODE>s should <b>not</b> be created with this field
	 ** as a parameter but through the <CODE>toFile</CODE> method instead.
	 **/
	public final String name;

	/** Actual last access time of this entry. */
	public final long date;


	/** Makes a new <CODE>Entry</CODE> from specified existing file */
	/* Used by List(String[]) */
	Entry(String fileName) {
		this(-1, fileName, null, System.currentTimeMillis() );
	}

	/** Makes a new <CODE>Entry</CODE> from specified existing file in specified directory.
	 ** <p>The <CODE>directory</CODE> is removed from actual name.
	 **/
	Entry(File file, File directory) throws IOException {
		this(file.length(), (file == null) ? null : file.getAbsolutePath(), directory, file.lastModified() ); // mangleFileName(file, directory)
	}

	private Entry(long length, String name, long date) {
		this(length, name, null, date);
	}

	private Entry(long length, String name, File directory, long date) {
		this.length = (length < 0) ? 0 : length;
		this.name = mangleFileName(name, directory); // Text.replace(name.replace('\\','/'), "./", "");  // For compatibility with Win32
		this.directory = directory;
		this.date = date;
	}

	/** Lists this entry in Unix-like format. **/
	public String list(ModusOperandi modusOperandi) {
		return "-r--r--r--\t1 " + modusOperandi.creatorUser() + " " + modusOperandi.creatorGroup()
				+ " " + Text.right("" + length, Settings.Constants.ENTRY_LENGTH_DIGITS) + " " + Text.getDateTime(date) + " " + name;
	}

	/** Returns a human-readable description of this entry
	 **/
	public String toString() {
		return name;
	}

	/**
	   Makes a verbose description of this entry including its name and size.
	   <p><CODE>"./foo" (25kb)</CODE>
	 */
	public String verbose() {
		return "\"" + name + "\" (" + Text.getSizeString(length) + ")";
	}

	/** Returns an <CODE>InputStream</CODE> for reading this entry, assumed it is a local file.
	 **/
	public InputStream getInputStream() throws IOException {
		return new FileInputStream( toFile() );
	}

	/**
	   Wether this entry denotes a file which is older than an already existing file ins specified directory.
	   @param outputDirectory the output directory to place the file in.
	   @return wether target file already exists an is newer than this entry.
	*/
	public boolean isOlderThanIn(File outputDirectory) {
		File existing = toFile(outputDirectory);
		if ( ! existing.exists() )
			return false;
		return date <= existing.lastModified();
	}

	/**
	   Restores an <CODE>Entry</CODE> instance.
	   @return <CODE>null</CODE> if no more entry can be read from the stream.
	*/
	static Entry read(DataInputStream input) throws IOException {
		if (input.available() <= 0)
			return null;
		try {
			long length = input.readLong();
			long date = input.readLong();
			String name = input.readUTF();
			return new Entry(length, name, date);
		} catch (EOFException e) {
			return null;
 			// throw new IOException("Corrupted catalog: " + Text.getMessage(e) );  // partial length/date/name
		}
 	}


	/**
	   Stores this <CODE>Entry</CODE> instance.
	*/
	void write(DataOutputStream output) throws IOException {
		output.writeLong(length);
		output.writeLong(date);
		output.writeUTF(name);
	}

	/** Mangles a native file name to an <COD>Entry</CODE> path.
	 ** @param name the native file name
	 ** @param basePath the base path, or <CODE>null</CODE>
	 ** @see ArchiveLayout.ENTRY_PATH_SEPARATOR
	 **/
	protected static String mangleFileName(String fileName, File basePath) {
		String name = (basePath == null) ? fileName : removeDirectory(fileName, basePath.getAbsolutePath() + File.separatorChar );
		return name.replace('\\', ArchiveLayout.ENTRY_PATH_SEPARATOR);  // For Win32 compatibility
	}

	/** Unmangles this <COD>Entry</CODE> path to native file name.
	 ** @see ArchiveLayout.ENTRY_PATH_SEPARATOR
	 **/
	public File toFile() { return toFile(directory); }

	/** Unmangles this <COD>Entry</CODE> path to native file name.
	 ** @param basePath the base path, or <CODE>null</CODE>
	 ** @see ArchiveLayout.ENTRY_PATH_SEPARATOR
	 **/
	public File toFile(File basePath) {
		String fileName = name.replace(ArchiveLayout.ENTRY_PATH_SEPARATOR, File.separatorChar);
		return (basePath == null) ? new File(fileName) : new File(basePath, fileName);
	}

	static String removeDirectory(File name, File path) {
		return (path == null) ? name.getPath() :
			removeDirectory(name.getAbsolutePath(), path.getAbsolutePath() + File.separatorChar );
	}

	private static String removeDirectory(String name, String path) {
		return name.startsWith(path) ? name.substring( path.length() ) : name;
	}


	/** Gets an <CODE>InputStream</CODE> ready for reading the file corresponding the this entry.
	 ** This entry is supposed to either have an absolute path or to designate a path relative
	 ** to the <em>current</em> working directory.
	 ** @param notifier the <CODE>FileEvent</CODE> notifier to report events through
	 **/
	public InputStream read(final FileEventNotifier notifier) {
		notifier.notifyAll(FileEventNotifier.PROCESSED, Entry.this, "read");
		final File localFile = toFile();
		try {
			final InputStream input = new FileInputStream(localFile);
			return new InputStream() {

			public int read() throws EOFException {
				try {
					return input.read();
				} catch (EOFException e) {
					throw e;
				} catch (IOException e) {
					fail(notifier, "read", localFile, e);
					return -1;
				}
			}

			public int read(byte[] data, int begin, int length) throws EOFException {
				if (length <= 0)
					return 0;
				try {
					return input.read(data, begin, length);
				} catch (EOFException e) {
					throw e;
				} catch (IOException e) {
					fail(notifier, "read", localFile, e);
					return 0;
				}
			}

			public long skip(long length) throws EOFException {
				if (length <= 0)
					return 0;
				try {
					return input.skip(length);
				} catch (EOFException e) {
					throw e;
				} catch (IOException e) {
					fail(notifier, "skip " + length + " bytes from", localFile, e);
					return 0;
				}
			}

			public void close() throws EOFException {
				try {
					input.close();
				} catch (EOFException e) {
					throw e;
				} catch (IOException e) {
					fail(notifier, "close", localFile, e);
				}
			}

		};
		} catch (FileNotFoundException e) {
			fail(notifier, "find", localFile, e);
			return null;
		}
	}

	/** Gets an <CODE>OutputStream</CODE> ready for writing this entry into specific directory.
	 ** If the target file already exists, this will ask the user wether it should be overwritten.
	 ** @param notifier the <CODE>FileEvent</CODE> notifier to report events through
	 ** @param actualDirectory the <em>base</em> directory to extract the file into (relative path
	 ** is preserved).
	 **/
	public OutputStream write(final FileEventNotifier notifier, File actualDirectory) {
		if (Settings.Constants.ASK_BEFORE_OVERWRITING_LOCAL_FILES) {
			File localFile = toFile(actualDirectory);
			if (localFile.exists() &&	( ! UserInterface.get().running().yesNo("File " + localFile + " already exists, overwrite ?") ) )
				return skip(notifier, actualDirectory);
		}
		return overwrite(notifier, actualDirectory);
	}

	/** Gets an <CODE>OutputStream</CODE> ready for freshening this entry into specific directory.
	 ** If the target file already exists <b>and is older</b> then it will be overwritten.
	 ** @param notifier the <CODE>FileEvent</CODE> notifier to report events through
	 ** @param actualDirectory the <em>base</em> directory to extract the file into (relative path
	 ** is preserved).
	 **/

	public OutputStream freshen(final FileEventNotifier notifier, File actualDirectory) {
		return isOlderThanIn(actualDirectory) ? skip(notifier, actualDirectory) : overwrite(notifier, actualDirectory);
	}

	/** Gets an <CODE>OutputStream</CODE> ready for <em>skipping</em> this entry.
	 ** @param notifier the <CODE>FileEvent</CODE> notifier to report events through
	 ** @param actualDirectory the <em>base</em> directory to extract the file into (relative path
	 ** is preserved).
	 **/
	public OutputStream skip(final FileEventNotifier notifier, File actualDirectory) {
		notifier.notifyAll(FileEventNotifier.PROCESSED, Entry.this, "skip");
		notifier.notifyAll(FileEventNotifier.SKIPPED, Entry.this, "skip");
		return new OutputStream() {
			public void write(int value) throws IOException {	}
			public void write(byte[] data, int begin, int length) throws IOException { }
			public void flush() throws IOException { }
			public void close() throws IOException { }
		};
	}

	/** Gets an <CODE>OutputStream</CODE> ready for writing this entry into specific directory.
	 ** The target file is overwritten if it already exists.
	 ** @param notifier the <CODE>FileEvent</CODE> notifier to report events through
	 ** @param actualDirectory the <em>base</em> directory to extract the file into (relative path
	 ** is preserved).
	 **/
	public OutputStream overwrite(final FileEventNotifier notifier, File actualDirectory) {
		notifier.notifyAll(FileEventNotifier.PROCESSED, Entry.this, "overwrite");
		final File localFile = toFile(actualDirectory);
		if ( localFile.exists() )
			notifier.notifyAll(FileEventNotifier.OVERWRITTEN, Entry.this, "overwrite");
		try {
			IO.makePathIfRequired(localFile);
			final OutputStream output = new FileOutputStream(localFile);
			return new OutputStream() {

			public void write(int value) throws IOException {
				try {
					output.write(value);
				} catch (IOException e) {
					fail(notifier, "write", localFile, e);
				}
			}

			public void write(byte[] data, int begin, int length) throws IOException {
				if (length <= 0)
					return;
				try {
					output.write(data, begin, length);
				} catch (IOException e) {
					fail(notifier, "write", localFile, e);
				}
			}

			public void flush() throws IOException {
				try {
					output.flush();
				} catch (IOException e) {
					fail(notifier, "flush", localFile, e);
				}
			}

			public void close() throws IOException {
				try {
					output.close();
				} catch (IOException e) {
					fail(notifier, "close", localFile, e);
				}
			}

		};
		} catch (FileNotFoundException e) {
			fail(notifier, "find", localFile, e);
			return null;
		} catch (IOException e) {
			fail(notifier, "create path for ", localFile, e);
			return null;
		}
	}

	/** Gets an <CODE>OutputStream</CODE> ready for comparing this entry to the file stored into specific directory.
	 ** @param notifier the <CODE>FileEvent</CODE> notifier to report events through
	 ** @param actualDirectory the <em>base</em> directory where the file to compared against is stored (relative path
	 ** is preserved).
	 **/
	public OutputStream compare(final FileEventNotifier notifier, final File actualDirectory) {
		notifier.notifyAll(FileEventNotifier.PROCESSED, Entry.this, "compare");
		final File localFile = toFile(actualDirectory);
		if ( ! localFile.exists() )
			fail(notifier, "read", localFile, new IOException("Local file does not exist") );
		try {
			final InputStream input = new FileInputStream(localFile);
			return new OutputStream() {

				public void write(int value) throws IOException {
					int source = -1;
					try {
						source = input.read();
					} catch (IOException e) {
						fail(notifier, "read", localFile, e);
					}
					if (source < 0)
						fail(notifier, "read enough bytes from", localFile, null);
					if (source != (value & 0xff))  // '0xff' to fix what seems to be a bug ...
						notifier.notifyAll(FileEventNotifier.MISMATCHING, Entry.this, "Archived file does not match local file " + localFile +
															 ": 0x" + Integer.toHexString(source) + ", 0x" + Integer.toHexString(value & 0xff) );
				}

				public void write(byte[] data, int begin, int length) throws IOException {
					if (length <= 0)
						return;
					byte[] localData = new byte[length];
					try {
						int numberOfReadBytes = input.read(localData, 0, length);
						if (numberOfReadBytes < length)
							fail(notifier, "read " + length + " bytes from", localFile, null);
						if ( ! sam.toolkit.Text.compare(data, begin, length, localData, 0, length) )
							notifier.notifyAll(FileEventNotifier.MISMATCHING, Entry.this, "Archived file does not match local file " + localFile);
					} catch (IOException e) {
						fail(notifier, "read", localFile, e);
					}
				}

				public void flush() throws IOException {	}
							
				public void close() throws IOException {
						try {
							input.close();
						} catch (IOException e) {
							fail(notifier, "close", localFile, e);
						}
				}

			};
		} catch (FileNotFoundException e) {
			fail(notifier, "find", localFile, e);
			return null;
		}
	}

	/** Notifies a failure.
	 ** This first dispatches a new <CODE>FileEventNotifier.FAILED</CODE> event asynchronously,
	 ** and interactively notifies the user of the failure.
	 ** Note that this might choose <b>not to kill</b> the application according to the
	 ** <CODE>sam.ui.RulingInterface.handleException</CODE> semantics.
	 ** @param notifier the <CODE>FileEvent</CODE> notifier to report the failure through
	 ** @param type the type of failure, e.g. <CODE>"read"</CODE> or <CODE>"write"</CODE>
	 ** @param file an optinnal local file involved in the failure, or <CODE>null</CODE>
	 ** @param throwable an optionnal <CODE>Throwable</CODE> which caused the failure, or <CODE>null</CODE>
	 **/
	private void fail(FileEventNotifier notifier, String type, File file, Throwable throwable) {
		String message = "Could not " + ((type == null) ? "process" : type) + " file";
		if (file != null)
			message += " (" + file.getPath() + ")";
		notifier.notifyAll(FileEventNotifier.FAILED, Entry.this, message + ": " + ((throwable == null) ? "" : throwable.getMessage()) );
		UserInterface.get().error().handleException(message, throwable);
	}

}
