
package sam.entry;

import sam.ui.UserInterface;

import java.util.*;
import java.io.*;


public abstract class Enumerator implements Enumeration {

	/** Empty enumerator **/
	public static final Enumerator EMPTY = new Enumerator() {
		public Entry nextEntry() throws NoSuchElementException { return null; }
		public boolean hasMoreElements() { return false; }
	};


	/** Gets next entry, or <CODE>null</CODE> if none */
	public abstract Entry nextEntry() throws NoSuchElementException;

	public Object nextElement() throws NoSuchElementException {
		Object next = nextEntry();
		if (next == null)
			throw new NoSuchElementException("(next element was null)");
		else
			return next;
	}

	/** Wether this <CODE>Enumerator</CODE> has more elements **/
	public abstract boolean hasMoreElements();

	/** Wether this <CODE>Enumerator</CODE> has no more elements **/
	public boolean isEmpty() { return ! hasMoreElements(); }

		/**
		   Lists entries names
		 */
	    public void println(PrintWriter out) throws IOException {
			while ( hasMoreElements() )
				out.println( nextEntry() );
	    }

		/**
		   Lists entries names
		 */
	    public void println(PrintStream out) throws IOException {
			while ( hasMoreElements() )
				out.println( nextEntry() );
	    }

		/**
		   Discards unspecified entries from current enumeration.
		   <p>Default filter is <CODE>NameFilter.WILDCARDS</CODE>
		 */
		public Enumerator filter(Enumerator specifications) {
			return filter(specifications, NameFilter.WILDCARDS);
		}

		/**
		   Discards unspecified entries from current enumeration.
		   @param specifications entry names specifications.
		   @param filter method for selecting an entry name.
		 */
		public Enumerator filter(Enumerator specifications, final NameFilter filter) {
			final List specificationsList = specifications.statify();
			return new Reload() {
					protected Entry reload() {
						Entry next = null;
						while ( ((next = Enumerator.this.nextEntry()) != null) &&
								( ! specificationsList.specifies(next, filter) ) )
								;
						return next;
					}
				};
		}



		/**
		   Stores remaining entries
		*/
	    public void store(OutputStream output) throws IOException {
			while ( hasMoreElements() )
				nextEntry().write( new DataOutputStream(output) );

	    }


		/**
		   Restores an <CODE>Enumerator</CODE> and statifies it
		*/
		public static List load(InputStream input) {
			return new StreamBased(input).statify();
		}


		/**
		   Builds a static list from remaining entries
		   <p>Missing (<CODE>null</CODE>) entries are ignored until <CODE>hasMoreElements</CODE> returns <CODE>null</CODE>
		*/
		public List statify() {
			if (this instanceof List)
				return (List) this;
			Vector buffer = new Vector();
			long skipped = 0;
			while ( hasMoreElements() )
				try {
					Entry next = nextEntry(); 
					if (next != null)
						buffer.addElement(next);
					else
						skipped++;
				} catch (NoSuchElementException e) {
					UserInterface.get().log().log(2, Enumerator.class, "Could not access next entry: " + e.getMessage() );
				}
			if (skipped > 0)
				UserInterface.get().log().log(3, Enumerator.class, "Skipped " + skipped + " entries, kept " + buffer.size() );
			Entry[] list = new Entry[ buffer.size() ];
			buffer.copyInto(list);
			return new List(list);
		}

}



/** This class is for easily coding an <CODE>Enumerator</CODE>: simply have a <CODE>Entry reload()</CODE>
 **   method which returns <CODE>null</CODE> when exhausted.
 **/
abstract class Reload extends Enumerator {
	private Entry next = null;
		
	public boolean hasMoreElements() {
		if (next == null)
			next = reload();
		return next != null;
	}
			
	public Entry nextEntry() {
		if (next == null)
			next = reload();
		Entry result = next;
		next = null;
		return result;
	}

	/** Actually loads the next entry from internal resource **/			
	protected abstract Entry reload();

}


/** Enumerator which loads its content from an input-stream **/
class StreamBased extends Reload {
	private final DataInputStream input;

	public StreamBased(InputStream input) { this.input = new DataInputStream(input); }

	protected Entry reload() {  // throws NoSuchElementException {
		try {
			return Entry.read(input);  // 'null' if exhausted
		} catch (IOException e) {
			return null;
		}
	}

}

