
package sam.entry;

import sam.ui.UserInterface;

import java.io.File;
import java.io.IOException;


/** Enumerates local files and recurse sub-directories **/
public class FileSystem extends Enumerator {

	/** Base directory from which to scan files **/
	private java.io.File baseDirectory;

	/** Prefix to be removed from file names when listing (most often <tt>baseDirectory</tt> itself) **/
	private java.io.File excludeDirectory;

	/** Statically listed files in the directory, never null **/
	private java.io.File[] files;

	/** Marker of the next-to-be-listed file **/
	private int next;

	/** Wether this instance was built recursively by another one **/
	private Enumerator recursed;

	/** Wether to skip directories in file enumeration **/
	private final boolean skipDirectories;

	/** Scans for file names specifications (wildcards allowed).
	 ** Scans all files matching <CODE>specifications</CODE> recursively from <CODE>directory</CODE>.
	 ** @see NameFilter.WILDCARDS
	 ** Moreover the file-system scanning is lazy until a call to <CODE>nextEntry()</CODE> or to <CODE>statify()</CODE>
	 **/
	public static Enumerator scan(File directory, String[] specifications) {
		return new FileSystem(directory, false, false).filter( new List(specifications).removeDuplicates(NameFilter.WILDCARDS) );
	}

	/** Enumerates specified files.
	 ** Each member of <CODE>specifications</CODE> must be either a proper file name or a proper readable directory.
	 ** In particular wildcards are <b>not</b> allowed here.
	 ** Moreover the file-system scanning is lazy until a call to <CODE>nextEntry()</CODE> or to <CODE>statify()</CODE>
	 **/
	public static Enumerator scan(String[] specifications) {
		if (specifications == null)
			return null;
		Enumerator enumerator = Enumerator.EMPTY;
		for (int i=0; i<specifications.length; ++i)
			enumerator = new Multiple(new FileSystem(specifications[i]), enumerator);
		return enumerator;
	}

	/** Builds a new file system enumerator which enumerates files in specified directory
	 ** <p>Note that directories are not listed themselves
	 **/
	public FileSystem(String baseDirectory) {	this( new java.io.File(baseDirectory) ); }

	/** Builds a new file system enumerator which enumerates files in specified directory
	 ** <p>Note that directories are not listed themselves
	 **/
	public FileSystem(java.io.File baseDirectory) {	this(baseDirectory, false);	}

	/** Builds a new file system enumerator which enumerates files in specified directory
	 ** @param listDirectories wether directories are to be listed themselves
	 **/
	public FileSystem(java.io.File baseDirectory, boolean listDirectories) { this(baseDirectory, listDirectories, false); }  // true;

	/** Builds a new file system enumerator.
	 ** @param baseDirectory the single file or directory to scan recursively.
	 ** @param listDirectories wether or not to include directories themselves in the enumeration.
	 ** @param keepDirectoryName wether or not to keep the base directory name in enumerated names
	 **/
	public FileSystem(java.io.File baseDirectory, boolean listDirectories, boolean keepDirectoryName) {
		this(baseDirectory, listDirectories, keepDirectoryName ? null : baseDirectory);
	}

	/** Builds a new file system enumerator.
	 ** @param baseDirectory the single file or directory to scan recursively.
	 ** @param listDirectories wether or not to include directories themselves in the enumeration.
	 ** @param excludeDirectory the directory part to be removed from the enumeration
	 **/
	private FileSystem(java.io.File baseDirectory, boolean listDirectories, java.io.File excludeDirectory) {
		this.baseDirectory = baseDirectory;
		this.skipDirectories = ! listDirectories;
		this.excludeDirectory = excludeDirectory;
		this.files = baseDirectory.isDirectory() ? baseDirectory.listFiles() : new File[] { baseDirectory };
		if (this.files == null) {
			UserInterface.get().log().log(3, FileSystem.class, "Directory listing of " + baseDirectory + " denied, skipped");
			this.files = new java.io.File[0]; // This directory is set to empty
		}
		this.next = 0;
		this.recursed = null;
	}

	public Entry nextEntry() {
		if (recursed != null)  // Listing files in directory 'files[next]'
			if ( recursed.hasMoreElements() )
				return recursed.nextEntry();
			else {  // Sub-directory exhausted, returning to current one
				recursed = null;
				if (skipDirectories)
					next++;
					// Otherwise, the next item that will be displayed is precisely 'files[next]'
			}
		// Listing next file
		if (next < files.length) {  // Some more files in current dir
			if ( files[next].isDirectory() ) {
				recursed = new FileSystem(files[next], ! skipDirectories, excludeDirectory);  // Recursive build
				return nextEntry();  // and call this method again, this time with a prepared <tt>recursed</tt> member
			} else  // Next file is not a directory
				try {
					return new Entry(files[ next++ ], excludeDirectory);
				} catch (IOException e) {
					UserInterface.get().log().log(3, FileSystem.class, "Can not acces " + files[next-1] + ", file skipped: " + e.getMessage()  );
					return nextEntry();
				}
		} else
				return null;  // Should hopefully not happen as long as <tt>hasMoreElements()</tt> was called before
	}

	public boolean hasMoreElements() {
		return (recursed != null)	? recursed.hasMoreElements() ? true : (next + (skipDirectories ? 1 : 0) < files.length)
				: (next < files.length);  // No need to list <tt>recursed</tt> files, since there is no <tt>recursed</tt>
	}

	public String toString() { return Entry.removeDirectory(baseDirectory, excludeDirectory);	}

}
