
package sam.entry;

import sam.toolkit.Text;

import java.util.*;
import java.io.*;


/** Static enumerator.
 ** <p>Can be reset.
 ** @see reset
 **/
public class List extends Enumerator {
	private final Entry[] entries;
	private int index;

	/** Make a list from file names list read from an input stream. */
	public List(InputStream input) throws IOException {
		this( Text.splitLines( sam.io.IO.getFromStream(input) ) );
	}

	/** Make a list from file names list. */
	public List(String[] names) {	this( makeEntries(names) );	}

	/** Makes a list from specified entries array. */
	public List(Entry[] entries) {
		this.entries = entries;
		reset();
	}

	/** Makes a list from specified entries array portion. */
	protected List(Entry[] entries, int start, int end) {	this( getEntries(entries, start, end) ); }

	public synchronized Entry nextEntry() {
		if (index >= entries.length)
			return null;
		return entries[ index++ ];
	}

	public synchronized boolean hasMoreElements() {
		return (index < entries.length);
	}

	/** Computes the total size of all entries of this enumeration.	**/
	public long size() {
		long size = 0;
		for (int i=0; i<entries.length; i++)
			size += entries[i].length;
		return size;
	}

	/** Counts the total number of remaining entries in this enumeration. **/
	public int count() {
		return entries.length - index;
	}

	/** Check all entries of this enumeration for reading. **/
	public void checkRead() throws IOException {
		for (int i=0; i<entries.length; i++)
			if ( ! entries[i].toFile().canRead() )
				throw new IOException("Can not read " + entries[i]);
	}

	public String toString() {
		return (entries.length > 10) ? (entries.length + " files") : Text.toString(entries, "", ", ", "");
	}

	/** Restarts the enumeration from the beginning	**/
	public synchronized void reset() {
		index = 0;
	}

	/** Tests wether all included entries are of length 0, or if there is no more entry at all.	**/
	public synchronized boolean isEmpty() {
		for (int i=index; i<entries.length; i++)
			if (entries[i].length > 0)
				return false;
		return true;
	}

	/** Keeps only entries from this enumeration that are newer than in specified enumeration. **/
	public Enumerator keepNewerThan(Enumerator others) {
		return new Diff(this, others, true);
	}


	/** Excludes specified entries from current list.	**/
	public Enumerator exclude(Enumerator others) {
		return new Diff(this, others, false);
	}


	/** Returns the enumeration of all entries appearing in <CODE>other</CODE> but not in <CODE>this</CODE>. **/
	public Enumerator getRemovedSince(Enumerator others) {
		return new Diff(others, this, false);
	}


	/** Appends all newer entries to this enumeration. **/
	public Enumerator appendNewerFrom(List others) {
		return new Multiple(this, others.keepNewerThan(this) );
	}


	/** Sorts this enumeration in lexicographical order.
	 ** @return <CODE>this</CODE> (<b>Beware !</b>)
	 **/
	public List sort() {
		return sort(Comparator.LEXICOGRAPHICAL);
	}

	/** Sorts this enumeration.
	 ** @return <CODE>this</CODE> (<b>Beware !</b>)
	 **/
	public List sort(java.util.Comparator comparator) {
		Arrays.sort(entries, index, entries.length, comparator);
		return this;
	}

	public List sortAndRemoveDuplicates() {
		return sortAndRemoveDuplicates(Comparator.LEXICOGRAPHICAL);
	}

	/** Sorts and remove duplicates in a <b>sorted</b> files list.
	 ** <p>The complexity is that of sorting, i.e. <CODE>n*ln(n)</CODE>.
	 ** @return a new <CODE>List</CODE> instance with duplicates removed.
	 **/
	public List sortAndRemoveDuplicates(java.util.Comparator comparator) {
		return new List(entries, index, entries.length).sort(comparator).removeDuplicates(comparator);
	}

	/** Removes duplicate entries from a <b>sorted</b> files list.
	 ** <p>The complexity is <b>linear</b>, because any entry has to be compared to
	 ** the only adjacent one for equality.
	 ** @param comparator the files comparator to be used for ordering and equality testing.
	 ** @return a new <CODE>List</CODE> instance with duplicates removed.
	 ** @see NameFilter
	 ** @throws IllegalArgumentException if this list is <b>not</b> sorted.
	 **/
	protected List removeDuplicates(java.util.Comparator comparator) {
		if (index >= entries.length)
			return new List( new Entry[0] );
		Vector buffer = new Vector(entries.length - index);
		int ordering = 0;
		buffer.addElement( entries[index] );
		for (int i=index + 1; i<entries.length; i++) {
			int comparison = comparator.compare(entries[i-1], entries[i]);
			if (comparison != 0) {
				if (comparison != ordering) {  // Local ordering not similar to previous order
					if (ordering == 0)  // Just because global ordering was not set (iteration 'i=index+1')
						ordering = comparison;
					else  // List was NOT ordered !
						throw new IllegalArgumentException("Can not remove duplicates in not ordered list");
				}
				buffer.addElement(entries[i]);
			}
		}
		Entry[] newEntries = new Entry[ buffer.size() ];
		buffer.copyInto(newEntries);
		return new List(newEntries);
	}

	/** Removes duplicates file specification.
	 ** <p>Reduces <CODE>{ "a*", "?", "*" }</CODE> to <CODE>{ "*", "?" }</CODE>
	 ** <p>The complexity is <b>quadratic</b>, because any entry has to be compared to
	 ** all others.
	 ** @param filter the name filter to be used. Default is <CODE>NameFilter.BASIC</CODE>
	 ** @return a new <CODE>List</CODE> instance with duplicates removed.
	 ** @see NameFilter
	 **/
	public List removeDuplicates(final NameFilter filter) {
		Entry[] buffer = getEntries(entries, index, entries.length);
		if (buffer.length == 0)
			return new List(buffer);
		// First move less restrictive filters to the beginning
		Arrays.sort(buffer, 0, buffer.length, filter.new Comparator() );
		// Then add only less restrictive ones
		int count = 1;
		for (int i=count; i<buffer.length; ) {
			Entry next = buffer[i++];
			if ( ! doesSpecify(next, filter, buffer, 0, count) )
				buffer[count++] = next;
		}
		return new List(buffer, 0, count);
	}

	/** Tests this regular expression against specified entry
	 ** @return <CODE>true</CODE> if, and only if at least one of these entries is a regular expression
	 ** that expands to <CODE>entry</CODE>
	 **/
	public boolean specifies(Entry entry) {
		return specifies(entry, NameFilter.WILDCARDS);
	}

	/** Tests this regular expression against specified entry
	 ** @return <CODE>true</CODE> if, and only if at least one of these entries is a regular expression
	 ** that expands to <CODE>entry</CODE>
	 ** @param filter the name filtering rule to be used
	 **/
	public boolean specifies(Entry entry, NameFilter filter) {
		return doesSpecify(entry, filter, entries, index, entries.length);
	}

	private static boolean doesSpecify(Entry entry, NameFilter filter, Entry[] entries, int start, int end) {
		for (int i=start; i<end; )
			if ( filter.specifies(entries[i++].name, entry.name) )
				return true;
		return false;
	}
				


	/** Turns an array of <CODE>String</CODE> into an array of <CODE>Entry</CODE> **/
	private static Entry[] makeEntries(String[] names) {
		Entry[] entries = new Entry[ names.length ];
		for (int i=0; i<entries.length; i++)
			entries[i] = new Entry(names[i]);
		return entries;
	}

	/** Gets a copy of a portion of this enumerator **/
	private static Entry[] getEntries(Entry[] source, int start, int end) {
		if (start >= end)
			return new Entry[0];
		Entry[] newEntries = new Entry[end - start];
		System.arraycopy(source, start, newEntries, 0, end - start);
		return newEntries;
	}


}



/** This class enumerates the differences between two lists. **/
class Diff extends Reload {
	private final List newers, olders;
	private final boolean checkDate;
	private Entry limit;

	/** Any valid comparator may be used, assuming it is consistent with <CODE>equals</CODE> */
	private final java.util.Comparator comparator = Comparator.LEXICOGRAPHICAL;

	/** Builds a new <CODE>Enumerator</CODE> listing <CODE>newers</CODE> minus <CODE>olders</CODE>
	 **	The lists are sorted first to speed up comparisons: complexity is thus <CODE>nlog(n)</CODE>
	 * */
	protected Diff(Enumerator newers, Enumerator olders, boolean checkDate) {
		this.newers = newers.statify().sortAndRemoveDuplicates(comparator);
		this.olders = olders.statify().sortAndRemoveDuplicates(comparator);
		this.checkDate = checkDate;
		limit = olders.hasMoreElements() ? olders.nextEntry() : null;
	}
						
	protected Entry reload() {
		while (true) {
			Entry next = newers.nextEntry();
			if ( (limit == null) || (next == null) )  // No more entry to be removed, or end of list
				return next;  // 'next' may be 'null'
			while (comparator.compare(limit, next) < 0) {
				limit = olders.nextEntry();
				if (limit == null)
					return next;
			}
			int comparison = comparator.compare(limit, next);
			if (  (comparison > 0)  ||  ( checkDate && (comparison == 0) && (limit.date < next.date) )  )
				return next;
		}
	}

}
