

package sam.entry;


	/** Name filtering class. **/
public abstract class NameFilter {

	/** Specification is reduced to equality */
	public static final NameFilter EQUALITY = new  NameFilter() {
		protected final boolean specifies(String specification, String entry) {
			return (specification == null) ? (entry == null) : specification.equals(entry);
		}
	};

	/** Standard <CODE>?</CODE> and <CODE>*</CODE> characters regular expansion. */
	public static final Wildcards WILDCARDS = new Wildcards('*', '?', false);

	/** @return <CODE>true</CODE> if and only if <CODE>specification</CODE> expands to <CODE>entry</CODE> */
	protected abstract boolean specifies(String specification, String entry);

	public boolean specifies(Entry a, Entry b) { return specifies(a.name, b.name);	}

	/**  Basic derivation of an order.
	 ** <p><CODE>a < b</CODE> if <CODE>a</CODE> specifies <CODE>b</CODE>,
	 ** <CODE>a > b</CODE> if <CODE>b</CODE> specifies <CODE>a</CODE>,
	 ** and <CODE>a = b</CODE> (in the sense that <CODE>compare(a,b)</CODE> returns <CODE>0</CODE>) otherwise
	 ** (even if this especially means that <b>a and b differ</b>).
	 **/
	protected class Comparator implements java.util.Comparator {
		public int compare(Object a, Object b) {
			if ( specifies( (Entry) a, (Entry) b ) )
				return -1;
			if ( specifies( (Entry) b, (Entry) a ) )
				return +1;
			return 0;
		}
	}

	/** Standard wild-characters regular expansion.
	 ** <p>Assuming <CODE>?</CODE> and <CODE>*</CODE> be the wilcards,
	 ** <CODE>WILDCARDS.specifies("a*c", "abc") = true</CODE>,
	 ** <CODE>WILDCARDS.specifies("*.txt", "foo.bar.txt") = true</CODE>,
	 ** <CODE>WILDCARDS.specifies("?", "a") = true</CODE>,
	 ** <CODE>WILDCARDS.specifies("?*", "a?c") = true</CODE>,
	 **/
	protected static class Wildcards extends NameFilter {
		private final char any, one;
		private final boolean caseMatters;

		/** Builds a new wildcards-based name filter
		 ** @param anyChar usually <CODE>*</CODE>
		 ** @param oneChar usually <CODE>?</CODE>
		 ** @param caseMatters wether case matters (UNIX-like) or not (Win32-like)
		 **/
		public Wildcards(char anyChar, char oneChar, boolean caseMatters) {
			this.any = anyChar;
			this.one = oneChar;
			this.caseMatters = caseMatters;
		}

		protected boolean specifies(String specification, String entry) {
			if ( (specification == null) || (entry == null) )
				return (specification == null) && (entry == null);
			return specifies(specification, 0, specification.length(), entry, 0, entry.length() );
		}

		/** Recursive implementation of a simple pattern-matching test **/
		private boolean specifies(String a, int x, int X, String b, int y, int Y) {
			while (true) {
				if (x >= X)
					return (y == Y);
				if (y >= Y) {
					while ((x < X) && (charAt(a, x) == any))
						x++;
					return (x == X);
				}
				char aChar = charAt(a, x);
				if (aChar == one) {
					x++;
					y++;
					continue;
				}
				else if (aChar == any) {
					do
						if (++x >= X)
							return true;
					while ( (aChar = charAt(a, x)) == any)
						;
					while (y < Y)
						if ( (charAt(b, y) == aChar) && ( specifies(a, x+1, X, b, y+1, Y) ) )
							return true;
						else
							y++;
				} else
					if (charAt(a, x++) != charAt(b, y++))
						return false;
			}
		}

		private char charAt(String s, int i) { return caseMatters ? s.charAt(i) : Character.toUpperCase(s.charAt(i));	}
		
	}


}


