

package sam.event;

import sam.entry.Entry;

import java.util.Set;
import java.util.Iterator;


/** Class for dispatching <CODE>FileEvent</CODE> through registered <CODE>FileEventListener</CODE>s.
 **/
public class FileEventNotifier {

	/** Class describing a notification type. **/
	public static abstract class NotificationType {
		/** Notifies specified listener of specified event. **/
		protected abstract void notify(FileEvent event, FileEventListener listener);
	}

	/** Notification type for failed entry events **/
	public static final NotificationType PROCESSED = new NotificationType() {
		protected void notify(FileEvent event, FileEventListener listener) { listener.processed(event); }
	};

	/** Notification type for processed entry events **/
	public static final NotificationType FAILED = new NotificationType() {
		protected void notify(FileEvent event, FileEventListener listener) { listener.failed(event); }
	};

	/** Notification type for missing entry events **/
	public static final NotificationType MISSING = new NotificationType() {
		protected void notify(FileEvent event, FileEventListener listener) { listener.missing(event); }
	};

	/** Notification type for mismatching entry events **/
	public static final NotificationType MISMATCHING = new NotificationType() {
		protected void notify(FileEvent event, FileEventListener listener) { listener.mismatching(event); }
	};

	/** Notification type for overwritten entry events **/
	public static final NotificationType OVERWRITTEN = new NotificationType() {
		protected void notify(FileEvent event, FileEventListener listener) { listener.overwritten(event); }
	};

	/** Notification type for skipped entry events **/
	public static final NotificationType SKIPPED = new NotificationType() {
		protected void notify(FileEvent event, FileEventListener listener) { listener.skipped(event); }
	};

	private final Set listeners = new java.util.HashSet();

	/** Registers a new <CODE>FileEvent</CODE> listener **/
	public void register(FileEventListener listener) { if (listener != null) listeners.add(listener); }

	/** Unregisters a previously registered <CODE>FileEvent</CODE> listener **/
	public void unregister(FileEventListener listener) { if (listener != null) listeners.remove(listener); }


	/** Notifies all registered listeners of specified event **/
	public void notifyAll(NotificationType type, Entry entry, String message) {
		notifyAll(type, new FileEvent(entry, message) );
	}

	/** Notifies all registered listeners of specified event **/
	public void notifyAll(NotificationType type, FileEvent event) {
		for (Iterator iterator = listeners.iterator(); iterator.hasNext(); )
			type.notify(event, (FileEventListener) iterator.next());
	}

}
