
package sam.event;

import sam.toolkit.Text;
import sam.entry.Entry;


/** Class summarizing informations about the current progress state.
 ** Note that this class has <b>no public constructor</b> as it should
 ** only be instantiated through a <CODE>ProgressEventNotifier</CODE>.
 ** @see ProgressEventNotifier
 **/
public class ProgressEvent {

	/** Partial statistics **/
	public static class Statistics {

		/** Number of bytes actually achieved **/
		public final long achieved;

		/** Number of bytes to be achieved **/
		public final long limit;

		/** Time elapsed since the beginning **/
		public final long elapsedTimeMilliseconds;

		/** The ratio between encoded data and plain data **/
		public final float encodingRatio;

		/** Gets the number of bytes already achieved in a human-readable format **/
		public String achievedSizeString() { return Text.getSizeString(achieved);	}

		/** Gets the number of bytes already achieved in a human-readable format **/
		public String limitSizeString() { return Text.getSizeString(limit);	}

		/** Gets the achieved percentage, as an <CODE>int</CODE> in the range 0..100 inclusive **/
		public int achievedPercentage() { return Text.getPercentage(achieved, limit); }

		/** Gets the processing rate in a human-readable format **/
		public String getRate() {
			return (elapsedTimeMilliseconds == 0) ? "" : Text.getSizeString( (long) Math.ceil(1000 * achieved / elapsedTimeMilliseconds) )+ "/s";
		}

		/** Gets the compression ratio in a human-readable format **/
		public String getCompressionRatioString() { return Text.getOneDecimalString(1.0 - encodingRatio) + "%"; }

		/** Gets the estimated total duration. **/
		public long getTotalDuration() {
			return (achieved == 0L) ? 0 : (long) Math.ceil((double)( (limit * (double)elapsedTimeMilliseconds / (double)achieved)));
		}

		/** Gets the estimated total duration, as a human-readable string **/
		public String getTotalDurationString() { return (achieved == 0L) ? "" : Text.getDuration( getTotalDuration() );	}

		/** Gets the estimated remaining duration. **/
		public long getRemainingDuration() { return getTotalDuration() - elapsedTimeMilliseconds;	}

		/** Gets the estimated remaining duration, as a human-remaining string **/
		public String getRemainingDurationString() { return Text.getDuration( getRemainingDuration() );	}

		/** Wether this job is finished **/
		public boolean isFinished() {	return achieved == limit; }

		/** Wether this job is finished **/
		public boolean isAtBeginning() { return achieved == 0L; }

		private Statistics(long achieved, long limit, long elapsedTimeMilliseconds, float encodingRatio) {
			this.achieved = achieved;
			this.limit = limit;
			this.elapsedTimeMilliseconds = elapsedTimeMilliseconds;
			this.encodingRatio = encodingRatio;
		}

	}

	/** Dummy entry for testing/debugging purpose **/
	public static final ProgressEvent DUMMY = new ProgressEvent(1234L, 2345L, 1000L, 0.50f, "dummy-entry",
																															3456L, 4567L, 10000L, 0.60f, 10, 20);

	/** Statistics on current entry **/
	public final Statistics entry;

	/** Global statistics **/
	public final Statistics global;

	/** The name of the entry being currently processed **/
	public final String currentEntryName;

	/** The number of the current entry **/
	public final int numberOfCurrentEntry;

	/** The total number of entries **/
	public final int totalEntriesCount;

	/** Gets a shorter name for the entry being currently processed
	 ** @param length the length of the desired shorter name
	 **/
	public String currentEntryShortName(int length) { return sam.toolkit.Text.left(currentEntryName, length); }

	/** Builds a new progress event
	 **/
	ProgressEvent(long entryAchieved, long entryLimit, long entryDuration, float entryCompressionRatio, String name,
								long globalAchived, long globalLimit, long globalDuration, float globalCompressionRatio,
								int numberOfCurrentEntry, int totalEntriesCount) {
		this.entry = new Statistics(entryAchieved, entryLimit, entryDuration, entryCompressionRatio);
		this.global = new Statistics(globalAchived, globalLimit, globalDuration, globalCompressionRatio);
		this.currentEntryName = name;
		this.numberOfCurrentEntry = numberOfCurrentEntry;
		this.totalEntriesCount = totalEntriesCount;
	}

	/** Wether this is the first event concerning the current entry,
	 ** i.e. if this is an event corresponding to a new entry
	 **/
	public boolean isNextEntry() { return (entry.achieved == 0); }

	public String toString() {
		return new StringBuffer()
				.append(numberOfCurrentEntry).append("/").append(totalEntriesCount)
				.append(" \"").append(currentEntryName).append("\"")
				.append(" at ").append(entry.achieved).append("/").append(entry.limit)
				.append(", total ").append(global.achieved).append("/").append(global.limit)
				.toString();
	}

}
