
package sam.event;

import sam.ui.UserInterface;
import sam.core.Settings;
import sam.entry.Entry;
import sam.toolkit.Text;
import sam.toolkit.Mailbox;

import java.util.Set;
import java.util.Iterator;
import java.io.IOException;


/** Class for dispatching <CODE>ProgressEvent</CODE> through registered <CODE>ProgressEventListener</CODE>s.
 **
 ** Typical usage is:
 ** <pre><codeblock>
 ** ProgressEventNotifier notifier = new ProgressEventNotifier();
 ** notifier.register( new ProgressEventListener() {
 **    public void progress(Event event) { System.out.println(event); }  // Or anything else
 ** } );
 ** OutputStream output = notifier.new EncodedOutputStream( new FileOutputStream("encoded.dat") );
 ** output = ... ; // which includes various filtering operations called "encoding"
 ** output = notifier.new PlainOutputStream(output);
 ** ... ; // Write plain bytes of data to output
 ** </codeblock></pre>
 **
 ** Note that event posting is <em>asynchronous</em> in order to preserve computationnal resources of the
 ** main thread.
 **
 ** @see Settings.Constants#PROGRESS_NOTIFICATION_PRIORITY
 **/
public class ProgressEventNotifier {

	/** The set of all registered <CODE>ProgressEvent</CODE> listeners **/
	private final Set listeners = new java.util.HashSet();

	/** The mailbox containing to-be-handled events **/
	private final Mailbox mailbox = new Mailbox();

	/** The time at which the <em>first</em> entry began to be processed **/
	private final long globalStartTime = System.currentTimeMillis();

	/** The time at which the current entry began to be processed **/
	private long entryStartTime = globalStartTime;

	/** The current number of plain bytes that have been processed **/
	private long plain = 0;

	/** The limit above which a new event shall be fired **/
	private long plainLimit = Settings.Constants.PROGRESS_PLAIN_GRANULARITY;

	/** The total number of bytes that have to be processed **/
	private long totalNumberOfPlainBytes = 0;

	/** The current number of encoded bytes that have been processed **/
	private long encoded = 0;

	/** The limit above which a new event shall be fired **/
	private long encodedLimit = Settings.Constants.PROGRESS_ENCODED_GRANULARITY;

	private long lastEntryPlain = 0;

	private long lastEntryEncoded = 0;

	/** The <CODE>Entry</CODE> being currently processed **/
	private Entry currentEntry = null;

	/** The number of <CODE>Entry</CODE> that have already been processed **/
	private int numberOfProcessedEntries = 0;

	/** The total number of <CODE>Entry</CODE> that have to be processed **/
	private int totalNumberOfEntries = 0;

	/** Builds a new <CODE>ProgressEvent</CODE> notifier.
	 ** Note that a call to <CODE>setTotalNumberOfPlainBytes</CODE> and one to <CODE>setTotalNumberOfEntries</CODE>
	 ** are required before proper use.
	 ** This also starts one postmaster thread, which will be stopped once this <CODE>ProgressEventNotifier</CODE>
	 ** is elected by the GC for finalization.
	 **/
	public ProgressEventNotifier() { start();	}

	/** Sets the total number of bytes that have to be processed **/
	public void setTotalNumberOfPlainBytes(long totalNumberOfPlainBytes) {
		this.totalNumberOfPlainBytes = totalNumberOfPlainBytes;
		UserInterface.get().log().log(5, ProgressEventNotifier.class, "There are " + totalNumberOfPlainBytes + " bytes (" +
																	Text.getSizeString(totalNumberOfPlainBytes) + ") to be processed");
	}

	/** Sets the total number of <CODE>Entry</CODE> that have to be processed **/
	public void setTotalNumberOfEntries(int totalNumberOfEntries) {
		this.totalNumberOfEntries = totalNumberOfEntries;
		UserInterface.get().log().log(5, ProgressEventNotifier.class, "There are " + totalNumberOfEntries + " entries to be processed");
	}

	/** Registers a new <CODE>ProgressEvent</CODE> listener **/
	public void register(ProgressEventListener listener) { if (listener != null) listeners.add(listener); }

	/** Unregisters a previously registered <CODE>ProgressEvent</CODE> listener **/
	public void unregister(ProgressEventListener listener) { if (listener != null) listeners.remove(listener); }

	/** Notifies all registered listeners of the current progress state
	 ** Note that notification is asynchronous.
	 **/
	protected synchronized void notifyListeners() {
		final ProgressEvent event = createEvent();
		if (event != null)
			mailbox.post(event);
	}

	/** Starts the event postmaster thread.
	 ** Note that calling this several times is safe although of no use.
	 **/
	public synchronized void start() {
		UserInterface.get().log().log(4, ProgressEvent.class, "Starting event postmaster");
		mailbox.start( new Mailbox.Reader() {
			public void read(Object mail) { doNotifyListeners( (ProgressEvent) mail ); }
		} );
	}

	/** Notifies the termination of the process.
	 ** This is a blocking call that waits for all <CODE>ProgressEvent</CODE> to have been handled.
	 **/
	public synchronized void terminate() { notifyListeners(); mailbox.terminate(); }

	/** Actually does notifies all listening threads of specified event.
	 ** Note that this is a synchronous call, i.e. this is run inside the current thread.
	 **/
	private void doNotifyListeners(ProgressEvent event) {
		if (event == null) {
			UserInterface.get().log().log(5, ProgressEventNotifier.class, "Attempt to notify null progress event !");
			return;
		}
		for (Iterator iterator = listeners.iterator(); iterator.hasNext(); )
			try {
				((ProgressEventListener) iterator.next()).progress(event);
			} catch (Throwable t) {
				UserInterface.get().error().warn("Could not deliver progress event: " + Text.getMessage(t) );
			}
	}

	/** Notifies more bytes of plain data to have been processed
	 ** This is called automatically by inner input/output streams
	 **/
	protected void notifyPlain(long plainIncrement) {
		plain += plainIncrement;
		if (plain > plainLimit) {
			plainLimit = plain + Settings.Constants.PROGRESS_PLAIN_GRANULARITY;
			notifyListeners();
		}
	}

	/** Notifies more bytes of encoded data to have been processed.
	 ** This is called automatically by inner input/output streams
	 **/
	protected void notifyEncoded(long encodedIncrement) {
		encoded += encodedIncrement;
		if (encoded > encodedLimit) {
			encodedLimit = encoded + Settings.Constants.PROGRESS_ENCODED_GRANULARITY;
			notifyListeners();
		}
	}

	/** Notifies a new entry being processed.
	 ** This is to be called manually at each new entry.
	 ** Note that even entries that are to be skipped or ignored shall be notified here.
	 **/
	public void notifyNextEntry(sam.entry.Entry entry) {
		if (currentEntry != null)
			notifyListeners();
		entryStartTime = System.currentTimeMillis();
		lastEntryPlain = plain;
		lastEntryEncoded = encoded;
		currentEntry = entry;
		++ numberOfProcessedEntries;
		notifyListeners();
	}

	/** Creates a new <CODE>ProgressEvent</CODE> describing the current progress state
	 ** @see ProgressEvent
	 **/
	protected final ProgressEvent createEvent() {
		if (currentEntry == null)
			return null;
		final long now = System.currentTimeMillis();
		return new ProgressEvent( (plain - lastEntryPlain), currentEntry.length, (now - entryStartTime),
															getEncodingRatio(plain - lastEntryPlain, encoded - lastEntryEncoded), currentEntry.name,
															plain, totalNumberOfPlainBytes, (now - globalStartTime), getEncodingRatio(encoded, plain),
															numberOfProcessedEntries, totalNumberOfEntries);
	}

	/** Gets a failsafe compression ratio
	 ** @return <CODE>1.0</CODE> if <CODE>plain</CODE> is negative or <CODE>encoded</CODE> is non-positive
	 **/
	private static final float getEncodingRatio(long plain, long encoded) {
		return ((plain < 0L) || (encoded <= 0L) ) ? 1.0f : (float) plain / (float) encoded;
	}

	/** On destruction the postmaster thread must be stopped.
	 ** Note that this implementation does <b>not</b> call <CODE>Thread.stop()</CODE>
	 **/
	protected void finalize() { terminate(); }

	/** General-purpose static inner-class for notifying the number of bytes read from an <CODE>InputStream</CODE>. **/
	private static abstract class InputStream extends java.io.InputStream {
 	  /** To-be-filtered <CODE>InputStream</CODE> to read from **/
  	private final java.io.InputStream delegee;

		public InputStream(java.io.InputStream delegee) { this.delegee = delegee;	}

		/** Notifies <CODE>numberOfReadBytes</CODE> to have been skipped **/
		protected abstract void notifyRead(long numberOfReadBytes);

	  /** @throws EOFException if no more bytes are available **/
 	  public int read() throws IOException {
			final int data = delegee.read();
			notifyRead(1);
			return data;
		}
					
	  /** @return the number of bytes actually read
		 ** @throws EOFException if no single byte is available
		 **/
		public int read(byte[] data, int begin, int length) throws IOException {
			if (length <= 0)
				return 0;
			int read = delegee.read(data, begin, length);
			notifyRead(read);
			return read;
		}


 		/** Skips <CODE>length</CODE> bytes.
		 **/
		public long skip(long length) throws IOException {
			long skipped = delegee.skip(length);
			notifyRead(skipped);
			return skipped;
		}

 		public void close() throws IOException { delegee.close();	}

	}

	/** Instance inner-class for notifying the number of bytes read from a plain <CODE>InputStream</CODE>. **/
	public class PlainInputStream extends InputStream {
		public PlainInputStream(java.io.InputStream delegee) { super(delegee); }
		protected void notifyRead(long numberOfReadBytes) { notifyPlain(numberOfReadBytes); }
	}

	/** Instance inner-class for notifying the number of bytes read from an encoded <CODE>InputStream</CODE>. **/
	public class EncodedInputStream extends InputStream {
		public EncodedInputStream(java.io.InputStream delegee) { super(delegee); }
		protected void notifyRead(long numberOfReadBytes) { notifyEncoded(numberOfReadBytes); }
	}


	/** General-purpose static inner-class for notifying the number of bytes written to an <CODE>OutputStream</CODE>. **/
	private static abstract class OutputStream extends java.io.OutputStream {

  	/** To-be-filterd <CODE>OutputStream</CODE> to write to **/
 	  private final java.io.OutputStream delegee;

		public OutputStream(java.io.OutputStream delegee) { this.delegee = delegee; }

		/** Notifies <CODE>numberOfReadBytes</CODE> to have been skipped **/
		protected abstract void notifyWritten(long numberOfWrittenBytes);

	  public void write(int value) throws IOException {
			delegee.write(value);
			notifyWritten(1);
		}

		public void write(byte[] data, int begin, int length) throws IOException {
			if (length <= 0)
				return;
			delegee.write(data, begin, length);
			notifyWritten(length);
		}

		public void flush() throws IOException { delegee.flush();	}

 		public void close() throws IOException {
			delegee.close();
		}

	}

	/** Instance inner-class for notifying the number of plain bytes written to an <CODE>OutputStream</CODE>. **/
	public class PlainOutputStream extends OutputStream {
		public PlainOutputStream(java.io.OutputStream delegee) { super(delegee); }
		protected void notifyWritten(long numberOfWrittenBytes) { notifyPlain(numberOfWrittenBytes); }
	}

	/** Instance inner-class for notifying the number of encoded bytes written to an <CODE>OutputStream</CODE>. **/
	public class EncodedOutputStream extends OutputStream {
		public EncodedOutputStream(java.io.OutputStream delegee) { super(delegee); }
		protected void notifyWritten(long numberOfWrittenBytes) { notifyEncoded(numberOfWrittenBytes); }
	}

}

