
package sam.io;

import java.io.*;


/** Bit-level output stream.
 ** @see BitInputStream
 **/
public class BitOutputStream {
		private final OutputStream delegee;
		private int buffer;
		private int mask = 0x80;
		private boolean isClosed = false;

		protected BitOutputStream(final OutputStream delegee) {
			this.delegee = delegee;
		}

		public void write(boolean bit) throws IOException {
				if (mask < 1)
					store();
				if (bit)
					buffer = buffer | mask;
				mask >>= 1;
		}

		/**
			 To be called <b>only</b> on exit, because of possibly remaining bytes !
		*/
		protected void storeAndflush() throws IOException {
				store();
				flush();
		}


		/**
			 This does NOT write current buffer but simply forwards the <CODE>flush</CODE> call to the underlying stream.
			 <p>Flushing current buffer would lead to inconsistencies in case of subsequent writes, and is thus disabled.
		*/
		public void flush() throws IOException {
				delegee.flush();
		}


		public void close() throws IOException {
				if (isClosed)
						throw new IOException("Can not close already closed bit-outputstream !");
				storeAndflush();
				delegee.close();
				isClosed = true;
		}

		/**
			 Flushes the buffer to the underlying stream.
			 <p><b>Disclaimer</b>: to be done only as the last call, since up to 7 unattributed leading bits
			 may be stored as well as part of the atomic one-byte = 8-bits flush. Consequent writes from the
			 stream becomes inconsistent.
		*/
		private void store() throws IOException {
				if (isClosed)
						throw new IOException("Can not write to closed bit-outputstream !");
				delegee.write(buffer);
				buffer = 0;
				mask = 0x80;
		}

}

