
package sam.io;

import sam.ui.UserInterface;

import java.io.*;
import sam.toolkit.Random;


/** Abstract class for filtering <CODE>byte[]</CODE> arrays **/
public abstract class ByteArrayFilter {

	/** Length of a block size **/
	protected final int length;

	protected ByteArrayFilter(int length) { this.length = length; }

	/** Filters the <CODE>data</CODE> <CODE>byte[]</CODE> array
	 ** from position <CODE>offset</CODE> up to <CODE>offset + length - 1</CODE>.
	 **/
	abstract protected void filter(byte[] data, int offset);

	public void filter(byte[] data) {	filter(data, 0, data.length);	}

	public void filter(byte[] data, int begin, int end) {
		for (int offset = begin; offset<end; offset += length)
			filter(data, offset);
	}

	/** Copies the biggest possible chunk of bytes from one <CODE>byte[]</CODE> array to another.
	 ** @return the number of bytes effectively copied.
	 **/
	private int fillAsMuchAsPossible(byte[] source, int sourceBegin, int sourceLimit, byte[] target, int targetBegin, int targetLimit) {
		final int sourceLength = sourceLimit - sourceBegin;
		final int targetLength = targetLimit - targetBegin;
		final int length = (sourceLength < targetLength) ? sourceLength : targetLength;  // <CODE>min(sourceLength, targetLength)</CODE>
		System.arraycopy(source, sourceBegin, target, targetBegin, length);
		return length;
	}


	/** Called whenever the whole buffer can not be filled up with fresh data.
	 ** <p>To be overriden as required.
	 ** <p>This implementation fills up remaining bytes with random data.
	 **/
	protected void fillRandomly(byte[] data, int start, int end) {
    for (int i=start; i<end; )
			data[i++] = Random.nextByte();  // Not so efficient actually ...
	}

	/**
	 ** Instance inner-class for filtering an <CODE>InputStream</CODE> with outer-instance's filter,
	 ** as in:
	 ** <pre><codeblock>
	 ** ByteArrayFilter filter = new MyByteArrayFilter(myKey);
	 ** InputStream input = filter.new InputStream(toBeFilteredInputStream);
	 ** </codeblock></pre>
	 ** @see sam.io.ByteArrayFilter.OutputStream
	 **/
	public class InputStream extends java.io.InputStream {

 	  /** To-be-filtered <CODE>InputStream</CODE> to read from **/
  	private final java.io.InputStream delegee;
    /** Buffered data, stored in filtered form **/
		private byte[] data;
		/** Index of the next-to-be-read byte inside <CODE>data</CODE> **/
		private int index;
		/** Limit of the readable bytes inside <CODE>data</CODE> **/
		private int limit;

		/** Builts a new <CODE>InputStream</CODE> to read data that will be transparently filtered.
		 ** @param delegee the underlying output stream to read to-be-filtered bytes from
		 **/
		public InputStream(java.io.InputStream delegee) { this(1, delegee);	}

		/** Builts a new <CODE>InputStream</CODE> to read data that will be transparently filtered.
		 ** @param size the desired buffer size <b>in atomic blocks</b> (i.e. the underlying
		 ** <CODE>byte[]</CODE> buffer will be of total length <b>size * length</b>, where <CODE>length</CODE>
		 ** is the length of the enclosing <CODE>ByteArrayFilter</CODE>)
		 ** @param delegee the underlying output stream to read to-be-filtered bytes from
		 **/
		public InputStream(int size, java.io.InputStream delegee) {
			this.delegee = delegee;
			this.data = new byte[size * length];
			this.limit = this.index = data.length;
		}

		/** @throws EOFException if no more bytes are available **/
		public int read() throws IOException {
			reloadIfRequired();
			return (int) data[index++] & 0xff;  // The <CODE>0xff</CODE> mask definetely <b>is</b> necessary
		}
					
		/** @return the number of bytes actually read
		 ** @throws EOFException if no single byte is available
		 **/
		/* BUG corrected: EOF exception have to be ignored during <CODE>reloadIfRequired</CODE>
		 */
		public int read(byte[] data, int begin, int length) throws IOException {
			if (length <= 0)
				return 0;
			int offset = begin;
			final int end = begin + length;
			while (offset < end) {
				try {
					reloadIfRequired();
				} catch (EOFException e) {
					UserInterface.get().log().log(8, ByteArrayFilter.InputStream.class, "End of data reached: " + e.getMessage() );
				}
				final int numberOfCopiedBytes = fillAsMuchAsPossible(this.data, this.index, this.limit, data, offset, end);
				if (numberOfCopiedBytes <= 0)
					break;
				offset += numberOfCopiedBytes;
				this.index += numberOfCopiedBytes;
			}
			return offset - begin;
		}


		/** Trivial although very efficient optimization:
		 ** do not filter to-be-skipped bytes !
		 **/
		public long skip(long length) throws IOException {
			final int toBeReadBytes = limit - index;  // Number of bytes still to be read in buffer
			if (length <= toBeReadBytes) { // Which includes <CODE>length == 0</CODE>
				index += (int) length;  // Previous test implies that <CODE>length <= data.length</CODE> and thus fills into an <CODE>int</CODE>
				return length;
			} else {
				index = limit;  // This is equivalent to skipping <CODE>toBeReadBytes</CODE> bytes
				return delegee.skip(length - toBeReadBytes) + toBeReadBytes;
			}
		}

		public void close() throws IOException { delegee.close();	}

		private void reloadIfRequired() throws IOException {
			if (index >= limit)
				reload();
		}

 		/** @throws EOFException if not more bytes are available **/
  	private void reload() throws IOException {
			int read = delegee.read(data);
			index = 0;
			if (read <= 0) { // EOF Reached, no single byte could be read
				limit = 0;
				throw new EOFException("No more to-be-filtered data could be read !");
			}
			limit = length * (int) Math.ceil( (double)read / (double)length);
			try {
				filter(data, 0, limit);
			} catch (Throwable t) {
				UserInterface.get().error().handleException("Fatal error while filtering data", t);
				throw new IOException( t.getMessage() );
			}
		}

	}

	/**
	 ** Instance inner-class for filtering an <CODE>OutputStream</CODE> with outer-instance's filter,
	 ** as in:
	 ** <pre><codeblock>
	 ** ByteArrayFilter filter = new MyByteArrayFilter(myKey);
	 ** OutputStream output = filter.new OutputStream(toBeFilteredOutputStream);
	 ** </codeblock></pre>
	 ** @see sam.io.ByteArrayFilter.InputStream
	 **/
	public class OutputStream extends java.io.OutputStream {

  	/** To-be-filterd <CODE>OutputStream</CODE> to write to **/
 	  private final java.io.OutputStream delegee;

    /** Buffered data, stored in filtered form **/
	  private byte[] data;

	  /** Index of the next-to-be-written byte inside <CODE>data</CODE> **/
		private int index;

		/** Builts a new <CODE>OutputStream</CODE> to write data that will be transparently filtered.
		 ** @param delegee the underlying output stream to write filtered bytes to
		 **/
		public OutputStream(java.io.OutputStream delegee) { this(1, delegee); }

		/** Builts a new <CODE>OutputStream</CODE> to write data that will be transparently filtered.
		 ** @param size the desired buffer size <b>in atomic blocks</b> (i.e. the underlying
		 ** <CODE>byte[]</CODE> buffer will be of total length <b>size * length</b>, where <CODE>length</CODE>
		 ** is the length of the enclosing <CODE>ByteArrayFilter</CODE>)
		 ** @param delegee the underlying output stream to write filtered bytes to
		 **/
		public OutputStream(int size, java.io.OutputStream delegee) {
			this.delegee = delegee;
			this.data = new byte[size * length];
			this.index = 0;
		}

		public void write(int value) throws IOException {
			saveIfRequired();
			data[ index++ ] = (byte) value;  // The <CODE>0xff</CODE> mask is <b>not</b> necessary here
		}

		public void write(byte[] data, int begin, int length) throws IOException {
			final int end = begin + length;
			int offset = begin;
			while  (offset < end) {
				saveIfRequired();
				final int numberOfCopiedBytes = fillAsMuchAsPossible(data, offset, end, this.data, this.index, this.data.length);
				offset += numberOfCopiedBytes;
				this.index += numberOfCopiedBytes;
			}
		}

		public void flush() throws IOException {
			// save();  // THIS IS A BUG: the 'fillRandomly' will cause random bytes be inserted if flush() is called before a 'write' !!
			delegee.flush();
		}

 		public void close() throws IOException {
			save();
			delegee.close();
		}

		private void saveIfRequired() throws IOException {
			if (index >= data.length)
				save();
		}

		private void save() throws IOException {
			if (index <= 0)
				return;
			int limit = length * (int) Math.ceil( (double)index / (double)length);
			fillRandomly(data, index, limit);
			try {
				filter(data, 0, limit);
			} catch (Throwable t) {
				UserInterface.get().error().handleException("Fatal error while filtering data", t);
				throw new IOException( t.getMessage() );
			}
			delegee.write(data, 0, limit);
			index = 0;
		}

	}

}

