
package sam.io;

import sam.toolkit.Text;

import java.io.IOException;
import java.util.zip.Checksum;
import java.util.zip.CRC32;
import java.util.zip.CheckedInputStream;


/** Class for automatically matching the <CODE>Checksum</CODE> of an <CODE>InputStream</CODE>
 ** against a predefined value.
 ** <p>This allows for transparent checksum-checking like in:
 ** <pre><codeblock>
 ** CheckedOutputStream output = new CheckOutputStream( new FileOutputStream("foo") );
 ** ...
 ** Checksum expectedChecksum = output.getChecksum();
 ** InputStream input = new ChecksumCheckerInputStream( new FileInputStream("bar"), expectedChecksum);
 ** ...
 ** </codeblock></pre>
 ** @deprecated
 **/
public class ChecksumCheckerInputStream extends CheckedInputStream {
	/** The required checksum **/
	private final Checksum requiredChecksum;

	/** Exception thrown whenver checksums do not match  **/
	public class MismatchingChecksumException extends IOException {
		MismatchingChecksumException(Checksum computed, Checksum required) {
			super("Mismatching " + required.getClass().getName() + "checksum: got 0x" +
						Text.toHexa(computed.getValue()) + " instead of 0x" + Text.toHexa(required.getValue()) );
		}
	}

	/** Builds a new <CODE>CheksumCheckerInputStream</CODE>
	 ** @param delegee the underlying input stream
	 ** @param requiredChecksum the required checksum to compare to
	 ** <p>Note that the current implementation is limitated to <CODE>CRC32</CODE> checksums.
	 **/
	public ChecksumCheckerInputStream(java.io.InputStream delegee, CRC32 requiredChecksum) {
		super(delegee, new CRC32() );
		this.requiredChecksum = requiredChecksum;
		getChecksum().reset();
	}

	/** Checks the computed checksum.
	 ** @throws MismatchingChecksumException if the current checksum does not match the required one.
	 **/
	public void check() throws MismatchingChecksumException {
		if ( getChecksum().getValue() != requiredChecksum.getValue() )
			throw new MismatchingChecksumException(getChecksum(), requiredChecksum);
	}

	/** Closes the underlying stream and checks the checksums.
	 ** @see #check
	 **/
	public void close() throws IOException {
		super.close();
		check();
	}

	/** Makes sure the checksums are checked on exit.
	 ** <p>Please understand that throwing a <CODE>MismatchingChecksumException</CODE> inside the
	 ** <em>Garbage Collector</em>'s thread might <b>not</b> be so desirable ...
	 **/
	protected void finalize() throws IOException {
		close();
	}

}


