
package sam.io;

import sam.ui.UserInterface;

import java.util.zip.Checksum;
import java.util.zip.CRC32;
import java.io.*;


/** Tracker that adds a <CODE>CRC32</CODE> checksum trace to processed data.
 ** This allows a per-block consistency checking during extraction.
 **/
public class ChecksumTracker extends Tracker {

	/** Up-to-date checksum for current block **/
	private Checksum checksum = new CRC32();

	/** Required checksum value if any,
	 ** or <CODE>0</CODE> if none.
	 **/
	private long requiredValue = 0;

	public ChecksumTracker() { super("Checksum", "CRC32 block checksum"); }

	public void update(byte[] data, int begin, int length) { checksum.update(data, begin, length); }

	/** Checks the computed checksum and loads that of the next one.
	 **/
	protected void load(DataInputStream input) throws IOException {
		requiredValue = input.readLong();
		if (requiredValue == checksum.getValue())
			UserInterface.get().log().log(7, ChecksumTracker.class, "Checksum 0x" + Long.toHexString(requiredValue) + " verified.");
		else {
			UserInterface.get().error().warn("Mismatching checksum: got 0x" + Long.toHexString(checksum.getValue()) +
																			 " instead of 0x" + Long.toHexString(requiredValue));
			throw new IOException("Mismatching checksum");
		}
		checksum.reset();
	}

	protected void save(DataOutputStream output) throws IOException {
		final long checksumValue = checksum.getValue();
		UserInterface.get().log().log(7, ChecksumTracker.class, "Setting block checksum 0x" + Long.toHexString(checksumValue) );
		output.writeLong(checksumValue);
		checksum.reset();
	}

}
