/*
  This whole file is distributed under the terms of the General Public License.

  Last revision date: 2002/05/11
 */

package sam.io;

import java.util.*;
import java.io.*;


/* For details on the implementation, please see class <CODE>Node</CODE> below.
 */

/** Implementation of the auto-adaptive Huffman-tree coding input/output-streams
 ** with variable pattern length.
 ** This allows for transparent Huffman input/output filtering, like in
 ** <pre><codeblock>
 ** InputStream source = new FileInputStream("foo.dat");
 ** OutputStream target = new Huffman8().output( new FileOutpuStream("bar.dat") );
 ** </codeblock></pre>
 ** for compression and similarly
 ** <pre><codeblock>
 ** InputStream source = new Huffman8().input( new FileInputStream("bar.dat") );
 ** OutputStream target = new FileOutpuStream("foo.dat");
 ** </codeblock></pre>
 **
 ** @see sam.io.Huffman#input
 ** @see sam.io.Huffman#output
 ** @see sam.io.Huffman.Leaf
 **/
// <CODE>public</CODE> just to allow acces to the <CODE>main</CODE> method
public abstract class Huffman extends Node {

	/** List of all leaves of the whole tree.
	 ** There should be exactly at most {@link numberOfPatterns} of those,
	 ** or exactly <CODE>numberOfPatterns</CODE> if this tree is full.
	 **/
	private Leaf[] leaves;

	private Node last;

	private final int numberOfPatterns;

	/** Mask of patterns coded by this huffman tree.
	 ** This is <CODE>1 << patternWidth</CODE>.
	 ** In particular this is equal to {@link numberOfPatterns} - 1.
	 **/
	private final int mask;

	/** The granularity of updates of this tree.
	 ** A value of <CODE>1</CODE> enforces update at each new pattern processed,
	 ** which is the most accurate but also the most expensive.
	 **/
	private final int granularity;

	/** Main executable code.
	 ** This program (de)compresses a file with either 8-bits or 16-bits patterns.
	 ** Syntax is <CODE>java sam.Huffman {+|-}{8|16} input output</CODE>
	 ** <b>Warning</b>: 16-bits [en|de]coding is extremly slow !
	 **/
	public static void main(String[] args) {
		if (args.length != 3) {
			System.err.println("Syntax: java " + Huffman.class.getName() + " {+|-}{8|16} input output");
			System.exit(1);
		}
		try {
			java.io.InputStream input = new FileInputStream(args[1]);
			java.io.OutputStream output = new FileOutputStream(args[2]);
			final boolean encode = (args[0].charAt(0) == '+');
			try {
				final int length = Integer.valueOf( args[0].substring(1) ).intValue();
				System.err.print( (encode ? "Encoding " : "Decoding") + " " + args[1] + " to " + args[2] + " with a " + length + " bits Huffman.");
				final Huffman huffman = (length == 8) ? (Huffman) new Huffman8() : (Huffman) new Huffman16();
				if (encode)
					output = huffman.output(output);
				else
					input = huffman.input(input);
				try {
					IO.bufferedDump(input, output);
					System.err.print("done.");
				} finally {
					output.close();
					input.close();
					System.err.println();
				}
				System.err.println("Shortest path: " + huffman.shortestPath() + ", longest path: " + huffman.longestPath() + ", expected ratio: " +
												 ((int) Math.floor(100 * huffman.ratio()) - 100) + "%");
				System.exit(0);
			} catch (NumberFormatException e) {
				System.err.println("Invalid Huffman pattern size: " + args[0] + ", expected either 8 or 16.");
				System.exit(1);
			}
		} catch (IOException e) {
			System.err.println("I/O exception occured: " + e.getMessage() );
			System.exit(2);
		}
	}


	/** Stream for decoding huffman-encoded data
	 ** @param delegee the stream to decode from
	 **/
	public abstract InputStream input(InputStream encodedStream);

	/** Stream for encoding 8 bits huffman-encoded data
	 ** @param delegee the 8 bits stream to encode to
	 **/
	public abstract OutputStream output(OutputStream encodedStream);

	/** Builds a new generic Huffman tree of specified pattern width.
	 ** @param patternWidth the pattern width.
	 ** This can be <b>any positive integer up to 32</b>, but for
	 ** obvious reasons multiples of 8 shall be of wider use.
	 **/
	protected Huffman(int patternWidth) {
		this(patternWidth, 1);
	}

	/** Creates a new well-balanced Huffman tree.
	 ** <p>Each of the <CODE>numberOfPatterns</CODE> patterns is initially weighted by 1.
	 ** This is an implementation requirement: a node is required to be strictly heavier than any of its sons,
	 ** which would not be granted in case of zero-weighted leaves.
	 ** @param granularity the granularity of the frequency updates. Higher means faster but less accurate (and thus
	 ** lower compression ratios). In particular, a sequence of at most <CODE>granularity</CODE> consecutive bytes will
	 ** <b>not</b> be compressed.
	 ** @param patternWidth the pattern width.
	 ** This can be <b>any positive integer up to 32</b>, but for
	 ** obvious reasons multiples of 8 shall be of wider use.
	 ** @param granularity the granularity of updates of this tree.
	 ** A value of <CODE>1</CODE> enforces update at each new pattern processed,
	 ** which is the most accurate but also the most expensive.
	 **/
	protected Huffman(int patternWidth, int granularity) {
		super(null);  // Builds the <CODE>Node</CODE> with no father (root node)
    this.numberOfPatterns = 2 << (patternWidth - 1);
		this.mask = numberOfPatterns - 1;
		this.granularity = granularity;
		left = null;
		right = null;
		last = this;
		buildBitsSequence();
		reset();
	}

	/** @return the total number of leaves of this Huffman tree **/
	public int numberOfLeaves() {	return numberOfPatterns; }

	/** @return the total number of nodes of this Huffman tree **/
	public int numberOfNodes() { return 2 * numberOfPatterns - 1; }

	/** @return the size in bits of the smallest encoded pattern **/
	public int shortestPath() {
		int length = numberOfPatterns;
		for (int i=0; i<leaves.length; ++i)
			if (leaves[i].bits.length < length)
				length = leaves[i].bits.length;
		return length;
	}

	/** @return the size in bits of the largest encoded pattern **/
	public int longestPath() {
		int length = 0;
		for (int i=0; i<leaves.length; ++i)
			if (leaves[i].bits.length > length)
				length = leaves[i].bits.length;
		return length;
	}

	/** @return the expected compression ratio associated to this Huffman tree **/
	public float ratio() {
		long expectedLength = 0;
		long totalWeight = 0;
		for (int i=0; i<leaves.length; ++i) {
			totalWeight += leaves[i].weight;
			expectedLength += leaves[i].weight * leaves[i].bits.length;
		}
		return (float) ((double)expectedLength / ((double)8.0 * totalWeight));
	}


	/** Resets the Huffman tree into a fully well-balanced one.
	 **/
 	public void reset() {
		leaves = new Leaf[numberOfPatterns];  // Sets up a fresh new tree.
		buildWellBalancedSubTree();
		resetWellBalancedOrder();
		buildBitsSequence();
	}

	private Node buildWellBalancedSubTree() {
		final int n = numberOfPatterns / 2;
		left = buildWellBalancedSubTree(0, n - 1, this, n);
		right = buildWellBalancedSubTree(n, numberOfPatterns - 1, this, n);
		return this;
	}

	/** Builds a well-balanced subtree, which may happen to be a leaf
	 ** @param begin pattern at the most-left leaf of the subtree
	 ** @param end pattern at the most-right leaf of the subtree
	 ** @param father node to which this subtree will be connected
	 ** @param weight <b>total</b> weight of the subtree
	 ** Note that a single leaf is built whenever <CODE>begin == end</CODE>, in which case
	 ** the weight should be one.
	 ** Note that the ordering linked list is <b>not</b> updated during this process
	 ** @see #resetWellBalancedOrder
	 **/
	private Node buildWellBalancedSubTree(int begin, int end, Node father, int weight) {
		if (begin == end) { // We are building a leaf here
		  Leaf leaf = leaves[begin] = new Leaf(begin, father);
			leaf.weight = 1;
			return leaf; 
		}	else {
			Node node = new Node(father, weight, null, null);
			node.buildBitsSequence();
			int middle = (end + begin) / 2;
			node.left = buildWellBalancedSubTree(begin, middle, node, weight / 2);
			node.right = buildWellBalancedSubTree(middle + 1, end, node, weight / 2);
			return node;
		}
	}

	/** Rebuilds the ordering link list stored in the <CODE>upper</CODE> and <CODE>lower</CODE> files of <CODE>Node</CODE> **/
	private void resetWellBalancedOrder() {
		upper = null;
		lower = left; left.upper = this;
		left.lower = right; right.upper = left;
		resetWellBalancedOrder(left, null, right, left, this);
		resetWellBalancedOrder(right, left, null, left, this);
	}

	/** Rebuilds the ordering link list of a subtree recursively.
	 ** @param a the desired subtree
	 ** @param leftNeighboor the left neighboor of a when looking on a single level; null if a is the most-left of its level
	 ** @param rightNeighboor the right neighboor of a when looking on a single level; null if a is the most-right of its level
	 ** @param mostLeftHere the most-left node of this level (never null)
	 ** @param mostRightUpper the most-right node of the <b>upper</b> level (never null)
	 **/
	private void resetWellBalancedOrder(Node a, Node leftNeighboor, Node rightNeighboor, Node mostLeftHere, Node mostRightUpper) {
		a.upper = (leftNeighboor == null) ? mostRightUpper : leftNeighboor;
		a.lower = (rightNeighboor == null) ? mostLeftHere.left : rightNeighboor;
		if (a.left != null) {  // We are not at a leaf here
			resetWellBalancedOrder(a.left, (leftNeighboor == null) ? null : leftNeighboor.right, a.right,
														 mostLeftHere.left, mostRightUpper.right);
			resetWellBalancedOrder(a.right, a.left, (rightNeighboor == null) ? null : rightNeighboor.left,
														 mostLeftHere.left, mostRightUpper.right);
		}

	}

	/** Gets the leaf coding for specified value.
	 **  Does <b>not</b> update the Huffman tree.
	 ** @return the leaf associated to byte value <CODE>value</CODE>.
	 ** @see sam.io.Huffman.Leaf#oneMore
	 **/
	protected Leaf get(int value) {
		Leaf leaf = leaves[value & mask];  // <CODE>&</CODE> is definetely different from <CODE>%</CODE>
		if (leaf == null) {
			if (last.left == null) {  // Special case for the first new byte value
				last.left = leaf = new Leaf(value, last);
				last.right = null;  // <!> Because 'new Node' forces be the right son <!>
				connect(last, leaf);
			} else {  // General case: ensure 'last' is a node, and append 'Leaf(value)' as its right son
				if ( last.right != null )
					last = last.right.insert();
				leaf = new Leaf(value, last);  // Leaf set as <b>right</b> of 'last', then 'last' becomes this leaf
				connect(last.left, leaf);
			}
			// System.err.println("Added leaf for 0x" + Integer.toHexString(value) + ": " + this + ", last = " + last +
			//   "\nInternal: " + super.toString() );
		}
		return leaf;
	}


	/** Gets a formatted representation of this huffman tree.
	 ** This is well-suited for debugging but is rather unreadable.
	 ** This only prints nodes and node the whole tree structure.
	 **/
	public String toString() {
		StringBuffer buffer = new StringBuffer().append("Huffman:[");
		boolean hasSome = false;
		for (int i=0; i<leaves.length; i++) {
			if (leaves[i] != null) {
				if (hasSome)
					buffer.append(',');
				buffer.append( leaves[i] );
				hasSome = true;
			}
		}
		return buffer.append(']').toString();
	}

	/** Class for reading leaves from an Huffman-compressed <CODE>InputStream</CODE>
	 **/
	protected class LeafInputStream {
		private final BitInputStream delegee;

		public LeafInputStream(java.io.InputStream delegee) {	this.delegee = new BitInputStream(delegee);	}

		/**
		   Uncompresses next pattern from a Huffman-compressed stream.
		   @return <CODE>null</CODE> if end of stream
		   @throws EOFException if the compressed stream did <b>not</b> end after a whole compressed bits pattern
		 */
		public Leaf read() throws IOException {
			Node node = Huffman.this;
			try {
				while ( ! node.isLeaf() ) {
					node = delegee.read() ? node.right : node.left;
				}
				Leaf leaf = (Leaf) node;
				leaf.oneMore();
				return leaf;
			} catch (EOFException e) {
				if (node != Huffman.this)
					throw new IOException( "Unexpected end of Huffman stream: at " + node.bits + ": " +	sam.toolkit.Text.getMessage(e) );
				return null;  // Everything fine, we just reached the end of the stream
			}
		}

		public void close() throws IOException {  delegee.close(); }

	}


	/** Class for writing Huffman-compressed leaves to an <CODE>OutputStream</CODE>
	 **/
	protected class LeafOutputStream {
		private final BitOutputStream delegee;

		public LeafOutputStream(java.io.OutputStream delegee) {
			this.delegee = new BitOutputStream(delegee);
		}

		public synchronized void write(Leaf leaf) throws IOException {
			for (int i=0; i < leaf.bits.length; i++)
				delegee.write( leaf.bits[i] );
			leaf.oneMore();
		}

		public void flush() throws IOException {	delegee.flush();	}
		public void close() throws IOException {	flush(); delegee.close();	}
	}


	/** Special <CODE>Node</CODE> representing a leaf.
	 **/
	protected class Leaf extends Node {

		/** The pattern encoded by this leaf.
		 ** Note that the bits present in {@link Huffman#mask} might be non-null.
		 **/
		public final int value;

		/** The total number of times this pattern has been processed.
		 ** It always holds that {@linl Huffman#granularity} times the <em>weight</em>
		 ** of this node equals <CODE>count</CODE>.
		 **/
		private int count = 0;

		/** Builds a new leaf.
		 ** @param value the pattern encoded by this leaf
		 ** @param father the father node of this leaf
		 **/
		Leaf(int value, Node father) {
			super(father);
			this.value = value;
			this.count = granularity;  // Force update at first creation
			left = null;
			right = null;
			leaves[value] = this;
			buildBitsSequence();
		}

		/** Gets a formatted representation of this huffman tree.
		 ** This is well-suited for debugging but is rather unreadable.
		 **/
		public String toString() {
			return ("0x" + Integer.toHexString(value) + "(" + weight + ")=" + Node.toString(bits) );
		}

		/** Increments the weight of this leaf by one.
		 ** Note that this also updates the whole huffman tree.
		 **/
		public void oneMore() {
			if (count++ >= granularity) {
				count = 0;
				last = super.increments(last);
			}
		}

	}

}


/** Main node class for representing a binary tree.
 ** <p>The Huffman tree consists of one instance named 'root', possibly linked to other nodes which form so-called
 ** 'subtrees' with their own sons. The root class is in fact only instantiated as the superclass of a
 ** <CODE>Huffman</CODE> instance, and so are leaves (node with no son at all) with <CODE>Huffman.Leaf</CODE>.
 ** <p>Each node has a weight, defined as the total weight of any of its sons (and thus the total weight of the
 ** subtree, by recursion). The weight of leaves as used by Huffman's algorithm stores the frequency of the
 ** corresponding value.
 **
 ** <p>Each node links to its father and its left and right sons through its according <CODE>father</CODE>,
 ** <CODE>left</CODE> and <CODE>right</CODE> field. <CODE>null</CODE> is used to code no such link.
 ** <p>Each node links to the lightest heavier node of the whole tree through its <CODE>upper</CODE> field,
 ** and to the heaviest lighter node of the whole tree through its <CODE>lower</CODE> field. <CODE>null</CODE>
 ** is used to denote no such node.
	 * */
class Node {

	/** Father node of this node.
	 ** This is <CODE>null</CODE> only for the root node which is always an actual {@link Huffman} instance.
	 ** This has package access as it is used by <CODE>Huffman</CODE>
	 **/
	Node father;

	/** Right son of this node.
	 ** This is <CODE>null</CODE> only if this is an actual {@link sam.io.Huffman.Leaf} instance,
	 ** and moreover <b>at most one</b> node among the whole tree which is the top-right leaf.
	 ** This has package access as it is used by <CODE>Huffman</CODE>
	 **/
	Node right;

	/** Left son of this node.
	 ** This is <CODE>null</CODE> only if this is the root and the whole tree is empty.
	 ** In particular in such a case this is an actual {@link Huffman} instance and
	 ** <CODE>right<CODE> is also <CODE>null</CODE>.
	 ** This has package access as it is used by <CODE>Huffman</CODE>
	 **/
	Node left;

	/** Lightest node that is heavier than this node.
	 ** In case of equality
	 ** This is <CODE>null</CODE> only if this is the root,
	 ** in which case this is an actual {@link Huffman} instance and
	 ** This has package access as it is used by <CODE>Huffman</CODE>
	 **/
	Node upper;

	/** Heaviest node that is lighter than this node.
	 ** This is <CODE>null</CODE> only if this is the top-right leaf
	 ** in which case this is an actual {@link sam.io.Huffman.Leaf} instance and
	 ** This has package access as it is used by <CODE>Huffman</CODE>
	 **/
	Node lower;

	/** Weight of this node.
	 ** By definition if this <CODE>Node</CODE> is a leaf then this is the number of occurrences
	 ** of the corresponding pattern
	 ** Otherwise this is the sum of the weights of its sons, where an empty son counts for zero.
	 **/
	long weight;

	/** Precomputed bits pattern of this node.
	 ** <p>Mainly used for leaves, bu also maintained for node for efficiency reasons.
	 ** <p>This can become expensive in memory.
	 ** This has package access as it is used by <CODE>Huffman</CODE>
	 **/
	boolean[] bits = null;


	/** Builds a new <em>leaf</em> as a <CODE>Node</CODE> instance
	 ** Note that this always sets it as the <b>right</b> son of <CODE>father</CODE>.
	 ** @param father the father of this leaf.
	 **/
	protected Node(Node father) {
		this(father, 0, null, null);
		if (father != null)
			father.right = this;
	}


	/** Builds a new <CODE>Node</CODE>.
	 ** Note that this always sets it as the <b>right</b> son of <CODE>father</CODE>.
	 ** @param father the father of this node
	 ** @param weight the initial weight of this node
	 ** @param left the left son of this node
	 ** @param right the right son of this node
	*/
	Node(Node father, long weight, Node Left, Node right) {
		this.father = father;
		this.weight = weight;
		this.left = left;
		this.right = right;
		if (father != null)
			father.right = this;
		// System.err.println("new Node(0x" + Integer.toHexString(value) + ":" + value + "/" + 256 + ")");
	}

	/** Gets the enumeration of all nodes <em>that are lighter</em> than this node
	 ** in decreasing order.
	 ** This is mostly usefull when called from the root node which is the heaviest,
	 ** i.e. from an <CODE>Huffman</CODE> instance.
	 */
	public Enumeration ordered() {
		return new Enumeration() {
				private Node current = Node.this;

				public Object nextElement() throws NoSuchElementException {
					if (current.lower != null)
						return (current = current.lower);
					else
						throw new NoSuchElementException();
				}

				public boolean hasMoreElements() {
					return current.lower != null;
				}

			};
	}


	/** Rebuilds the bits sequence for this node, <b>and all its sons</b>.
	 ** <p>Computation based on the <CODE>father</CODE>'s bits, if any.
	 **/
	void buildBitsSequence() {
		if (father != null)
			buildBitsSequence(father.bits, (father.right == this) );
		else {
			bits = new boolean[0];
			recurseBuildBitsSequence();
		}
	}

	/** Rebuilds the bits sequence for this node, <b>and all its sons</b>.
	 ** @param prefix the prefix bits sequence pre-computation
	 ** @param direction wether this is a right son.
	 **/
	private void buildBitsSequence(boolean[] prefix, boolean direction) {
		bits = new boolean[prefix.length + 1];
		System.arraycopy(prefix, 0, bits, 0, prefix.length);
		bits[ prefix.length  ] = direction;
		recurseBuildBitsSequence();
	}

	/** Recursively builts the bit sequences of the left then of the right son.
	 **/
	private void recurseBuildBitsSequence() {
		if (left != null)
			left.buildBitsSequence(bits, false);
		if (right != null)
			right.buildBitsSequence(bits, true);
	}


	/** Increments the weight of this node.
	 ** Maintains the ordered list in order, and updates the father's weight recursively.
	 ** Note that the produced tree
	 ** @param last the lightest leaf
	 ** @return the lightest leaf
	 **/
	Node increments(Node last) {
		weight++;
		while ( (upper != null) && (upper.upper != null) && (upper.weight < weight) ) {
			swapTree(upper);
			swapOrder();
		}
        if (father != null) 
			father.increments(last);
		return (last != this) ? last : (lower == null) ? this : lower;
	}

	/** Wether this node is a pure leaf, i.e. if it has <b>no</b> son. **/
	public boolean isLeaf() {
		return (left == null) && (right == null);
	}

	/** Gets a formatted representation of this node.
	 ** This is well-suited for debugging but is rather unreadable.
	 ** Moreover, this is a bi-recursive call i.e. calling it on the root
	 ** will produce a <CODE>String</CODE> representation of the <b>whole</b> tree.
	 **/
	public String toString() {
		return ("Node(" + weight + ")#" + hashCode() + "( " +  left + " , " + right + " )" );
	}

	/** Inserts a new node.
	 ** <b>Warning</b>: This is consistent only if <CODE>this</CODE> is a <b>right</b> son.
	 ** In particular insertions can only be done from the top-right node.
	 **/
	Node insert() {
		Node node = new Node(father);
		father = node;
		node.left = this;
		node.weight = this.weight;
		connect(upper, node);
		connect(node, this);
		node.buildBitsSequence();
		buildBitsSequence();  // Shoudl have been recursively done already
		// System.err.println("Inserted: " + node + " from " + father + ", upper = " + upper + 
		//			   " , node.upper = " + node.upper );
		return node;
	}

	/** Swaps <CODE>this</CODE> and <CODE>target</CODE> in the tree
	 ** Maintains the <CODE>bits</CODE> field up-to-date
	 **/
	private void swapTree(Node target) {
		// System.err.println("   > " + this + ".swapTree( " + target + " ) , father = " + father );
		if (target == null)
			return;
		Node newFather = target.father;
		if ( (father == null) || (newFather == null) )
			return;
		replaceSon(father, target, father.left == this);  // 'upper' takes my place in the tree
		replaceSon(newFather, this, newFather.left == target);  // and I take its old place in the tree
		target.buildBitsSequence();
		buildBitsSequence();
		// System.err.println("    <<< " + this + ".swapTree( " + target + " , father = " + father );
	}



	/** Swaps <CODE>this</CODE> and <CODE>upper</CODE> in the ordered list
	 **/
	private void swapOrder() {
		// System.err.println("   > " + this + ".swapOrder() , upper = " + upper + " , lower = " + lower );
		if (upper == null)
			return;
		Node uppest = upper.upper;
		connect(upper, lower);
		connect(this, upper);
		connect(uppest, this);
		// System.err.println("    <<< " + this + ".swapOrder() , upper = " + upper + " , lower = " + lower );
	}

	/** Connects <CODE>up</CODE> to <CODE>down</CODE> in that order
	 **/
	static void connect(Node up, Node down) {
		if (up != null)
			up.lower = down;
		if (down != null)
			down.upper = up;
	}


	/** Replaces son <CODE>oldSon</CODE> in <CODE>father</CODE> with <CODE>newSon</CODE>
	 **/
	private static void replaceSon(Node father, Node son, boolean left) {
		if (father != null)
			if (left)
				father.left = son;
			else
				father.right = son;
		son.father = father;
	}


	/** Builds a human-readable representation of a bits sequence.
	 **/
	static String toString(boolean[] bits) {
		if (bits == null)
			return "null";
		StringBuffer buffer = new StringBuffer();
		for (int i=0; i<bits.length; i++)
			buffer.append( bits[i] ? '1' : '0' );
		return buffer.toString();
	}

}

