
package sam.io;

import sam.toolkit.Random;
import sam.ui.UserInterface;

import java.io.*;


/** Implementation of a mixer that is well-suited for Huffman-encoded streams
 ** It inserts pseudo-randomized characters drawn in a way that is equivalent
 ** to draw an arbitrary Huffman tree among all possible ones.
 ** Beware of confusions with the <CODE>sam.io.Huffman</CODE> class !
 **/
public class HuffmanMixer extends Mixer {
	private final Huffman huffman;
	private final int numberOfPatterns;

	public HuffmanMixer(Huffman huffman) {
		this.huffman = huffman;
		this.numberOfPatterns = huffman.numberOfLeaves();
		UserInterface.get().log().log(4, HuffmanMixer.class, "Initializing Huffman-" + numberOfPatterns + " mixer");
	}
			
 	public String verbose() { return "drawing the Huffman tree randomly at each block" ; }

	protected long getDataChunkLength(long blockSize) { return blockSize; }

	public void writeGarbageChunk(java.io.OutputStream delegee) throws IOException {
		UserInterface.get().log().log(6, HuffmanMixer.class, "Unbalancing Huffman-" + numberOfPatterns + " tree");
		huffman.reset();
		Random.IntegerSetRandomizer patternsDrawer = new Random.IntegerSetRandomizer(numberOfPatterns);
		// 1) Computes each pattern weight
		UserInterface.get().log().log(6, HuffmanMixer.class, "Computing new weights");
		int[] patternsWeights = new int[numberOfPatterns];
		int weight = 1;  // Hopefully this would not require a <CODE>long</CODE>, since that many bytes will be written !
		for (int i=0; i<numberOfPatterns; ++i) {
			int pattern = patternsDrawer.next();
			patternsWeights[pattern] = weight;
			if ( Random.nextBoolean() )
				weight *= 2;  // This definetely makes sense
		}
		// 2) Writes those patterns arbitrarily into the target
		if (numberOfPatterns == 256) {  // 8-bits encoding
			UserInterface.get().log().log(6, HuffmanMixer.class, "Writing single-byte patterns");
			for (Random.IntegerSetRandomizer fullDrawer = new Random.IntegerSetRandomizer(patternsWeights); fullDrawer.hasMoreElements(); )
				delegee.write( fullDrawer.next() );
		} else {  // Here we assume <CODE>numberOfPatterns == 65536</CODE>
			UserInterface.get().log().log(6, HuffmanMixer.class, "Writing double-byte patterns");
			final DoubleByte.Writer doubleByteDelegee = new DoubleByte.OutputStreamWriter (delegee);
			for (Random.IntegerSetRandomizer fullDrawer = new Random.IntegerSetRandomizer(patternsWeights); fullDrawer.hasMoreElements(); )
				doubleByteDelegee.write( fullDrawer.next() );
		} 
	}

	public void readGarbageChunk(java.io.InputStream stream, long length) throws IOException {
		UserInterface.get().log().log(6, HuffmanMixer.class, "Resetting Huffman-" + numberOfPatterns + " tree");
		huffman.reset();  // This is the very important !
		UserInterface.get().log().log(6, HuffmanMixer.class, "Skipping " + length + " garbage bytes");
		stream.skip(length);
	}

}
