

package sam.io;

import sam.ui.UserInterface;
import sam.toolkit.Process;
import sam.toolkit.Text;
import sam.core.Settings;

import java.io.*;
import java.util.zip.*;
import java.util.jar.*;
import java.awt.Frame;


/** Input/output toolkit
 **/
public class IO  {
		private IO() {	}  // Purely static class

		/** Basic null output.
		 ** All bytes stored into this <CODE>OutputStream</CODE> are simply dropped.
		 **/	
		public static final java.io.OutputStream NULL_OUTPUT = new OutputStream() {
			public void write(int b) {  }
			public void write(byte[] b, int start, int end) { }
			public void flush() { }
			public void close() { }
		};

		/** Creates any necessary directory inside which a specified file is contained.
		 ** @param file a <b>file</b> (not directory) name
		 ** @throws <CODE>IOException</CODE> if an I/O <b>or security</b> exception occured
		 **/
		public static void makePathIfRequired(File file) throws IOException {
			if (file == null)
				return;
			File requiredDirectory = file.getParentFile();
			try {
				if ( (requiredDirectory == null) || requiredDirectory.isDirectory() )  // Nothing to be done there
					return;
				if ( ! requiredDirectory.mkdirs() )  // Tries to create missing directories
					throw new IOException("Failed to create path " + requiredDirectory );
			} catch (SecurityException e) {
				throw new IOException("Not allowed to create path " + requiredDirectory + ": " + Text.getMessage(e) );
			}
		}

		/** Turns a <CODE>String</CODE> to a readable stream
		 **/	
		public static InputStream getAsStream(String content) {
			return getAsStream( new String[] { content } );
		}

		/** Turns a list of <CODE>String</CODE>s to a readable stream
		 **/
		public static InputStream getAsStream(String[] lines) {
			ByteArrayOutputStream output = new ByteArrayOutputStream();
			try {
				OutputStreamWriter writer = new OutputStreamWriter(output);
				for (int i=0; i<lines.length; i++) {
					if (i > 0)
						writer.write('\n');
					writer.write(lines[i], 0, lines[i].length());
				}
				writer.flush();
			} catch (IOException e) {
			}
			return new ByteArrayInputStream( output.toByteArray() );
		}

		/** Turns a stream to a <CODE>String</CODE>
		 **/
		public static String getFromStream(InputStream stream) throws IOException {
			return getFromStream(stream, true);
		}

		/** Tokenizes a stream.
		 ** Note that this implementation does <b>not</b> use the standard <CODE>java.io.StreamTokenizer</CODE> class.
		 **/
		public static String[] getTokens(InputStream stream, String delimiters) throws IOException {
			return Text.split(IO.getFromStream(stream), delimiters);
		}

		/** Turns a stream to a <CODE>String</CODE>
		 ** @param stream the stream to read from
		 ** @param close wether to close the input stream once exhausted
		 **/
		public static String getFromStream(InputStream stream, boolean close) throws IOException {
			if (stream == null)
				return null;
			Reader reader = new InputStreamReader(stream);
			StringBuffer string = new StringBuffer();
			char[] buffer = new char[1024];
			int read;
			try {
				while ( (read = reader.read(buffer, 0, 1024)) >= 0 )
					string.append(buffer, 0, read);
			} catch (EOFException e) {
				// Ignore EOF, simply exits the dump loop
			}
			if (close)
				stream.close();
			return string.toString();
		}

		/** Pre-extracts data.
		 ** This dumps a stream fully to a temporary file which is then re-opened for reading.
		 ** @param data the data stream to be pre-extracted
		 ** @param temporary the temporary <em>file</em> or <em>directory</em> to pre-extract data into
		 **/
		public static InputStream preextract(InputStream data, File temporary) throws IOException {
			if (temporary.isDirectory())
				temporary = createTemporaryFile(temporary);
			OutputStream output = new FileOutputStream(temporary);
			bufferedDump(data, output);
			output.close();
			return new FileInputStream(temporary);
		}

		/** Creates a new temporary file.
		 **/
		public static File createTemporaryFile(File directory) throws IOException {
			return File.createTempFile(Settings.Environment.PROGRAM_NAME, Settings.Environment.TEMPORARY_FILES_SUFFIX, directory);
		}


		/** Creates a new entry in a <CODE>Jar</CODE> archive
		 ** @return an <CODE>OutputStream</CODE> to be written into, which automatically closes the underlying entry on <CODE>close()</CODE>
		 ** <b>Warning:</b> forgetting the <CODE>close()</CODE> call will cause the entry to be closed once in the GC's thread, which
		 ** means at non-deterministic points, which might corrupt the underlying ZIP archive.
		 ** @param jar the <CODE>Jar</CODE> archive to write to
		 ** @param name the <CODE>Jar</CODE> entry name to create, which might include a <CODE>/</CODE>-separated path
		 ** @param compressionLevel the compression level ranging from <CODE>0</CODE> (no compression) to <CODE>9</CODE>
		 ** (best but slowest compression).
		 **/
		public static OutputStream putNextEntry(final JarOutputStream jar, final String name, int compressionLevel) throws IOException {
			UserInterface.get().log().log(4, IO.class,"Storing \"" + name + "\" with compression level " + compressionLevel);
			if (compressionLevel < 0)
				jar.setMethod(ZipEntry.STORED);
			else {
				jar.setMethod(ZipEntry.DEFLATED);
				jar.setLevel(compressionLevel);
			}
			jar.putNextEntry( new JarEntry(name) );
			return new java.io.OutputStream() {
				private boolean isClosed = false;
				public void write(int b) throws IOException { jar.write(b); }
				public void write(byte[] data, int begin, int end) throws IOException { jar.write(data, begin, end); }
				public synchronized void close() throws IOException {
					if (! isClosed) {
						UserInterface.get().log().log(4, IO.class, "Closing entry \"" + name + "\"");
						jar.closeEntry();
						isClosed = true;
					}
				}
				protected synchronized void finalize() throws IOException {
					if (! isClosed) {
						UserInterface.get().error().warn("closing Jar entry \"" + name + "\" on asynchronous finalize(),"
																						 + " expect broken archive; closing anyway");
						close();
					}
				}
			};
		}

		/** Dumps a stream to a new entry in a <CODE>Jar</CODE> archive
		 ** @param jar the <CODE>Jar</CODE> archive to write to
		 ** @param name the <CODE>Jar</CODE> entry name to create, which might include a <CODE>/</CODE>-separated path
		 ** Data is stored according to the default compression level
		 ** @see sam.core.Settings.Defaults#JAR_COMPRESSION_LEVEL
		 **/
		public static void jar(InputStream input, long length, String name, JarOutputStream jar) throws IOException {
			jar(input, length, name, jar, Settings.Defaults.JAR_COMPRESSION_LEVEL);
		}

		/** Dumps a stream to a new entry in a <CODE>Jar</CODE> archive
		 ** @param input the input data stream
		 ** @param length the expected length of the input stream
		 ** @param jar the <CODE>Jar</CODE> archive to write to
		 ** @param name the <CODE>Jar</CODE> entry name to create, which might include a <CODE>/</CODE>-separated path
		 ** @param compressionLevel the compression level ranging from <CODE>0</CODE> (no compression) to <CODE>9</CODE>
		 ** (best but slowest compression).
		 **/
		public static void jar(InputStream input, long length, String name, JarOutputStream jar, int compressionLevel) throws IOException {
			OutputStream output = putNextEntry(jar, name, compressionLevel);
			try {
				bufferedDump(input, output, length);
			} finally {
				input.close();
				output.close();
			}
		}

		/** Dumps a stream to a new entry in a <CODE>Jar</CODE> archive
		 ** @param input the input data stream
		 ** @param jar the <CODE>Jar</CODE> archive to write to
		 ** @param name the <CODE>Jar</CODE> entry name to create, which might include a <CODE>/</CODE>-separated path
		 ** Data is stored according to the default compression level
		 ** @see sam.core.Settings.Defaults#JAR_COMPRESSION_LEVEL
		 **/
		public static void jar(InputStream input, String name, JarOutputStream jar) throws IOException {
			jar(input, name, jar, Settings.Defaults.JAR_COMPRESSION_LEVEL);
		}

		/** Dumps a stream to a new entry in a <CODE>Jar</CODE> archive
		 ** @param input the input data stream
		 ** @param jar the <CODE>Jar</CODE> archive to write to
		 ** @param name the <CODE>Jar</CODE> entry name to create, which might include a <CODE>/</CODE>-separated path
		 ** @param compressionLevel the compression level ranging from <CODE>0</CODE> (no compression) to <CODE>9</CODE>
		 ** (best but slowest compression).
		 **/
		public static void jar(InputStream input, String name, JarOutputStream jar, int compressionLevel) throws IOException {
			OutputStream output = putNextEntry(jar, name, compressionLevel);
			try {
				bufferedDump(input, output);
			} finally {
				input.close();
				output.close();
			}
		}

		/** Dumps a <CODE>byte[]</CODE> array of data to a new entry in a <CODE>Jar</CODE> archive
		 ** @param data the input data
		 ** @param jar the <CODE>Jar</CODE> archive to write to
		 ** @param name the <CODE>Jar</CODE> entry name to create, which might include a <CODE>/</CODE>-separated path
		 ** Data is stored according to the default compression level
		 ** @see sam.core.Settings.Defaults#JAR_COMPRESSION_LEVEL
		 **/
		public static void jar(byte[] data, String name, JarOutputStream jar) throws IOException {
			jar(data, name, jar, Settings.Defaults.JAR_COMPRESSION_LEVEL);
		}

		/** Dumps a <CODE>byte[]</CODE> array of data to a new entry in a <CODE>Jar</CODE> archive
		 ** @param data the input data
		 ** @param jar the <CODE>Jar</CODE> archive to write to
		 ** @param name the <CODE>Jar</CODE> entry name to create, which might include a <CODE>/</CODE>-separated path
		 ** @param compressionLevel the compression level ranging from <CODE>0</CODE> (no compression) to <CODE>9</CODE>
		 ** (best but slowest compression).
		 **/
		public static void jar(byte[] data, String name, JarOutputStream jar, int compressionLevel) throws IOException {
			OutputStream output = putNextEntry(jar, name, compressionLevel);
			try {
				output.write(data);
			} finally {
				output.close();
			}
		}

		/** Extracts a whole entry to a file
		 ** @param jar the jar archive
		 ** @param name the <CODE>Jar</CODE> entry name to create, which might include a <CODE>/</CODE>-separated path
		 ** @param file the output file
		 **/
		public static void unjar(JarFile jar, String name, File file) throws IOException {
			unjar(jar, name, new FileOutputStream(file));
		}

		/** Extracts a whole entry to a stream
		 ** @param jar the jar archive
		 ** @param name the <CODE>Jar</CODE> entry name to create, which might include a <CODE>/</CODE>-separated path
		 ** @param output the output stream
		 **/
		public static void unjar(JarFile jar, String name, OutputStream output) throws IOException {
			ZipEntry entry = jar.getEntry(name);
			InputStream input = jar.getInputStream(entry);
			try {
				bufferedDump(input, output, entry.getSize());
			} finally {
				input.close();
				output.close();
			}
		}

		/** Loads a full stream to memory
		 ** @param input the input data stream
		 ** @return the <CODE>byte[]</CODE> array of data
		 **/
		public static byte[] readFully(InputStream input) throws IOException {
			if (input == null)
				return null;
			ByteArrayOutputStream buffer = new ByteArrayOutputStream();
			try {
				bufferedDump(input, buffer);
			} finally {
				return buffer.toByteArray();
			}
		}

		/** Dumps an input data stream to an output stream
		 ** This implementation loads and writes data in a bufferized fashion
		 ** @param input the input data stream
		 ** @param output the output stream
		 ** @param length the expected length of the data to be dumped
		 **/
		public static void bufferedDump(InputStream input, OutputStream output, long count) throws IOException {
			if (count <= 0)  // This test is important although trivial
				return;
			byte[] buffer = new byte[Settings.Constants.DUMP_BUFFER_SIZE];
			int read = 0;
			while (true)  // Retry as requested
				try {  // Try to read
					do  // As long as some more bytes
						while (true)  // Retry as requested
							try {  // Try to write
								Process.haltIfRequired();
								output.write(buffer, 0, read);
								count -= read;
								break;
							} catch (InterruptedException e) {
								UserInterface.get().error().handleException("I/O operation interrupted", e);
							} catch (IOException e) {
								askWetherToRetry(e);
							}
						while ((count > 0) && ( (read = input.read(buffer, 0, (count > buffer.length) ? buffer.length : (int)count) ) > 0));
					break;
				} catch (EOFException e) {  // Should actually not happen as <CODE>InputStream.read(byte[])</CODE> catches them ...
					break;
				} catch (IOException e) {
					askWetherToRetry(e);
				}
		}

		/** Dumps an input data stream to an output stream
		 ** This implementation loads and writes data byte-per-byte
		 ** @param input the input data stream
		 ** @param output the output stream
		 ** @param length the expected length of the data to be dumped
		 ** @deprecated
		 **/
		public static void unBufferedDump(InputStream input, OutputStream output, long count)	throws IOException {
			int read;
			while ( (count-- > 0) && ( (read = read(input)) >= 0 ) )
				while (true)
					try {
						output.write(read);
						break;
					} catch (EOFException e) {
						break;
					} catch (IOException e) {
						askWetherToRetry(e);
					}
			if (count >= 0)
				UserInterface.get().log().log(2, IO.class, "Truncated entry, missing #" + (count + 1) + " bytes");
			return;  // Everything fine, simpy exit.
		}

		/** Dumps an input data stream to an output stream
		 ** This implementation loads and writes data in a bufferized fashion
		 ** @param input the input data stream
		 ** @param output the output stream
		 **/
		public static void bufferedDump(InputStream input, OutputStream output) throws IOException {
			byte[] buffer = new byte[Settings.Constants.DUMP_BUFFER_SIZE];
			int read = 0;
			while (true) {  // Retry as requested
				try {  // Try to read
					do {  // As long as some more bytes
						while (true)  // Retry as requested
							try {  // Try to write
								output.write(buffer, 0, read);
								Process.haltIfRequired();
								break;
							} catch (InterruptedException e) {
								UserInterface.get().error().handleException("I/O operation interrupted", e);
							} catch (IOException e) {
								askWetherToRetry(e);
							}
					}	while ( (read = input.read(buffer)) > 0);
					break;
				} catch (EOFException e) {  // Should actually not happen as <CODE>InputStream.read(byte[])</CODE> catches them ...
					break;
				} catch (IOException e) {
					askWetherToRetry(e);
				}
			}
		}

		/** Dumps an input data stream to an output stream
		 ** This implementation loads and writes data byte-per-byte
		 ** @param input the input data stream
		 ** @param output the output stream
		 ** @deprecated
		 **/
		public static void unBufferedDump(InputStream input, OutputStream output) throws IOException {
			int read;
			while ( (read = read(input)) >= 0 )
				while (true)
					try {
						output.write(read);
						break;
					} catch (IOException e) {
						askWetherToRetry(e);
					}
		}

		/** Failsafe way of reading one byte of data
		 ** @param input the input data stream to read from
		 ** @deprecated
		 **/
		private static int read(InputStream input) throws IOException  {
			try {
				Process.haltIfRequired();
			} catch (InterruptedException e) {
				UserInterface.get().error().handleException("I/O operation interrupted", e);
			}
			while (true)
				try {
					return input.read();
				} catch (EOFException e) {
					UserInterface.get().log().log(2, IO.class, "Unexpected EOF. (" + Text.getMessage(e) + ")." );
					return -1;  // Error reported, escape.
				} catch (IOException e) {
					askWetherToRetry(e);
				}
		}


		/** Asks the user wether to retry an interrupted I/O operation
		 **/
		public static void askWetherToRetry(IOException e) throws IOException {
			UserInterface.get().log().log(2, IO.class, "I/O error: " + e);
			if (UserInterface.get().running().yesNo("I/O failure occured:\n" +
									 Text.getMessage(e) + "\nRetry ?"))
				throw e;  // User does not want do ignore this error
			UserInterface.get().log().log(2, IO.class, "Retrying" );
		}

}
