
package sam.io;

import sam.ui.UserInterface;

import java.io.*;
import java.awt.*;
import java.awt.event.*;


/** Class for mixing streams with garbage bytes.
 ** Mixing consists of
 ** <pre><codeblock>
 ** Mixer mixer = ...;
 ** OutputStream mixedOutput = mixer.new OutputStream(nativeOutput);
 ** ...  // Use 'mixedOutput' here
 ** mixedOutput.close();
 ** byte[] trace = mixer.getTrace();
 ** </codeblock></pre>
 ** then unmixing consists of
 ** <pre><codeblock>
 ** byte[] trace = ...
 ** Mixer mixer = ...
 ** mixer.setTrace(trace);
 ** InputStream unmixedInput = mixer.new InputStream(mixedInput);
 ** ...  // Use 'unmixedInput' here
 ** </codeblock></pre>
 **/
public abstract class Mixer {
	private byte[] trace;
	private long blockSize;
	private static final long UNLIMITED_BLOCKSIZE = Long.MAX_VALUE;  // Buggy if total size greater than 9223372036854775807 bytes ...
	private Tracker tracker;

	/** Sets the next data chunk length, in bytes.
	 ** @param blockSize the block size of the data stream,
	 ** <b>which might be zero</b> to specify infinite block size
	 ** A lazy implementation might just return <CODE>blockSize</CODE>
	 ** @return desired data chunk length, <b>or zero to specify unlimited data chunk</b>
	 **/
	protected abstract long getDataChunkLength(long blockSize);

	/** Sets the next data chunk length, in bytes.
	 ** @param blockSize the block size of the data stream,
	 ** <b>which might be zero</b> to specify infinite block size
	 ** A lazy implementation might just return <CODE>blockSize</CODE>
	 ** @return desired data chunk length, <b>or the maximum positive <CODE>long</CODE> to specify unlimited data chunk</b>
	 **/
	public long nextDataChunkLength(long blockSize) {
		final long dataChunkLength = getDataChunkLength(blockSize);
		return (dataChunkLength < 1) ? UNLIMITED_BLOCKSIZE : dataChunkLength;
	}

	/** Writes garbage data to the stream.
	 ** @param stream the output to write onto
	 ** Note that the number of bytes written will be kept back automatically
	 **/
	public abstract void writeGarbageChunk(java.io.OutputStream stream) throws IOException;

	/** Reads garbage data from the stream.
	 ** @param stream the input to write onto
	 ** @param length the number of garbage bytes to read
	 **/
	public abstract void readGarbageChunk(java.io.InputStream stream, long length) throws IOException;

	/** Gets the block size of this mixer. **/
	protected long blockSize() { return blockSize; }

	/** Sets the block size.
	 **/
	public void setBlockSize(long blockSize) {
		UserInterface.get().log().log(4, Mixer.class, "Setting block size to " + blockSize + " bytes");
		this.blockSize = blockSize;
	}

	/** Sets the tracker. **/
	public void setTracker(Tracker tracker) {
		UserInterface.get().log().log(4, Mixer.class, "Setting tracker to " + ((tracker == null) ? "<none>" : tracker.name));
		this.tracker = tracker;
	}


	/** Gets the current mixing trace since creation of the stream
	 **/
	public byte[] getTrace() { return trace; }

	/** Sets the current mixing trace
	 **/
	public void setTrace(byte[] trace) { this.trace = trace; }

	/** Inner-class for transparently unmixing an mixed stream
	 **/
	public class InputStream extends java.io.InputStream {
		private final java.io.InputStream delegee;
		private final java.io.DataInputStream trace;

		/** The number of plain bytes ready for reading **/
		protected long dataChunkLength;

		public InputStream(java.io.InputStream delegee) {
			this.delegee = delegee;
			this.trace = new DataInputStream( new ByteArrayInputStream(Mixer.this.trace) );
			this.dataChunkLength = 0;  // This forces unmixing at startup
		}

		public int read() throws IOException {
			if (dataChunkLength < 1)
				unmix();
			-- dataChunkLength;  // This is bug-prone if IO-failure and subsequent retry
			int read = delegee.read();
			tracker.update(read);
			return read;
		}

		/** Reads plain bytes of data.
		 ** Note that there might be less bytes actually read than requested.
		 ** @param data <CODE>byte[]</CODE> buffer to store read data into
		 ** @param begin offset of the buffer to start storing data at
		 ** @param length requested length of data to be read
		 **/
		public int read(byte[] data, int begin, int length) throws IOException {
			if (dataChunkLength < 1)
				unmix();
			final int actuallyRead = delegee.read(data, begin, (dataChunkLength < length) ? (int) dataChunkLength : length);
			dataChunkLength -= actuallyRead;
			tracker.update(data, begin, actuallyRead);
			return actuallyRead;
		}

		public void close() throws IOException {
			delegee.close();
		}

		/** Unmixes the stream so that next bytes are plain data bytes.
		 **/
		protected void unmix() throws IOException {
			try {
				UserInterface.get().log().log(5, Mixer.class, "Unmixing stream from its garbage data");
				readGarbageChunk(delegee, trace.readLong() );
			} finally {
				dataChunkLength = trace.readLong();
				try {
					tracker.load(trace);
				} catch (IOException e) {
					UserInterface.get().error().handleException("Error while checking block", e);
				}
				UserInterface.get().log().log(5, Mixer.class, "Ready for next " + dataChunkLength + " data bytes");
			}
		}

	}


	/** Inner-class for transparently mixing a stream
	 ** The trace is stored back in the outer <CODE>Mixer</CODE> instance
	 ** once the <CODE>close()</CODE> method is called
	 **/
	public class OutputStream extends java.io.OutputStream {
		private final ByteCountOutputStream delegee;         // Actual stream to write into
		private final java.io.ByteArrayOutputStream bytes;  // Mixing trace, in raw bytes form
		private final java.io.DataOutputStream trace;       // Mixing trace, in plain <CODE>long</CODE> form
		private long dataChunkLength;                       // Length of the currently being written data chunk

		public OutputStream(java.io.OutputStream delegee) {
			this.delegee = new ByteCountOutputStream(delegee);
			this.bytes = new ByteArrayOutputStream();
			this.trace = new DataOutputStream(bytes);
			this.dataChunkLength = 0;
		}

		public void write(int b) throws IOException {
			if (delegee.numberOfWrittenBytes() >= dataChunkLength)
				mix();
			delegee.write(b);
			tracker.update(b);
		}

		public void write(byte[] data, int begin, int length) throws IOException {
			while (length > 0) {
				long allowed = dataChunkLength - delegee.numberOfWrittenBytes();
				if (allowed <= 0) {
					mix();
					allowed = dataChunkLength;  // Since <CODE>delegee.numberOfWrittenBytes()</CODE> must equal zero here (not multi-threaded !)
				}
				if (allowed > length)
					allowed = length;
				delegee.write(data, begin, (int) allowed);  // No loss of precision here since (data.length - begin) fits into an <CODE>int</CODE>
				tracker.update(data, begin, (int) allowed);
				begin += allowed;
				length -= allowed;
			}
		}

		public void flush() throws IOException { delegee.flush();	}

		public void close() throws IOException {
			flush();
			try {
				mix();  // This might be buggy is called called several times
			} finally {
				Mixer.this.trace = bytes.toByteArray();  // Saves trace
				delegee.close();  // May throw 'IOException'
			}
		}

		protected void mix() throws IOException {
			UserInterface.get().log().log(5, Mixer.class, "Mixing stream with garbage data");
			delegee.reset();  // Resets the bytes count
			writeGarbageChunk(delegee);  // Abstract call
			final long garbageChunkLength = delegee.numberOfWrittenBytes();
			delegee.reset();
			dataChunkLength = nextDataChunkLength(blockSize);  // Abstract call
			trace.writeLong(garbageChunkLength);  // Saves garbage length first ...
			trace.writeLong(dataChunkLength);     // ... and then data length
			try {
				tracker.save(trace);
			} catch (IOException e) {
				UserInterface.get().error().handleException("Error while saving block summary", e);
			}
			UserInterface.get().log().log(5, Mixer.class, "Inserted " + garbageChunkLength
																		+ " garbage bytes, feeding back with " + dataChunkLength + " bytes of real data");
		}

	}

	/** Dummy mixer implementation.
	 ** This just does nothgin and serves as an example or for debugging purpose.
	 **/
	public static final Mixer DUMMY = new Mixer() {
		protected long getDataChunkLength(long blockSize) {	return blockSize;	}
		public void writeGarbageChunk(java.io.OutputStream delegee) throws IOException { }
		public void readGarbageChunk(java.io.InputStream stream, long length) throws IOException { stream.skip(length); }
		public String verbose() { return "NOT mixing data" ; }
	};

}


