
package sam.io;

import sam.entry.Entry;
import sam.io.*;
import sam.toolkit.*;
import sam.ui.UserInterface;

import java.io.*;
import java.util.*;
import java.util.zip.*;


/** Generic class for packing and unpacking data.
 **/
public abstract class Packer extends StreamFilter {

	/** The registry of all available packers
	 ** Note that in this case it is a registry of <CODE>Packer</CODE> <em>instances</em>
	 ** instead of <em>classes</em>.
	 **/
	private static final Map registry = new HashMap();

	/** The name of this packer **/
	public final String name;

	/** A human-readable description of this packer **/
	public final String description;

	/** Builds a new <CODE>Packer</CODE>
	 ** @param name the desired packer name
	 ** @param description a human-readable description
	 **/
	protected Packer(String name, String description) {
		this.name = name;
		this.description = description;
		registry.put(name, this);
	}

	/** Gets the packer of specified name.
	 ** @param name the name of the desired packer
	 ** @return <CODE>null</CODE> if no packer has specified name
	 **/
	public static Packer get(String name) {
		try {
			return (Packer) registry.get(name);
		} catch (ClassCastException e) {
			UserInterface.get().error().warn("BUG: inconsistant Packer registry: \"" + name + "\" !");
			return null;
		}
	}

	/** Gets the list of all registered packers. **/
	public static String[] getPackers() { return Text.getKeyStrings(registry); }

	/** Creates a mixer that is well-suited for this <CODE>Packer</CODE>
	 ** Note that this will be called at most <b>once</b> for each <CODE>Packer</CODE> instance.
	 **/
	protected abstract Mixer buildMixer();

	/** The singleton <CODE>Mixer</CODE> instance associated to this <CODE>Packer</CODE> **/
	private Mixer mixer = null;

	/** Gets the <CODE>Mixer</CODE> associated to this <CODE>Packer</CODE>
	 ** Note that the <CODE>Mixer</CODE> instance is always the same.
	 **/
	public final Mixer getMixer() {
		if (mixer == null)
			mixer = buildMixer();
		return mixer;
	}

	// Force automatic registration of some packers
	static {
		final Packer[] dummy = new Packer[] {
			new Packer("Store", "") {
	      public InputStream filterInput(InputStream input) { return input; }
	      public OutputStream filterOutput(OutputStream output) {	return output;	}
				protected Mixer buildMixer() { return Mixer.DUMMY; }
   		}, new Packer("Huffman", "using the 8-bits adaptive Huffman algorithm") {
				private final Huffman huffman = new Huffman8();
	      public InputStream filterInput(InputStream input) { return huffman.input(input); }
	      public OutputStream filterOutput(OutputStream output) {	return huffman.output(output);	}
				protected Mixer buildMixer() { return new HuffmanMixer(huffman); }
			}, new Packer("Mixer", "mixing data with random chunks") {
	      public InputStream filterInput(InputStream input) { return input; }
	      public OutputStream filterOutput(OutputStream output) {	return output;	}
				protected Mixer buildMixer() { return RandomMixer.DEFAULT; }
			}, new Packer("Huffman16", "using the 16-bits adaptive Huffman algorithm") {
				private final Huffman huffman = new Huffman16();
	      public InputStream filterInput(InputStream input) { return huffman.input(input); }
	      public OutputStream filterOutput(OutputStream output) {	return huffman.output(output);	}
				protected Mixer buildMixer() { return new HuffmanMixer(huffman); }
			}, new Packer("Huffman^2", "using the 8-bits adaptive Huffman algorithm *twice*") {
				private final Huffman huffman = new Huffman8();
 				public final InputStream filterInput(InputStream input) { return new Huffman8().input( huffman.input(input) ); }
	    	public final OutputStream filterOutput(OutputStream output) { return huffman.output( new Huffman8().output(output) ); }
				protected Mixer buildMixer() { return new HuffmanMixer(huffman); }
			}, new Packer("Deflate", "using the deflate algorithm") {
				public InputStream filterInput(InputStream input) throws IOException {	return new InflaterInputStream(input); }
				public OutputStream filterOutput(OutputStream output) throws IOException {	return new DeflaterOutputStream(output); }
  			public void finishFilter(OutputStream output) throws IOException {	((DeflaterOutputStream) output).finish();		}
				protected Mixer buildMixer() { return RandomMixer.DEFAULT; }
			}, new Packer("GZip", "using the GZip algorithm") {
			  public InputStream filterInput(InputStream input) throws IOException {	return new GZIPInputStream(input);	}
			  public OutputStream filterOutput(OutputStream output) throws IOException {return new GZIPOutputStream(output);	}
			  public void finishFilter(OutputStream output) throws IOException {	((GZIPOutputStream) output).finish();	}
				protected Mixer buildMixer() { return RandomMixer.DEFAULT; }
			}
		};
	}

}

