
package sam.io;

import sam.toolkit.Workshop;
import sam.toolkit.Text;
import sam.ui.UserInterface;
import sam.core.Settings;

import java.io.*;
import java.util.*;


/** Multi-threaded implementation of <CODE>InputStream</CODE> and <CODE>OutputStream</CODE>
 **/
/* Sucessfully tested on 20020121-18h45 by glacote
 */
public class ParallelIO {

		/** Exception thrown in case of errors in <CODE>ParallelIO</CODE> **/
		public static class Exception extends IOException {
				protected Exception() { this("Parallel I/O failed !"); }
				protected Exception(String message) { super(message); }
		}

		/** Maps <CODE>Thread</CODE>s to <CODE>IOProcessor</CODE>s **/
		private static Map PROCESSORS = new Hashtable();

		/** Registers specified <CODE>IOProcessor</CODE> as the current one **/
		public static synchronized void register(IOProcessor ioprocessor) throws Exception {
				final Thread currentThread = Thread.currentThread();
				if (Settings.Constants.PARALLELIO_PARANOIAC && (PROCESSORS.get(currentThread) != null) )
					throw new Exception("Current IOProcessor already registered !");
				PROCESSORS.put(currentThread, ioprocessor);
		}

		/** Gets the current <CODE>IOProcessor</CODE> **/
		public static synchronized IOProcessor getCurrentProcessor() throws Exception {
			if (PROCESSORS.isEmpty())
				return DUMMY_PROCESSOR;
			final Thread currentThread = Thread.currentThread();
			final Object currentIOProcessor = PROCESSORS.get(currentThread);
			if (Settings.Constants.PARALLELIO_PARANOIAC && ((currentIOProcessor == null) || (! (currentIOProcessor instanceof IOProcessor))) )
					throw new Exception("Current IOProcessor unregistered !");
			return (IOProcessor) currentIOProcessor;
		}

		/** The next expected token **/
		private long nextExpectedToken = 0;

		/** The number of <CODE>IOProcessor</CODE>s built inside this <CODE>ParallelIO</CODE> instance **/
		private int numberOfIOProcessors = 0; // -1 Because <CODE>DUMMY_PROCESSOR</CODE> increments it

		private static IOProcessor DUMMY_PROCESSOR = new ParallelIO().new IOProcessor() {
			public void requestToken() {}
			public void grantToken() {}
			public void process() {}
			public void run() {}
		};


		/** Waits for specified token be ready to be processed **/
		protected synchronized void requestToken(long token) throws Exception {
				if (Settings.Constants.PARALLELIO_PARANOIAC && (token < nextExpectedToken))
						throw new Exception("Attempt to re-use outdated token " + token + " at " + nextExpectedToken);
				try {
						while (token > nextExpectedToken)
								wait();
				} catch (InterruptedException e) {
						throw new Exception("Interrupted while awaiting token " + token + ": " + Text.getMessage(e) );
				}
		}

		protected synchronized void grantToken(long token) throws Exception {
				if (Settings.Constants.PARALLELIO_PARANOIAC && (token != nextExpectedToken))
						throw new Exception("Attempt to grant token " + token + " at " + nextExpectedToken);
				++ nextExpectedToken;
				notifyAll();  // It is critical to <CODE>notifyAll()</CODE> and not only to <CODE>notify()</CODE>
		}

		/** Abstraction for processors which actually <em>do</em> the job	 **/
		protected abstract class IOProcessor implements Runnable {
				private long token;

				protected IOProcessor() {
					synchronized(ParallelIO.this) {
						this.token = numberOfIOProcessors ++; 
						UserInterface.get().log().log(5, ParallelIO.class, "Built IOProcessor(" + token + "," + numberOfIOProcessors + ")");
					}
				}

				public void run() {
					try {
						register(this);  // Registers the threads which actually <b>runs</b> the <CODE>IOProcessor</CODE>, not the one which built it
						process();
					} catch (Exception e) {
						UserInterface.get().error().handleException("Multi-threading exception occured in " + Thread.currentThread(), e);
						// Here a multi-threading exception occured
					} catch (Throwable t) {
						UserInterface.get().error().fail("Exception occured while processing: " + Text.getMessage(t) );
					}
				}

				/** Here the real job **/
				abstract protected void process() throws Throwable;

				/** Requests this processor's token be allowed to be processed **/
				public synchronized void requestToken() throws Exception { ParallelIO.this.requestToken(token); }

				/** Grants this processor's token having been processed **/
				public synchronized void grantToken() throws Exception {
					synchronized(ParallelIO.this) {
						ParallelIO.this.grantToken(token);
						token += numberOfIOProcessors;
					}
				}
		}

		/** Splits <CODE>n</CODE> bytes into <CODE>p</CODE> parts
		 ** @return the part size
		 **/
		public static int split(int n, int p) throws Exception {
				if (n < 1) return 0;
				if (p < 1) throw new Exception("Can not split " + n + " into " + p + " parts");
				return (n % p == 0) ? n / p : (n % (p - 1) != 0) ? n / (p - 1) : (n / (p - 1) - p) ;
		}

		/** Guess what ... **/
		private static int min(int a, int b) { return (a < b) ? a : b; }


		/** Class for scattering buffered writes accross multiple threads.
		 ** @see sam.io.ParallelIO.GatherOutputStream
		 **/
		public class ScatterOutputStream extends java.io.OutputStream {
				private final OutputStream delegee;
				private final Workshop workshop;

				public ScatterOutputStream(OutputStream delegee, Workshop workshop) { this.delegee = delegee; this.workshop = workshop; }

				/** Simply forwards the flushing call to the underlying stream **/
				public void flush() throws IOException { delegee.flush(); }

				/** Simply forwards the closing call to the underlying stream **/
				public void close() throws IOException { delegee.close(); }

				/** Simply writes the single byte on the underlying stream **/
				public void write(int b) throws IOException { delegee.write(b); }

				/** Writes data using multiple threads to the underlying stream **/
				public void write(final byte[] data, int begin, int length) throws IOException {
						int p = workshop.numberOfWorkers();
						int m = split(length, p);
						for (int offset = begin; offset < begin+length; ) {
								final int chunkBegin = offset;
								final int chunkLength = min(m, length - (offset - begin) );
								workshop.startAsSoonAsPossible(new IOProcessor() {
												protected void process() throws IOException { delegee.write(data, chunkBegin, chunkLength);	}
								} );
								offset += chunkLength;
						}
						workshop.waitForCompletion(); // Here we do <b>not</b> allow to start threads on a later call as long as this one is not finished
				}

		}

		/** Class for gathring buffered writes from multiple threads.
		 ** @see sam.io.ParallelIO.ScatterOutputStream
		 **/
		public class GatherOutputStream extends java.io.OutputStream {
				private final OutputStream delegee;

				public GatherOutputStream(OutputStream delegee) { this.delegee = delegee; }

				/** Simply forwards the flushing call to the underlying stream **/
				public void flush() throws IOException { delegee.flush(); }

				/** Simply forwards the closing call to the underlying stream **/
				public void close() throws IOException { delegee.close(); }

				/** Simply writes the single byte on the underlying stream **/
				public void write(int b) throws IOException { delegee.write(b); }

				/** Gathers data written by possibly multiple threads (and thus out-of-order) to the underlying stream **/
				public void write(byte[] data, int begin, int length) throws IOException {
						IOProcessor currentProcessor = getCurrentProcessor();
						currentProcessor.requestToken();  // This is to restore ...
						delegee.write(data, begin, length);
						currentProcessor.grantToken();   // ... the sequentiality of writes
				}

		}


		/** Class for scattering buffered reads accross multiple threads.
		 ** @see sam.io.ParallelIO.GatherInputStream
		 **/
		public class ScatterInputStream extends java.io.InputStream {
				private final InputStream delegee;
				private Workshop workshop;

				public ScatterInputStream(InputStream delegee, Workshop workshop) { this.delegee = delegee; this.workshop = workshop; }

				/** Simply forwards the closing call to the underlying stream **/
				public void close() throws IOException { delegee.close(); }

				/** Simply reads one byte from the underlying stream **/
				public int read() throws IOException { return delegee.read(); }

				/** Scatters reads on multiple threads from the underlying stream **/
				public int read(final byte[] data, int begin, int length) throws IOException {
					int p = workshop.numberOfWorkers();
					int m = split(length, p);
					/** Used to track back the number of bytes read accross several threads **/
					class NumberOfBytesTracker {
						private int length = 0;
						protected synchronized int length() { return length; }
						protected synchronized void add(int n) {
							if (n > 0)  // Do <b>not</b> take negative lengthes into account !
								length += n;
						}
					}
					final NumberOfBytesTracker tracker = new NumberOfBytesTracker();
					for (int offset = begin; offset < begin+length; ) {
						final int chunkBegin = offset;
						final int chunkLength = min(m, length - (offset - begin) );
						workshop.startAsSoonAsPossible(new IOProcessor() {
							protected void process() throws IOException {
								tracker.add( delegee.read(data, chunkBegin, chunkLength) );
							}
						} );
						offset += chunkLength;
					}
					workshop.waitForCompletion(); // Here we do <b>not</b> allow to start threads on a later call as long as this one is not finished
					return tracker.length();  // Which hopefully matches <CODE>length</CODE>
				}

		}

		/** Class for gathring buffered writes from multiple threads.
		 ** @see sam.io.ParallelIO.ScatterOutputStream
		 **/
		public class GatherInputStream extends java.io.InputStream {
				private InputStream delegee;

				public GatherInputStream(InputStream delegee) { this.delegee = delegee; }
				
				/** Simply forwards the closing call to the underlying stream **/
				public void close() throws IOException { delegee.close(); }

				/** Simply reads one byte from the underlying stream **/
				public int read() throws IOException { return delegee.read(); }

				/** Scatters reads on multiple threads from the underlying stream **/
				public int read(byte[] data, int begin, int length) throws IOException {
					final IOProcessor currentProcessor = getCurrentProcessor();
					currentProcessor.requestToken();  // This is to restore ...
					final int realLength = delegee.read(data, begin, length);
					currentProcessor.grantToken();   // ... the sequentiality of writes
					return realLength;
				}

		}

	public static void main(String[] args) {
		if (args.length != 4) {
			System.err.println("Syntax: java " + ParallelIO.class.getName() + " n p input output");
			System.exit(1);
		}
		try {
			java.io.InputStream input = new FileInputStream(args[2]);
			java.io.OutputStream output = new FileOutputStream(args[3]);
			final int n = parseNumberOfThreads(args[0]);
			final int p = parseNumberOfThreads(args[1]);
			System.err.print("Dumping \"" + args[2] + "\" to \"" + args[3] + "\" with " + n + " reading and " + p + " writing threads");
			if (n > 1) {
				ParallelIO parallelIO = new ParallelIO();
				input = parallelIO.new GatherInputStream(input);
				input = parallelIO.new ScatterInputStream(input, new Workshop(n));
			}
			if (p > 1) {
				ParallelIO parallelIO = new ParallelIO();
				output = parallelIO.new GatherOutputStream(output);
				output = parallelIO.new ScatterOutputStream(output, new Workshop(p));
			}
			final long startTime = System.currentTimeMillis();
			try {
				IO.bufferedDump(input, output);
			} finally {
				final long endTime = System.currentTimeMillis();
				output.close();
				input.close();
				System.err.println(" done in " + (endTime - startTime) + " ms.");
			}
			System.exit(0);
		} catch (java.lang.Exception e) {
			System.err.println("Exception occured: " + Text.getMessage(e));
			e.printStackTrace();
			System.exit(2);
		}
	}

	private static int parseNumberOfThreads(String s) {
		try {
			int n = Integer.valueOf(s).intValue();
			if (n < 1)
				throw new NumberFormatException();
			return n;
		} catch (NumberFormatException e) {
			System.err.println("Invalid number of threads " + s);
			System.exit(1);
			return 0; // Fake for compiler
		}
	}

}
