
package sam.io;

import sam.ui.UserInterface;
import sam.ui.RulingInterface;
import sam.core.Settings;
import sam.toolkit.Random;

import java.io.*;


/** Standard output trace.
 ** This is a simple encapsulation of a named <CODE>PrintWriter</CODE>
 ** whose initialization is lazy.
 **/
public class PrintTrace {

	/** Name of this trace, e.g. the underlying file name **/
	public final String name;

	/** Type of this trace  **/
	public final String type;

	/** The output to print to. **/
	private PrintWriter output;

	/** The underlying file to write to, if any **/
	private final File file;

	private PrintTrace(String type, File file) {
		this.name = file.getPath();
		this.type = type;
		this.file = file;
		this.output = output;
	}

	private PrintTrace(String name, String type, OutputStream output) {
		this.name = name;
		this.type = type;
		this.file = null;
		this.output = new PrintWriter(output);
	}

	public String toString() {
		return name;
	}

	/** Creates an adequate printable trace.
	 ** @param trace either <CODE>"stdout"</CODE>, <CODE>"stderr"</CODE> or any valid writable file name
	 ** @param type a single-word description of the trace type, e.g. "log" or "skipped_entries"
	 ** Additionnally blank specification is an alias for <CODE>Settings.Default.TRACE(type)</CODE>
	 ** @return a lazy initialization <CODE>PrintTrace</CODE> or <CODE>null</CODE> if error
	 **/
	public static PrintTrace make(String trace, String type) {
		if (trace == null)
			return null;
		trace = trace.trim();
		if ("".equals(trace))
			trace = Settings.Constants.DEFAULT_TRACE;
		if (trace.equalsIgnoreCase(Settings.Constants.STDOUT_NAME))
			return new PrintTrace("standard output", type, System.out);
		else if (trace.equalsIgnoreCase(Settings.Constants.STDERR_NAME))
			return new PrintTrace("standard error", type, System.err);
		else {
			File file = UserInterface.get().ruling().getFileOrDirectory(type, trace, RulingInterface.WRITABLE_FILE);
			return (file == null) ? null : new PrintTrace(type, file);
		}
	}

	/** Draws an example from given file name and standard streams. **/
	public static String example(String example) {
		return Random.drawString( new String[] {Settings.Constants.STDOUT_NAME, Settings.Constants.STDERR_NAME, example} );
	}

	/** Gets the output writer of this trace.
	 ** Note that its initialization is delayed to the first call of this method,
	 ** which avoids creating log files that will never be used.
	 **/
	public PrintWriter out() {
		if (output == null)
			try {
				output = (file == null) ? null : new PrintWriter( new FileWriter(file, true) );
				// <CODE>true</CODE means append new data instead of overwriting
			} catch (IOException e) {  // Should not happen actually since <CODE>file</CODE> was said to be writable
				UserInterface.get().error().handleException("Fatal error while opening " + type, e);
			}
		return output;
	}

}


