
package sam.io;

import sam.ui.UserInterface;

import java.io.*;
import java.awt.*;
import java.awt.event.*;


/** Basic pseudo-random implementation of a mixer.
 **/
public class RandomMixer extends Mixer {

		/** Immutable default value, defined as <CODE>new RandomMixer(0.05, 0.05, 0.50, 0.50)</CODE> */
		public static final RandomMixer DEFAULT = new RandomMixer();

		/** Actual density factor */
		public final float density;
		/** Actual frequency factor */
		public final float frequency;
		/** Actual variety factor */
		public final float variety;
		/** Actual redundancy factor */
		public final float redundancy;

		private final float randomLength;
		private final float mainLength;
		private final byte[] bytes;
		private int bytesCount;

		/** Default constructor.
		 ** @see DEFAULT
		 **/
		private RandomMixer() {
			this.frequency  = (float) 0.05;
			this.density    = (float) 0.05;
			this.variety    = (float) 0.5;
			this.redundancy = (float) 0.5;
			this.randomLength = (float) density * frequency;
			this.mainLength   = (float) (1.0 / density - 1.0) * randomLength;
			this.bytes = new byte[256];
			this.bytesCount = 0;
		}


		public RandomMixer(String frequency) throws NumberFormatException {
			this( parsePercent(frequency), DEFAULT.density, DEFAULT.variety, DEFAULT.redundancy );
		}

		public RandomMixer(String frequency, String density) throws NumberFormatException {
			this( parsePercent(frequency), parsePercent(density), DEFAULT.variety, DEFAULT.redundancy );
		}

		public RandomMixer(String frequency, String density, String variety) throws NumberFormatException {
			this( parsePercent(frequency), parsePercent(density), parsePercent(variety), DEFAULT.redundancy );
		}

		/**
		 Builds a new <CODE>RandomMixer</CODE> instance, with values as <CODE>String</CODE> coding integers in the
	   range <CODE>1..99</CODE>
		*/
		public RandomMixer(String frequency, String density, String variety, String redundancy) throws NumberFormatException {
			this( parsePercent(frequency), parsePercent(density), parsePercent(variety), parsePercent(redundancy) );
		}

		/**
		 Builds a new <CODE>RandomMixer</CODE> instance, with values as <CODE>float</CODE> in the range <CODE>0.01..0.99</CODE>
		*/
		public RandomMixer(float frequency, float density, float variety, float redundancy) throws NumberFormatException {
			this.frequency  = checkRange(frequency, -0.01, 99.0, "frequency" );
			this.density    = checkRange(density, 0.01, 0.99, "density" );
			this.variety    = checkRange(variety, 0.01, 1.0, "variety" );
			this.redundancy = checkRange(redundancy, 0.01, 1.0, "redundancy" );
			this.randomLength = (float) density * frequency;
			this.mainLength   = (float) (1.0 / density - 1.0) * randomLength;
			this.bytes = new byte[256];
			this.bytesCount = 0;
			UserInterface.get().log().log(4, RandomMixer.class, "Initializing RandomMixer(" + frequency +
																		"," + density + "," + variety + "," + redundancy + ") mixer");
		}

		public String toString() {
			return "RandomMixer(" + frequency +  "," + density + "," + variety + "," + redundancy + ")";
		}

  	public String verbose() {
			return new StringBuffer().append("interleaving data ").append( factored(frequency, 100) )
				.append("% of time with random data chunks each of length ").append( factored(randomLength, 100) )
				.append("% of the total length, by ").append( factored(redundancy, 100) )
				.append("% self-redundant and composed of bytes within ").append( factored(variety, 256) )
				.append(" randomly choosen bytes").toString();
		}

		private static float parsePercent(String value) throws NumberFormatException{
			return (float) ( Integer.valueOf(value).intValue() / 100.0);
		}

		private static float checkRange(float value, double min, double max, String name) throws NumberFormatException {
			if ( (value < min) || (value > max))
				throw new NumberFormatException(name + " " + value + " not in range [" + min + "," + max + "]");
			return value;
		}

		public static int factored(float value, int factor) {
			return (int) Math.ceil(value * factor) % (factor + 1);
		}

		protected long getDataChunkLength(long blockSize) {
			return sam.toolkit.Random.nextLong( (long) Math.ceil(mainLength * blockSize) );
		}

		public void writeGarbageChunk(java.io.OutputStream delegee) throws IOException {
			final long length = sam.toolkit.Random.nextLong( (long) Math.ceil(randomLength * blockSize()) );
			int bytesCount = (int) Math.ceil(variety * 256);  // Assumes 'bytesCount >= 2', i.e. 'variety >= 0.01'
			if (bytesCount < 2)
				bytesCount = 2;
			bytesCount = sam.toolkit.Random.nextInt(bytesCount);
			if (bytesCount <= 0)
				bytesCount = 1;
			final byte[] bytes = new byte[bytesCount];
			for (int i=0; i<bytesCount; )
				bytes[i++] = sam.toolkit.Random.nextByte();  // Possibly several times the same byte in the 'bytes' array
			long written = 0;
			for ( ; written < randomLength; )  {
				final byte nextRandomValue = bytes[ sam.toolkit.Random.nextInt(bytesCount) ];
				final long repeat = sam.toolkit.Random.nextLong( (long) Math.ceil(redundancy * length) );
				for (int i=0; i<repeat; ++i)
					delegee.write(nextRandomValue);  // may throw 'IOException'
				written += repeat;
			}
		}


		public void readGarbageChunk(java.io.InputStream stream, long length) throws IOException {
			UserInterface.get().log().log(6, RandomMixer.class, "Skipping " + length + " garbage bytes");
			stream.skip(length);
		}

}



