
package sam.io;

import sam.toolkit.*;
import sam.ui.UserInterface;

import java.io.*;
import java.util.*;


/** This class allows customizable mixing tracking.
 ** This is a block-oriented trace that can be used to
 ** record checksums or redundancy data.
 **/
public abstract class Tracker {

	/** The registry of all available trackers
	 ** Note that in this case it is a registry of <CODE>Tracker</CODE> <em>instances</em>
	 ** instead of <em>classes</em>.
	 **/
	private static final Map registry = new HashMap();

	/** The name of this tracker **/
	public final String name;

	/** A human-readable description of this tracker **/
	public final String description;

	/** Builds a new <CODE>Tracker</CODE>
	 ** @param name the desired packer name
	 ** @param description a human-readable description
	 **/
	protected Tracker(String name, String description) {
		this.name = name;
		this.description = description;
		registry.put(name, this);
	}

	/** Gets the tracker of specified name.
	 ** @param name the name of the desired tracker
	 ** @return <CODE>null</CODE> if no tracker has specified name
	 **/
	public static Tracker get(String name) {
		try {
			return (Tracker) registry.get(name);
		} catch (ClassCastException e) {
			UserInterface.get().error().warn("BUG: inconsistant Tracker registry: \"" + name + "\" !");
			return null;
		}
	}

	/** Gets the list of all registered trackers. **/
	public static String[] getTrackers() { return Text.getKeyStrings(registry); }

	/** Updates this tracker state with specified plain data
	 ** <b>Warning</b> This method is not supposed to alter the supplied data
	 ** although for efficiency reason it is still not prevented from doing it.
	 ** @param data the plain data to be considered
	 ** @param begin the beginning of the portion of data considered
	 ** @param length the length of the portion of data considered
	 **/
	public abstract void update(byte[] data, int begin, int length);

	/** Updates this tracker state with specified plain data
	 ** @param data the plain data to be considered
	 **/
	public void update(byte[] data) { if (data != null)  update(data, 0, data.length); }

	/** Updates this tracker state with one byte of plain data
	 ** @param data the plain data to be considered
	 **/
	public void update(int data) { update(new byte[] { (byte) (data & 0xff) }); }

	/** Saves this tracker state **/
	protected abstract void save(DataOutputStream output) throws IOException;

	/** Reloads this tracker state **/
	protected abstract void load(DataInputStream input) throws IOException;

	// Force automatic registration of some trackers
	static {
		final Tracker[] dummy = new Tracker[] {
			new Tracker("None", "empty tracker that keeps back no information at all") {
				public void update(byte[] data, int begin, int length) { }
				protected void load(DataInputStream input) { }
				protected void save(DataOutputStream output) { }
			},
			new ChecksumTracker()
		};
	}

}


