
package sam.rules;

import sam.ui.UserInterface;

import java.lang.reflect.*;


/**
 ** Wrapper for java command-lines
 **/
public class JavaProgram {
	private String className = null;
	private final String[][] args;

	/** Creates a new Java command wrapper.
	 ** @param classPathAndClass class file name and path. For example <CODE>"./path1/path2/com.package.class"</CODE>
	 ** will load class named <CODE>class</CODE> from package <CODE>com.package</CODE> searching binaries from base directory
	 ** <CODE>./path1/path2/</CODE
	 ** Note that the class binary file is allowed to be created during the extraction process.
	 **/
	JavaProgram(String classPathAndClass, String[] args) throws SyntaxError {
		int cut = classPathAndClass.lastIndexOf('/');
		if (cut > 0) {
			sam.toolkit.Process.appendClassPath( classPathAndClass.substring(0, cut) );
			className = classPathAndClass.substring(cut + 1);
		} else
			className = classPathAndClass;
		this.args = new String[][] { args };
		}


	/**
	 **  Runs the application in the current JVM.
	 ** @return exit code of the application, or <CODE>0</CODE> if none.
	 ** @throw Throwable, any pending exception thrown by the application.
	 ** Failure to launch application does <b>not</b> involve throwing an exception.
	 **/
	public int execute() throws SyntaxError {
		// 1) Loads the <CODE>main(String[])</CODE> method
		Method main = null;
		try {
			main = Class.forName(className).getMethod("main", new Class[] { String[].class } );
		} catch (ClassNotFoundException e) {
			throw new SyntaxError("Could not load class \"" + className + "\": " + sam.toolkit.Text.getMessage(e) );
		} catch (NoSuchMethodException e) {
			throw new SyntaxError("Could not find \"main(String[])\" method from class \"" + className + "\": " +
														sam.toolkit.Text.getMessage(e) );
		} catch (SecurityException e) {
			throw new SyntaxError("Does not have the right to access \"main(String[])\" method from class \"" + className + "\": " +
														sam.toolkit.Text.getMessage(e) );
		}
		// 2) Executes it
		Object result = null;
		try {
			result = main.invoke(null, args);  // <CODE>null</CODE> since <CODE>main</CODE> is a static method
		} catch (IllegalAccessException e) {
			throw new SyntaxError("Attempt to make illegal access \"main(String[])\" method from class \"" + className + "\": " +
														sam.toolkit.Text.getMessage(e) );
		} catch (InvocationTargetException e) {
			UserInterface.get().log().log(1, JavaProgram.class, "SAM was successfull but launched application FAILED in turn: " +
																		sam.toolkit.Text.getMessage(e.getTargetException()) );
		}
		UserInterface.get().log().log(1, JavaProgram.class, "Application completed sucessfully.");
		// 3) Returns its exit code
		if ((result != null) && (result instanceof Number)) {
			int returnCode = ((Number)result).intValue();
			UserInterface.get().log().log(1, JavaProgram.class, "Return code: " + returnCode +
																		( (returnCode > 0) ? " (failure)" : " (success)" ));
			return returnCode;
		} else  // <CODE>main</CODE> did not return an integer !!
			throw new SyntaxError("WARNING: " + main + " returned " +
														((result == null) ? ("null") : ("instance of " + result.getClass().getName())) +
														" instead of an integer as expected");
	}

	/** Makes a command-line like representation of this application and parameters. **/
	public String toString() { return "JavaProgram(" + toSwitch() + ")"; }

	/** Makes a command-line like representation of this application and parameters.
	 **/
	public String toSwitch() { return className + sam.toolkit.Text.toString(args[0], " ", " ",""); }

}


