
package sam.rules;

import sam.cipher.Specification;
import sam.cipher.Cipher;
import sam.ui.UserInterface;
import sam.ui.RulingInterface;
import sam.io.IO;
import sam.io.Packer;
import sam.io.Mixer;
import sam.io.Tracker;
import sam.core.Settings;
import sam.toolkit.Text;
import sam.event.EventNotifier;
import sam.event.FileEventNotifier;
import sam.event.FileEventListener;
import sam.event.ProgressEventNotifier;
import sam.event.ProgressEventListener;
import sam.entry.Entry;

import java.util.Stack;
import java.util.jar.JarOutputStream;
import java.io.*;


/** Class describing what has to be done and how to do it.
 **/
public class ModusOperandi {

	/** Class defining which mode of operation is required:
	 ** typically one of "archiving", "extracting", etc.
	 **/
	public static abstract class Opus {
		public final String name;
		protected Opus(String name) { this.name = name; }
		abstract protected JarOutputStream process(ModusOperandi modusOperandi);
	}

	/** Archiving mode of operation.
	 ** This is the default mode, which consists in storing files into a new archive.
	 **/
	public static final Opus ARCHIVING = new sam.core.Storer("archiving") {
		protected void storeEntry(InputStream source, OutputStream archive, Entry entry) throws IOException {
			IO.bufferedDump(source, archive, entry.length);
		}
	};

	/** Extracting mode of operation.
	 ** Existing local files might be overwritten.
	 ** @see Settings.Constants#ASK_BEFORE_OVERWRITING_LOCAL_FILES
	 **/
	public static final Opus EXTRACTING = new sam.core.Unstorer("extracting") {
		protected void unstoreEntry(InputStream archive, Entry entry, File actualDirectory, FileEventNotifier fileEventNotifier)
		throws IOException {
			IO.bufferedDump(archive, entry.write(fileEventNotifier, actualDirectory), entry.length);
		}
	};

	/** Listing mode of operation.
	 ** Only lists the entries contained in the archive.
	 ** @see sam.entry.Entry#list
	 **/
	public static final Opus LISTING = new sam.core.Unstorer("listing") {
		protected void unstoreEntry(InputStream archive, Entry entry, File actualDirectory, FileEventNotifier fileEventNotifier) { }
		protected boolean notifyProgress() { return false; }
	};

	/** Freshening mode of operation.
	 ** Entries which are <em>newer</em> than local files are extracted.
	 **/
	public static final Opus FRESHENING = new sam.core.Unstorer("freshening") {
		protected void unstoreEntry(InputStream archive, Entry entry, File actualDirectory, FileEventNotifier fileEventNotifier)
		throws IOException {
			IO.bufferedDump(archive, entry.freshen(fileEventNotifier, actualDirectory), entry.length);
		}
	};

	/** Overwriting mode of operation.
	 ** Entries which are <em>newer</em> than local files are extracted.
	 **/
	public static final Opus OVERWRITING = new sam.core.Unstorer("overwriting") {
		protected void unstoreEntry(InputStream archive, Entry entry, File actualDirectory, FileEventNotifier fileEventNotifier)
		throws IOException {
			IO.bufferedDump(archive, entry.overwrite(fileEventNotifier, actualDirectory), entry.length);
		}
	};

	/** Comparing mode of operation.
	 ** Entries stored in the archive are compared against local files.
	 ** No file is actually extracted onto disk.
	 **/
	public static final Opus COMPARING = new sam.core.Unstorer("comparing") {
		protected void unstoreEntry(InputStream archive, Entry entry, File actualDirectory, FileEventNotifier fileEventNotifier)
		throws IOException {
			IO.bufferedDump(archive, entry.compare(fileEventNotifier, actualDirectory), entry.length);
		}
	};

	/** Benchmarking mode of operation.
	 ** No archive is actually created.
	 **/
	public static final Opus BENCHMARKING = new sam.core.Storer("benchmarking") {
		/** No actual byte will ever be written on disk **/
		protected JarOutputStream createArchive(File archive) {
			try {
				return new JarOutputStream(IO.NULL_OUTPUT);
			} catch (IOException e) {	return null; }
		}
		protected void storeEntry(InputStream source, OutputStream archive, Entry entry) throws IOException {
			IO.bufferedDump(source, archive, entry.length);
		}
	};

	/** Probing mode of operation.
	 ** This checks wether an archive is not damaged nor broken.
	 ** No file is actually extracted onto disk, but the block trackers are probed.
	 **/
	public static final Opus CHECKING = new sam.core.Unstorer("checking") {
		protected void unstoreEntry(InputStream archive, Entry entry, File actualDirectory, FileEventNotifier fileEventNotifier)
		throws IOException {
			IO.bufferedDump(archive, entry.skip(fileEventNotifier, actualDirectory), entry.length);
		}
	};

	/** Helping mode of operation.
	 ** This prints help.
	 **/
	public static final Opus HELPING = new Opus("helping") {
		protected java.util.jar.JarOutputStream process(ModusOperandi modusOperandi) {
			sam.ui.UserInterface.get().running().print("Syntax", sam.rules.SyntaxSet.get().help() );
			return null;
		}
	};

	/** Describing mode of operation.
	 ** This prints information on the considered archive.
	 **/
	public static final Opus DESCRIBING = new Opus("describing") {
		protected java.util.jar.JarOutputStream process(ModusOperandi modusOperandi) {
			sam.ui.UserInterface.get().running().print("Content", modusOperandi.infos() );
			return null;
		}
	};

	/** Processes specified rule set **/
	public void process(RuleSet ruleSet) {
		JarOutputStream output = null;
		try {
			ruleSet.setTo(this);
			UserInterface.get().log().log(1, ModusOperandi.class, "Entering " + opus.name + " mode");
			output = opus.process(this);
		} finally {
			notifier().progress().terminate();
		}
		if (output != null) {
			ruleSet.saveTo(output);
			try {
				UserInterface.get().log().log(1, ModusOperandi.class, "Closing archive");
				output.finish();  // Writes the Zip central directory
				output.close();
			} catch (IOException e) {
				UserInterface.get().error().handleException("Could not close the archive properly", e);
			}
		}
	}

	/** The mode of operation of the application
	 ** @see sam.ruleSet.ModusOperandi.Opus
	 ** @see #ARCHIVING
	 ** @see #EXTRACTING
	 **/
	private Opus opus = ARCHIVING;

	/** @return the program is run in archiving mode **/
	public boolean isArchiving() { return opus == ARCHIVING; }
	/** @return the program is run in extracing mode **/
	public boolean isExtracting() { return ! isArchiving(); }

	/** @return wether this is a special opus (i.e. neither archiving nor extracting) **/
	public boolean isSpecial() { return (opus == HELPING) || (opus == DESCRIBING); }

	/** Sets the mode of operation of the application **/
	void setOpus(Opus opus) { this.opus = opus; }

	/** Sets wether the program is run in archiving opus **/
	void setArchivingMode(boolean isArchiving) { opus = isArchiving ? ARCHIVING : EXTRACTING; }

	/** Sets wether the program is run in archiving mode
	 ** @param archive the archive file to be processed.
	 ** The mode defaults to archiving if file does not exist, and to extracting otherwise.
	 **/
	void setArchivingMode(File archive) { setArchivingMode( archive.exists() ); }

	/** The event dispatcher for this modus operandi. **/
	private EventNotifier notifier = new EventNotifier(new FileEventNotifier(), new ProgressEventNotifier());

	/** Registers a new <CODE>ProgressEvent</CODE> listener **/
	void register(FileEventListener listener) {
		notifier.file().register(listener);
	}

	/** Registers a new <CODE>ProgressEvent</CODE> listener **/
	void register(ProgressEventListener listener) {
		notifier.progress().register(listener);
	}

	/** @return the event dispatcher for this modus operandi. **/
	public EventNotifier notifier() { return notifier; }

	/** The archive to be written to or read from **/
	private File archive = null;

	// Note that the <CODE>archive</CODE> field is set through the <CODE>addFile</CODE> method below

	/** Sets the archive to be written to or read from.
	 **/
	void setArchive(File archive) {
		this.archive = archive;
	}

	/** @return the archive to be written to or read from **/
	public File archive() {
		if (archive == null) {
			if ( isArchiving() ) {
				archive = sam.ui.UserInterface.get().ruling().getFileOrDirectory("archive", new File("MyArchive.jar"), RulingInterface.FILE);
				setArchivingMode(archive);
			} else
				archive = sam.toolkit.Process.getJarSourceArchive(Settings.Environment.BOOT_CLASS);
		}
		return archive;
	}

	/** The output directory to be written to or read from **/
	private File outputDirectory = null;

	/** Sets the output directory to be written to or read from.
	 **/
	void setOutputDirectory(File outputDirectory) {
		this.outputDirectory = outputDirectory;
	}

	/** @return the output directory to be written to or read from.
	 **/
	public File outputDirectory() { return outputDirectory; }

	/** The imported archive file for differential archiving **/
	private File importedArchive = null;
	/** @return the imported archive file for differential archiving **/
	public File importedArchive() { return importedArchive; }
	/** @return wether this archive is part of a multiple archive differential backup **/
	public boolean hasImportedArchive() { return importedArchive != null; }
	/** Sets the imported archive file for differential archiving **/
	void setImportedArchive(File importedArchive) {
		this.importedArchive = importedArchive;
		this.importedFiles = sam.core.ArchiveLayout.loadFullCatalog(importedArchive).statify();
	}

	/** The imported catalog for differential archiving
	 ** Automatically set through the <CODE>setImportedArchive</CODE> method
	 **/
	private sam.entry.List importedFiles = null;
	/** @return the imported catalog for differential archiving **/
	public sam.entry.List importedFiles() { return importedFiles; };

	/** The directory were temporary files are created **/
	private File temporaryDirectory = null;
	/** @return directory were temporary files are created **/
	public File temporaryDirectory() { return temporaryDirectory; }
	/** Sets directory were temporary files are created **/
	void setTemporaryDirectory(File temporaryDirectory) { this.temporaryDirectory = temporaryDirectory; }

	/** The packer used for (de)compression **/
	private Packer packer = null;

	/** @return the packer used for (de)compression
	 ** Note that the returned packer is never <CODE>null</CODE>
	 **/
	public Packer packer() {
		if (packer == null)
			packer = Settings.Defaults.PACKER;
		return packer;
	}

	/** Sets the packer used for (de)compression and the mixer associated to it **/
	void setPacker(Packer packer) { this.packer = packer; }

	/** The tracker used for (de)compression **/
	private Tracker tracker = null;

	/** @return the tracker used for (de)compression
	 ** Note that the returned tracker is never <CODE>null</CODE>
	 **/
	public Tracker tracker() {
		if (tracker == null)
			tracker = Settings.Defaults.TRACKER;
		return tracker;
	}

	/** Sets the tracker used for (de)compression and the mixer associated to it **/
	void setTracker(Tracker tracker) { this.tracker = tracker; }

	/** The block size to be used (in <b>bytes</b>) **/
	private int blockSize = Settings.Defaults.BLOCK_SIZE;
	/** @return the block size to be used (in <b>bytes</b>) **/
	public int blockSize() { return blockSize; }
	/** Sets the block size to be used (in <b>bytes</b>) **/
	void setBlockSize(int blockSize) { this.blockSize = blockSize; }

	/** The number of vertical threads **/
	private int verticalThreads = Settings.Defaults.VERTICAL_THREADS;
	/** @return the number of vertical threads **/
	public int verticalThreads() { return verticalThreads; }
	/** Sets the number of vertical threads **/
	void setVerticalThreads(int verticalThreads) { this.verticalThreads = verticalThreads; }

	/** The number of horizontal threads **/
	private int horizontalThreads = Settings.Defaults.HORIZONTAL_THREADS;
	/** @return the number of horizontal threads **/
	public int horizontalThreads() { return horizontalThreads; }
	/** Sets the number of horizontal threads **/
	void setHorizontalThreads(int horizontalThreads) { this.horizontalThreads = horizontalThreads; }

	/** The archive comment/description (stored in decrypted text) **/
	private String comment = Settings.Defaults.COMMENT;
	/** @return the archive comment/description (stored in decrypted text) **/
	public String comment()  { return comment; }
	/** @return the archive comment/description (stored in decrypted text) **/
	void setComment(String comment)  { this.comment = comment; }

	/** The version number of the program which created the application (in extracting mode) **/
	private int archiveSamVersion = 0;
	/** Sets the version number of the program which created the application (in extracting mode) **/
	void setArchiveSamVersion(int archiveSamVersion) { this.archiveSamVersion = archiveSamVersion; }
	/** Checks wether the data stored in the archive can be processed by this version of the application **/
	public boolean checkSamVersion() { return archiveSamVersion <= Settings.Environment.PROGRAM_VERSION; }

	/** The version number of the layout of the archive (in extracting mode) **/
	private int archiveLayoutVersion = 0;
	/** Sets the version number of the layout of the archive (in extracting mode) **/
	void setArchiveLayoutVersion(int archiveLayoutVersion) { this.archiveLayoutVersion = archiveLayoutVersion; }
	/** Checks wether the layout of the archive can be handled by this version of the application **/
	public boolean checkLayoutVersion() { return archiveLayoutVersion == sam.core.ArchiveLayout.LAYOUT_VERSION_NUMBER;	}

	/** The serial number of the archive **/
	private long archiveSerialNumber = 0;
	/** Sets the serial number of the archive **/
	void setArchiveSerialNumber(long serialNumber) { this.archiveSerialNumber = serialNumber;  }
	/** Sets the serial number of the archive **/
	public long archiveSerialNumber() { return archiveSerialNumber;  }

	/** The user creating the archive  **/
	private String creatorUser = null;
	/** @return the user creating the archive  **/
	public String creatorUser() { return creatorUser; }
	/** Sets the user creating the archive  **/
	void setCreatorUser(String creatorUser) { this.creatorUser = creatorUser; }

	/** The group of the user creating the archive  **/
	private String creatorGroup = null;
	/** @return group of the the user creating the archive  **/
	public String creatorGroup() { return creatorGroup; }
	/** Sets group of the the user creating the archive  **/
	void setCreatorGroup(String creatorGroup) { this.creatorGroup = creatorGroup; }

	/** The class to be executed after program termination **/
	private JavaProgram javaProgram = null;
	/** @return the class to be executed after program termination **/
	public int executePostCommand() throws SyntaxError { return (javaProgram == null) ? 0 : javaProgram.execute(); }

	/** Sets the class to be executed after program termination **/
	void setPostCommand(String className, String[] arguments) throws SyntaxError {
		this.javaProgram = new JavaProgram(className, arguments);
	}

	/** Sets the user interface and registers its <CODE>FileEvent</CODE> and <CODE>ProgressEvent</CODE> listeners
	 ** @param userInterface the user interface to be used.
	 ** If <CODE>null</CODE> then the default user interface from <CODE>UserInterface.get()</CODE> is used instead
	 **/
	void setUserInterface(String userInterface) {
		if (userInterface != null)
			UserInterface.set(userInterface);
		notifier.file().register( UserInterface.get().running().getFileEventListener() );
		notifier.progress().register( UserInterface.get().running().getProgressEventListener() );
	}

	/** Prints a short summary of all settings
	 **/
	public String infos() {
		return "Archive: \"" + archive + "\" Serial: " + Long.toHexString(archiveSerialNumber) + "\n"
				+ "Created by " + creatorUser + " (" + creatorGroup + ")\n"
				+ files.count() + "  total files = " + files.size() + " Bytes\n"
				+ "Packer: " + packer.name + " Block size: " + blockSize + " bytes Tracker: " + tracker.name
				+ "\nPacked size " + archive.length() + " bytes\n"
				+ ((cipherNames.length > 0) ? ("Ciphers: " + Text.toString(cipherNames, "", " + ", "")) : "No cipher") + "\n"
				+ (hasImportedArchive() ? ("Differential archive following " + importedArchive.getName() + "\n") : "")
					;
	}

	/**	Sets the cipher stack
	 ** @return the actually reentered stack of cipher names.
	 **/
	String[] setCiphers(Specification[] specifications) {
			Specification[] fullFilledSpecifications = isSpecial() ? specifications :
					isArchiving() ? UserInterface.get().ruling().chooseCiphers(specifications) :
					UserInterface.get().ruling().enterPasswords(specifications);
		this.ciphers = Specification.getCiphers(fullFilledSpecifications, isArchiving() );
		return ( cipherNames = Specification.getNames(fullFilledSpecifications) );
	}

	/** The ciphers stack **/
	private Cipher[] ciphers = null;

	/** The stack of cipher names **/
	private String[] cipherNames = null;

	/** Gets the ciphers stack.
	 ** Note that instantiation is <em>lazy</em>.
	 **/
	public Cipher[] ciphers() {	return ciphers;	}

	/** The wildcards filtering files to be extracted **/
	private sam.entry.List wildcards = null;
	/** @return the wildcards filtering files to be extracted **/
	public sam.entry.List wildcards() { return wildcards; }
	/** @return the wildcards filtering files to be extracted **/
	void setWildcards(sam.entry.Enumerator wildcardsSpecifications) {
		wildcards = (wildcardsSpecifications == null) ? null : wildcardsSpecifications.statify().sortAndRemoveDuplicates();
	}

	/** The list of files to be put into/contained in the archive **/
	private sam.entry.List files = null;
	/** @return the list of files to be put into/contained in the archive **/
	public sam.entry.List files() { return files; }
	/** Sets the list of files to be put into/contained in the archive **/
	void setFiles(sam.entry.List files) { this.files = files; }

	/** @return an <CODE>sam.entry.Enumerator</CODE> of all the files that have been <em>removed</em>
	 ** since the previous backup of a differential set
	 ** <p>The returned <CODE>sam.entry.Enumerator</CODE> is independant of the internal enumeration of files
	 **/
	public sam.entry.Enumerator removedFiles() {
		UserInterface.get().log().log(3, ModusOperandi.class, "Scanning for removed files");
		sam.entry.Enumerator removedFiles = files.getRemovedSince(importedFiles).statify();
		files.reset();
		importedFiles.reset();
		return removedFiles;
	}

	/** @return an <CODE>sam.entry.Enumerator</CODE> of all the files that have been <em>modified</em>
	 ** since the previous backup of a differential set
	 ** <p>The returned <CODE>sam.entry.Enumerator</CODE> is independant of the internal enumeration of files
	 **/
	public sam.entry.List newerFiles() {
		UserInterface.get().log().log(3, ModusOperandi.class, "Scanning for newer files");
		sam.entry.List newerFiles = files.keepNewerThan(importedFiles).statify();
		files.reset();
		importedFiles.reset();
		return newerFiles;
	}

	/** Sets the list of files that actually will be processed.
	 ** @return the properly sorted list of to-be-processed files
	 **/
	public sam.entry.List setToBeProcessedFiles(sam.entry.List files) {
		if ( isArchiving() ) {
			UserInterface.get().log().log(2, ModusOperandi.class, "Sorting " + files.count() + " files");
			files = files.sortAndRemoveDuplicates().statify();
			UserInterface.get().log().log(2, ModusOperandi.class, "Checking " + files.count() + " for reading");
			try {
				files.checkRead();
			} catch (IOException e) {
				UserInterface.get().error().handleException("Failed to read entries, expect files be skipped", e);
			}
		}
		notifier.progress().setTotalNumberOfEntries( files.count() );
		notifier.progress().setTotalNumberOfPlainBytes( files.size() );
		return files.statify();
	}

}
