
package sam.rules;

import sam.ui.UserInterface;
import sam.core.Settings;
import sam.io.IO;
import sam.toolkit.Text;

import java.io.*;
import java.util.zip.ZipFile;
import java.util.jar.JarOutputStream;


/** Class describing a rule.
 ** <p>This includes parsing, verbose explainations, etc.
 **/
public abstract class Syntax {

	/** The name of this single <CODE>Syntax</CODE> element **/
	public final String name;

	/** A human-readable description of this single <CODE>Syntax</CODE> element **/
	public final String description;

	/** The switch character of this <CODE>Syntax</CODE> element **/
	final char switchType;

	/** Wether this syntax has ever been specified, or not **/
	private boolean hasEverBeenSpecified = false;

	/** Builds a new <CODE>Syntax</CODE> element
	 ** @param switchType the switchType character
	 ** @param name the name
	 ** @param description a human-readable description
	 **/
	protected Syntax(char switchType, String name, String description) {
		this.name = name;
		this.description = description;
		this.switchType = switchType;
		UserInterface.get().log().log(21, Syntax.class, "Defined syntactic element \"" + name + "\" (" + description +
																	") with type '" + switchType + "'");
	}

	/** Wether this syntax element recognizes proposed switch type **/
	public boolean recognizesSwitch(char proposedSwitchType) { return switchType == proposedSwitchType; }

	/** Builds a switch prefix for this syntax, e.g. <CODE>-i</CODE> or <CODE>-?</CODE> **/
	String getSwitchPrefix() {
		return (switchType == FILENAME_SWITCH_TYPE) ? "" : ("-" + switchType);
	}

	/** Really parses a command-line supplied switch **/
	protected abstract void doParse(String command);

	/** Parses a command-line supplied switch
	 ** This is a failsafe wrapper to {@link #doParse}
	 **/
	public final void parse(String command) {
		if (command == null)
			return;
		hasEverBeenSpecified = true;
		command = command.trim();
		if ((command.length() > 0) && (command.charAt(0) == Settings.Constants.COMMAND_LINE_FILE_ALIAS))
			try {  // Loads command-line parameter from file
				command = Text.toString(IO.getTokens( new FileInputStream(command.substring(1)), ",\n"), "", ",", "");
			} catch (IOException e) {
				UserInterface.get().error().handleException("Failed to load command line parameter from file " + command.substring(1), e);
			}
		doParse(command);
	}

	/** Loads a <CODE>Rule</CODE> from specified archive **/
	public abstract void load(ZipFile archive);

	/** Gets the <CODE>Rule</CODE> specified by this <CODE>RuleSpecification</CODE>.
	 ** This might require prompting the user for a correct specification through
	 ** the current <CODE>RulingInterface</CODE>
	 **/
	public abstract Rule getRule();

	/** Builds an example for this single <CODE>Syntax</CODE> element **/
	public abstract String example();

	/** Special switch type for files **/
	static final char FILENAME_SWITCH_TYPE = (char) 0xf5;

	/** Special switch type for abstract syntactic elements that are not really user-selectable
	 ** @see Resource
	 **/
	protected static final char NON_SELECTABLE_TYPE = 0;

	/** Class for specifications of pure resources that are not user-selectable **/
	protected static abstract class Resource extends Syntax {
		protected Resource(String name) { super(NON_SELECTABLE_TYPE, name, null); }

		abstract public void load(ZipFile archive);
		abstract protected void doSetTo(ModusOperandi modusOperandi);
		abstract protected void doSaveTo(JarOutputStream output);

		public final String example() { return null; }
		protected final void doParse(String argument) {	}
		public final Rule getRule() {
			return new Rule() {
				public void setTo(ModusOperandi modusOperandi) { Resource.this.doSetTo(modusOperandi);	}
				public void saveTo(JarOutputStream output) { Resource.this.doSaveTo(output); }
				public String toSwitch() { return null; }
				public String verbose() { return ""; }
			};
		}
	}

	/** Shortcut for easily parsing an integer **/
	protected int parseInteger(String valueSpecifiation, String title, int min, int max) {
		try {
			int value = Integer.valueOf(valueSpecifiation).intValue();
			if ((value < min) || (value > max))
				throw new NumberFormatException("Value out of range");
			return value;
		} catch (NumberFormatException e) {
			return UserInterface.get().ruling().getInteger(title, min, min, max);  // Default value is the minimum
		}
	}

}

