
package sam.rules;

import sam.ui.UserInterface;
import sam.core.Settings;
import sam.io.IO;
import sam.toolkit.Text;
import sam.toolkit.ClassRegistry;

import java.util.Enumeration;
import java.io.FileInputStream;
import java.io.IOException;


/** Class representing a whole syntax set.
 ** This defines how the command-line is parsed into the actual <CODE>RuleSet</CODE>
 ** @see Syntax
 ** @see RuleSet
 **/
public abstract class SyntaxSet {

	protected SyntaxSet() { }

	/** Registry of all defined <CODE>SyntaxSet</CODE>
	 ** Subclasses should include a <em>static</em> statement like
	 ** <pre><codeblock>
	 ** static {
	 **   REGISTRY.register(MySyntaxSet.class, "my syntax set");
	 ** }
	 ** </codeblock></pre>
	 **/
	protected static final ClassRegistry REGISTRY = new ClassRegistry(SyntaxSet.class, "syntax set");

	static {
		// IMPORTANT NOTE:
		// Defining the default syntax set as a <CODE>Class</CODE> member into <CODE>sam.core.Settings</CODE>
		// with value <CODE>UnixLikeSyntaxSet.class</CODE> will cause the cinit of <CODE>Settings</CODE> load
		// class <CODE>SyntaxSet</CODE> which will access that member here whereas it is still <CODE>null</CODE> !!
		REGISTRY.register(UnixLikeSyntaxSet.class, "default");
		REGISTRY.set("default");
	}

	/** Sets the syntax set
	 ** @param name an already registered <CODE>SyntaxSet</CODE> name
	 **/
	public static boolean set(String syntaxSetName) { return REGISTRY.set(syntaxSetName); }

	/** Builds a new <CODE>SyntaxSet</CODE> instance.
	 **/
	public static SyntaxSet get() {
		SyntaxSet syntaxSet = (SyntaxSet) REGISTRY.get();
		if (syntaxSet == null) {
			UserInterface.get().error().warn("Could not initialize syntax set, defaulting to Unix-like");
			syntaxSet = new UnixLikeSyntaxSet();
		}
		return syntaxSet;
	}

	/** Gets the name of this syntax set **/
	public abstract String name();

	/** Enumerates all <CODE>Syntax</CODE> pieces composing this syntax set. **/
	public abstract Enumeration syntaxes();

	/** Parses a command-line and imports specified fields **/
	public void parse(String[] arguments) {
		for (int i=0; i<arguments.length; ++i) {
			String argument = arguments[i].trim();
			if ((argument.length() > 0) && (argument.charAt(0) == Settings.Constants.COMMAND_LINE_FILE_ALIAS))
				try {
					parse( IO.getTokens( new FileInputStream(argument.substring(1)), " \t\n") );
				} catch (IOException e) {
					UserInterface.get().error().fail("Failed to read command line from " + argument.substring(1), e);
				}
			else if ( ! parseArgument(getSwitchType(argument), getSwitchValue(argument)) )
				UserInterface.get().error().unhandledParameter(argument);
		}
	}

	/** Loads syntaxes from specified archive **/
	public void load(java.util.zip.ZipFile archive) {
		for (Enumeration syntaxes = syntaxes(); syntaxes.hasMoreElements(); )
			((Syntax) syntaxes.nextElement()).load(archive);
	}

	/** Hopefully helps. **/
	public String help() {
		return syntax() + "\n\tExample:\n" + example() + "\n";
	}

	/** Gets the list of all valid one-char parameter prefixes, as a <CODE>String</CODE> **/
	public String getDeclaredTypes() {
		StringBuffer buffer = new StringBuffer();
		for (Enumeration syntaxes = syntaxes(); syntaxes.hasMoreElements(); ) {
			buffer.append( ((Syntax) syntaxes.nextElement()).switchType );
		}
		return buffer.toString();
	}


	/** Gets the syntax in an abstract description **/
	public String syntax() {
		StringBuffer buffer = new StringBuffer();
		for (Enumeration syntaxes = syntaxes(); syntaxes.hasMoreElements(); ) {
			Syntax syntax = (Syntax) syntaxes.nextElement();
			if (syntax.description != null)
				buffer.append(" ").append( syntax.getSwitchPrefix() ).append(syntax.name);
		}
		buffer.append('\n').append("\tWhere:\n");
		for (Enumeration syntaxes = syntaxes(); syntaxes.hasMoreElements(); ) {
			Syntax syntax = (Syntax) syntaxes.nextElement();
			if (syntax.description != null)
				buffer.append("\t").append(syntax.name).append(":\t").append(syntax.description).append('\n');
		}
		return buffer.toString();
	}

	/** Builds an example of this syntax set. **/
	public String example() {
		StringBuffer buffer = new StringBuffer();
		buffer.append(Settings.Environment.PROGRAM_BINARY);
		for (Enumeration syntaxes = syntaxes(); syntaxes.hasMoreElements(); ) {
			Syntax syntax = (Syntax) syntaxes.nextElement();
			if (syntax.description != null) {
				String example = syntax.example();
				if (example != null)
					buffer.append(" ").append( syntax.getSwitchPrefix() ).append(example);
			}
		}
		return buffer.toString();
	}

	/** Tool for parsing one command-line argument against all syntaxes of this syntax set. **/
	private boolean parseArgument(char switchType, String switchValue) {
		for (Enumeration syntaxes = syntaxes(); syntaxes.hasMoreElements(); ) {
			Syntax syntax = (Syntax) syntaxes.nextElement();
			if (syntax.recognizesSwitch(switchType)) {
				UserInterface.get().log().log(22, Syntax.class, "Parsing \"" + switchValue + "\" as \"" + syntax.name
																			+ "\" (" + syntax.description + ")");
				syntax.parse(switchValue);
				return true;
			}
		}
		UserInterface.get().log().log(22, SyntaxSet.class, "Failed to parse \"" + switchValue + "\"");
		return false;
	}

	/** Tool for getting the switch type of a command-line argument **/
	private static final char getSwitchType(String name) {
		if ((name.length() < 2) || (name.charAt(0) != '-'))
			return Syntax.FILENAME_SWITCH_TYPE;
		return name.charAt(1);
	}

	/** Tool for getting the switch value part of a command-line argument **/
	private static final String getSwitchValue(String name) {
		if ((name.length() < 2) || (name.charAt(0) != '-'))
			return name;
		return name.substring(2);
	}

}
