

package sam.rules;

import sam.io.Packer;
import sam.io.Mixer;
import sam.io.Tracker;
import sam.io.PrintTrace;
import sam.toolkit.Text;
import sam.core.Settings;
import sam.toolkit.Random;
import sam.core.ArchiveLayout;
import sam.ui.UserInterface;
import sam.ui.RulingInterface;
import sam.cipher.Specification;
import sam.event.FileEvent;
import sam.event.FileEventAdapter;

import java.io.*;
import java.util.*;
import java.util.zip.ZipFile;
import java.util.jar.JarOutputStream;


/** Unix-like implementation of the syntax.
 **/
public class UnixLikeSyntaxSet extends SyntaxSet {

	public final String name() { return "Unix-like syntax set"; }

	public final Enumeration syntaxes() {	return sam.toolkit.Text.enumerate(syntaxes);	}

	private boolean isNotSpecialMode = true;

	/** The actual list of all syntaxes **/
	private final Syntax[] syntaxes = new Syntax[] {
	new Syntax('u', "user_interface", "the user interface") {
		// private String ui = Settings.Defaults.USER_INTERFACE;
		public String example() { return sam.toolkit.Random.drawString( UserInterface.getKnwonInterfaces() ); }
		protected void doParse(String argument) { UserInterface.set(argument);	}
		public void load(ZipFile archive) {	}
		public Rule getRule() {
			return new Rule() {
				public void setTo(ModusOperandi modusOperandi) { modusOperandi.setUserInterface(null); }
				public void saveTo(JarOutputStream output) {  }
				public String toSwitch() { String ui = UserInterface.REGISTRY.name(); return (ui == null) ? "" : ui; }
				public String verbose() { String ui = UserInterface.REGISTRY.name();
				return (ui == null ) ? "" : ("with the " + ui + " user interface"); }
			};
		}
	}, new Syntax('g', "log_level", "the log level") {
		private int logLevel = Settings.Defaults.LOG_LEVEL;
		public String example() { return ("" + sam.toolkit.Random.nextInt(Settings.Constants.MAXIMUM_LOG_LEVEL + 1)); }
		protected void doParse(String argument) {
			logLevel = ("".equals(argument)) ? Settings.Defaults.LOG_LEVEL : Text.parseInteger(argument, Settings.Constants.MINIMUM_LOG_LEVEL-1);
			while (logLevel < Settings.Constants.MINIMUM_LOG_LEVEL)
				logLevel = UserInterface.get().ruling().getInteger(description, Settings.Defaults.LOG_LEVEL,
																													 Settings.Constants.MINIMUM_LOG_LEVEL, Settings.Constants.MAXIMUM_LOG_LEVEL);
			if (Settings.Constants.EARLY_LOG_LEVEL_SETTING)
				UserInterface.get().log().setLevel(logLevel);
		}
		public void load(ZipFile archive) {	}
		public Rule getRule() {
			return new Rule() {
				public void setTo(ModusOperandi modusOperandi) {
					if ( ! Settings.Constants.EARLY_LOG_LEVEL_SETTING)
						UserInterface.get().log().setLevel(logLevel);
				}
				public void saveTo(JarOutputStream output) {  }
				public String toSwitch() { return (logLevel > 0) ? ("" + logLevel) : null; }
				public String verbose() { return (logLevel > 0) ? ("with log level " + logLevel) : ""; }
			};
		}
	// <b>Warning<b>: the file syntax must be the FIRST resource-rule in the list, in order to properly reload other ones in extracting mode
	}, new Syntax(Syntax.FILENAME_SWITCH_TYPE, "filename", "a file name") {
		private boolean isArchiving = true;
		private File archive = Settings.Defaults.ARCHIVE;
		private Set files = new HashSet();  // Set<String>
		private sam.entry.Enumerator catalog = null;
		public String example() { return "MyArchive.jar MyFile1 MyFile2"; }
		protected void doParse(String argument) {
			if (requiresArchive()) // First file specification must be the archive in archiving mode (default)
				enterArchive(argument);
			else
				files.add(isArchiving ? enterInputFile(argument) : argument);
		}
		private boolean requiresArchive() { return isNotSpecialMode && isArchiving && (archive == null); }
		private void enterArchive(String suggested) {
			archive = UserInterface.get().ruling().getFileOrDirectory(description, suggested, RulingInterface.FILE);
			isArchiving = ! archive.canRead();  // If archive file already exists and is readable, then assume extracting mode
			if (! isArchiving)
				UserInterface.get().log().log(21, UnixLikeSyntaxSet.class, "Setting \"" + archive +"\" as to-be-extracted archive");
		}
		private String enterInputFile(String suggested) {
			return UserInterface.get().ruling().getFileOrDirectory("input directory", suggested,
																														 RulingInterface.READABLE_FILE_OR_DIRECTORY).getPath();
		}
		public void load(ZipFile archive) {
			this.archive = new File( archive.getName() );
			this.isArchiving = false;
			try {
				this.catalog = sam.entry.Enumerator.load( ArchiveLayout.CATALOG.get(archive) );  // This is lazy until <CODE>statify</CODE>
			} catch (IOException e) {
				UserInterface.get().error().handleException("Could not load the catalog of " + archive, e);
			}
		}
		public Rule getRule() {
      if (requiresArchive())
				enterArchive( (Settings.Defaults.ARCHIVE == null) ? null : Settings.Defaults.ARCHIVE.getPath() );
			if (files.isEmpty())
				files.add( isArchiving ? enterInputFile(Settings.Defaults.TO_BE_ARCHIVED_FILES) : Settings.Defaults.TO_BE_EXTRACTED_FILES);
			final String[] filesSpecifications = new String[ files.size() ];
			files.toArray(filesSpecifications);
			if (isArchiving) {
				UserInterface.get().log().log(4, UnixLikeSyntaxSet.class, "Scanning and sorting files from " +
																			sam.toolkit.Text.toString(filesSpecifications) + "");
				catalog = sam.entry.FileSystem.scan(filesSpecifications).statify().sortAndRemoveDuplicates();
			}
			// Loads all rules from archive now if in extracting mode
			// This is why this syntax element must be the first, in order to load the other ones
			// <em>before</em> their <CODE>getRule()</CODE> method is called
			if ( ! isArchiving )
				try {
					UnixLikeSyntaxSet.this.load( new ZipFile( archive.getPath() ) );
				} catch (IOException e) {  // thrown by <CODE>new ZipFile(...)</CODE> and <b>not</b> by <CODE>load(...)</CODE> itself
					UserInterface.get().error().handleException("Could not access archive \"" + archive + "\", " +
																											"(ensure it is not broken with Winzip or alike)", e);
				}
			return new Rule() {
				public void setTo(ModusOperandi modusOperandi) {
					modusOperandi.setArchivingMode(isArchiving);
					modusOperandi.setArchive(archive);
					modusOperandi.setFiles( catalog.statify() );
					modusOperandi.setWildcards(isArchiving ? null : new sam.entry.List(filesSpecifications) );
				}
				public void saveTo(JarOutputStream archive) {
					OutputStream output = ArchiveLayout.CATALOG.put(archive);
					try {
						catalog.store(output);
					} catch (IOException e) {
						UserInterface.get().error().handleException("Could not save archive catalog", e);
					} finally {
						try {
							output.close();
						} catch (IOException e) {
							UserInterface.get().error().handleException("Could not store the catalog of " + archive, e);
						}
					}
				}
				public String toSwitch() { return (isArchiving ? archive.getPath() : "") + " " + Text.toString(filesSpecifications, " ", "" , ""); }
				public String verbose() { 
					return isArchiving ? ("into file " + archive + " storing files in " + Text.toString(filesSpecifications, " ", "" , "") ) :
							("from file " + archive + " unstoring files matching " + Text.toString(filesSpecifications, " ", "" , "") );
				}
			};
		}
	}, new Syntax('?', "help", null) {
		private boolean helpRequired = false;
		public String example() { return null;}
		protected void doParse(String argument) { helpRequired = true; isNotSpecialMode = false; }
		public void load(ZipFile archive) { }
		public Rule getRule() {
			return helpRequired ? new Rule() {
				public void setTo(ModusOperandi modusOperandi) { modusOperandi.setOpus(ModusOperandi.HELPING); }
				public void saveTo(java.util.jar.JarOutputStream archive) { }
				public String toSwitch() { return "-?"; }
				public String verbose() { return "prints some help"; }
			} : null;
		}
	}, new Syntax('B', "benchmark", null) {
		private boolean isBenchmarking = false;
		public String example() { return null;}
		protected void doParse(String argument) { isBenchmarking = true; isNotSpecialMode = false; }
		public void load(ZipFile archive) { }
		public Rule getRule() {
			return isBenchmarking ? new Rule() {
				public void setTo(ModusOperandi modusOperandi) { modusOperandi.setOpus(ModusOperandi.BENCHMARKING); }
				public void saveTo(java.util.jar.JarOutputStream archive) { }
				public String toSwitch() { return ""; }
				public String verbose() { return "benchmarks the application"; }
			} : null;
		}
	}, new Syntax('C', "check", null) {
		private boolean isProbing = false;
		public String example() { return null;}
		protected void doParse(String argument) { isProbing = true; }
		public void load(ZipFile archive) { }
		public Rule getRule() {
			return isProbing ? new Rule() {
				public void setTo(ModusOperandi modusOperandi) { modusOperandi.setOpus(ModusOperandi.CHECKING); }
				public void saveTo(java.util.jar.JarOutputStream archive) { }
				public String toSwitch() { return ""; }
				public String verbose() { return "checks the archive"; }
			} : null;
		}
	}, new Syntax('i', "imported_archive", "a previous archive to make a differential backup against") {
		// This imported archive rule specification <b>must</b> be before the files rule specification
		private File importedArchive = null;
		public String example() { return "myPreviousArchive.jar";}
		protected void doParse(String argument) {
			importedArchive = UserInterface.get().ruling().getFileOrDirectory(description, argument, RulingInterface.READABLE_FILE);
		}
		public void load(ZipFile archive) {
			String importedArchiveName = ArchiveLayout.IMPORTED_ARCHIVE.getString(archive);
			importedArchive = (importedArchiveName == null) ? null : new File(importedArchiveName);
		}
		public Rule getRule() {
			return new Rule() {
				public void setTo(ModusOperandi modusOperandi) { modusOperandi.setImportedArchive(importedArchive); }
				public void saveTo(JarOutputStream output) {
					if (importedArchive != null)
						ArchiveLayout.IMPORTED_ARCHIVE.putString(importedArchive.getPath(), output);
				}
				public String toSwitch() { return (importedArchive == null) ? null : importedArchive.getName(); }
				public String verbose() { return (importedArchive == null) ?  "" : ("importing files from archive " + importedArchive.getName()); }
			};
		}
	}, new Syntax('t', "temporary_directory", "directory were temporary files are created") {
		private File temporaryDirectory = Settings.Defaults.TEMPORARY_DIRECTORY;
		public String example() { return Random.drawString(new String[] { Settings.Defaults.TEMPORARY_DIRECTORY.getPath(), ".",
																																					Settings.Environment.HOME} );}
		protected void doParse(String argument) {
			temporaryDirectory = UserInterface.get().ruling().getFileOrDirectory(description, argument, RulingInterface.WRITABLE_DIRECTORY);
		}
		public void load(ZipFile archive) {	}
		public Rule getRule() {
			return new Rule() {
				public void setTo(ModusOperandi modusOperandi) { modusOperandi.setTemporaryDirectory(temporaryDirectory); }
				public void saveTo(JarOutputStream output) {  }
				public String toSwitch() { return temporaryDirectory.toString(); }
				public String verbose() { return "creating temporary files in " + temporaryDirectory; }
			};
		}
	}, new Syntax('b', "block_size", "the block size to be used (in kilobytes) zero meaning unlimited blocks") {
		private int blockSize = Settings.Defaults.BLOCK_SIZE;
		public String example() { return ("" + sam.toolkit.Random.nextInt(Settings.Constants.MAXIMUM_BLOCK_SIZE));}
		protected void doParse(String argument) {	blockSize = parseInteger(argument, "block size", 0, Settings.Constants.MAXIMUM_BLOCK_SIZE);	}
		public void load(ZipFile archive) {	blockSize = ArchiveLayout.BLOCK_SIZE.getInt(archive);	}
		public Rule getRule() {
			if (blockSize < 0)
					blockSize = UserInterface.get().ruling().getInteger(description, Settings.Defaults.BLOCK_SIZE,
																															0, Settings.Constants.MAXIMUM_BLOCK_SIZE);
			return new Rule() {
				public void setTo(ModusOperandi modusOperandi) { modusOperandi.setBlockSize(blockSize * 1024); }
				public void saveTo(JarOutputStream output) { ArchiveLayout.BLOCK_SIZE.putLong(blockSize, output); }
				public String toSwitch() { return ("" + blockSize); }
				public String verbose() { return "with a block size of " + blockSize + "kB"; }
			};
		}
	}, new Syntax('m', "packer", "the packer to be used") {
		private Packer packer = Settings.Defaults.PACKER;
		public String example() { return Random.drawString( Packer.getPackers() );}
		protected void doParse(String argument) {
			packer = Packer.get(argument);  // Which may return <CODE>null</CODE> itself
		}
		public void load(ZipFile archive) {
			packer = Packer.get( ArchiveLayout.PACKER.getString(archive) );
			packer.getMixer().setTrace( ArchiveLayout.MIXER.getBytes(archive) );
		}
		public Rule getRule() {
			if (packer == null)
				packer = Packer.get( UserInterface.get().ruling()
														 .chooseString("packer", Packer.getPackers(),
																					 (Settings.Defaults.PACKER == null) ? null : Settings.Defaults.PACKER.name) );
			return new Rule() {
				public void setTo(ModusOperandi modusOperandi) { modusOperandi.setPacker(packer); }
				public void saveTo(JarOutputStream output) {
					ArchiveLayout.PACKER.putString(packer.name, output);
					ArchiveLayout.MIXER.putBytes(packer.getMixer().getTrace(), output);
				}
				public String toSwitch() { return (packer == null) ? null : packer.name; }
				public String verbose() { return (packer == null) ? null : packer.description; }
			};
		}
	}, new Syntax('z', "comment", "plain-text comment of the archive") {
		private String comment = null;
		public String example() { return "\"my comment here\""; }
		protected void doParse(String argument) { comment = argument; }
		public void load(ZipFile archive) { comment = ArchiveLayout.COMMENT.getString(archive); }
		public Rule getRule() {
			return new Rule() {
				public void setTo(ModusOperandi modusOperandi) { if (comment != null) modusOperandi.setComment(comment); }
				public void saveTo(JarOutputStream output) { if (comment != null) ArchiveLayout.COMMENT.putString(comment, output); }
				public String toSwitch() { return comment; }
				public String verbose() { return (comment == null) ? "" : (" with comment \"" + Text.summarize(comment, 35) + "...\""); }
			};
		}
	}, new Syntax('k', "tracker", "the tracker of the archive") {
		private Tracker tracker = Settings.Defaults.TRACKER;
		public String example() { return Random.drawString( Tracker.getTrackers() );}
		protected void doParse(String argument) {
			tracker = Tracker.get(argument);  // Which may return <CODE>null</CODE> itself
		}
		public void load(ZipFile archive) {	tracker = Tracker.get( ArchiveLayout.TRACKER.getString(archive) );	}
		public Rule getRule() {
			if (tracker == null)
				tracker = Tracker.get( UserInterface.get().ruling()
														 .chooseString("tracker", Tracker.getTrackers(),
																					 (Settings.Defaults.TRACKER == null) ? null : Settings.Defaults.TRACKER.name) );
			return new Rule() {
				public void setTo(ModusOperandi modusOperandi) { modusOperandi.setTracker(tracker); }
				public void saveTo(JarOutputStream output) { ArchiveLayout.TRACKER.putString(tracker.name, output); }
				public String toSwitch() { return (tracker == null) ? null : tracker.name; }
				public String verbose() { return (tracker == null) ? null : tracker.description; }
			};
		}
	}, new Syntax('c', "mismatching", "compares stored entries to local files") {
		private PrintTrace trace = null;
		public String example() { return PrintTrace.example("mismatching_entries.list");}
		protected void doParse(String argument) { trace = PrintTrace.make(argument, "mismatching_entries"); }
		public void load(ZipFile archive) { }
		public Rule getRule() {
			return new Rule() {
				public void setTo(ModusOperandi modusOperandi) {
					if (trace != null) {
						modusOperandi.setOpus(ModusOperandi.COMPARING);
						modusOperandi.notifier().file().register(new FileEventAdapter() {
							protected void mismatching(FileEvent event) { event.printTo( trace.out() ); }
						} );
					}
				}
				public void saveTo(JarOutputStream output) {  }
				public String toSwitch() { return (trace == null) ? null : trace.name; }
				public String verbose() {
					return (trace == null) ? null : "comparing stored entries to local files and reporting mismatching ones to " + trace.name;
				}
			};
		}
	}, new Syntax('n', "freshened", "freshens local files with newer entries found in archive") {
		private PrintTrace trace = null;
		public String example() { return PrintTrace.example("freshened_entries.list");}
		protected void doParse(String argument) { trace = PrintTrace.make(argument, "freshened_entries"); }
		public void load(ZipFile archive) { }
		public Rule getRule() {
			return new Rule() {
				public void setTo(ModusOperandi modusOperandi) {
					if (trace != null) {
						modusOperandi.setOpus(ModusOperandi.FRESHENING);
						modusOperandi.notifier().file().register(new FileEventAdapter() {
							protected void overwritten(FileEvent event) { event.printTo( trace.out() ); }
						} );
					}
				}
				public void saveTo(JarOutputStream output) {  }
				public String toSwitch() { return (trace == null) ? null : trace.name; }
				public String verbose() {
					return (trace == null) ? null :
							"freshening local files from stored entries and reporting effectively extracted ones to " + trace.name;
				}
			};
		}
	}, new Syntax('l', "listed", "list entries") {
		private PrintTrace trace = null;
		public String example() { return PrintTrace.example("entries.list");}
		protected void doParse(String argument) { trace = PrintTrace.make(argument, "entries"); }
		public void load(ZipFile archive) { }
		public Rule getRule() {
			return new Rule() {
				public void setTo(final ModusOperandi modusOperandi) {
					if (trace != null) {
						modusOperandi.setOpus(ModusOperandi.LISTING);
						modusOperandi.notifier().file().register(new FileEventAdapter() {
							protected void processed(FileEvent event) {
								trace.out().println( event.entry.list(modusOperandi) );
								trace.out().flush();
							}
						} );
					}
				}
				public void saveTo(JarOutputStream output) {  }
				public String toSwitch() { return (trace == null) ? null : trace.name; }
				public String verbose() {
					return (trace == null) ? null :
							"freshening local files from stored entries and reporting effectively extracted ones to " + trace.name;
				}
			};
		}
	}, new Syntax('d', "", "describe archive") {
		private boolean isSpecified = false;
		public String example() { return ""; }
		protected void doParse(String argument) { isSpecified = true; }
		public void load(ZipFile archive) { }
		public Rule getRule() {
			return new Rule() {
				public void setTo(final ModusOperandi modusOperandi) { if (isSpecified) modusOperandi.setOpus(ModusOperandi.DESCRIBING); }
				public void saveTo(JarOutputStream output) {  }
				public String toSwitch() { return isSpecified ? "" : null; }
				public String verbose() { return isSpecified ? null : "describin archive"; }
			};
		}
	}, new Syntax('o', "overwritten", "overwrites local files") {
		private PrintTrace trace = null;
		public String example() { return PrintTrace.example("overwritten_entries.list");}
		protected void doParse(String argument) { trace = PrintTrace.make(argument, "overwritten_entries"); }
		public void load(ZipFile archive) { }
		public Rule getRule() {
			return new Rule() {
				public void setTo(ModusOperandi modusOperandi) {
					if (trace != null) {
						modusOperandi.setOpus(ModusOperandi.OVERWRITING);
						modusOperandi.notifier().file().register(new FileEventAdapter() {
							protected void overwritten(FileEvent event) { event.printTo( trace.out() ); }
						} );
					}
				}
				public void saveTo(JarOutputStream output) {  }
				public String toSwitch() { return (trace == null) ? null : trace.name; }
				public String verbose() {
					return (trace == null) ? null :
							"freshening local files from stored entries and reporting effectively extracted ones to " + trace.name;
				}
			};
		}
	}, new Syntax('s', "skipped", "reports skipped files") {
		private PrintTrace trace = null;
		public String example() { return PrintTrace.example("skipped_entries.list");}
		protected void doParse(String argument) { trace = PrintTrace.make(argument, "skipped_entries"); }
		public void load(ZipFile archive) { }
		public Rule getRule() {
			return new Rule() {
				public void setTo(ModusOperandi modusOperandi) {
					if (trace != null) {
						modusOperandi.notifier().file().register(new FileEventAdapter() {
							protected void skipped(FileEvent event) { event.printTo( trace.out() ); }
						} );
					}
				}
				public void saveTo(JarOutputStream output) {  }
				public String toSwitch() { return (trace == null) ? null : trace.name; }
				public String verbose() { return (trace == null) ? null : "reporting skipped entries to " + trace.name;	}
			};
		}
	}, new Syntax('f', "failed", "reports failed files") {
		private PrintTrace trace = null;
		public String example() { return PrintTrace.example("failed_entries.list");}
		protected void doParse(String argument) { trace = PrintTrace.make(argument, "failed_entries"); }
		public void load(ZipFile archive) { }
		public Rule getRule() {
			return new Rule() {
				public void setTo(ModusOperandi modusOperandi) {
					if (trace != null) {
						modusOperandi.notifier().file().register(new FileEventAdapter() {
							protected void failed(FileEvent event) { event.printTo( trace.out() ); }
						} );
					}
				}
				public void saveTo(JarOutputStream output) {  }
				public String toSwitch() { return (trace == null) ? null : trace.name; }
				public String verbose() { return (trace == null) ? null : "reporting skipped entries to " + trace.name;	}
			};
		}
	}, new Syntax('j', "threads", "the number of threads") {
		private int verticalThreads = Settings.Defaults.VERTICAL_THREADS;
		private int horizontalThreads = Settings.Defaults.HORIZONTAL_THREADS;
		public String example() { return (1 +sam.toolkit.Random.nextInt(5)) + "x" + (1 +sam.toolkit.Random.nextInt(5));}
		protected void doParse(String argument) {
			final int cut = argument.indexOf('/');
			if (cut < 0) {  // Total number of threads specified
				try {
					int N = Integer.valueOf(argument).intValue();
					if (N < 1)
						throw new NumberFormatException(N + " < 1");
					horizontalThreads = N;
					verticalThreads = (int) Math.ceil( Math.sqrt( (double) N ) );
				} catch (NumberFormatException e) {
					UserInterface.get().error().warn("Expected integer number of threads, got \"" + argument + "\" instead");
				}
			} else {  // Specifies both the number of horizontal and of vertical threads
				try {
					horizontalThreads = Text.parseInteger(argument.substring(0, cut), 1);
					verticalThreads = Text.parseInteger(argument.substring(cut + 1), 1);
				} catch (NumberFormatException e) {
					UserInterface.get().error().warn("Expected \"<n>x<n>\" integer numbers of threads, got \"" + argument + "\" instead");
				}
			}
		}
		public void load(ZipFile archive) {	}
		public Rule getRule() {
			return new Rule() {
				public void setTo(ModusOperandi modusOperandi) {
					modusOperandi.setVerticalThreads(verticalThreads);
					modusOperandi.setHorizontalThreads(horizontalThreads);
				}
				public void saveTo(JarOutputStream output) {  }
				public String toSwitch() { return verticalThreads + "x" + horizontalThreads; }
				public String verbose() { return "using " + verticalThreads + " vertical and " + horizontalThreads + " horizontal threads"; }
			};
		}
	}, new Syntax('p', "ciphers", "the ciphers and passwords") {
		private final Stack cipherNames = new Stack();
		private final transient Stack passwords = new Stack();
		private boolean isSpecified = false;
		public String example() { return Random.drawString( new String[] { "cipher1:passwd1,cipher2:passwd2,cipher3:passwd3" ,
																																					 "cipher1,cipher2,cipher3:passwd1,passwd2,passwd3",
																																					 "cipher1,cipher2", ":passwd1,passwd2" } ); }
		protected void doParse(String argument) {
			if (argument == null)
				return;
			isSpecified = true;
			argument = argument.trim();
			if ("".equals(argument))
				return;
			final int firstComma = argument.indexOf(',');
			final int firstColon = argument.indexOf(':');
			if (firstComma < firstColon) {
				if (Text.countCharacter(argument, ':') > 1) {
					UserInterface.get().error().unhandledParameter("Cipher syntax: two many colons in \"" + argument + "\"");
					return ;
				}
				Text.addArray(cipherNames, Text.split(argument.substring(0, firstColon), ","));
				Text.addArray(passwords, Text.split(argument.substring(firstColon + 1), ","));
			} else if (firstComma == -1)  // Which implies (firstColon == -1)
				cipherNames.push(argument);
			else {
				final String[] ciphersAndPasswords = Text.split(argument, ",");
				for (int i=0; i<ciphersAndPasswords.length; ++i) {
					cipherNames.push( Text.firstWord(ciphersAndPasswords[i], ':') );
					passwords.push( Text.secondWord(ciphersAndPasswords[i], ':') );
				}
			}
		}
  	private String[] getCiphers() {
			Stack stack = new Stack();
			String[] newCiphers = (String[]) cipherNames.toArray(new String[0]);
			for (int i=0; i<newCiphers.length; ) {
				String cipher = newCiphers[i++].trim();
				if ( Settings.Constants.CIPHER_WILDCARD.equals(cipher) )
					Text.addArray(stack, sam.cipher.Cipher.REGISTRY.getCiphers() );
				else
					stack.push(cipher);
			}
			return (String[]) stack.toArray(new String[0]);
		}
		public void load(ZipFile archive) {
			if (Settings.Constants.SUPPLIED_CIPHERS_IGNORED_IN_EXTRACTING)
				cipherNames.removeAllElements();
			Text.addArray(cipherNames, ArchiveLayout.CIPHERS.getStrings(archive) );
			isSpecified = true;
		}
		public Rule getRule() {
			if (cipherNames.size() < passwords.size())
				UserInterface.get().error().warn("Two many passwords (" + passwords.size() + ") supplied for " + cipherNames.size() + " ciphers");
			final Specification[] specifications =
					isSpecified ? Specification.fillIn(getCiphers(), (String[]) passwords.toArray(new String[0]) )
					: Settings.Defaults.CIPHER_SPECIFICATIONS;
			return new Rule() {
				// The actual cipher stack must be remembered from <CODE>ModusOperandi</CODE>
				// This is buggy if <CODE>setTo</CODE> is not followed by a <CODE>saveTo</CODE> on the archive
				// created by the same <CODE>ModusOperandi</CODE>
				String[] ciphers = null;
				public void setTo(ModusOperandi modusOperandi) { ciphers = modusOperandi.setCiphers(specifications); }
				public void saveTo(JarOutputStream output) { ArchiveLayout.CIPHERS.putStrings(ciphers, output);	}
				public String toSwitch() { return Text.toString(ciphers, "", ",", ""); }
				public String verbose() {
					return ((ciphers == null) || (ciphers.length == 0)) ? "" :
						"ciphering data with the " + Text.toString(ciphers, "", "algorithm, the ", "") + " algorithm" ;
				}
			};
		}
	}, new Syntax.Resource("user_group_information") {
			private String user = Settings.Environment.USER;
			private String group = Settings.Environment.GROUP;
			public void load(ZipFile archive) {
				user = ArchiveLayout.USER.getString(archive);
				group = ArchiveLayout.GROUP.getString(archive);
			}
			protected void doSetTo(ModusOperandi modusOperandi) {
				modusOperandi.setCreatorUser(user);
				modusOperandi.setCreatorGroup(group);
			}
			protected void doSaveTo(JarOutputStream output) {
				ArchiveLayout.USER.putString(user, output);
				ArchiveLayout.GROUP.putString(group, output);
			}
	}, new Syntax.Resource("serial_number") {
			private long serialNumber = Random.nextLong();
			public void load(ZipFile archive) { serialNumber = ArchiveLayout.SERIAL_NUMBER.getLong(archive); }
			protected void doSetTo(ModusOperandi modusOperandi) { modusOperandi.setArchiveSerialNumber(serialNumber);	}
			protected void doSaveTo(JarOutputStream output) { ArchiveLayout.SERIAL_NUMBER.putLong(serialNumber, output); }
	}, new Syntax.Resource("version_control") {
			private int samVersion = Settings.Environment.PROGRAM_VERSION;
			private int layoutVersion = ArchiveLayout.LAYOUT_VERSION_NUMBER;
			public void load(ZipFile archive) {
				samVersion = ArchiveLayout.SAM_VERSION.getInt(archive);
				layoutVersion = ArchiveLayout.LAYOUT_VERSION.getInt(archive);
			}
			protected void doSetTo(ModusOperandi modusOperandi) {
				modusOperandi.setArchiveSamVersion(samVersion);
				modusOperandi.setArchiveLayoutVersion(layoutVersion);
			}
			protected void doSaveTo(JarOutputStream output) {
				ArchiveLayout.SAM_VERSION.putInt(samVersion, output);
				ArchiveLayout.LAYOUT_VERSION.putInt(layoutVersion, output);
			}
	}, new Syntax.Resource("runtime") {
			public void load(ZipFile archive) {	}
			protected void doSetTo(ModusOperandi modusOperandi) {
			}

			/** Saves the runtime into the archive.
			 ** @see ArchiveLayout.RUNTIME
			 **/
			protected void doSaveTo(JarOutputStream archive) {	ArchiveLayout.RUNTIME.putRuntime(archive);	}
	}

	};

}

