
package sam.toolkit;

import sam.ui.UserInterface;
import sam.core.Settings;

import java.util.Map;
import java.util.HashMap;
import java.util.NoSuchElementException;
import java.lang.reflect.Constructor;


/** General-purpose <CODE>Class</CODE> registry
 ** This serves as both a registry and a factory.
 ** This class allows a user to a specific sub-class of a certain
 ** base type under a given name, and later instantiate it based on that
 ** name.
 **/
public class ClassRegistry {

	/** The current user interface.
	 ** Might be <CODE>null</CODE> as long as the user interface is never queried.
	 **/
	private Object object = null;

	/** The base type of this registry **/
	private final Class baseType;

	/** Dictionnary of all registered targets **/
	private final Map registry = new HashMap();

	/** The description of targets **/
	private final String description;

	/** The current target name.
	 **/
	private String name = null;

	/** The current target.
	 **/
	private Object target = null;

	/** Builds a new <CODE>ClassRegistry</CODE> **/
	public ClassRegistry() { this(null, null); }

	/** Builds a new <CODE>ClassRegistry</CODE>
	 ** @param baseType the base type of all instantiated targets
	 **/
	public ClassRegistry(Class baseType) { this(baseType, null); }

	/** Builds a new <CODE>ClassRegistry</CODE>
	 ** @param baseType the base type of all instantiated targets
	 ** @param description the human-readable name of all instantiated targets
	 **/
	public ClassRegistry(String description) { this(null, description);	}

	/** Builds a new <CODE>ClassRegistry</CODE>
	 ** @param baseType the base type of all instantiated targets
	 ** @param description the human-readable name of all instantiated targets
	 **/
	public ClassRegistry(Class baseType, String description) {
		this.baseType = (baseType == null) ? Object.class : baseType;
		this.description = (description == null) ? (baseType == null) ? "target" : baseType.getName() : description;
		reset();
	}

	/** Sets <CODE>name</CODE> as the current target.
	 ** @return <CODE>true</CODE> iff <CODE>name</CODE> is already registered 
	 **/
	public boolean set(String name) {
		if (name == null)
			return false;
		if (name.equals(this.name))
			return true;
		if ( ! registry.containsKey(name) ) {
			if (UserInterface.REGISTRY != this)
				UserInterface.get().error().warn("No such " + description + ": \"" + name + "\"");
			else
				; // System.err.println("No such " + description + ": \"" + name + "\"");
			return false;
		}
		this.name = name;
		reset();
		return true;
	}

	/** Wether any target has ever been set.
	 ** @see #set
	 **/
	public boolean isSet() { return name != null ; }

	/** Gets the name of the gurrent target
	 ** @return <CODE>""</CODE> if no target is currently set
	 ** @see #set
	 ** @see #isSet
	 **/
	public String name() { return isSet() ? name : ""; }

	/** Registers a new target as <CODE>name</CODE>
	 ** @return <CODE>true</CODE> iff the registration was successfull.
	 ** This is the case iff <CODE>target</CODE> is a subtype of <CODE>baseType</CODE>,
	 ** its <CODE>ClassLoader</CODE> allows introspection and it defines a <CODE>public</CODE>
	 ** no-argument constructor.
	 **/
	public boolean register(Class target, String name) throws ClassCastException {
		if (target == null)
			return false;
		if ( ! baseType.isAssignableFrom(target) ) {
			UserInterface.get().error().warn(target + " is not a valid " + description);
			return false;
		}
		Constructor constructor;
		try {
			constructor = sam.toolkit.Process.getConstructor(target, new Class[0] );
		} catch (NoSuchMethodException e) {
			UserInterface.get().error().warn("Could not find public no-argument constructor in class " + target);
			return false;
		}
		if (constructor == null) {
			UserInterface.get().error().warn("Could not find public no-argument constructor in class " + target);
			return false;
		}
		registry.put(name, constructor);
		if (Settings.Constants.NEWLY_REGISTERED_CLASS_IS_AUTOMATICALLY_SET && (this.name == null))
			set(name);
		return true;
	}

	/** Gets the list of all registered objects **/
	public String[] registeredTargets() { return Text.getKeyStrings(registry); }

	/** Instantiates a new target registered as <CODE>name</CODE>
	 ** Note that dynamic casting to <CODE>baseType</CODE> is <em>guaranteed</em> to be successfull.
	 ** @return <CODE>null</CODE> if the building failed.
	 **/
	protected Object build(String name) {
		if (name == null)
			return null;
		Constructor constructor = null;
		try {
			constructor = (Constructor) registry.get(name);
		} catch (NoSuchElementException e) {
			UserInterface.get().error().warn("Could not find " + description + " named \"" + name + "\"");
			return null;
		} catch (ClassCastException e) {
			UserInterface.get().error().warn("Bug: " + description + " named \"" + name + "\" is not of type Constructor !");
			return null;
		}
		try {
			return constructor.newInstance( new Object[0] );
		} catch (Throwable t) {
			UserInterface.get().error().warn("Instantiation of " + description + " named \"" + name + "\" failed: "
																			 + sam.toolkit.Text.getMessage(t) );
			// Catches ClassCastException, InstantiationException, IllegalAccessException, IllegalArgumentException, InvocationTargetException
			return null;
		}
	}

	/** Resets the current target.
	 ** Subsequent calls to <CODE>get</CODE> will thus cause <em>reinstantiation</em> of the target class.
	 **/
	public void reset() {	this.target = null; }

	/** Gets the current user interface.
	 ** <p>Note that instanciation is lazy and occurs only once.
	 ** @return the current user interface
	 ** @see reset
	 **/
	public Object get() {
		if (target == null)
			target = build(name);
		if (target == null)
			if (this != UserInterface.REGISTRY)
				UserInterface.get().error().warn("Could not initialize " + description + " \"" + name + "\"");
			else
				; // System.err.println("Could not initialize " + description + " \"" + name + "\"");
		else
			;
		return target;
	}

	/** Sets and gets specified user interface.
	 ** @return the specified user interface
	 ** @see set
	 ** @see get
	 **/
	public Object get(String name) {
		set(name);
		return get();
	}

}
