
package sam.toolkit;

import sam.core.Settings;
import sam.toolkit.Fifo;


/** This class implements an asynchronous mailbox.
 **/
public class Mailbox {

	/** The firs-in first-out list of to-be-handled events **/
	private final Fifo mails = new Fifo();

	/** Wether the postmaster is to be running or not **/
	private volatile boolean continueRunning = true;

	/** General type for mail readers **/
	public static abstract interface Reader {
		/** Reads one mail **/
		public abstract void read(Object mail);
	}

	/** Builds a new <CODE>Mailbox</CODE>. **/
	public Mailbox() { 	}

	/** Starts the event postmaster thread.
	 ** Note that spawning multiple readers is allowed.
	 **/
	public synchronized void start(final Reader reader) {
		final Thread mailPostMaster = new Thread(reader + "(" + this + ")") {
			public void run() {
				try {
					while (true) {
						Object mail = null;
						synchronized(mails) {
							while ( mails.isEmpty() )
								if (continueRunning)
									mails.wait();
								else
									throw new InterruptedException("Finishing");
							mail = mails.get();
						}  // This bracket is a key to performance:
						// it allows another thread to post messages while this threads handles <CODE>mail</CODE>
						if (mail != null)
							reader.read(mail);
					}
				} catch (InterruptedException e) {
				}
			}
		};
		mailPostMaster.setPriority( Settings.Constants.MAIL_POSTMASTER_PRIORITY );
		mailPostMaster.start();
	}

	/** Posts one more mail **/
	public synchronized void post(Object mail) {
		mails.put(mail);
		notifyReaders();
	}

	/** Waits for all readers to have read all pending mails.
	 ** Note that this might be a blocking call.
	 **/
	public synchronized void waitForReaders() {
		while ( ! mails.isEmpty() )
			notifyReaders();
	}

	/** Stops the mail postmaster thread.
	 ** This actually stops <em>all</em> if more than one was launched.
	 ** Note that calling this several times is safe although of no use.
	 **/
	public synchronized void stop() {
		if (Settings.Constants.ALLOW_MULTIPLE_MAIL_STOPS || continueRunning) {
			continueRunning = false;
			notifyReaders();
		}
	}

	private void notifyReaders() {
		synchronized(mails) {
			mails.notifyAll();
		}
	}

	/** Notifies the termination of the process.
	 ** This is a blocking call that waits for all <CODE>Mailbox</CODE> to have been handled,
	 ** and then stops all mail readers.
	 ** Please understand that it is <b>not</b> equivalent to use to consecutives calls to
	 ** <CODE>mailbox.waitForReaders(); mailbox.stop()</CODE> since this would release the
	 ** lock between the two statements whereas it should <b>not</b>.
	 **/
	public synchronized void terminate() {
		waitForReaders();
		stop();
	}

	/** On destruction waits for all readers to finish reading any pending mail.
	 ** Note that this implementation does <b>not</b> call <CODE>Thread.stop()</CODE>
	 **/
	protected void finalize() { stop();	}

	/** Sample mailbox test code.
	 ** Posts all command-line arguments and spawns own reader.
	 **/
	public static void main(String[] args) {
		Mailbox mailbox = new Mailbox();
		System.err.println("Posting " + args.length + " mails");
		for (int i=0; i<args.length; ++i)
			mailbox.post(args[i]);
		System.err.println("Asynchronously reading mail");
		mailbox.start(new Reader() {
			public void read(Object mail) {
				System.err.println("Read: " + mail);
			}
		});
		System.err.println("Waiting for readers");
		mailbox.waitForReaders();
		System.err.println("Stopping readers.");
		mailbox.stop();
		System.err.println("Done.");
	}


}

