

package sam.toolkit;

import sam.ui.UserInterface;

import java.util.*;
import java.util.zip.*;
import java.util.jar.*;
import java.net.URL;
import java.awt.*;
import java.io.*;
import java.lang.reflect.Constructor;


/** General-purpose utility class grouping various tools concerning
 ** processes, threads and VM.
 **/
public class Process {
		private Process() {	}  // Purely static class

		private static final Object lock = new int[0];
		private static boolean halted = false;
		private static boolean interrupted = false;

		public static void suspend() {
			UserInterface.get().log().log(2, Process.class, "Suspended on " + new Date() );
			synchronized(lock) {
				halted = true;
			}
		}

		public static void interrupt() {
			UserInterface.get().log().log(2, Process.class, "Interrupted on " + new Date() );
			synchronized(lock) {
				interrupted = true;
			}
		}

		public static void restart() {
			synchronized(lock) {
				halted = false;
				lock.notifyAll();
			}
			UserInterface.get().log().log(2, Process.class, "Restart on " + new Date());
		}

		public static void haltIfRequired() throws InterruptedException {
			synchronized(lock) {
				if (interrupted) {
					interrupted = false;
					throw new InterruptedException("Interrupted by user");
				}
				if (halted)
					lock.wait();
			}
		}

		/** Gets the current call stack.
		 ** Note that this is <b>very slow</b> since it is implemented through exception throwing.
		 ** Moreover this might be system-dependant, VM-dependant and even compiler-dependant.
		 ** Thus it should only be used for debugging or possibly logging but <b>never</b> alter the program semantic.
		 **/
		public static String[] currentCallStack() {
			try {
				String stackTrace = null;
				// 1) Gets stack trace
				try {
					throw new Throwable("dummy throwable");
				} catch (Throwable t) {
					final CharArrayWriter charArrayWriter = new CharArrayWriter();
					t.printStackTrace( new PrintWriter(charArrayWriter) );
					stackTrace = charArrayWriter.toString();
				}
				// 2) Parses its elements
				StringTokenizer tokenizer = new StringTokenizer(stackTrace, "\n");
				String[] callStack = new String[tokenizer.countTokens() - TO_BE_SKIPPED_STACK_TRACE_LINES];
				for (int i=0 /* ONE */ ; i<TO_BE_SKIPPED_STACK_TRACE_LINES; ++i)
					tokenizer.nextToken();
				for (int i=0; i<callStack.length; ++i)
					callStack[i] = tokenizer.nextToken().trim().substring( TO_BE_SKIPPED_STACK_TRACE_LINE_PREFIX.length() );
				return callStack;
			} catch (Throwable t) {
				return new String[0];
			}
		}

		/** Gets the list of all active threads.
		 **/
		public static Thread[] getCurrentThreads() {
			ThreadGroup group = Thread.currentThread().getThreadGroup();
			try {
				while (group.getParent() != null)
					group = group.getParent();
			} catch (SecurityException e) {
				UserInterface.get().error().handleException("Could not get top-level thread group", e);
			}
			// Due to implementation concerns the <CODE>activeCount</CODE> does not return the <em>exact</em>
			// number of active threads. Thus we have to build a (hopefully) too big one, fill it and then extracts non-null elements
			Thread[] tooLargeArray = new Thread[ 2 * group.activeCount() ];
			group.enumerate(tooLargeArray);
			int count = 0;
			for ( ; count < tooLargeArray.length; ++count)
				if (tooLargeArray[count] == null)
					break;
			Thread[] wellSizedArray = new Thread[ count ];
			System.arraycopy(tooLargeArray, 0, wellSizedArray, 0, count);
			return wellSizedArray;
		}


		/** Number of top call stack lines to be skipped by <CODE>currentCallStack()</CODE> **/
		private static final int TO_BE_SKIPPED_STACK_TRACE_LINES = 2;

		/** Number of top call stack lines to be skipped by <CODE>currentCallStack()</CODE> **/
		private static final String TO_BE_SKIPPED_STACK_TRACE_LINE_PREFIX = "at ";

		/** Default search path for temporary directories
		 **	@see getDefaultTemporary
		 ** @deprecated
		 **/
		private static final String[] TEMPORARY_DIRECTORIES = { "TMP", "TEMP", "tmp", "temp", "user.home"  };

		/** @deprecated **/
		private static final String CLASSPATH_PROPERTY = "java.classpath";
		
		public static File getJarSourceArchive(Class clazz) {
	    return getBaseJarArchive( getClassUrl(clazz) );
		}
		
		public static File getClassDirectory(Class clazz) {
	    return getBaseDirectory( new File( getClassUrl(clazz).getFile() ), clazz.getPackage() );
		}
	    
		public static File getBaseDirectory(File directory, Package aPackage) {
	    if (directory == null)
				return null;
	    int packageSubpartsCount = new StringTokenizer( aPackage.getName(), "." ).countTokens();
			for (int i=0; i < packageSubpartsCount; i++)
				directory = directory.getParentFile();
			return directory;
		}
		
		public static File getBaseJarArchive(URL url) {
	    if (url == null)
				return null;
	    String classEntry = url.getPath();
	    int cut = classEntry.indexOf('!');
	    return (cut == -1) ? new File("") : new File( classEntry.substring(0, cut) );
		}
		
		
		public static URL getClassUrl(Class clazz) {
	    URL url = getClassUrl(clazz, '/');
	    return ( (url != null) || ('/' == File.separatorChar) ) ? url :	getClassUrl(clazz, File.separatorChar);
		}
		
		public static URL getClassUrl(Class clazz, char separatorChar) {
	    return clazz.getClassLoader().getResource( clazz.getName().replace('.', separatorChar) + ".class" );
		}
		
		/** @deprecated **/		
		public static String getDefaultTemporary() {
			String directory;
			for (int i=0; i<TEMPORARY_DIRECTORIES.length; i++)
				if ( (directory = System.getProperty(TEMPORARY_DIRECTORIES[i])) != null)
					return directory;
			return null;
		}

		/** @deprecated **/		
		public static void appendClassPath(String classPath) {
			if (classPath == null)
				return;
			try {
				String oldClassPath = System.getProperty(CLASSPATH_PROPERTY);
				System.setProperty(CLASSPATH_PROPERTY,
								   (oldClassPath != null) ? classPath : (classPath + ':' + oldClassPath) );
			} catch (SecurityException e) {
				UserInterface.get().error().handleException("Could not expand class path by \"" + classPath +	"\" due to security restrictions: ",
																										e);
			}
		}


		public static void exit() {
			exit(0);
		}

		public static void exit(int exitCode) {
			if (exitCode > 1)
				System.err.println("Process failed, exit code:" + exitCode);
			UserInterface.disposeFrames();
			System.gc();
			System.exit(exitCode);
		}

		/** Finds out a suitable constructor inside specified class that accepts specified arguments types.
		 **/
			public static Constructor getConstructor(Class target, Class[] args) throws NoSuchMethodException {
				try {
					Constructor[] constructors = target.getConstructors();
					for (int i=0; i<constructors.length; i++) {
						Class[] params = constructors[i].getParameterTypes();
						if (params.length == args.length) {
							if (params.length == 0)
								return constructors[i];
							int j=0;
							while ( args[j].isAssignableFrom(params[j]) )
								if (++j == params.length)
									return constructors[i];
						}
					}
					throw new NoSuchMethodException("No public " + target.getName() + "(" + args[0].getName() + ") found !");
				} catch (java.lang.Exception e) {
					throw new NoSuchMethodException("Could not find suitable " + target.getName() + "(" + Text.toString(args, "", ",", "") + ") : "
																					+ e.getMessage() );
				}
			}


}
