
/*
  This whole file is distributed under the terms of the General Public License.

  Last revision date: 2002/05/12
 */

package sam.toolkit;

import sam.core.Settings;

import java.util.*;


/** Random numbers generator.
 ** This is basically an enhanced version of the {@link java.util.Random} generator.
 ** It ties to be even more unpredictable by spawning arbitrarily scheduled consumer threads.
 ** <p>Instances of this class are fully immutable.
 **/
public class Random  {

	/** The basic number generator. **/
	private static final java.util.Random random = new java.util.Random();
		
	static {
		// Spawns a new thread which every random(10) seconds consumes the randomizers, for more randomness
		if (Settings.Constants.RANDOM_GENERATOR_CONSUMER_LIFETIME > 0)
			consume(Settings.Constants.RANDOM_GENERATOR_CONSUMER_LIFETIME);
	}

	/** Draws a boolean randomly. **/
	public static boolean nextBoolean() {
		return random.nextBoolean();
	}

	/** Draws a byte randomly.
	 **/
	public static byte nextByte() {
		return (byte) random.nextInt(256);
	}

 	/** Draws an integer randomly. **/
	public static int nextInt() {
		return (((int) nextByte() & 0xff) << 24) | (((int) nextByte() & 0xff) << 16)
				| (((int) nextByte() & 0xff) << 8) | ((int) nextByte() & 0xff);
	}

 	/** Draws a long integer randomly. **/
	public static long nextLong() {
    return (nextInt() & 0xffffffffL) << 32 | (nextInt() & 0xffffffffL);  // Note the <CODE>long</CODE> mask
	}

 	/** Draws an integer randomly in the range <CODE>0, limit - 1</CODE>.
	 ** @param limit the upper limit of the allowed range.
	 ** If <CODE>limit</CODE> is less than <CODE>1</CODE> then <CODE>0</CODE> is returned.
	 **/
	public static int nextInt(int limit) {
		return random.nextInt( (limit < 1) ? 0 : limit);
	}

 	/** Draws a long integer randomly in the range <CODE>0, limit - 1</CODE>.
	 ** @param limit the upper limit of the allowed range.
	 ** If <CODE>limit</CODE> is less than <CODE>1</CODE> then <CODE>0</CODE> is returned.
	 **/
	public static long nextLong(long limit) {
		if (limit < 1)
			limit = 1;
		long value = random.nextLong() % limit;
		return (value < 0) ? (value + limit) : value;
	}

 	/** Draws an hexadecimal sequence of specified length.
	 ** @param digits number of digits required, possibly arbitrarily large
	 ** @return an hexadecimal string of length <CODE>digits</CODE>
	 **/
	public static String nextHex(int digits) {
		StringBuffer buffer = new StringBuffer();
		for (int i=0; i< (digits / 8); i++)
			buffer.append( nextHex8(8) );
		return buffer.append( nextHex8(digits % 8) ).toString();
	}

	private static String nextHex8(int digits) {
		String rand = Integer.toHexString( new java.util.Random().nextInt() & ((0x1 << (4*digits-1)) - 1) );
		while (rand.length() < digits)
			rand = "0" + rand;
		return rand;
	}

	/** Draws one of specified <CODE>String</CODE>s randomly
	 ** @param strings the list of allowed strings
	 ** @return <CODE>null</CODE> if <CODE>strings</CODE> is <CODE>null</CODE> or a randomly chosen long among <CODE>strings</CODE> otherwise
	 **/
	public static String drawString(String[] strings) {
			return ((strings == null) || (strings.length == 0)) ? null : strings[ nextInt(strings.length) ];
	}

	/** Draws one of specified <CODE>Long</CODE>s randomly
	 ** @param longs the list of allowed longs
	 ** @return 0 if <CODE>longs</CODE> is <CODE>null</CODE> or a randomly chosen long among <CODE>longs</CODE> otherwise
	 **/
	public static long drawLong(long[] longs) {
			return ((longs == null) || (longs.length == 0)) ? 0 : longs[ nextInt(longs.length) ];
	}



	/** Spawns a new thread which at random times consums the random generator */
	public static void consume(int maximum) {
		consume(maximum / 10, maximum);
	}

	/** Spawns a new thread which at random times consums the random generator */
	public static void consume(int minimum, int maximum) {
		consume(minimum, maximum, minimum);
	}

	/** Spawns a new thread which at random times consums the random generator
	 ** <p>Multi-threading is an excellent source of random-ness, since any process (even
	 ** non-<CODE>Java</CODE> ones) of the system is involved.
	 ** @param minimum minimum duration (in milliseconds) between two consumptions. Default is <CODE>maximum / 10</CODE>
	 ** @param maximum maximum duration (in milliseconds) between two consumptions.
	 ** @param stride stride (in milliseconds) of randomly-choosen sleep times . Default is equal <CODE>minimum</CODE>.
	 **/
	public static void consume(final int minimum, int maximum, final int stride) {
		int x = (maximum - minimum) / stride;
		final int amplitude = (x < 2) ? 2 : x;
		new Thread("RandomDrawer") {
				{
					setDaemon(true);
				}
				public void run() {
					while (true) {
						try {
							sleep( nextInt(amplitude) * stride + minimum );
						} catch (InterruptedException e) {
						}
					}
				}

			}.start();
	}

	/** Class for randomizing over fixed integer sets.
	 ** This is a randomizer that randomly chooses a <em>permutation</em> on a set of integers.
	 ** Note that all <CODE>n!</CODE> permutations on <CODE>n</CODE> integers are <b>equi-probable</b>.
	 ** This class implements such a randomly chosen permutation as a standard {@link java.util.Enumeration}
	 ** on those allows integers.
	 **/
	public static class IntegerSetRandomizer implements java.util.Enumeration {

		/** Weights of each integer, decreased each time it is drawn **/
		private int[] credits;
		/** Number of still eligible (non-distinct) integers **/
		private int eligible;

		/** Builds a new randomly choosen permutation on a range of integers.
		 ** @param length the largest integer: the integer range thus is <CODE>[0 , length-1]</CODE>.
		 **/
		public IntegerSetRandomizer(int length) {
			this( buildCredits(length, 1) );
		}

		/** Builds a new randomly choosen permutation on a range of weighted integers.
		 ** @param the initial <em>weight</em> of each integer
		 ** The integer range thus is <CODE>[0 , credits.length-1]</CODE> <em>with repetitions</em>;
		 ** each <CODE>i</CODE> is repeated exactly <CODE>credits[i]</CODE> times in the whole allowed set.
		 **/
		public IntegerSetRandomizer(int[] credits) {
			this.credits = credits;
			this.eligible = 0;
			for (int i=0; i<credits.length; ++i)
				if (credits[i] > 0)
					++eligible;
		}

		/** Wether this enumeration has more elements to enumerate. **/
		public boolean hasMoreElements() {
			return eligible > 0;
		}

		/** Gets the next element to enumerate
		 ** @return an instance of {@link java.lang.Integer} containing the next integer
		 ** @throws NoSuchElementException if no more such integer is to be enumerated
		 **/
		public Object nextElement() throws NoSuchElementException {
			return new Integer( next() );
		}

		/** Gets the next integer to enumerate
		 ** @throws NoSuchElementException if no more such integer is to be enumerated
		 **/
		public int next() throws NoSuchElementException {
			if (eligible < 1)
				throw new NoSuchElementException("Can not draw any more integer, credits exhausted !");
			int elected = 0;
			try {
				for (int n = Random.nextInt(eligible); n > 0; --n)  // Skips to the n-th non-exhausted number
					while (credits[elected] <= 0)  // Skip exhausted numbers
						++elected;
			} catch (ArrayIndexOutOfBoundsException e) {
				throw new NoSuchElementException("BUG: got non-positive credits array !");
			}
			if (--credits[elected] <= 0)
				--eligible;  // One less eligible number
			return elected;
		}

		/** Builds up an uniform array **/
		private static final int[] buildCredits(int length, int value) {
			int[] credits = new int[length];
			for (int i=0; i<length; ++i)
				credits[i] = value;
			return credits;
		}

	}

	public static void main(String[] args) {
		for (int i=0; i<10; ++i)
			System.err.println("Draw " + i + ": 0x" + Text.toHexa( nextInt() ) + ", 0x" + Text.toHexa( nextLong() ) );
	}


} 
