

package sam.toolkit;

import java.util.*;
import java.io.*;
import java.text.DateFormat;
import java.lang.reflect.*;


/** Repository for various low-level utilities functions.
 ** This class only has static fields or methods.
 **/
public class Text {
		private Text() {	}  // Purely static class

		/** @deprecated **/
		public static final String USER_NAME = getUserName();
		/** @deprecated **/
		public static final String USER_GROUP = "sam";  // System-dependant and thus not available under Java; this is a conventional workaround

		/** @deprecated **/
		private static String getUserName() {
			try {
				return System.getProperty("user.name");
			} catch (SecurityException e) {
				return "user";
			}
		}

		/** Incredibly enough this is not part of the standard API ... **/
		public static int countCharacter(String s, char c) {
			if (s == null)
				return 0;
			int count = 0;
			for (int i=0; i<s.length(); )
				if (s.charAt(i++) == c)
					++ count;
			return count;
		}

		/** Drops the <CODE>null</CODE> elements of an array
		 ** @return <CODE>null</CODE> if passed array is itself <CODE>null</CODE>
		 ** Note that the runtime type of the returned array is that of passed array
		 **/
		public static Object[] dropNulls(Object[] array) {
			if (array == null)
				return null;
			// 1) Counts non-null elements
			int count = 0;
			for (int i=0; i<array.length; ++i)
				if (array[i] != null)
					++ count;
			// 2) Stores non-null elements
			Object[] nonNullArray = (Object[]) allocateNewArray(array, count);  // This cast will always succeed
			if (nonNullArray == null)
				return null;
			for (int i=0, j=0; i<array.length; ++i)
				if (array[i] != null)
					nonNullArray[ j++ ] = array[i];
			return nonNullArray;
		}

		/** Allocates a new array with the same base type as the specified array.
		 ** @return <CODE>null</CODE> if passed type is itself <CODE>null</CODE> or if instantiation fails
		 **/
		public static Object allocateNewArray(Object array, int size) {
			return (array == null) ? null : allocateNewArray(array.getClass().getComponentType(), size);
		}

		/** Allocates a new array of specified <em>base</em> type.
		 ** @return <CODE>null</CODE> if passed type is itself <CODE>null</CODE> or if instantiation fails
		 **/
		public static Object allocateNewArray(Class type, int size) {
			if (type == null)
				return null;
			try {
				return Array.newInstance(type, size);
			} catch (Throwable t) {
				return null;
			}
		}

		/** Gets a sub-array of an array.
		 ** @param array the source array
		 ** @param begin the beginning of the sub-array
		 ** @param length the sub-array length
		 ** @return a newly allocated array <b>of same runtime type</b> as passed array
		 **/
		public static Object subArray(Object array, int begin, int length) {
			if (array == null)
				return null;
			Object subarray = allocateNewArray(array, length);
			System.arraycopy(array, begin, subarray, 0, length);
			return subarray;
		}

		/** Gets the first word of a two-word string **/
		public static String firstWord(String pairOfWords, char delimiter) {
			if (pairOfWords == null)
				return null;
			int delimiterIndex = pairOfWords.indexOf(delimiter);
			return (delimiterIndex < 0) ? pairOfWords : pairOfWords.substring(0, delimiterIndex);			
		}

		/** Gets the second word of a two-word string **/
		public static String secondWord(String pairOfWords, char delimiter) {
			if (pairOfWords == null)
				return null;
			int delimiterIndex = pairOfWords.indexOf(delimiter);
			return (delimiterIndex < 0) ? "" : pairOfWords.substring(delimiterIndex + 1);			
		}


		/** Splits a <CODE>String</CODE> accross carriage-return delimiters **/
		public static String[] splitLines(String text) {
			return split(text, "\n");
		}

		/** Splits a <CODE>String</CODE> accross specified delimiters
		 **/
		public static String[] split(String text, String split) {
			if (text == null)
				return new String[0];
			if (split == null)
				return new String[] { text };
			Vector buffer = new Vector();
			for (Enumeration tokens = new StringTokenizer(text, split); tokens.hasMoreElements(); )
				buffer.add( tokens.nextElement() );
			String[] lines = new String[ buffer.size() ];
			buffer.copyInto(lines);
			return lines;
		}

		public static String[] merge(String[][] stringArrays) {
			if (stringArrays == null)
				return null;
			// 1) Counts total size
			int length = 0;
			for (int i=0; i<stringArrays.length; ++i)
				if (stringArrays[i] != null)
					length += stringArrays[i].length;
			// 2) Creates the array and copies contents into it
			String[] array = new String[length];
			for (int i=0, offset=0; i<stringArrays.length; ++i)
				if (stringArrays[i] != null) {
					System.arraycopy(stringArrays[i], 0, array, offset, stringArrays[i].length);
					offset += stringArrays[i].length;
				}
			return array;
		}

		/** Gets a snapshot of the valid key set of a given <CODE>Map</CODE>.
		 ** @param map the map
		 ** @param runtimeType an array of same runtime array type as the one desired, e.g. <CODE>new String[0]</CODE>
		 ** @param an array of the same runtime type as <CODE>runtimeType</CODE>
		 **/
		public static Object[] getKeys(final Map map, Object[] runtimeType) {
			return (map == null) ? null : (String[]) map.keySet().toArray(runtimeType);
		}

		/** Gets a snapshot of the valid key set of a given <CODE>Map</CODE>.
		 ** @param map the map containing <b>only <CODE>String</CODE></b> instances or <CODE>null</CODE>s.
		 **/
		public static String[] getKeyStrings(final Map map) { return (String[]) getKeys(map, new String[0]); }

		/** Converts one integer to corresponding character according to the local <CODE>Charset</CODE>
		 ** <b>Warning:</b> this method is highly inefficient.
		 ** Moreover, it is intrinsequently buggy since some charsets require <CODE>char</CODE> be encoded
		 ** onto <b>several</b> <CODE>byte</CODE>s
		 **/
		public static char intToChar(int value) {
			return (value == -1) ? (char) 0 : bytesToChar(new byte[] { (byte) value } );
		}

		/** Converts one integer to corresponding character according to the local <CODE>Charset</CODE>
		 ** <b>Warning:</b> this method is highly inefficient.
		 **/
		public static char bytesToChar(byte[] value) {
			try {
				return (char) new InputStreamReader( new ByteArrayInputStream(value) ).read();
			} catch (IOException e) {
				return (char) 0;
			}
		}

		/** Turns a vector supposedly containing <CODE>String</CODE>s only to a <CODE>String[]</CODE> array.
		 ** @throws ClassCastException if the vector does not contain only <CODE>String</CODE> instances
		 **/
		public static String[] toStringArray(Vector vector) {
			try {
				return (String[]) vector.toArray(new String[0]);
			} catch (ClassCastException e) {
				throw new ClassCastException("Could not extract String[] array from vector containing incompatible objects: " + e.getMessage() );
			}
		}

		/** Appends specified elements to the end of the vector.
		 ** @return the original vector, or a new one if this was <CODE>null</CODE> and <CODE>additionnalElements</CODE> was not.
		 **/
		public static Vector addArray(Vector vector, Object[] additionnalElements) {
			if (additionnalElements == null)
				return vector;
			if (vector == null)
				vector = new Vector();
			vector.ensureCapacity(vector.size() + additionnalElements.length);  // Performance hack: grows the vector once only
			for (int i=0; i<additionnalElements.length; )
				vector.addElement( additionnalElements[i++] );
			return vector;
		}


		/** Compares two <CODE>byte</CODE> arrays
		 ** @param x the first <CODE>byte[]</CODE> array
		 ** @param xBegin the beginning of the first array region
		 ** @param y the second <CODE>byte[]</CODE> array
		 ** @param yBegin the beginning of the second array region
		 ** @return <CODE>true</CODE> iff both arrays are <CODE>null</CODE> or if both regions match
		 **/
		public static boolean compare(byte[] x, byte[] y) {
			return compare(x, y, 0);
		}


		/** Compares two <CODE>byte</CODE> arrays
		 ** @param x the first <CODE>byte[]</CODE> array
		 ** @param xBegin the beginning of the first array region
		 ** @param y the second <CODE>byte[]</CODE> array
		 ** @param yBegin the beginning of the second array region
		 ** @return <CODE>true</CODE> iff both arrays are <CODE>null</CODE> or if both regions match
		 **/
		public static boolean compare(byte[] x, byte[] y, int yBegin) {
			return compare(x, 0, y, yBegin);
		}

		/** Compares two <CODE>byte</CODE> arrays
		 ** @param x the first <CODE>byte[]</CODE> array
		 ** @param xBegin the beginning of the first array region
		 ** @param xLength the length of the first array region
		 ** @param y the second <CODE>byte[]</CODE> array
		 ** @param yBegin the beginning of the second array region
		 ** @return <CODE>true</CODE> iff both arrays are <CODE>null</CODE> or if both regions match
		 **/
		public static boolean compare(byte[] x, int xBegin, byte[] y, int yBegin) {
			return (y == null) ? (x == null) : compare(x, xBegin, y, yBegin, y.length - yBegin);
		}

		/** Compares two <CODE>byte</CODE> arrays
		 ** @param x the first <CODE>byte[]</CODE> array
		 ** @param xBegin the beginning of the first array region
		 ** @param y the second <CODE>byte[]</CODE> array
		 ** @param yBegin the beginning of the second array region
		 ** @param yLength the length of the second array region
		 ** @return <CODE>true</CODE> iff both arrays are <CODE>null</CODE> or if both regions match
		 **/
		public static boolean compare(byte[] x, int xBegin, byte[] y, int yBegin, int yLength) {
			return compare(x, xBegin, yLength, y, yBegin, yLength);
		}

		/** Compares two <CODE>byte</CODE> arrays
		 ** @param x the first <CODE>byte[]</CODE> array
		 ** @param xBegin the beginning of the first array region
		 ** @param xLength the length of the first array region
		 ** @param y the second <CODE>byte[]</CODE> array
		 ** @param yBegin the beginning of the second array region
		 ** @param yLength the length of the second array region
		 ** @return <CODE>true</CODE> iff both arrays are <CODE>null</CODE> or if both regions match
		 **/
		public static boolean compare(byte[] x, int xBegin, int xLength, byte[] y, int yBegin, int yLength) {
			if ((x == null) || (y == null))
				return (x == null) & (y == null);
			if (xLength != yLength)
				return false;
			try {
				for (int i=xBegin, j=yBegin, t = xBegin + xLength; i<t; ++i, ++j)
					if (x[i] != y[j])
						return false;
				return true;
			} catch (ArrayIndexOutOfBoundsException e) {
				return false;
			}
		}


		/** Enumerates elements of an array.
		 ** <p>Incredibly enough this is not part of the standard API.
		 **/
		public static Enumeration enumerate(Object[] array) {	return addArray(null, array).elements(); }


		/** Parses an integer, or returns zero if incorrectly formatted **/
		public static int parseInteger(String value) {
			return parseInteger(value, 0);
		}

		/** Parses an integer, or returns <CODE>defaultValue</CODE> if incorrectly formatted **/
		public static int parseInteger(String value, int defaultValue) {
			if (value == null)
				return defaultValue;
			try {
				return Integer.valueOf(value).intValue();
			}	catch (NumberFormatException e) {
				return defaultValue;
			}
		}

		/** Converts a <CODE>long</CODE> integral number to its hexadecimal representation, without trailing zeros **/
		public static String toHexa(long n) { return Long.toHexString(n);	}

		/** Converts an <CODE>int</CODE> integral number to its hexadecimal representation, without trailing zeros **/
		public static String toHexa(int n) { return Integer.toHexString(n);	}

		/** Recovers an <CODE>int</CODE> integral number from its hexadecimal representation **/
		public static long fromHexa(String s) {
			if (s.length() < 16) // 16 * 4 = 64 bit
				return Long.parseLong(s, 16);
			// Otherwise, there is a problem (a bug ?!) in the jdk
			return (Long.parseLong(s.substring(0, 8), 16) << 32) | Long.parseLong(s.substring(8), 16);
		}

		/** Shortens a <CODE>String</CODE> by cutting it on the right
		 ** @param text the text to be shortened
		 ** @param length the desired length
		 **/
		public static String summarize(String text, int length) {
			return summarize(text, length, true);
		}

		public static String summarize(String text, int length, boolean cutOnTheRight) {
			String summary = (text == null) ? "(N/A)" : splitLines(text)[0];
			if (summary.length() > length)
				summary = cutOnTheRight ? summary.substring(0, length) : summary.substring(summary.length() - length);
			return summary;
		}

		/** Shrinks or expands specified text do desired length
		 ** @param text the text to be modified
		 ** @param length the new desired length
		 ** If <CODE>text</CODE> is <CODE>null</CODE> of too small, spaces are appended at the <em>end</em>.
		 ** If <CODE>>text</CODE> is too long then <em>leading</em> characters ar dropped.
		 **/
		public static String left(String text, int length) {
			if (length <= 0)
				return "";
			if (text == null)
				text = "";
			return (text.length() < length) ? ( text + StringFactory.SPACES.get(length - text.length()) )
					:	text.substring(text.length() - length);
		}

		/** Shrinks or expands specified text do desired length
		 ** @param text the text to be modified
		 ** @param length the new desired length
		 ** If <CODE>text</CODE> is <CODE>null</CODE> of too small, spaces are appended at the <em>beginning</em>.
		 ** If <CODE>>text</CODE> is too long then <em>leading</em> characters ar dropped.
		 **/
		public static String right(String text, int length) {
			if (length <= 0)
				return "";
			if (text == null)
				text = "";
			return (text.length() < length) ? (StringFactory.SPACES.get(length - text.length()) + text)	:	text.substring(0, length);
		}


		/** Counts the number of lines of a text.
		 ** @param text the desired text
		 ** @return the number of lines
		 **/
		public static int countLines(String text) {
			int count = new StringTokenizer(text, "\n").countTokens();
			return (count < 1) ? 1 : count;
		}
	
		public static String getMessage(Throwable t) {
			if (t == null)
				return "";
			String message = t.getMessage();
			return ( (message == null) || ("".equals(message)) ) ? getLastClassName( t.getClass() ) :
				( getLastClassName( t.getClass() ) + ":      " + message );
		}

		/** Gets the <CODE>Java</CODE> class name without its package qualification.
		 ** This is consistent with non-anonymous inner classes as long as they are named with <CODE>'$'</CODE>
		 ** as a separator.
		 **/	
		public static String getLastClassName(Class clazz) {
			String name = clazz.getName();
			return name.substring(name.lastIndexOf('.') + 1, name.length()).replace('$', '.');
		}

		/** Gets the <CODE>String</CODE> representation of an array.
		 ** It is incredible that this does <b>not</b> exist in the standard <CODE>jre</CODE> ...
		 **/
		public static String toString(Object[] array) {
			return toString(array, ", ");
		}

		/** Gets the <CODE>String</CODE> representation of an array.
		 ** It is incredible that this does <b>not</b> exist in the standard <CODE>jre</CODE> ...
		 ** @param delimiter the delimiter printed between each pair of elements of the array,
		 ** e.g. <CODE>", "</CODE>
		 **/
		public static String toString(Object[] array, String delimiter) {
			return toString(array, "[", delimiter, "]");
		}

		/** Gets the <CODE>String</CODE> representation of an array.
		 ** It is incredible that this does <b>not</b> exist in the standard <CODE>jre</CODE> ...
		 ** @param delimiter the delimiter printed between each pair of elements of the array,
		 ** e.g. <CODE>", "</CODE>
		 ** @param prefix the prefix first printed, e.g. <CODE>"["</CODE>
		 ** @param postfix the postfix printed at last, e.g. <CODE>"]"</CODE>
		 **/
		public static String toString(Object[] array, String prefix, String delimiter, String postfix) {
			if (array == null)
				return "null";
			StringBuffer buffer = new StringBuffer().append(prefix);
			for (int i=0; i<array.length; i++) {
				if (i>0)
					buffer.append(delimiter);
				buffer.append( array[i] );
			}
			return buffer.append(postfix).toString();
		}

		/** Turns an array of integers to the array of their <CODE>String</CODE> representation
		 ** @return <CODE>null</CODE> if <CODE>array</CODE> is <CODE>null</CODE>
		 **/
		public static String[] toStrings(int[] array) {
			if (array == null)
				return null;
			String[] strings = new String[array.length];
			for (int i=0; i<strings.length; ++i)
				strings[i] = Integer.toString( array[i] );
			return strings;
		}

		/** Turns an array of integers to the array of their <CODE>String</CODE> representation
		 ** @return <CODE>null</CODE> if <CODE>array</CODE> is <CODE>null</CODE>, or in case of
		 ** invalid number format
		 **/
		public static int[] fromStrings(String[] strings) {
			if (strings == null)
				return null;
			int[] ints = new int[strings.length];
			try {
				for (int i=0; i<strings.length; ++i)
					ints[i] = Integer.valueOf( strings[i] ).intValue();
				return ints;
			} catch (NumberFormatException e) {
				sam.ui.UserInterface.get().error().warn("Invalid integer format: " + e.getMessage() );
				return null;
			}
		}

		/** Wether specified array contains specified object.
		 ** It is incredible that this does <b>not</b> exist in the standard <CODE>jre</CODE> ...
		 **/
		public static boolean contains(Object[] array, Object object) {
			if ( (array == null) || (array.length == 0))
				return false;
			for (int i=0; i<array.length; ++i)
				if ( (array[i] == object) || ((array[i] != null) && (array[i].equals(object))) )
					return true;
			return false;
		}

		/** Replaces all occurences of a pattern in a text
		 ** @param text the text to be replaced in
		 ** @param pattern the pattern to be replaced
		 ** @param alternative the desired substitution to <CODE>pattern</CODE>
		 **/
		public static String replace(String text, String pattern, String alternative) {
			char[] chars = text.toCharArray();
			StringBuffer buffer = new StringBuffer(chars.length);
			int index = 0;
			int cut;
			while ( (cut = text.indexOf(pattern, index)) >= 0 ) {
				buffer.append(chars, index, cut - index);
				buffer.append(alternative);
				index = cut + pattern.length();
			}
			return buffer.append(chars, index, chars.length - index).toString();
		}

		/** Gets a human-readable <CODE>String</CODE> representation of a size.
		 ** This is basically what a <CODE>du -h</CODE> would do to <CODE>du</CODE>
		 ** Note that the multiplicator is <CODE>1024</CODE> for "kilo" and
		 ** <CODE>1048576</CODE> for "mega"
		 **/
		public static String getSizeString(long bytes) {
			if (bytes > 1048576)
				return getOneDecimalString(bytes, 1048576) + " MB";
			else if (bytes > 1024)
				return getOneDecimalString(bytes, 1024) + " kB";
			else
				return bytes + " bytes";
		}

		/** Gets the representation of <CODE>ratio</CODE> with a one-decimal-digit precision.
		 **/
		public static String getOneDecimalString(double ratio) { return getOneDecimalString((long) (ratio * 1000), 1000);	}

		/** Gets the representation of <CODE>value / divisor</CODE> with a one-decimal-digit precision.
		 **/
		public static String getOneDecimalString(long value, long divisor) {
			if (divisor == 0)
				return "";
			String sign = ( ((value >= 0) && (divisor >= 0)) || ((value <= 0) && (divisor <= 0)) ) ? "" : "-";
			value = (value < 0) ? (- value) : value;
			divisor = (divisor < 0) ? (- divisor) : divisor;
			if (value > 10 * divisor)
				return sign + (value / divisor);
			else
				return sign + (value / divisor) + "." + (10 * value / divisor % 10);
		}

		/** Gets the percentage corresponding to specified ratio.
		 ** @param achieved the numerator supposedly non-negative
		 ** @param limit the divisor supposedly non-negative
		 ** @return 100 times the ratio (rounded)
		 **/
		public static int getPercentage(long achieved, long limit) {
			if ((achieved < 0L) || (limit < 0L))
				return 0;
			return (limit == 0L) ? 100 : (int) ((achieved * 100) / limit);
		}

		/** @deprecated **/
		public static int getPercent(int done, int total, boolean direction) {
			return getPercent((long) done, (long) total, direction);
		}

		/** @deprecated **/
		public static int getPercent(long done, long total, boolean direction) {
			return 100 - (int) Math.floor(100 * ( direction ? ((double)done / (double)total) :
											 ((double)total / (double)done) ) );
		}

		/** @deprecated **/
		private static final DateFormat DATE = DateFormat.getDateInstance(DateFormat.SHORT),
			DATE_TIME = DateFormat.getDateTimeInstance(DateFormat.MEDIUM, DateFormat.MEDIUM),
			TIME = DateFormat.getTimeInstance(DateFormat.MEDIUM);

		/** @deprecated **/
		public static String getDateTime(long date) {
			return DATE_TIME.format( new Date(date) );
		}

		/** @deprecated **/
 		public static String getDate() {
			return getDate( System.currentTimeMillis() );
		}

		/** @deprecated **/
 		public static String getDate(long date) {
			return DATE.format( new Date(date) );
		}

		/** @deprecated **/
		public static String getTime() {
			return getTime( System.currentTimeMillis() );
		}

		/** @deprecated **/
		public static String getTime(long date) {
			return TIME.format( new Date(date) );
		}

		/** @deprecated **/
		public static String getDuration(long duration) {
			duration = duration / 1000;
			StringBuffer buffer = new StringBuffer();
			long days = duration / 86400;
			if (days > 0)
				if (days > 1)
					buffer.append(days).append(" days ");
				else
					buffer.append("one day ");
			duration %= 86400;
			append2Digits(buffer, duration / 3600);
			buffer.append(':');
			duration %= 3600;
			append2Digits(buffer, duration / 60);
			buffer.append(':');
			duration %= 60;
			append2Digits(buffer, duration);
			return buffer.toString();
		}

		// Incredible that this does NOT exist in the standard API ...
		private static void append2Digits(StringBuffer buffer, long value) {
			buffer.append(value / 10);
			buffer.append(value % 10);
		}

}
