
package sam.toolkit;

import sam.ui.UserInterface;


/** This class implements a multi-threading workshop.
 **
 ** It basically allows to dispatch any number of
 ** <CODE>Runnable</CODE>tasks accross a <b>fixed</b>
 ** number of working <CODE>Thread>/CODE>s.
 **/
/* Sucessfully tested on 20020121-18h30 by glacote
 */
public class Workshop {
	/** Enable very strict runtime checks **/
	private static final boolean PARANOIA = true;
	/** Wether an already started <CODE>Worker</CODE> can be restarted after having finished **/
	private static final boolean REUSE_WORKERS = false;


	public static class IllegalWorkerStateException extends IllegalThreadStateException {
		public IllegalWorkerStateException(String s) { super(s); }
	}

	/** The list of all workers. @see Worker **/
	private Worker[] workers;
	/** The highest position in <CODE>workers</CODE> of a
	 ** ready (not currently running) worker, which is
	 ** a number between <CODE>-1</CODE> and <CODE>workers.length - 1</CODE>.
	 **/
	private int lastReadyWorker;

	/** @param numberOfWorkers the <b>fixed</b> maximum number of simultaneous active tasks **/
	public Workshop(int numberOfWorkers) {
		if (numberOfWorkers < 1)
			throw new IllegalArgumentException("You can not manage a workshop with " + numberOfWorkers + " workers !");
		this.workers = new Worker[numberOfWorkers];
		for (int i=0; i<workers.length; ++i)
			workers[i] = new Worker();
		lastReadyWorker = workers.length - 1;
	}

	/** @return the maximum number of simultaneous active tasks **/
	public int numberOfWorkers() {
		return workers.length;
	}


	/** Starts specified task as soon as a worker is ready to run it.
	 ** Note that this turns to a <b>blocking call</b> whenever no
	 ** such worker exists, until one active threads completes its actual job.
	 **/
	public synchronized void startAsSoonAsPossible(final Runnable task) throws IllegalWorkerStateException {
		// Note the <CODE>while</CODE> instead of an <CODE>if</CODE> (<CODE>notifyAll()</CODE> ...)
		try {
			while (lastReadyWorker < 0)
				wait();  // Which actually releases the lock
			workers[ lastReadyWorker -- ].startNow(task);
		} catch (InterruptedException e) {
			throw new IllegalWorkerStateException("BUG: Interrupted while waiting for a worker to be ready: " + e.getMessage() );
		}
	}

	/** Waits for all <b>currently running</b> tasks be completed,
	 ** which excludes any pending and not currently running task.
	 ** Thus this method should only be called <b>inside the same thread</b>
	 ** as that which calls <CODE>startAsSoonAsPossible</CODE>
	 ** (which is a blocking call). In this way, calling <CODE>waitForCompletion</CODE>
	 ** actually leads to waiting for all <em>pending</em> tasks be completed.
	 **/
	public synchronized void waitForCompletion() throws IllegalWorkerStateException {
		// Note the <CODE>while</CODE> instead of an <CODE>if</CODE> (<CODE>notifyAll()</CODE> ...)
		try {
			while (lastReadyWorker < workers.length - 1)
				wait();  // Which actually releases the lock
		} catch (InterruptedException e) {
			throw new IllegalWorkerStateException("Interrupted while waiting for all tasks to be completed: " + e.getMessage() );
		}
	}



	/** Internal class for really running tasks.
	 **/
	protected class Worker extends Thread {
		private Runnable task = null;

		protected synchronized void startNow(Runnable task) throws IllegalWorkerStateException {
			if (PARANOIA && (this.task != null))
				throw new IllegalWorkerStateException("Attempt to assign a task to a busy worker (REVOLUTION) !");
			this.task = task;
			start();
		}

		public synchronized void run() {
			task.run();  // Here the real job; could catch and forward exceptions here ...
			synchronized(Workshop.this) {
				this.task = null;
				workers[ ++ lastReadyWorker ] = REUSE_WORKERS ? this : new Worker();  // Free this worker
				Workshop.this.notify();
			}
		}

	}

	/** Sample test code.
	 **/
	public static void main(String[] args) throws Throwable {
		Workshop workshop = new Workshop( Integer.valueOf(args[0]).intValue() );
		final int n = Integer.valueOf(args[1]).intValue();
		System.err.println("Spawning " + n + " tasks across " + workshop.numberOfWorkers() + " threads");
		for (int i=0; i<n; ++i) {
			final int _i = i;
			workshop.startAsSoonAsPossible(new Runnable() {
				public String toString() { return "Task" + _i; }
				public void run() {
					System.err.println(" " + this + " going to sleep.");
					try {
						Thread.currentThread().sleep(500);
					} catch (InterruptedException e) {
					}
					System.err.println(" " + this + " awoken.");
					System.err.println(" " + this + " completed.");
				}
			});
		}
		workshop.waitForCompletion();
	}


}
