
package sam.ui;


/** Abstract error interface
 **/
public abstract class ErrorInterface {

	/** Called in case of an unhandled parameter being parsed from command-line.
	 ** This could either ignore it (and warns the user about that) through the <CODE>warn</CODE> method
	 ** or fail through the <CODE>fail</CODE> method.
	 ** Note that it is <b>not</b> possible to <em>handle</em> the parameter (e.g. setting adequate value
	 ** into <CODE>ModusOperandi>/CODE>) since this would be a design mistake.
	 **/
	public abstract void unhandledParameter(String parameter);

	/** Called whenever an archive of incompatible layout version is submitted for extraction.
	 ** This could either ignore the problem (and warns the user about that) through the <CODE>warn</CODE> method
	 ** or fail through the <CODE>fail</CODE> method.
	 ** Note that attempts to continue extraction are subjects to subsequent errors as long as the layout
	 ** versioning is what it is meant to be.
	 ** A conservative workaround is to always <b>use the included runtime of the archive</b> (which is
	 ** by construction up-to-date) instead of external binaries.
	 **/
	public abstract void incompatibleArchiveLayout(int version);

	/** Attempts to handle specified exception.
	 ** Note that this might default to failing through the <CODE>fail</CODE> methods.
	 ** However the program is expected to pursue its execution as the exception is not believed
	 ** to be fatal.
	 **/
	public abstract void handleException(String message, java.lang.Throwable e);

	/** Warns the user about a dangerous, possibly inconsistant or unusal event.
	 **/
	public abstract void warn(String message);

	/** @throws Abort whenever the whole program should abort. **/
	protected abstract void failure(String message) throws UserInterface.Abort;

	/** Aborts the program due to specified <CODE>Throwable</CODE>.
	 ** Note that aborting is implemented through exception throwing rather than abrupt VM interruption
	 ** (aka <CODE>System.exit()</CODE>) for obvious reasons.
	 ** @param error the <CODE>Throwable</CODE> that caused the fatal error
	 **/
	public final void fail(Throwable error) throws UserInterface.Abort {
		fail("Fatal error: " + sam.toolkit.Text.getMessage(error) );
	}

	/** Aborts the program due to specified <CODE>Throwable</CODE>.
	 ** Note that aborting is implemented through exception throwing rather than abrupt VM interruption
	 ** (aka <CODE>System.exit()</CODE>) for obvious reasons.
	 ** @param message a human-readable explanation of the error
	 ** This might include a workaround suggestion
	 ** @param error the <CODE>Throwable</CODE> that caused the fatal error
	 **/
	public final void fail(String message, Throwable error) throws UserInterface.Abort {
		fail(message + ": " + sam.toolkit.Text.getMessage(error) );
	}

	/** Aborts the program.
	 ** Note that aborting is implemented through exception throwing rather than abrupt VM interruption
	 ** (aka <CODE>System.exit()</CODE>) for obvious reasons.
	 ** @param message a human-readable explanation of the error
	 ** This might include a workaround suggestion
	 **/
	public final void fail(String message) throws UserInterface.Abort {
		failure(message);
		throw new UserInterface.Abort(message);
	}

	/** Simple test code.
	 **/
	public void test() {
		try {
			System.out.println("Testing error interface implemented by " + getClass().getName() );

			try {
				System.out.println("\tTesting \"unhandledParameter\"");
				unhandledParameter("dummy-test-parameter");
				System.out.println("\t> \"unhandledParameter\" did not abort");
			} catch (UserInterface.Abort abort) {
				System.out.println("\tCall to \"unhandledParameter\" made the program abort: " + abort.getMessage() );
			}

			try {
				System.out.println("\tTesting \"incompatibleArchiveLayout\"");
				incompatibleArchiveLayout(12345);
				System.out.println("\t> \"incompatibleArchiveLayout\" did not abort");
			} catch (UserInterface.Abort abort) {
				System.out.println("\tCall to \"incompatibleArchiveLayout\" made the program abort: " + abort.getMessage() );
			}

			try {
				System.out.println("\tTesting \"handleException\"");
				handleException("Test exception", new java.lang.Throwable("This is a test exception !"));
				System.out.println("\t> \"handleException\" did not abort");
			} catch (UserInterface.Abort abort) {
				System.out.println("\tCall to \"handleException\" made the program abort: " + abort.getMessage()
													 + ", but it should try not to !");
			}

			try {
				System.out.println("\tTesting \"warn\"");
				warn("This is a test warning");
				System.out.println("\t> \"warn\" did not abort, OK.");
			} catch (UserInterface.Abort abort) {
				System.err.println("\tWARNING: Call to \"warn\" made the program abort: " + abort.getMessage() + " whereas this is forbidden !");
				return;
			}

			try {
				System.out.println("\tTesting \"failure\"");
				fail("Test failure");
				System.err.println("\t> \"failure\" did not abort, whereas it should have !");
				return;
			} catch (UserInterface.Abort abort) {
				System.out.println("\tCall to \"fail\" made the program abort: " + abort.getMessage() );
			}

		} catch (Throwable t) {
			System.err.println("Test of ruling interface failed: " + sam.toolkit.Text.getMessage(t) );
		}
	}



}
