
package sam.ui;

import sam.core.Settings;
import sam.toolkit.Mailbox;
import sam.toolkit.StringFactory;


/** General purpose log interface.
 ** This is an abstraction for all log events : it defines how to select the log level, to conditionnally log a new message,
 ** or to disable logging.
 ** The effective way log messages are displayed is <em>not</em> defined in this <CODE>abstract</CODE> but rather in its subclasses.
 ** @see UserInterface
 **/
public abstract class LogInterface {

	/** The repository of all log messages.
	 ** This is a <CODE>static</CODE> member as it is shared by <em>all</em> <CODE>LogInterface</CODE>
	 ** instances.
	 ** @see #start
	 ** @see #stop
	 **/
	private static final Mailbox MESSAGES = new Mailbox();

	// This statement MUST be *after* the <CODE>Messages</CODE> initialization
	static {
		if (Settings.Constants.ALWAYS_STARTS_LOG)
			start();
	}

	/** The factory for efficiently building prefixes. **/
	private static final StringFactory PREFIX = new StringFactory(Settings.Constants.LOG_INDENTATION);

	/** The log level upper limit.
	 ** Negative values disable logging.
	 ** @see #setLevel
	 **/
	private int levelLimit = Settings.Defaults.LOG_LEVEL;

	protected LogInterface() { }

	/** Gets the current log level upper limit. **/
	protected int getLevel() { return levelLimit;	}

	/** Sets the log level upper limit.
	 ** Negative values disable logging.
	 ** @param level the desired log level for subsequent calls
	 ** which can not be greater than <CODE>Settings.Defaults.MAXIMUM_LOG_LEVEL</CODE>
	 **/
	public void setLevel(int level) {
		this.levelLimit = (level >= Settings.Constants.MAXIMUM_LOG_LEVEL) ? Settings.Constants.MAXIMUM_LOG_LEVEL :
				(level <= Settings.Constants.MINIMUM_LOG_LEVEL) ? Settings.Constants.MINIMUM_LOG_LEVEL : level;
	}

	/** The last set log level **/
	private int lastEnabledLevel = Settings.Constants.MAXIMUM_LOG_LEVEL;

	/** Disables any logging.
	 ** No more log message will be allowed until a call to <CODE>enable()</CODE>
	 ** @see #getLevel
	 **/
	public void disable() {
		lastEnabledLevel = levelLimit;
		setLevel( Settings.Constants.MINIMUM_LOG_LEVEL );
	}

	/** Enables logging again.
	 ** If a call to {@link setLevel} had been made before then that log level is reset back.
	 ** Otherwise <em>all</em> log events are unconditionnally allowed until a call to
	 ** either <CODE>setLevel()</CODE> or to <CODE>disable()</CODE>
	 ** @see #getLevel
	 **/
	public void enable() {
		setLevel(lastEnabledLevel);
	}

	/** Logs a new event.
	 ** The message is actually logged if, and only if the specified log level is <em>lower</em>
	 ** than or equal to the current log level.
	 ** <p>Note that the <CODE>Class</CODE> parameter is not used in this implementation, although
	 ** it is meant to control the logging authorizations on a per-class basis
	 ** @param level the log level of this event, higher meaning <em>lower</em> priority
	 ** @param clazz the class issuing this log eveng
	 ** @param message the log message
	 **/
	public void log(int level, Class clazz, String message) {
		if (level > levelLimit)
			return;
		post(PREFIX.get(level) + message);
	}

	/** Actually handles a log message.
	 ** This is the very method meant to be overridden and to do the logging job.
	 **/
	abstract protected void log(String logMessage);

	/** Posts a new log message
	 ** Message processing is asynchronous
	 ** @see #start
	 ** @see #stop
	 **/
	protected static void post(String logMessage) { MESSAGES.post(logMessage); } 

	/** Starts logging.
	 ** Spawns a new thread which will log all sent log messages.
	 ** Note that the actual <CODE>UserInterface</CODE> used to handle log messages
	 ** may vary according to calls made to <CODE>UserInterface.set</CODE>
	 ** @see sam.ui.UserInterface#set
	 **/
	public static void start() {
		Mailbox.Reader reader = new Mailbox.Reader() {
			public void read(Object mail) { UserInterface.get().log().log( (String) mail ); }
		};
		MESSAGES.start(reader);
	}

	/** Flushes any pending log message.
	 ** This a blocking call that waits for all pending log messages to have been properly handled.
	 ** <b>Warning:</b> in particular that this is a dead call if no call to <CODE>start()</CODE>
	 ** was previously made.
	 **/
	public static void flush() {
		MESSAGES.waitForReaders();
	}

	/** Stops logging.
	 ** This stops logging and flushes any pending log message.
	 ** This is a blocking call.
	 ** @see #flush
	 **/
	public static void stop() {
		flush();
		MESSAGES.stop();
	}

	/** Simple test code.
	 **/
	public void test() {
		try {
			System.out.println("Testing log interface implemented by " + getClass().getName() );
			for (int i=-1; i<10; ++i)
				log(i, LogInterface.class, "Test log #" + i);
		} catch (Throwable t) {
			System.err.println("Test of ruling interface failed: " + sam.toolkit.Text.getMessage(t) );
		}
		LogInterface.flush();
	}

}
