
package sam.ui;

import sam.cipher.Specification;
import sam.cipher.Cipher;
import sam.toolkit.Text;
import sam.core.Settings;

import java.io.File;


/** Generic ruling interface type.
 ** A ruling interface is in charge of parsing command-line
 ** and interactively prompt the user for missing parameters
 **/
public abstract class RulingInterface {

	/** Prompts the user for entering a string.
	 ** This is a basic and general-purpose method.
	 ** @param title the title (or <em>type</em>) of the value to be chosen
	 ** @param suggestedValue the suggested (yet invalid) value
	 ** In current implementation this will always be <CODE>null</CODE> since this is the
	 ** only "invalid" <CODE>String</CODE>
	 **/
	protected abstract String enterString(String title, String suggestedValue);

	/** Gets a user-supplied string.
	 ** This is a basic and general-purpose method.
	 ** @param title the title (or <em>type</em>) of the value to be chosen
	 ** @param suggestedValue the suggested value to be returned if not <CODE>null</CODE>
	 **/
	public String getString(String title, String suggestedValue) {
		String value = enterString( (title == null) ? "" : title, (suggestedValue == null) ? "" : suggestedValue);
		return (value == null) ? "" : value;
	}

	/** Prompts the user for entering an integral number.
	 ** @param title the title (or <em>type</em>) of the value to be chosen
	 ** @param suggestedValue the suggested number to be returned if valid
	 ** @param min the minimum allowed value, or <CODE>Integer.MIN_VALUE</CODE>
	 ** @param max the maximum allowed value, or <CODE>Integer.MAX_VALUE</CODE>
	 **/
	protected abstract int enterInteger(String title, int suggestedValue, int min, int max) throws NumberFormatException;

	/** Gets a user-supplied integral number.
	 ** @param title the title (or <em>type</em>) of the value to be chosen
	 ** @param suggestedValue the suggested number to be returned if valid
	 ** @param min the minimum allowed value, or <CODE>Integer.MIN_VALUE</CODE>
	 ** @param max the maximum allowed value, or <CODE>Integer.MAX_VALUE</CODE>
	 **/
	public final int getInteger(String title, int suggestedValue, int min, int max) {
		while (true) {
			try {
				final int integer = enterInteger(title, suggestedValue, min, max);
				if ((integer >= min) && (integer <= max))
					return integer;
				else
					UserInterface.get().running().print("Incorrect value", "integer " + integer + " is not in range [" + min + "," + max + "] !");
			} catch (NumberFormatException e) {
					UserInterface.get().running().print("Incorrect value", " not an integral number !");
			}
		}
	}

	/** File access mode **/
	public static abstract class AccessMode {

		/** Human-readable description of this access mode **/
		public final String description;

		private AccessMode(String description) { this.description = description; }

		public String toString() { return description; }

		/** Wether specified file grants this access mode **/
		protected abstract boolean accepts(File file);

	}

	/** Access mode for a readable or writable file **/
	public static final AccessMode FILE = new AccessMode("file") {
		protected boolean accepts(File file) { return ! file.isDirectory(); }
	};
	/** Access mode for a readable file **/
	public static final AccessMode READABLE_FILE = new AccessMode("readable file") {
		protected boolean accepts(File file) { return file.isFile() && file.canRead(); }
	};
	/** Access mode for a writable or creatable file **/
	public static final AccessMode WRITABLE_FILE = new AccessMode("writable file") {
		protected boolean accepts(File file) {
			if ( file.exists() && ! file.canWrite() ) {
				UserInterface.get().running().print("Invalid " + description, "Can not write " + description + " \"" + file + "\" !");
				return false;
			} else  // Confirms that the file shall be overwritten
				if ( file.exists() && ( ! UserInterface.get().running().yesNo(description + " \"" + file + "\" already exists, override") ) ) {
						UserInterface.get().running().print("Invalid " + description, description + " \"" + file + "\" NOT overridden");
						return false;
				}
			return true;
		}
	};
	/** Access mode for a readable and  traversable directory **/
	public static final AccessMode READABLE_DIRECTORY = new AccessMode("readable directory") {
		protected boolean accepts(File file) { return file.isDirectory() && file.canRead(); }
	};
	/** Access mode for a readable and  traversable directory **/
	public static final AccessMode WRITABLE_DIRECTORY = new AccessMode("writable directory") {
		protected boolean accepts(File file) { return file.isDirectory() && file.canWrite(); }
	};
	public static final AccessMode READABLE_FILE_OR_DIRECTORY = new AccessMode("readable file or directory") {
		protected boolean accepts(File file) { return READABLE_FILE.accepts(file) || READABLE_DIRECTORY.accepts(file); }
	};



	/** Prompts the user for entering a file name.
	 ** @param title the title (or <em>type</em>) of the directory to be chosen
	 ** @param suggestedValue the suggested (yet invalid) file or directory to be returned
	 ** @param mode the required access mode for this file
	 **/
	protected abstract File enterFileOrDirectory(String title, File suggestedValue, AccessMode mode);

	/** Gets a file or directory and ensures it can be read/written as required
	 ** @param title the title (or <em>type</em>) of the directory to be chosen
	 ** @param suggestedValue the suggested (yet invalid) file or directory to be returned
	 ** @param mode the required access mode for this file
	 **/
	public File getFileOrDirectory(String title, String suggestedValue, AccessMode mode) {
		return getFileOrDirectory(title, (suggestedValue == null) ? null : new File(suggestedValue), mode);
	}

	/** Gets a file or directory and ensures it can be read/written as required
	 ** @param title the title (or <em>type</em>) of the directory to be chosen
	 ** @param suggestedValue the suggested (yet invalid) file or directory to be returned
	 ** @param mode the required access mode for this file
	 **/
	public File getFileOrDirectory(String title, File suggestedValue, AccessMode mode) {
		File file = suggestedValue;
		while (true) {
			if (file == null)
				UserInterface.get().running().print("No file specified", "please specify a " + mode.description + " !");
			else
				if ( ! mode.accepts(file) )
						UserInterface.get().running().print("Invalid file", "\"" + file + "\" is not a " + mode.description + " !");
				else
					return file;
			;
			file = enterFileOrDirectory(title, suggestedValue, mode);
		}
	}

	/** Prompts the user for choosing an integer among several ones.
	 ** @param title the title (or <em>type</em>) of the value to be chosen
	 ** @param values the list of allowed values
	 ** @param suggestedValue the suggested value to be returned on cancel
	 **/
	public int chooseInteger(String title, int[] values, int suggestedValue) {
		if (values == null)
			return suggestedValue;
		try {
			final String value = chooseString(title, Text.toStrings(values), ("" + suggestedValue) );
			return (value == null) ? suggestedValue : Integer.valueOf(value).intValue();
		} catch (NumberFormatException e) {  // Proposed values where <b>not</b> all valid integer representation (this is a BUG)
			UserInterface.get().error().warn("Proposed values " + Text.toString(Text.toStrings(values)) + " were not valid integers");
			return suggestedValue;
		}
	}

	/** Prompts the user for choosing a string among several ones.
	 ** @param title the title (or <em>type</em>) of the value to be chosen
	 ** @param values the list of allowed values
	 ** @param suggestedValue the suggested value to be returned on cancel
	 ** @return the chosen value, or <CODE>null</CODE> if cancelled or
	 ** if <CODE>values</CODE> is <CODE>null</CODE>
	 **/
	public String chooseString(String title, String[] values, String suggestedValue) {
		if (values == null)
			return null;
		String chosenValue = (suggestedValue == null) ? "" : suggestedValue;
		while ( (chosenValue != null) && ( ! Text.contains(values, chosenValue)) )
			chosenValue = choose(title, values);
		return chosenValue;
	}

	/** Prompts the user for entering a string.
	 ** @param title the title (or <em>type</em>) of the value to be chosen
	 ** @param values the list of allowed values
	 ** @param suggestedValue the suggested value to be returned on cancel
	 ** @param the chosen value, or possibly <CODE>null</CODE> if cancelled
	 **/
	protected abstract String choose(String title, String[] values);

	/** Prompts the user to enter the password for specified ciphers in deciphering mode.
	 ** @param ciphers the suggested ciphers list
	 **/
	public final Specification[] enterPasswords(Specification[] ciphers) {
		if ((ciphers == null) || (ciphers.length == 0))
			return new Specification[0];
		if (Settings.Constants.DO_NOT_REENTER_FULLY_SPECIFIED_CIPHERS && Specification.areFullySpecified(ciphers))
			return ciphers;
		return Specification.dropEmpty(enterCiphers("Enter passwords", ciphers, false));
	}

	/** Prompts the user to choose the ciphers and enter the corresponding passwords in ciphering mode.
	 ** @param ciphers the suggested ciphers list
	 **/
	public final Specification[] chooseCiphers(Specification[] ciphers) {
		if (Settings.Constants.DO_NOT_REENTER_FULLY_SPECIFIED_CIPHERS && Specification.areFullySpecified(ciphers))
			return ciphers;
		Specification[] specifications = null;
		while (true) {
			specifications = Specification.dropEmpty( enterCiphers("Choose ciphers", ciphers, true) );
			if ((specifications == null) || (specifications.length == 0)) {
				UserInterface.get().log().log(5, RulingInterface.class, "No cipher specified, data will NOT be encrypted");
				return new Specification[0];
			}
			Specification[] verifications =
				enterCiphers("Reenter passwords", Specification.fillIn(Specification.getNames(specifications), null), false);
			if ((verifications == null) || (specifications.length != verifications.length)) {
				UserInterface.get().running().print("Incorrect passwords", "missing passwords");
				continue;
			}
			boolean passwordsMatch = true;
			for (int i=0; i<specifications.length; ++i) {
				if (specifications[i].name != verifications[i].name)
					UserInterface.get().error().fail("BUG: reentered passwords are not in the same order !");
				if ( ! specifications[i].password.equals(verifications[i].password) ) {
					UserInterface.get().running().print("Incorrect password " + i, "passwords for " + specifications[i].name + " do not match !");
					passwordsMatch = false;
				}
			}
			if (passwordsMatch)
				return specifications;
		}
 }

	/** Prompts the user for entering the ciphers/passwords.
	 ** @param title the non-<CODE>null</CODE> title
	 ** @param ciphers the suggested ciphers list <b>or <CODE>null</CODE> if none specified</b>
	 ** @param canAddCiphers wether the user can add/remove ciphers or simply supply passwords
	 ** @param the ordered list of cipher specifications
	 ** <p>If <CODE>canAddCiphers</CODE> is <CODE>true</CODE> then the user should be allowed to select several among
	 ** registered ciphers to form
	 ** a stack of ciphers. Note also that empty specifications (those with empty passwords) are automatically
	 ** discarded.
	 ** <p>If <CODE>canAddCiphers</CODE> is <CODE>false</CODE> then <CODE>ciphers</CODE> is the ordered stack of
	 ** the ciphers that have actually be chosen by the user, may it be user-selected in ciphering mode or
	 ** recovered from archive in deciphering mode. Thus the user should <b>not</b> be able to alter the list,
	 ** but instead just (re)enter the corresponding passwords.
	 **/
	protected abstract Specification[] enterCiphers(String title, Specification[] ciphers, boolean canAddCiphers);

	/** Simple test code.
	 **/
	public void test() {
		try {
			System.out.println("Testing ruling interface implemented by " + getClass().getName() );
			System.out.println("\tInteger: " + getInteger("any integer in [0,25]", 5, 0, 25) );
			System.out.println("\tReadFile: \"" + getFileOrDirectory("any readable file", (File)null, READABLE_FILE) + "\"");
			System.out.println("\tWriteFile: \"" + getFileOrDirectory("any writable file", (File)null, WRITABLE_FILE) + "\"");
			System.out.println("\tChoice: \"" + chooseInteger("a number", new int[] { 1, 2, 3}, 2) + "\"");
			System.out.println("\tCiphers: \"" + Text.toString( chooseCiphers(new Specification[0]) ) );
		} catch (Throwable t) {
			System.err.println("Test of ruling interface failed: " + Text.getMessage(t) );
			t.printStackTrace();
		}
	}

}
