
package sam.ui;

import sam.event.*;


/** Generic running interface type.
 ** A running interface is in charge of defining
 ** all the events handlers.
 ** @see UserInterface
 **/
public abstract class RunningInterface {

	/** Builds a new <CODE>FileEvent</CODE> listener.
	 ** Note that instantiation is lazy and occurs at most once.
	 ** Moreover <CODE>FileEvent</CODE> are delivered <b>synchronously</b>.
	 ** @return <CODE>null</CODE> if no <CODE>FileEvent</CODE> listener should be defined.
	 ** @see sam.event.FileEventListener
	 **/
	public abstract FileEventListener getFileEventListener();

	/** Builds a new <CODE>ProgressEvent</CODE> listener.
	 ** Note that instantiation is lazy and occurs at most once.
	 ** Moreover <CODE>ProgressEvent</CODE> are delivered <b>asynchronously</b> but remained ordered (and are all delivered).
	 ** @return <CODE>null</CODE> if no <CODE>ProgressEvent</CODE> listener should be defined.
	 ** @see sam.event.ProgressEventListener
	 **/
	public abstract ProgressEventListener getProgressEventListener();

	/** Prints a message to inform the user of something normal.
	 ** @param title the title (or <em>type</em>) of the message delivered,
	 ** <b>or <CODE>null</CODE> if no title is specified</b>.
	 ** @param message the actual message to be delivered.
	 ** Note that neither <CODE>title</CODE> nor <CODE>message</CODE> will ever be <CODE>null</CODE>.
	 **/
	public abstract void print(String title, String message);

	/** Prints a message to inform the user of something normal.
	 ** @param message the actual message to be delivered.
	 ** Note that neither <CODE>title</CODE> nor <CODE>message</CODE> will ever be <CODE>null</CODE>.
	 **/
	public void print(String message) { print(null, message); }

	/** Asks a question to the user.
	 **/
	public abstract boolean yesNo(String title, String question);

	/** Asks a question to the user. **/
	public boolean yesNo(String question) { return yesNo(null, question); }

	/** Simple test code.
	 **/
	public void test() {
		try {
			System.out.println("Testing running interface implemented by " + getClass().getName() );
			String message = "Hello World !";
			System.out.println("Printing \"" + message + "\":");
			print("Test message", message);
			if ( yesNo("Did you see \"" + message + "\" properly") )
				System.out.println("OK, then printing/asking seems to work properly");
			else
				System.out.println("Oops, printing/asking seems NOT to work properly !");

			System.out.println("\tTesting (synchronous) file event listener");
			FileEventNotifier fileNotifier = new FileEventNotifier();
			fileNotifier.register( getFileEventListener() );
			fileNotifier.notifyAll(FileEventNotifier.PROCESSED, sam.entry.Entry.DUMMY, "\"Processed\" notification works !");
			fileNotifier.notifyAll(FileEventNotifier.SKIPPED, sam.entry.Entry.DUMMY, "\"Skipped\" notification works !");
			fileNotifier.notifyAll(FileEventNotifier.FAILED, sam.entry.Entry.DUMMY, "\"Failed\" notification works !");
			fileNotifier.notifyAll(FileEventNotifier.OVERWRITTEN, sam.entry.Entry.DUMMY, "\"overwritten\" notification works !");
			fileNotifier.notifyAll(FileEventNotifier.MISMATCHING, sam.entry.Entry.DUMMY, "\"mismatching\" notification works !");
			fileNotifier.notifyAll(FileEventNotifier.MISSING, sam.entry.Entry.DUMMY, "\"Missing\" notification works !");

			System.out.println("\tTesting (asynchronous) progress event listener synchronously");
			getProgressEventListener().progress(ProgressEvent.DUMMY);

		} catch (Throwable t) {
			System.err.println("Test of ruling interface failed: " + sam.toolkit.Text.getMessage(t) );
		}
	}

}


