
package sam.ui;

import sam.toolkit.ClassRegistry;
import sam.core.Settings;

import java.awt.Frame;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;


/** Pluggable user interface abstraction
 ** A user interface defines all the interactions between the user and the program;
 ** they are grouped by categories as follows:
 ** <ul>
 ** <li>the <CODE>RulingInterface</CODE> groups the interactions concerning the program ruling, i.e. command-line parsing</li>
 ** <li>the <CODE>RunningInterface</CODE> groups the output made by the program concerning its running (progress reports etc.)</li>
 ** <li>the <CODE>ErrorInterface</CODE> groups the interactions made in case of errors, which may lead the program either to terminate
 ** or to ignore them.</li>
 ** <li>the <CODE>LogInterface</CODE> defines the log level and where to output log events</li>
 ** </ul>
 **
 ** <p>By now three different user interfaces are proposed:
 ** <ul>
 ** <li>the <CODE>sam.ui.basic.UserInterface</CODE> is a basic non-interactive user interface</li>
 ** <li>the <CODE>sam.ui.text.UserInterface</CODE> is a simple text-based user interface</li>
 ** <li>the <CODE>sam.ui.awt.UserInterface</CODE> is a simple <CODE>awt</CODE> based graphical user interface</li>
 ** </ul>
 **
 ** @see RulingInterface
 ** @see RunningInterface
 ** @see ErrorInterface
 ** @see LogInterface
 **/
public abstract class UserInterface {


	/** Class registry in charge of registring and lazily instantiating <CODE>UserInterface</CODE>s **/
	public static final ClassRegistry REGISTRY = new ClassRegistry(UserInterface.class, "user interface");

	private static UserInterface UI = null;

	protected UserInterface() { }

	/** Exception thrown whenever the whole program should abort. **/
	public static class Abort extends java.lang.Error {
		public Abort(String message) { super("Aborted: " + message); }
	}

	/** Sets <CODE>name</CODE> as the current user interface.
	 ** @throws NoSuchElementException if <CODE>name</CODE> is not a registered <CODE>UserInterface</CODE>
	 **/
	public static boolean set(String name) {
		if (name == null)
			return false;
		if (name.equals( REGISTRY.name() ))
			return true;
		// if ( isSet() )  // This is buggy: might be set whereas still not initializable !
		//	get().log().log(1, UserInterface.class, "Switching from " + REGISTRY.name() + " to " + name + " user interface");
		boolean successfullySet = REGISTRY.set(name);
		if (successfullySet)
			UI = null;  // Resets the <CODE>UI</CODE> cache so that the new UI is instantiated on next <CODE>get()</CODE> call
		return successfullySet;
	}

	/** Wether a user interface has ever been set.
	 ** @see #set
	 **/
	public static boolean isSet() { return REGISTRY.isSet(); }

	/** Sets the best user interface.
	 ** @see sam.core.Settings.Constants.UI#PREFERENCES
	 ** @see sam.core.Settings.Constants.UI#FAILSAFE
	 **/
	protected static void setBest() {
		for (int i=0; i<sam.core.Settings.Constants.UI.PREFERENCES.length; ++i) {
			try {
				set( sam.core.Settings.Constants.UI.PREFERENCES[i] );
				if ( ((UserInterface)REGISTRY.get()).log() != null )  // This tries to instantiate the just-set UI
					return;
			} catch (Throwable t) {
			}
		}
		UserInterface.set(sam.core.Settings.Constants.UI.FAILSAFE);
	}

	/** Gets the list of all known user interfaces **/
	public static String[] getKnwonInterfaces() { return REGISTRY.registeredTargets(); }

	/** Registers a new <CODE>UserInterface</CODE> as <CODE>name</CODE>
	 ** <p>This is also set as the current one if none currently is.
	 **/
	protected static boolean register(Class UI, String name) { return (UI == null) ? false : REGISTRY.register(UI, name); }

	/** Gets the current user interface.
	 ** <p>Instanciation is lazy and occurs only once.
	 ** @return the current user interface
	 **/
	public static UserInterface get() {
		if (UI == null) {
			if ( ! isSet() )
				setBest();
			UI = (UserInterface) REGISTRY.get();  // This dynamic cast will always succeed
			if (UI == null)
				UI = new sam.ui.basic.UserInterface(System.err);
			UI = new Hook(UI);
		}
		return UI;
	}

	/** Abruptly disposes <b>all</b> frames of the application. **/
	public static void disposeFrames() {
		Frame[] frames = Frame.getFrames();
		for (int i=0; i<frames.length; ++i) {
			frames[i].hide();
			frames[i].removeNotify();
			frames[i].dispose();
		}
	}

	/** @return the <CODE>RulingInterface</CODE> associated with this <CODE>UserInterface</CODE> **/
	public abstract RulingInterface ruling();

	/** @return the <CODE>RunningInterface</CODE> associated with this <CODE>UserInterface</CODE> **/
	public abstract RunningInterface running();

	/** @return the <CODE>ErrorInterface</CODE> associated with this <CODE>UserInterface</CODE> **/
	public abstract ErrorInterface error();

	/** @return the <CODE>LogInterface</CODE> associated with this <CODE>UserInterface</CODE> **/
	public abstract LogInterface log();

	/** Hook that allows for lazy and singleton instanciation of the components
	 ** of a <CODE>UserInterface</CODE>
	 **/
 	private static class Hook extends UserInterface {
		private final UserInterface realInterface;
		private RunningInterface running = null;
		private RulingInterface ruling = null;
		private ErrorInterface error = null;
		private LogInterface log = null;

		/** @param realInterface non-<CODE>null</CODE> real user interface to be hooked **/
		private Hook(UserInterface realInterface) { this.realInterface = realInterface; }

		public final RulingInterface ruling() {
			if (ruling == null) {
				try {
					log().log(1, UserInterface.class, "Activating " + REGISTRY.name() + " ruling interface");
					ruling = realInterface.ruling();
				} catch (Throwable t) {
					error().handleException("Failed to activate "+ REGISTRY.name() + " ruling interface", t);
					set(Settings.Constants.UI.FAILSAFE);
					return get().ruling();
				}
			}
			return ruling;
		}

		public final RunningInterface running() {
			if (running == null) {
				try {
					log().log(1, UserInterface.class, "Activating " + REGISTRY.name() + " running interface");
					running = realInterface.running();
				} catch (Throwable t) {
					error().handleException("Failed to activate "+ REGISTRY.name() + " running interface", t);
					set(Settings.Constants.UI.FAILSAFE);
					return get().running();
				}
			}
			return running;
		}

		public final ErrorInterface error() {
			if (error == null) {
				try {
					log().log(1, UserInterface.class, "Activating " + REGISTRY.name() + " error interface");
					error = realInterface.error();
				} catch (Throwable t) {
					set(Settings.Constants.UI.FAILSAFE);
					get().error().handleException("Failed to activate "+ REGISTRY.name() + " error interface", t);
					return get().error();
				}
			}
			return error;
		}

		public final LogInterface log() {
			if (log == null) {
				try {
					log = realInterface.log();
				} catch (Throwable t) {
					error().handleException("Failed to activate "+ REGISTRY.name() + " log interface", t);
					set(Settings.Constants.UI.FAILSAFE);
					return get().log();
				}
			}
			return log;
		}

	}

	/** Simple test code
	 ** Loads a user interface and performs basic tests on it
	 ** Syntax is <CODE>java sam.ui.UserInterface [user-interface]</CODE>
	 **/
	public static void main(String[] arguments) {
		try {
			Settings.Startup.load();
			for (int i=0; i<arguments.length; ++i) {
				System.out.println("Testing user interface \"" + arguments[i] + "\"");
				if ( ! set(arguments[i]) )
					System.err.println("Failed to set \"" + arguments[i] + "\" as the user interface !");
				else {
					UserInterface.get().ruling().test();
					UserInterface.get().running().test();
					UserInterface.get().error().test();
					UserInterface.get().log().test();
				}
			}
		} finally {
			LogInterface.stop();
			System.out.println("Done (please wait for pending logs).");
		}
	}


}
