
package sam.ui.awt;

import sam.toolkit.*;
import sam.cipher.Specification;
import sam.core.Settings;

import java.awt.*;
import java.awt.event.*;


/** Cipher specification input field. **/
public class CipherInputField extends Frame {
	private static final char PASSWORD_HIDE_CHAR = '*';

	private final SpecificationsList specificationsList;
	private final boolean canAddCiphers;
	private final Object lock = new int[0];
	private ButtonBar buttons;
	private boolean isCanceled = false;
	private Container passwordArea;
	private final CipherBar cipherBar;

	/** Opens a new cipher stack window.
	 ** @param title the window title
	 ** @param specifications the initial cipher stack
	 ** @param canAddCiphers wether the user can add ciphers to the supplied cipher stack
	 **/
	public CipherInputField(String title, Specification[] specifications, boolean canAddCiphers) {
			super(title);
			setLocation(200, 150);
			setBackground(Color.lightGray);
			setLayout( new BorderLayout() );
			this.canAddCiphers = canAddCiphers;
			if (canAddCiphers) {
				Specification[] registeredCiphers = Specification.getRegisteredCiphers();
				add( ( cipherBar = new CipherBar(registeredCiphers) ), BorderLayout.WEST);
				if (specifications == null)
					specifications = registeredCiphers;
				setSize(600, registeredCiphers.length * 37 + 110);
			} else {
				cipherBar = null;
				setSize(400, specifications.length * 37 + 70);
			}
			setResizable(false);
			specificationsList = new SpecificationsList(specifications.length);
			for (int i=0; i<specifications.length; ++i)
				specificationsList.add( specifications[i] );
			makePasswordArea();
			try {
				synchronized(lock) {
					lock.wait();  // Wait for the password values be entered
				}
			} catch (InterruptedException e) {
			} 
			setVisible(false);
			removeNotify();
			dispose();
		}

		/** Returns <CODE>false</CODE> if, and only if the user exited with the "OK" button. **/
		public boolean isCanceled() {	return isCanceled; }

		/** Gets the specifications selected byt the user
		 ** @see #isCanceled
		 **/
		public Specification[] getSpecifications() { return specificationsList.getSpecifications();	}

		/** Rebuilts the password input area **/
		private void makePasswordArea() {
			passwordArea = new Container();
			passwordArea.setLayout( new BorderLayout() );
			passwordArea.add(specificationsList , BorderLayout.CENTER);
			buttons = new ButtonBar();
			passwordArea.add(buttons, BorderLayout.SOUTH);
			add(passwordArea, BorderLayout.CENTER);
			show();
		}

		/** Field for entering a whole list of specicications	**/
		protected class SpecificationsList extends ScrollPane {
			private char echoChar = PASSWORD_HIDE_CHAR;
			private Grid group;

			/** Builds a new SpecificationsList field.
			 ** @param numberOfSpecifications the number of specifications to be entered
			 ** <p> Any negative value (e.g. <CODE>-1</CODE>) means unspecified
			 ** and initializes a scrollable (and growable at will) component.
			 **/
			protected SpecificationsList(int numberOfSpecifications) {
				this.group = canAddCiphers ?  new GrowableGrid(numberOfSpecifications, true) : new Grid(numberOfSpecifications, 1);
				add(group);
			}

			/** Stacks a new single {@link SingleSpecification} for specified cipher.
			 ** @param name the name of a cipher
			 ** @param password the possibly supplied password
			 **/
			public void add(Specification specification) {
				group.append( new SingleSpecification(specification) );
			}

			/** Clears the cipher stack. **/
			public void removeAllSpecifications() { group.removeAll();	}

			/** Sets the echo character of passwords input fields
			 ** In particular the <CODE>'\0'</CODE> value turns masking off (which treatens security).
			 ** @see java.awt.TextField#setEchoChar
			 **/
			protected void setEchoChar(char echoChar) {
				this.echoChar = echoChar;
				Component[] singles = group.getComponents();
				if (singles != null)
					for (int i=0; i<singles.length; i++) {
						TextField text = ((SingleSpecification) singles[i]).password;
						text.setEchoChar(echoChar);
						text.setText( text.getText() );  // Force refresh, sorry ...
					}
			}

			/** Gets the stack of specification currently built. **/
			public Specification[] getSpecifications() {
				Component[] singleSpecifications = group.getComponents();
				Specification[] specifications = new Specification[ singleSpecifications.length ];
				for (int i=0; i<specifications.length; ++i)
						specifications[i] = ((SingleSpecification) singleSpecifications[i]).getSpecification();
				return specifications;
			}

			/** GUI Component for one single cipher/password input
			 ** (roughly speaking a <CODE>TextField</CODE>)
			 **/
			protected class SingleSpecification extends Container {
				private final Label name;
				private final TextField password;

				/** Builds a new cipher/password input field.
				 ** @param name the cipher name
				 ** @param password the possibly supplied password
				 **/
				protected SingleSpecification(Specification specification) {
					setLayout( new BorderLayout() );
					this.name = new Label(specification.name);
					SingleSpecification.this.add(this.name, BorderLayout.CENTER);
					this.password = new TextField(20);
					this.password.setEchoChar(echoChar);
					this.password.setText(specification.password);
					Container passwordInputZone = new Container();
					passwordInputZone.setLayout( new BorderLayout() );
					passwordInputZone.add(password, BorderLayout.CENTER);
					if (canAddCiphers) {
						SingleSpecification.this.add( new Grid(1,1).
							appendPanel(new Button("X", new ActionListener() {
									public void actionPerformed(ActionEvent e) {
										SpecificationsList.this.group.remove(SingleSpecification.this);
										CipherInputField.this.makePasswordArea();
									}
								}) ), BorderLayout.WEST);
						passwordInputZone.add(new Grid(1,1).
							appendPanel(new Button("<<", new ActionListener() {
									public void actionPerformed(ActionEvent e) {
										password.setText( password.getText() + Random.nextHex(8) );
									}
								}) ), BorderLayout.EAST);
						passwordInputZone.add(new Grid(1,1).
							appendPanel(new Button(">>", new ActionListener() {
									public void actionPerformed(ActionEvent e) {
										password.setText( Random.nextHex(8) + password.getText() );
									}
								}) ), BorderLayout.WEST);
					}
					SingleSpecification.this.add(passwordInputZone, BorderLayout.EAST);
				}

				/** Gets the specification entered inside this SingleSpecification
				 **/
				public Specification getSpecification() {
					return new Specification(name.getText(), password.getText() );
				}

			}
		
		}


		/** Stack of buttons for pushing new ciphers **/
		protected class CipherBar extends Grid {

			protected CipherBar(final Specification[] specifications) {
				super(specifications.length + 1, 1);
				append(new SwitchButton("All ", "Clear ") {
						public boolean on() {
							for (int i=0; i<specifications.length; i++)
								specificationsList.add( specifications[i] );
							CipherInputField.this.makePasswordArea();
							return true;
						}
						public boolean off() {
							specificationsList.removeAllSpecifications();
							return true;
						}
					} );
				for (int i=0; i<specifications.length; i++)
					append( new CipherEntry(specifications[i]) );
			}

			/** Entry that allows adding one new cipher **/
			private class CipherEntry extends Container {
				private CipherEntry(final Specification specification) {
					setLayout( new BorderLayout() );
					add(new Grid(1,1).append( new Button(specification.name, new ActionListener() {
							public void actionPerformed(ActionEvent e) {
								specificationsList.add(specification);
								CipherInputField.this.makePasswordArea();
							}
						}  ) ), BorderLayout.CENTER);
					add(new Grid(1,1).appendPanel( new Button("?", new ActionListener() {
							public void actionPerformed(ActionEvent e) {
								new InfoBox(specification.name, specification.getDescription() );
							}
						}  ) ), BorderLayout.EAST);
				}
			}

		}

		/** Buttons bar implementation
		 **/
		private class ButtonBar extends Grid {

			protected ButtonBar() {
				super(1, 5);
				appendPanel(new Button("About", new ActionListener() {
						public void actionPerformed(ActionEvent e) {
							new InfoBox("About", Settings.Environment.programDescription() );
						}
					}));
				appendPanel(new Button("Cancel", new ActionListener() {
						public void actionPerformed(ActionEvent e) {
							synchronized(lock) {
								isCanceled = true;
								lock.notify();
							}
						}
					}));
				appendPanel(new SwitchButton("Hide", "Show") {
						public boolean on() {
							specificationsList.setEchoChar(PASSWORD_HIDE_CHAR);
							return true;
						}
						public boolean off() {
							specificationsList.setEchoChar('\0' );
							return true;
						}
					} );
				appendPanel(new Button("OK", new ActionListener() {
						public void actionPerformed(ActionEvent e) {
							synchronized(lock) {
								isCanceled = false;
								lock.notify();
							}
						}
					}));
			}

		}

}




