
package sam.ui.awt;

import sam.core.Settings;

import java.awt.*;
import java.awt.event.*;


/** General-purpose dialog box.
 **/
public class Dialog extends java.awt.Dialog {

	/** Special time duration specifying no time limit at all **/
	public static final long NO_TIME_LIMIT = -1;

	/** Marker of non-blocking (non-modal) dialogs */
	public static final int SPAWN    = 1 << 1;

	/** Result state specifying that no button has been choosen **/
	public static final int NOTHING  = 0;

 	/** Result state specifying that the dialog timed out **/
	public static final int TIME_OUT = 1 << 2;

 	/** Result state specifying that the dialog was killed (window closed) **/
	public static final int KILLED   = 1 << 3;

	/** Flag specifying that the content of this dialog should be editable by the user **/
	public static final int EDITABLE = 1 << 4;

	/** The first option, set to <CODE>0x100</CODE> instead of e.g. <CODE>0x1</CODE> to help detect bugs. **/
	private static final int FIRST_OPTION = 0x100;

	/** The labels of the buttons, <b>in the same order as they shall appear</b> and simultaneously
	 ** <b>in the same order as corresponding button flags</b>.
	 **/
	private static final String[] NAMES = { "OK", "Cancel", "Yes", "No", "Always", "Never", "Open",	"Close", "GO", "Ignore" };

	private static final int OPTIONS_COUNT = NAMES.length;

	/** Alias of the "OK" button **/
	public static final int OK       = FIRST_OPTION << 0;

	/** Alias of the "CANCEL" button **/
	public static final int CANCEL   = FIRST_OPTION << 1;

	/** Alias of the "YES" button **/
	public static final int YES      = FIRST_OPTION << 2;

	/** Alias of the "NO" button **/
	public static final int NO       = FIRST_OPTION << 3;

	/** Alias of the "ALWAYS" button **/
	public static final int ALWAYS   = FIRST_OPTION << 4;

	/** Alias of the "NEVER" button **/
	public static final int NEVER    = FIRST_OPTION << 5;

	/** Alias of the "OPEN" button **/
	public static final int OPEN     = FIRST_OPTION << 6;

	/** Alias of the "CLOSE" button **/
	public static final int CLOSE    = FIRST_OPTION << 7;

	/** Alias of the "GO" button **/
	public static final int GO       = FIRST_OPTION << 8;

	/** Alias of the "IGNORE" button **/
	public static final int IGNORE   = FIRST_OPTION << 9;

	/** Combined mode of for input dialog
	 ** This is the combination of <CODE>EDITABLE</CODE>, <CODE>OK</CODE>
	 ** and <CODE>CANCEL</CODE> buttons
	 **/
	public static final int INPUT    = EDITABLE | OK | CANCEL;

	/** Combined mode of for question dialog
	 ** This is the combination of <CODE>YES</CODE> and <CODE>NO</CODE> buttons
	 **/
	public static final int QUESTION = YES | NO;

	/** The component to be displayed **/
	private final Component content;

	/** The choice made by the user **/
	private int choice = NOTHING;

	/** Pop-ups a new information box.
	 **	<p>The text can be multi-lined.
	 ** @param title the title of the dialog.
	 ** @param content the content to be displayed in this dialog
	 ** @param mode a combination of options for this dialog, e.g. <CODE>Dialog.OK | Dialog.CANCEL</COD>
	 ** @see #EDITABLE
	 **/
	protected Dialog(String title, Component content, int mode) {
		this(title, content, mode, (mode & EDITABLE) == 0 ? Settings.Constants.UI.AWT.DIALOG_DEFAULT_LIFETIME : 0);
	}

	/** Pop-ups a new information box.
	 **	<p>The text can be multi-lined.
	 ** @param title the title of the dialog.
	 ** @param content the content to be displayed in this dialog
	 ** @param mode a combination of options for this dialog, e.g. <CODE>Dialog.OK | Dialog.CANCEL</COD>
	 ** @param lifeTime time <b>in seconds</b> after which this box automatically gets dismissed.
	 **/
	protected Dialog(String title, Component content, int mode, long lifeTime) {
		super(new Frame(title), title, (mode & SPAWN) == 0);  // Modal dialog if not 'SPAWN'
		this.content = content;
		setBackground(Color.lightGray);
		add(content, BorderLayout.CENTER);
		add( makeButtons(mode), BorderLayout.SOUTH );
		ensureLocation(this, new Dimension(200, 85), Settings.Constants.UI.AWT.DEFAULT_LOCATION);
		dismissAfter(lifeTime);
		show();
	}

	protected void finalize() {
		hide();
		dispose();
	}

	/** Returns current text.
	 ** <p>Useful only if this infobox is editable !
	 **/
	protected Component getContent() { return content; }

	/** Returns user selected button.
	 ** Note that the choice might contain several states combined in addition to the chosen button.
	 ** @see #NOTHING
	 ** @see #KILLED
	 **/
	public int getChoice() { return choice;	}

	/** Kills this dialog. **/
	public void kill() { exit(KILLED); }

	/** Exits with specified choice. **/
	private void exit(int choice) {
		this.choice |= choice;
		setVisible(false);
		setModal(false);
		hide();
		dispose();
	}

	/** Builds the component containing specified buttons.
	 ** @param mode the combination of buttons to be built
	 **/
	private Component makeButtons(int mode) {
		Grid buttons = new GrowableGrid(OPTIONS_COUNT, false);
		for (int i = 0, j = FIRST_OPTION; i < OPTIONS_COUNT; i++, j *= 2)
			if ( (j & mode) != 0) {
				final int option = j;
				buttons.appendPanel( new Button(NAMES[i], new ActionListener() {
					public void actionPerformed(ActionEvent e) { exit(option); }
				} ) );
			}
		return buttons;
	}

	/** Finds best location and dimension to fit the current screen's constraints.
	 ** @param source the component to be placed
	 ** @param required the required dimension
	 ** @param corner the upper left corner
	 **/
	private void ensureLocation(Component source, Dimension required, Point corner) {
		pack();
		Dimension current = getSize();
		if ( (current.width < required.width) || (current.height < required.height))
			setSize(required);
		Point start = source.getLocation();
		Dimension sourceSize = source.getSize();
		Dimension thisSize = getSize();
		int x = start.x + (sourceSize.height - thisSize.height) / 2;
		int y = start.y + (sourceSize.width - thisSize.height) / 2;
		setLocation( (x < corner.x) ? corner.x : x , (y < corner.y) ? corner.y : y );
	}

	/** Spawns a new thread that automatically dismisses this box after specified amount of time.
	 ** @param delay the delay in seconds, or <CODE>0</CODE> if none
	 **/
	private void dismissAfter(final long delay) {
		if (delay > 0)
			new Thread() {
				public void run() {
					try {
						sleep(1000 * delay);
					} catch (InterruptedException e) { }
					Dialog.this.exit(TIME_OUT);
				}
			}.start();
	}

}
