
package sam.ui.awt;

import sam.core.Settings;

import java.awt.*;
import java.awt.event.*;
import java.io.PrintWriter;


/** Pluggable <a href=http://java.sun.com/docs/books/tutorial/ui/>AWT</a>-based log interface
 **/
public class LogInterface extends sam.ui.LogInterface {

	private Component activator = null;
	private LogLevelInput logLevelInput = null;

	private PrintWriter out;

	private final SwitchButton button = new SwitchButton("Close Log", " View Log ") {
		protected boolean on() {
			LogInterface.this.hide();  // Which also changes this button's state
			return false;  // Don't rechange state
		}
		protected boolean off() {
			LogInterface.this.show(Settings.Constants.UI.AWT.DEFAULT_LOG_POSITION);  // Which also changes this button's state
			return false;  // Don't rechange state
		}
	};

	private Frame frame;

	/** Builds a new log interface **/
	public LogInterface() {
		TextAreaWriter textAreaWriter = new TextAreaWriter();
		this.frame = new Frame("Log");
		frame.add(textAreaWriter.display, BorderLayout.CENTER);
		this.logLevelInput = new LogLevelInput(getLevel(), Settings.Constants.MINIMUM_LOG_LEVEL, Settings.Constants.MAXIMUM_LOG_LEVEL);
		frame.add(logLevelInput, BorderLayout.NORTH);
		frame.addWindowListener( new WindowAdapter() {	public void windowClosing(WindowEvent e) {  hide(); }  }  );
		frame.pack();
		this.out = new PrintWriter(textAreaWriter);
		if (activator != null)
			frame.setSize( activator.getSize() );
		else
			frame.setSize(Settings.Constants.UI.AWT.LOG_WINDOW_SIZE);
	}

	/** Sets the log level.
	 ** This also updates graphical components.
	 **/
	public void setLevel(int level) {
    doSetLevel(level);
		logLevelInput.updateLevel();
	}

	private void doSetLevel(int level) { super.setLevel(level); }


	/** Simple log-level scroll bar. **/
	protected class LogLevelInput extends Container {
		private Scrollbar scrollbar;
		private Label label;

		/** Builds a new log level input **/
		protected LogLevelInput(int value, int min, int max) {
			setLayout( new BorderLayout() );
			label = new Label();
			add(label, BorderLayout.WEST);
			scrollbar = new Scrollbar(Scrollbar.HORIZONTAL, value, (max - min) / 20 + 5, min, max);
			add(scrollbar, BorderLayout.CENTER);
			scrollbar.addAdjustmentListener(new AdjustmentListener() {
				public void adjustmentValueChanged(AdjustmentEvent e) {	set( e.getValue() ); }
			} );
			set(value);
		}

		private void set(int value) {
			doSetLevel(value);
			updateLevel();
		}

		void updateLevel() {
			int level = getLevel();
			scrollbar.setValue(level);
			label.setText( (level < 0) ? "Log disabled" : ("Log: " +  + level) );
		}

	}


	protected void finalize() { hide(); }

	protected void log(String logMessage) {	out.println(logMessage); }

	/** Gets a proper switch button to activate/deactivate the log.
	 ** @param activator the component the button will be attached to
	 **/
	public synchronized java.awt.Button getButton(Component activator) {
		this.activator = activator;
		return button;
	}

	/** Shows the log window.
	 ** @param direction one of <CODE>BorderLayout.NORTH</CODE>, <CODE>BorderLayout.SOUTH</CODE>,
	 ** <CODE>BorderLayout.WEST</CODE>, <CODE>BorderLayout.EAST</CODE>.
	 **/
	protected void show(String direction) {
		Point location = getLocation(activator, frame, direction);
		if (location != null)
			frame.setLocation(location);
		frame.show();
		button.setState(false);
		if (Settings.Constants.UI.AWT.SHOWING_LOG_ENABLES_IT)
			enable();
	}

	/** Hides the log window. **/
	protected void hide() {
		frame.hide();
		frame.removeNotify();
		frame.dispose();
		button.setState(true);
		if (Settings.Constants.UI.AWT.HIDING_LOG_DISABLES_IT)
			disable();
	}

	/** Gets the best location to display the log adjacent to another.
	 ** @param direction one of <CODE>BorderLayout.NORTH</CODE>, <CODE>BorderLayout.SOUTH</CODE>,
	 ** <CODE>BorderLayout.WEST</CODE>, <CODE>BorderLayout.EAST</CODE>, <CODE>BorderLayout.CENTER</CODE>.
	 **/
	protected static Point getLocation(Component source, Component target, String direction) {
		if ((source == null) || (target == null))
			return null;
		if (direction == null)
			direction = BorderLayout.CENTER;
		Point topLeft = source.getLocation();
		Dimension sourceSize = source.getSize();
		Dimension targetSize = target.getSize();
		Point downRight = new Point( topLeft.x + sourceSize.width, topLeft.y + sourceSize.height );
		return
			direction.equals(BorderLayout.NORTH)  ? new Point( topLeft.x, topLeft.y - targetSize.height ) :
			direction.equals(BorderLayout.EAST)   ? new Point( topLeft.x + sourceSize.width, topLeft.y) :
			direction.equals(BorderLayout.SOUTH)  ? new Point( topLeft.x, topLeft.y  + sourceSize.height) :
			direction.equals(BorderLayout.WEST)   ? new Point( topLeft.x - targetSize.width, topLeft.y) :
			direction.equals(BorderLayout.CENTER) ? new Point(topLeft.x + (sourceSize.width - targetSize.width) / 2,
																												topLeft.y + (sourceSize.height - targetSize.height) / 2)
				: null;
	}

}

