
package sam.ui.awt;

import sam.core.Settings;

import java.awt.*;


/** Double-index progress-bar.
 ** <p>Reports both reading and writing progrees in bicolor growing horizontal bars.
 **/
public class ProgressBar extends Component {

	/** Nnumber of plain bytes processed. **/
	private float plain = 0;

	/** Nnumber of encoded bytes processed. **/
	private float encoded = 0;

	/** Any text to be displayed on the bar **/
	private String text = null;

	/** Builds a new progress bar. **/
	public ProgressBar() { }

	/** Resets this progress bar. **/
	public synchronized void reset() {
		this.plain = 0f;
		this.encoded = 0f;
		this.text = null;
	}

	/** Sets the number of plain bytes processed. **/
	public synchronized void setPlain(float plain) {
		this.plain = (plain < 0f) ? 0f : plain;
		repaint();
	}
	
	/** Sets the number of encoded bytes processed. **/
	public synchronized void setEncoded(float encoded) {
		this.encoded = (encoded < 0f) ? 0f : (encoded > plain) ? plain : encoded;
		repaint();
	}

	/** Sets the text to be displayed **/
	public synchronized void setText(String text) {
		this.text = text;
		repaint();
	}

	/** Paints the text onto the progress bar **/
	protected synchronized void paintText(Graphics graphic) {
		if (text == null)
			return;
		graphic.setFont( graphic.getFont().deriveFont(Font.BOLD) );
		graphic.setColor(Color.black);
		graphic.drawString(text, width / 2 - 9, height / 2 + 6);  // Custom settings adequate for short texts
		graphic.setColor(Settings.Constants.UI.AWT.TEXT_BAR_COLOR);
		graphic.drawString(text, width / 2 - 10, height / 2 + 5);  // Custom settings adequate for short texts
	}

	/** Gets the adapted width for specified ratio. **/
	protected synchronized int getWidth(float ratio) {
		int width = (int) Math.floor(ratio * this.width);
		return (width < 0) ? 0 : (width > this.width - 2) ? (this.width - 2) : width;
	}

	/** Paints the border of the progress bar. **/
	protected synchronized void paintBorders(Graphics graphic) {
		graphic.setColor(Color.lightGray);
		graphic.drawLine(0, 1, width - 2, 1);
		graphic.setColor(Color.black);
		graphic.drawRect(0, 0, width - 2, height - 1);
	}

	/** Paints the inner bars. **/
	protected synchronized void paintBars(Graphics graphic) {
		final int plainLimit = getWidth(encoded);
		final int encodedLimit = getWidth(plain);
		graphic.setColor(Settings.Constants.UI.AWT.ENCODED_BAR_COLOR);
		graphic.fillRect(2, 2, encodedLimit - 2, height - 4);
		graphic.setColor(Settings.Constants.UI.AWT.PLAIN_BAR_COLOR);
		graphic.fillRect(encodedLimit, 2, plainLimit - encodedLimit - 2, height - 4);
		graphic.setColor(Color.black);
		graphic.drawLine(0, height - 2, plainLimit - 2, height - 2);
		for (int i = 1 + width / 10; i < plainLimit - 2; i += width / 10)
			graphic.drawLine(i, 2, i, height - 3);
		graphic.setColor(Color.lightGray);
		for (int i = width / 10; i < plainLimit; i += width / 10)
			graphic.drawLine(i, 1, i, height - 2);
		graphic.setColor(Color.black);
	}

	private int width = 0;
	private int height = 0;

	private void setSize() {
		Dimension size = getSize();
		width = size.width;
		height = size.height;
	}

	/** Repaints this progress bar.
	 ** Note that althougth this is not <em>globally</em> synchronized to
	 ** enhance performance this is still thread-safe.
	 **/
	public void paint(Graphics graphic) {
		setSize();
		paintBorders(graphic);
		paintBars(graphic);
		paintText(graphic);
	}

	/** Repaints this progress bar. **/
	public void update(Graphics graphic) { paint(graphic); }

}

