

package sam.ui.basic;

import sam.ui.*;
import sam.event.*;
import sam.core.Settings;
import sam.cipher.Specification;

import java.io.*;


/** Pluggable basic user interface
 ** This interface allows no interaction with the user at all:
 ** any missing or misspecified parameter leads to fatal error.
 **/
public class UserInterface extends sam.ui.UserInterface {

	static {
		if ( ! register(UserInterface.class, "basic") )  // Should never happend actually unless very strange conditions (low memory ...)
			System.err.println("Warning: could not register basic user interface");
	}

	/** The desired <CODE>PrintWriter</CODE> where messages shall be printed **/
	private final PrintWriter output;

	/** Builds a new basic user interface.
	 ** Messages are printed onto standard output (which is <em>not</em> the standard <em>error</em> output)
	 ** @see java.lang.System#out
	 **/
	public UserInterface() { this( new PrintWriter(System.out) ); }

	/** Builds a new basic user interface.
	 ** Messages are printed onto specified output.
	 ** @param output the desired <CODE>OutputStream</CODE> where messages shall be printed
	 **/
	public UserInterface(OutputStream output) { this( new PrintWriter(new OutputStreamWriter(output)) ); }

	/** Builds a new basic user interface.
	 ** Messages are printed onto specified output.
	 ** @param output the desired <CODE>PrintWriter</CODE> where messages shall be printed
	 **/
	public UserInterface(final PrintWriter output) { this.output = output; }

	public final RulingInterface ruling() {
		return new RulingInterface() {
			protected String enterString(String title, String suggestedValue) {
				missing(null, title);
				return suggestedValue;  // Fake for compiler
			}
			protected int enterInteger(String title, int suggestedValue, int min, int max) {
				missing(null, title);
				return suggestedValue;  // Fake for compiler
			}
			protected File enterFileOrDirectory(String title, File suggestedValue, AccessMode mode) {
				missing(null, title);
				return suggestedValue;  // Fake for compiler
			}
			protected String choose(String title, String[] values) {
				missing(null, title);
				return null;  // Fake for compiler
			}
			protected Specification[] enterCiphers(String title, Specification[] ciphers, boolean canAddCiphers) {
				missing(canAddCiphers ? "ciphers specification" : "passwords", "the cipher stack");
				return null;  // Fake for compiler
			}
			private void missing(String type, String name) {
				print("Missing " + ((type == null) ? "argument" : type) + " for " + name, "program aborted." );
				terminate("Invalid syntax");
			}
		};
	}

	public final RunningInterface running() {
		return new RunningInterface() {
			public void print(String title, String message) {	UserInterface.this.print(title, message); }
			public FileEventListener getFileEventListener() { return null; }
			public ProgressEventListener getProgressEventListener() { return null; }
			public boolean yesNo(String title, String question) { return false; }
		};
	}

	public final ErrorInterface error() {
		return new ErrorInterface() {
			public void unhandledParameter(String parameter) { error("Unhandled parameter", parameter); }
			public void incompatibleArchiveLayout(int version) { error("Incompatible version", "" + version); }
			public void handleException(String message, java.lang.Throwable e) { error("Exception occured: ",
																																								 message + " since " + e.getMessage());	}
			public void warn(String message) { print("Warning:", message); }
			protected void failure(String message) { error("application failed", message); }
			private void error(String title, String message) {
				print("Error: " + title, message);
				terminate("Fatal error.");
			}
		};
	}

	public LogInterface log() {
		return new LogInterface() {
			protected void log(String message) { print(Settings.Defaults.LOG_PREFIX, message); }
		};
	}

	private void print(String title, String message) {
		output.println( ((title == null) ? "" : (title + ": ")) + message);
		output.flush();
	}

	private void terminate(String cause) throws Abort {
		output.println("Aborted");
		throw new Abort(cause);
	}

}
