
package sam.ui.text;

import sam.ui.*;
import sam.event.ProgressEvent;
import sam.core.Settings;
import sam.toolkit.StringFactory;


/** Text-based progress event listener
 ** This outputs progress information one line per entry.
 **/
public class ProgressEventListener implements sam.event.ProgressEventListener {

	/** The user interface to attach to **/
	private final UserInterface ui;

	/** The string factory for creating the "done" progress bar **/
	private static final StringFactory DONE_FACTORY = new StringFactory("=");

	/** The string factory for creating the "remaining" progress bar **/
	private static final StringFactory REMAINING_FACTORY = new StringFactory("-");

	/** <CODE>String</CODE> that erases the whole line once displayed **/
	private static final String DEL_LINE = new StringFactory(Settings.Constants.BACKSPACE).get(Settings.Constants.UI.Text.LINE_LENGTH);

	private boolean hasFinished = false;

	ProgressEventListener(UserInterface ui) { this.ui = ui; }

	/** This outputs progress information one line per entry.
	 ** The actual output is <CODE>apt-get</CODE> alike:
	 ** <pre><codeblock>
	 ** mydir/mydir/myfile1: 831kB (-31%) at 123kB/s
	 ** </codeblock></pre>
	 **/
	public void progress(ProgressEvent event) {
		if (hasFinished)
			return;
		if ( event.entry.isFinished() )
			ui.println(null, DEL_LINE + event.currentEntryName + " " + event.entry.limitSizeString()
								 + " (" + event.entry.getCompressionRatioString() + ") at " + event.entry.getRate() + 
								 StringFactory.SPACES.get(Settings.Constants.UI.Text.LINE_LENGTH - 40) );
		else {
			ui.print(DEL_LINE
							 + event.numberOfCurrentEntry + "/" + event.totalEntriesCount + " "
							 +	event.currentEntryShortName(Settings.Constants.UI.Text.ENTRY_SHORT_NAME_LENGTH) + ": "
							 + event.entry.achievedPercentage() + "% of " + event.entry.limitSizeString() + " (" + event.entry.getRate() + ") "
							 + getProgressBar(event.entry.achievedPercentage() / (float)100.0) + " (" + event.global.achievedPercentage() + "%) "
							 + event.global.getRemainingDurationString() + "    "
							 );
		}
		if (event.global.isFinished()) {
			hasFinished = true;
			ui.println(null, "");
			ui.println("Finished", event.totalEntriesCount + " files = " + event.global.limitSizeString() + " (" +
								 event.global.getCompressionRatioString() + ") at " + event.global.getRate() );
		}
	}

	/** Gets text-based a progress bar corresponding to specified achieved percentage **/
	protected static String getProgressBar(float achievedPercentage) {
		final int achievedLength = (int) Math.ceil(achievedPercentage * Settings.Constants.UI.Text.PROGRESS_BAR_LENGTH);
		return "[" + DONE_FACTORY.get(achievedLength) +
				REMAINING_FACTORY.get(Settings.Constants.UI.Text.PROGRESS_BAR_LENGTH - achievedLength) + "]";
	}


}
