

package sam.ui.text;

import sam.ui.*;
import sam.event.*;
import sam.core.Settings;
import sam.toolkit.Text;

import java.io.*;


/** Pluggable text user interface
 ** This interface allows no interaction with the user at all:
 ** any missing or misspecified parameter leads to fatal error.
 **/
public class UserInterface extends sam.ui.UserInterface {

	static {
		if ( ! register(UserInterface.class, "text") )  // Should never happend actually unless very strange conditions (low memory ...)
			System.err.println("Warning: could not register text user interface");
	}

	/** The desired <CODE>PrintWriter</CODE> where messages shall be printed **/
	private final PrintWriter output;

	/** The desired <CODE>InputStream</CODE> to read user-supplied informations from **/
	private final InputStream input;

	/** Builds a new text user interface.
	 ** Messages are printed onto standard output (which is <em>not</em> the standard <em>error</em> output)
	 ** @see System#out
	 **/
	public UserInterface() { this(new PrintWriter(System.out), System.in); }

	/** Builds a new text user interface.
	 ** Messages are printed onto specified output.
	 ** @param output the desired <CODE>OutputStream</CODE> where messages shall be printed
	 **/
	public UserInterface(OutputStream output) { this(new PrintWriter(new OutputStreamWriter(output)), System.in); }

	/** Builds a new text user interface.
	 ** Messages are printed onto specified output.
	 ** @param output the desired <CODE>PrintWriter</CODE> where messages shall be printed
	 **/
	public UserInterface(PrintWriter output, InputStream input) {
		this.output = output;
		this.input = input;
	}

	public final sam.ui.RulingInterface ruling() { return new RulingInterface(this); }

	public final RunningInterface running() {
		return new RunningInterface() {

			public void print(String title, String message) {	UserInterface.this.println(title, message); }

			public boolean yesNo(String title, String question) {
				UserInterface.this.print(title, question);
				char answer;
				do {
					UserInterface.this.print(" (Yes/No) ? ");
					answer = Character.toUpperCase( readChar() );
				} while ((answer != 'Y') && (answer != 'N'));
				return (answer == 'Y');
			}

			public FileEventListener getFileEventListener() { 
				return new FileEventAdapter() {
					public void failed(FileEvent event)      { println("Failed entry",      event.toString() ); }
					public void missing(FileEvent event)     { println("Missing entry",     event.toString() ); }
				};
			}

			public sam.event.ProgressEventListener getProgressEventListener() { return new ProgressEventListener(UserInterface.this); }

		};
	}

	public final ErrorInterface error() {
		return new ErrorInterface() {

			public void unhandledParameter(String parameter) { error("Unhandled parameter", parameter); }

			public void incompatibleArchiveLayout(int version) { error("Incompatible version", "" + version); }

			public void handleException(String message, java.lang.Throwable e) {
				println("Exception occured", message + " since " + Text.getMessage(e) );
e.printStackTrace();
				if ( UserInterface.get().running().yesNo("Kill program") )
					failure( (e == null) ? "" : e.getMessage() );
			}

			public void warn(String message) { println("Warning:", message); }

			protected void failure(String message) { error("application failed", message); }

			private void error(String title, String message) {
				println("Error: " + title, message);
				terminate("Fatal error.");
			}

		};
	}

	public LogInterface log() {
		return new LogInterface() {
			protected void log(String message) { println(Settings.Defaults.LOG_PREFIX, message); }
		};
	}

	/** Reads one line from the input **/
	String readString() {
		try {
			return new LineNumberReader( new InputStreamReader(input) ).readLine();
		} catch (IOException e) {
			System.err.println("I/O error reading user message aborted");
			return "";
		}
	}

	/** Reads one character from the input **/
	char readChar() {
		try {
			if (Settings.Environment.STDIN_IS_SYNCRHONOUS)
				return Text.intToChar( input.read() );
			else {
				String string = readString();
				return (string.length() < 1) ? (char) 0 :string.charAt(0);
			}
		} catch (IOException e) {
			System.err.println("I/O error reading user key aborted");
			return (char) 0;
		}
	}

	void print(String title, String message) {
		if (title != null) {
			output.print(title);
			output.print(": ");
		}
		if (message != null)
			output.print(message);
		output.flush();
	}

	void print(String message) { print(null, message); }

	void println(String title, String message) {
		print(title, message);
		output.println();
		output.flush();
	}

	void terminate(String cause) throws Abort {
		println(null, "Aborted");
		throw new Abort(cause);
	}

}
